/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.apps;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.ParseException;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.TimerTask;

import edu.uml.giro.gambit.core.GambitConstants;
import edu.uml.giro.gambit.core.GambitCore;
import edu.uml.giro.gambit.core.LocalDataOneSiteAllTimes;
import edu.uml.giro.gambit.database.GambitDatabase;
import edu.uml.giro.gambit.database.StationLocation;
import edu.uml.lgdc.appuserdb.api.DatabaseCredentials;
import edu.uml.lgdc.appuserdb.api.UserRole;
import edu.uml.lgdc.appuserdb.constants.Application;
import edu.uml.lgdc.database.ConnectList;
import edu.uml.lgdc.database.DatabaseContents;
import edu.uml.lgdc.geospace.CharChoice;
import edu.uml.lgdc.geospace.EarthGrid;
import edu.uml.lgdc.geospace.Geomap;
import edu.uml.lgdc.project.CommandLineParam;
import edu.uml.lgdc.project.Console;
import edu.uml.lgdc.project.SharedResourcesDir;
import edu.uml.lgdc.time.TimeScale;


/**
 * Example application to retrieve data records from GAMBIT database  
 * 
 * Uses GambitReader.ini to specify database connection details
 * 
 * 
 * @author Galkin
 *
 */
public class GambitReaderExample {
	
	public static final String INI_FILE_NAME = "GambitReader.ini";
	public static final String VERSION = "GAMBIT Database Reader Example v0.1";

	static {
		// read credentials for accessing GAMBIT database
		GambitReaderSettings settings = new GambitReaderSettings(INI_FILE_NAME);
		Map<DatabaseContents,ConnectList> connectorsToContent = new HashMap<>();
		connectorsToContent.put(DatabaseContents.IRTAM, settings.getGambitDbConnectSettings().getGambitConnectList());
		DatabaseCredentials.set(connectorsToContent);
	}
	
	protected static GambitCore gambitCore = new GambitCore( GambitDatabase.DBERROR_QUIT );
	static {
		gambitCore.setUserRole(UserRole.USER);
	}

	public GambitReaderExample() {
		
		System.out.println(VERSION);
		System.out.println("Online with Gambit Database at " + DatabaseCredentials.getDatabaseConnectItem(DatabaseContents.IRTAM).getURL());

		// reading example for 2017.01.01 00:00 
		TimeScale tov = new TimeScale(2017,1,1,0,0,0);
		
		// read GAMBIT database, fill core with data
		System.out.println("Reading GAMBIT record for "+tov.toHumanUT()+"...");
		gambitCore.setTimeOfValidity( tov, GambitDatabase.READ);
		System.out.println("done.");
		
		// core has synthesizers for climate, weather, and deviation maps
		// core has pre-computed maps for a default spatial resolution 
		// core has LocalData with source information used for assimilation
		
		// get weather map for foF2... see earthGrid for dimensions
		Geomap map1 = gambitCore.getChar(CharChoice.FOF2).weatherMap;
		EarthGrid earthGrid = map1.getEarthGrid();
		double[][] mapValues = map1.getMap();
		
		// get source observations of hmF2 for sensor location MHJ45
		
		LocalDataOneSiteAllTimes oneDayMHJ45 = gambitCore.getLocalData(CharChoice.HMF2).getSite("MHJ45");
		double[] observedValues = oneDayMHJ45.getObservations();
		// in order to compare to quiet-time hmF2:
		double[] climateValues = oneDayMHJ45.getClimateValues();
		
		// All sensor locations that were available at that TOV time
		System.out.print("Stations contributed to IRTAM computation at "+tov.toHumanUT()+": ");
		LocalDataOneSiteAllTimes[] sites = gambitCore.getLocalData(CharChoice.HMF2).getSites();
		int i = 0;
		for( LocalDataOneSiteAllTimes site : sites ) { 
			if( site != null && !site.isUnused() ) {
				if( i++%10 == 0 )
					System.out.println();
				StationLocation station = site.getLocation();
				System.out.print(station.ursi+" ");
			}
		}
		System.out.println("\nFinished.");
	}
	
	public static void main(String[] args) {
		CommandLineParam commandLineParam = new CommandLineParam(args);
		String errMsg = commandLineParam.checkParams(GambitConstants.KEYS, GambitConstants.KEY_TYPES, true);
		if (errMsg != null) {
			System.out.println(errMsg);
			System.exit(1);
		}
		String result = SharedResourcesDir.readAndSet(commandLineParam);
		if (result != null && result != SharedResourcesDir.NOT_SET) {
			Console.showError(result);
			System.exit(1);
		}
		
		GambitReaderExample grex = new GambitReaderExample();
	}
}
