/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.coefficients;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.MissingResourceException;

import edu.uml.giro.gambit.expansions.ExpansionBasis2D;
import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.project.Console;

/**
 * Contains coefficient related operations.
 *
 */
public abstract class Coefficients2D extends Coefficients<double[][]> {
	double[][] coeffs = null;

	@Override
	public double[][] get() {
		return coeffs;
	}

	@Override
	public void put(double[][] coeffs) {
		this.coeffs = coeffs;
	}

	public void putOne( int i, int j, double value ) {
		coeffs[i][j] = value;
	}
	
	public double[][] clone() {
		if (coeffs == null)
			return null;
		double[][] newc = new double[coeffs.length][coeffs[0].length];
		for (int i = 0; i < coeffs.length; i++)
			for (int j = 0; j < coeffs[i].length; j++)
				newc[i][j] = coeffs[i][j];
		return newc;
	}
	
	@Override
	public void reset() {
		if (coeffs != null)
			for (int i = 0; i < coeffs.length; i++)
				for (int j = 0; j < coeffs[i].length; j++)
					coeffs[i][j] = 0;
	}

	@Override
	public void exportCoefficientsToTextFile(PrintWriter writer) {
		int count = 0;
		for (int i = 0; i < coeffs.length; i++)
			for (int j = 0; j < coeffs[i].length; j++) {
				count++;
				writer.write(" " + FC.DoubleToString(coeffs[i][j], 15, 8));
				if (count == 4) {
					writer.write("\n");
					count = 0;
				}
			}
	}

	/**
	 * Though theoretically it returns <b>null</b> if IOException happens, but in practice it will never happened as all streams here 
	 * are based on limited length arrays.
	 */
	@Override public ByteArrayInputStream toInputStream() {
		if (coeffs == null)
			throw new MissingResourceException("Call to build a coefficients input stream without coefficients", this.getClass().getName(), null);
		if (basis == null)
			throw new MissingResourceException("Call to build a coefficients input stream without basis", this.getClass().getName(), null);

		ExpansionBasis2D basis2d = (ExpansionBasis2D) basis;
		ByteArrayOutputStream byteAOS = new ByteArrayOutputStream(basis2d.getCompactSpatialBasisLength() * basis2d.getTemporalBasisLength() * 8);
		try {
			DataOutputStream out = new DataOutputStream(byteAOS);
			for (int iSpatial = 0; iSpatial < basis2d.getCompactSpatialBasisLength(); iSpatial++)
				for (int iDiurnal = 0; iDiurnal < basis2d.getTemporalBasisLength(); iDiurnal++) {
					out.writeDouble(coeffs[iSpatial][iDiurnal]);
				}
			return new ByteArrayInputStream(byteAOS.toByteArray());
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Though theoretically it can be incomplete filling of array coeffs[][] in case of IOException, it never happens in practic as
	 *  
	 */
	@Override public void loadFromInputStream(InputStream is) {
		if (basis == null)
			throw new MissingResourceException("Call to load coefficients from input stream without basis", this.getClass().getName(), null);

		ExpansionBasis2D basis2d = (ExpansionBasis2D) basis;

		if (coeffs == null || coeffs.length != basis2d.getCompactSpatialBasisLength() || coeffs[0].length != basis2d.getTemporalBasisLength())
			coeffs = new double[basis2d.getCompactSpatialBasisLength()][basis2d.getTemporalBasisLength()];
		try {
			DataInputStream in = new DataInputStream(is);
			for (int iSpatial = 0; iSpatial < basis2d.getCompactSpatialBasisLength(); iSpatial++)
				for (int iDiurnal = 0; iDiurnal < basis2d.getTemporalBasisLength(); iDiurnal++) {
					coeffs[iSpatial][iDiurnal] = in.readDouble();
				}
			in.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
