/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.coefficients;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.MissingResourceException;

import edu.uml.giro.gambit.expansions.ExpansionBasis2D;
import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.project.CommonConst;
import edu.uml.lgdc.time.TimeScale;

/**
 * Contains coefficient related operations.
 * 
 * @author Artem Vesnin
 *
 */
public class Coefficients_foF2 extends Coefficients2D {
	// URSI-model file prefix
	static public final String URSI_MODEL_PREF = "URSI";
	// URSI-model file extension, without dot
	static public final String URSI_MODEL_EXT = "ASC";

	// file I/O optimization
	private int monthPrev = -1;
	private double rgPrev = -1;

	@Override
	public void loadFromClimateFile(int month, double sunspotNumber) {
		// this is URSI format
		final int LOW_R = 0;
		final int HIGH_R = 1;
		if (basis == null)
			throw new MissingResourceException("Call to read coefficients from climate file without basis", this.getClass().getName(), null);

		ExpansionBasis2D basis2d = (ExpansionBasis2D) basis;

		if (coeffs == null || coeffs.length != basis2d.getCompactSpatialBasisLength() || coeffs[0].length != basis2d.getTemporalBasisLength()) {
			coeffs = new double[basis2d.getCompactSpatialBasisLength()][basis2d.getTemporalBasisLength()];
		}
		// URSI files have two sets of coefficients for two levels of solar
		// activity
		double[][][] foF2 = new double[2][basis2d.getCompactSpatialBasisLength()][basis2d.getTemporalBasisLength()];
		double rr1, rr2;

		if (TimeScale.checkMonth(month)) {
			double rg = sunspotNumber > 150.0 ? 150.0 : sunspotNumber;
			if (month != monthPrev || rg != rgPrev) {
				// .....only for a new month and/or a new sunspot number,
				// read two ursi coefficient sets for given month and low/high
				// solar activity
				// then interpolate for the given solar activity rg
				File file = new File(CommonConst.getShareResourcesDir(), URSI_MODEL_PREF + (10 + month) + "." + URSI_MODEL_EXT);
				String fileURSI = file.getPath();
				BufferedReader in = null;
				String line;
				int i = 0, j = 0, k = 0;
				try {
					in = new BufferedReader(new FileReader(file));
					while (in.ready()) {
						line = in.readLine();
						if (line == null)
							break;
						for (int m = 0; m < 4; m++) {
							foF2[i][j][k] = FC.StringToDouble(line.substring(1 + m * 15, 1 + (m + 1) * 15));
							k++;
							if (k >= basis2d.getTemporalBasisLength()) {
								k = 0;
								j++;
								if (j >= basis2d.getCompactSpatialBasisLength()) {
									j = 0;
									i++;
								}
							}
						}
						if (i == 2)
							break;
					}
				} catch (IOException ex) {
					System.err.println("ERROR-Coefficients_foF2: can't read file " + fileURSI);
					System.err.println(ex.toString());
				} finally {
					if (in != null) {
						try {
							in.close();
						} catch (IOException ex) {
							ex.printStackTrace();
						}
					}
				}
				//
				// Linear interpolation in solar activity, for a new value of rg
				//
				rr2 = rg / 100.0;
				rr1 = 1.0 - rr2;
				for (i = 0; i < basis2d.getCompactSpatialBasisLength(); i++) {
					for (j = 0; j < basis2d.getTemporalBasisLength(); j++) {
						coeffs[i][j] = foF2[LOW_R][i][j] * rr1 + foF2[HIGH_R][i][j] * rr2;
					}
				}
				monthPrev = month;
				rgPrev = rg;
			}
		}
	}
}
