/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.coefficients;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.MissingResourceException;

import edu.uml.giro.gambit.core.GambitConstants;
import edu.uml.giro.gambit.expansions.ExpansionBasis2D;
import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.project.CommonConst;
import edu.uml.lgdc.time.TimeScale;

/**
 * Contains coefficient related operations.
 * 
 * @author Artem Vesnin
 *
 */
public class Coefficients_hmF2 extends Coefficients2D {
	// file I/O optimization
	private int monthPrev = -1;
	private double rgPrev = -1;

	@Override
	public void loadFromClimateFile(int month, double sunspotNumber) {
		final int COEF_LENGTH = 16;
		final int LOW_R = 0;
		final int HIGH_R = 1;
		sunspotNumber = sunspotNumber > 150.0 ? 150.0 : sunspotNumber;

		if (basis == null)
			throw new MissingResourceException("Call to read coefficients from climate file without basis", this.getClass().getName(), null);

		ExpansionBasis2D basis2d = (ExpansionBasis2D) basis;

		if (coeffs == null || coeffs.length != basis2d.getCompactSpatialBasisLength() || coeffs[0].length != basis2d.getTemporalBasisLength()) {
			coeffs = new double[basis2d.getCompactSpatialBasisLength()][basis2d.getTemporalBasisLength()];
		}

		// two sets of coefficients for two levels of solar activity
		double[][][] hmF2 = new double[2][basis2d.getCompactSpatialBasisLength()][basis2d.getTemporalBasisLength()];
		double rr1, rr2;

		if (TimeScale.checkMonth(month)) {
			double rg = sunspotNumber > 150.0 ? 150.0 : sunspotNumber;
			if (month != monthPrev || rg != rgPrev) {
				// .....only for a new month and/or a new sunspot number,
				// read two ursi coefficient sets for given month and low/high
				// solar activity
				// then interpolate for the given solar activity rg
				File file = new File(CommonConst.getShareResourcesDir(), GambitConstants.ARGENTINA_COEFFS_J6_FOLDER 
						+ GambitConstants.ARGENTINA_COEFFS_J6_PREF + FC.IntegerToString(month, 2, '0') 
						+ "." + GambitConstants.ARGENTINA_COEFFS_J6_EXT);
				BufferedReader in = null;
				String line;
				int iSpatial = 0, iDiurnal = 0, iSolarActivity = 0;
				int iActivity = 0;
				try {
					in = new BufferedReader(new FileReader(file));
					while (in.ready()) {
						line = in.readLine();
						if (line == null)
							break;
						for (int iColumn = 0; iColumn < 4; iColumn++) {
							if (iActivity == 1)
								hmF2[iSolarActivity][iSpatial][iDiurnal] = Double.valueOf(line.substring(iColumn * COEF_LENGTH, (iColumn + 1)
										* COEF_LENGTH));
							iDiurnal++;
							if (iDiurnal >= basis2d.getTemporalBasisLength()) {
								iDiurnal = 0;
								iSpatial++;
								if (iSpatial >= basis2d.getCompactSpatialBasisLength()) {
									iSpatial = 0;
									iSolarActivity++;
								}
							}
						}
						if (iSolarActivity == 2) {
							iSolarActivity = 0;
							iActivity++;
						}
						if (iActivity == 2)
							break;
					}
				} catch (IOException ex) {
					throw new MissingResourceException("Failed to read coefficients from climate file "+file.getAbsolutePath(), this.getClass().getName(), null);
				} finally {
					if (in != null)
						try {
							in.close();
						} catch (IOException ex) {
							ex.printStackTrace();
						}
				}

				// Linear interpolation in solar activity, for a new value of rg
				rr2 = rg / 100.0;
				rr1 = 1.0 - rr2;
				for (int i = 0; i < basis2d.getCompactSpatialBasisLength(); i++) {
					for (int j = 0; j < basis2d.getTemporalBasisLength(); j++) {
						coeffs[i][j] = hmF2[LOW_R][i][j] * rr1 + hmF2[HIGH_R][i][j] * rr2;
					}
				}
				monthPrev = month;
				rgPrev = rg;
			}
		}
	}
}
