/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.core;

import java.io.File;
import java.util.Calendar;
import java.util.HashMap;
import java.util.concurrent.TimeUnit;

import edu.uml.giro.gambit.coefficients.Coefficients2D;
import edu.uml.giro.gambit.coefficients.Coefficients_B0;
import edu.uml.giro.gambit.coefficients.Coefficients_B1;
import edu.uml.giro.gambit.coefficients.Coefficients_foF2;
import edu.uml.giro.gambit.coefficients.Coefficients_hmF2;
import edu.uml.giro.gambit.database.GambitDatabase;
import edu.uml.giro.gambit.database.GambitDatabaseReader;
import edu.uml.giro.gambit.expansions.ExpansionBasis;
import edu.uml.giro.gambit.expansions.ExpansionBasis2D;
import edu.uml.giro.gambit.expansions.ExpansionBasis_JonesGallet;
import edu.uml.giro.gambit.expansions.ExpansionBasis_SphericalCapHarmonicsK6;
import edu.uml.giro.gambit.synthesizers.GeoSynthesizer;
import edu.uml.giro.gambit.synthesizers.Synthesizer;
import edu.uml.giro.gambit.synthesizers.Synthesizer2D;
import edu.uml.giro.gambit.synthesizers.SynthesizerSCH;
import edu.uml.lgdc.appuserdb.api.UserRole;
import edu.uml.lgdc.datatype.RingBuffer;
import edu.uml.lgdc.geospace.CharChoice;
import edu.uml.lgdc.geospace.EarthGrid;
import edu.uml.lgdc.geospace.Geomap;
import edu.uml.lgdc.geospace.MapChoice;
import edu.uml.lgdc.geospace.MapTypeChoice;
import edu.uml.lgdc.geospace.PlasmaMath;
import edu.uml.lgdc.geospace.SunSpots;
import edu.uml.lgdc.geospace.TimeGrid;
import edu.uml.lgdc.project.Console;
import edu.uml.lgdc.project.settings.Setting;
import edu.uml.lgdc.quantities.NO_VALUE;
import edu.uml.lgdc.time.TimeScale;

abstract public class Core {

	protected TimeScale timeOfValidity = new TimeScale(2011, 8, 1, 0, 0, 0);
	private UserRole userRole = UserRole.GUEST;
	
	protected Characteristic[] chars;

	protected final EarthGrid defaultEarthGrid = new EarthGrid(GambitConstants.NUM_LATITUDE_NODES, GambitConstants.NUM_LONGITUDE_NODES);
	protected final TimeGrid defaultTimeGrid = new TimeGrid();

	public AlgorithmVersions VERSIONS;
	
	protected GeoSynthesizer weather_foF2 = null, climate_foF2 = null;
	protected GeoSynthesizer weather_hmF2 = null, climate_hmF2 = null;
	protected GeoSynthesizer weather_NmF2 = null, climate_NmF2 = null;
	protected GeoSynthesizer weather_B0 = null;
	protected GeoSynthesizer weather_B1 = null;

	public Core() {

		chars = new Characteristic[CharChoice.SIZE];
		defaultTimeGrid.setUTs(new TimeScale("1932.01.01"), GambitConstants.FORECAST_DURATION_MIN);
		
		// Chars related construction
		for (CharChoice cc : CharChoice.values()) {
			switch (cc) {
			case FOF2:
				// 1. Climate setup
				// Basis
				ExpansionBasis_JonesGallet basis_foF2 = new ExpansionBasis_JonesGallet();
				basis_foF2.setup(ExpansionBasis_JonesGallet.CCIR_DIURNAL_ORDER, ExpansionBasis_JonesGallet.SPATIAL_ORDER,
						ExpansionBasis_JonesGallet.SPATIAL_DEGREE);
				// Coeffs
				Coefficients2D coeffs_foF2c = new Coefficients_foF2();
				coeffs_foF2c.setExpansionBasis(basis_foF2);
				climate_foF2 = new GeoSynthesizer(CharChoice.FOF2, basis_foF2, coeffs_foF2c, null);
				climate_foF2.setTimeGrid(defaultTimeGrid);
				climate_foF2.setEarthGrid(defaultEarthGrid);
				Geomap climateMap_foF2 = new Geomap(MapChoice.FOF2_CLIMATE, NO_VALUE.MHZ);
				climateMap_foF2.setEarthGrid(defaultEarthGrid);
				// 2. Observations setup
				LocalData observations_foF2 = new LocalData(CharChoice.FOF2, defaultTimeGrid);
				// 3. WeathGeomapmap
				Geomap weatherMap_foF2 = new Geomap(MapChoice.FOF2_WEATHER, NO_VALUE.MHZ);
				weatherMap_foF2.setEarthGrid(defaultEarthGrid);
				// 4Geomapeviations
				Geomap deviationMap_foF2 = new Geomap(MapChoice.FOF2_DELTA, NO_VALUE.MHZ);
				Geomap deviationMap_foF2_percent = new Geomap(MapChoice.FOF2_DELTA_PERCENT, NO_VALUE.MHZ);
				deviationMap_foF2.setEarthGrid(defaultEarthGrid);
				deviationMap_foF2_percent.setEarthGrid(defaultEarthGrid);
				// 5. Construct the foF2 characteristic
				chars[CharChoice.FOF2.getIndex()] = new Characteristic(CharChoice.FOF2, climateMap_foF2, climate_foF2, observations_foF2, 
						weatherMap_foF2, weather_foF2, deviationMap_foF2, deviationMap_foF2_percent);
				break;
			case NMF2:
				// 1. Climate will use foF2 synthesizer
				GeoSynthesizer climate_NmF2 = new GeoSynthesizer(CharChoice.NMF2, climate_foF2.getExpansionBasis(), null, climate_foF2);
				Geomap	climateMap_NmF2 = new Geomap(MapChoice.NMF2_CLIMATE, NO_VALUE.DENSITY);
				climateMap_NmF2.setEarthGrid(defaultEarthGrid);
				// 2. Observations setup;
				LocalData observations_NmF2 = new LocalData(CharChoice.NMF2, defaultTimeGrid );
				// 3. WeatherGeomapp
				Geomap weatherMap_NmF2 = new Geomap(MapChoice.NMF2_WEATHER, NO_VALUE.DENSITY);
				weatherMap_NmF2.setEarthGrid(defaultEarthGrid);
				// 4. DeGeomaptions
				Geomap deviationMap_NmF2 = new Geomap( MapChoice.NMF2_DELTA, NO_VALUE.SPECIAL);
				Geomap deviationMap_NmF2_percent = new Geomap(MapChoice.NMF2_DELTA_PERCENT, NO_VALUE.SPECIAL);
				deviationMap_NmF2.setEarthGrid(defaultEarthGrid);
				deviationMap_NmF2_percent.setEarthGrid(defaultEarthGrid);
				// 5. Construct the NmF2 characteristic
				chars[CharChoice.NMF2.getIndex()] = new Characteristic(CharChoice.NMF2, climateMap_NmF2, climate_NmF2, observations_NmF2, weatherMap_NmF2, 
						weather_NmF2, deviationMap_NmF2, deviationMap_NmF2_percent);
				break;
			case HMF2:
				// 1. Climate setup
				// Basis
				ExpansionBasis_JonesGallet basis_hmF2 = new ExpansionBasis_JonesGallet();
				basis_hmF2.setup(ExpansionBasis_JonesGallet.CCIR_DIURNAL_ORDER, ExpansionBasis_JonesGallet.SPATIAL_ORDER,
						ExpansionBasis_JonesGallet.SPATIAL_DEGREE);
				// Coeffs
				Coefficients_hmF2 coeffs_hmF2c = new Coefficients_hmF2();
				coeffs_hmF2c.setExpansionBasis(basis_hmF2);
				GeoSynthesizer climate_hmF2 = new GeoSynthesizer(CharChoice.HMF2, basis_hmF2, coeffs_hmF2c, null);
				climate_hmF2.setTimeGrid(defaultTimeGrid);
				climate_hmF2.setEarthGrid(defaultEarthGrid);
				Geomap climateMap_hmF2 = new Geomap(MapChoice.HMF2_CLIMATE, NO_VALUE.KM);
				climateMap_hmF2.setEarthGrid(defaultEarthGrid);
				// 2. Observations setup
				LocalData observations_hmF2 = new LocalData(CharChoice.HMF2, defaultTimeGrid);
				// 3. Weather map
				Geomap weatherMap_hmF2 = new Geomap(MapChoice.HMF2_WEATHER, NO_VALUE.KM);
				weatherMap_hmF2.setEarthGrid(defaultEarthGrid);
				// 4. Deviations
				Geomap deviationMap_hmF2 = new Geomap(MapChoice.HMF2_DELTA, NO_VALUE.SPECIAL);
				Geomap deviationMap_hmF2_percent = new Geomap(MapChoice.HMF2_DELTA_PERCENT, NO_VALUE.SPECIAL);
				deviationMap_hmF2.setEarthGrid(defaultEarthGrid);
				deviationMap_hmF2_percent.setEarthGrid(defaultEarthGrid);
				// 5. Construct the hmF2 characteristic
				chars[CharChoice.HMF2.getIndex()] = new Characteristic(CharChoice.HMF2, climateMap_hmF2, climate_hmF2, observations_hmF2, 
						weatherMap_hmF2, weather_hmF2, deviationMap_hmF2, deviationMap_hmF2_percent);
				break;
			case B0:
				// 1. Climate setup
				// Basis
				ExpansionBasis_SphericalCapHarmonicsK6 basis_B0 = new ExpansionBasis_SphericalCapHarmonicsK6();
				basis_B0.setSpatialOrder(ExpansionBasis_SphericalCapHarmonicsK6.SPATIAL_DEGREE);
				// Coeffs (IMPORTANT: has to come after basis is initialized
				Coefficients2D coeffs_B0c = new Coefficients_B0();
				coeffs_B0c.setExpansionBasis(basis_B0);
				SynthesizerSCH climate_B0 = new SynthesizerSCH(CharChoice.B0, basis_B0, coeffs_B0c);
				climate_B0.setTimeGrid(defaultTimeGrid);
				climate_B0.setEarthGrid(defaultEarthGrid);
				Geomap climateMap_B0 = new Geomap(MapChoice.B0_CLIMATE, NO_VALUE.KM);
				climateMap_B0.setEarthGrid(defaultEarthGrid);
				// 2. Observations setup
				LocalData observations_B0 = new LocalData(CharChoice.B0, defaultTimeGrid);
				// 3. Weather map
				Geomap weatherMap_B0 = new Geomap(MapChoice.B0_WEATHER, NO_VALUE.KM);
				weatherMap_B0.setEarthGrid(defaultEarthGrid);
				// 4. Geomapiations
				Geomap deviationMap_B0 = new Geomap(MapChoice.B0_DELTA, NO_VALUE.SPECIAL);
				Geomap deviationMap_B0_percent = new Geomap(MapChoice.B0_DELTA_PERCENT, NO_VALUE.SPECIAL);
				deviationMap_B0.setEarthGrid(defaultEarthGrid);
				deviationMap_B0_percent.setEarthGrid(defaultEarthGrid);
				// 5. Construct the B0 characteristic
				chars[CharChoice.B0.getIndex()] = new Characteristic(CharChoice.B0, climateMap_B0, climate_B0, observations_B0, weatherMap_B0,
						weather_B0, deviationMap_B0, deviationMap_B0_percent);
				break;
			case B1:
				// 1. Climate setup
				// Basis
				ExpansionBasis_SphericalCapHarmonicsK6 basis_B1 = new ExpansionBasis_SphericalCapHarmonicsK6();
				basis_B1.setSpatialOrder(ExpansionBasis_SphericalCapHarmonicsK6.SPATIAL_DEGREE);
				// Coeffs (IMPORTANT: has to come after basis is initialized
				Coefficients2D coeffs_B1c = new Coefficients_B1();
				coeffs_B1c.setExpansionBasis(basis_B1);
				SynthesizerSCH climate_B1 = new SynthesizerSCH(CharChoice.B1, basis_B1, coeffs_B1c);
				climate_B1.setTimeGrid(defaultTimeGrid);
				climate_B1.setEarthGrid(defaultEarthGrid);
				Geomap climateMap_B1 = new Geomap(MapChoice.B1_CLIMATE, NO_VALUE.KM);
				climateMap_B1.setEarthGrid(defaultEarthGrid);
				// 2. Observations setup
				LocalData observations_B1 = new LocalData(CharChoice.B1, defaultTimeGrid);
				// 3. Weather map
				Geomap weatherMap_B1 = new Geomap( MapChoice.B1_WEATHER, NO_VALUE.KM);
				weatherMap_B1.setEarthGrid(defaultEarthGrid);
				// 4. DeviationsGeomap
				Geomap deviationMap_B1 = new Geomap(MapChoice.B1_DELTA, NO_VALUE.SPECIAL);
				Geomap deviationMap_B1_percent = new Geomap(MapChoice.B1_DELTA_PERCENT, NO_VALUE.SPECIAL);
				deviationMap_B1.setEarthGrid(defaultEarthGrid);
				deviationMap_B1_percent.setEarthGrid(defaultEarthGrid);
				// 5. Construct the B0 characteristic
				chars[CharChoice.B1.getIndex()] = new Characteristic(CharChoice.B1, climateMap_B1, climate_B1, observations_B1, 
						weatherMap_B1, weather_B1, deviationMap_B1, deviationMap_B1_percent);
				break;
			default:
				throw new RuntimeException("design error: overlooked characteristic, " + cc.getName());
			}
			chars[cc.getIndex()].reset();
		}
	}

	//
	//-----------------------------------    T i m e O f V a l i d i t y   -------------------------------------------
	//

	
	public TimeScale getTimeOfValidity() {
		return timeOfValidity;
	}

	public void setTimeOfValidity(TimeScale newTime, int databaseMode) {
//		if( newTime.equals(timeOfValidity) )
//			return;
		TimeScale now = new TimeScale();
		if (newTime.diffIn(Calendar.SECOND, now) > 75) {
			throw new RuntimeException("Attempt to set future time " + newTime + " as validity time");
		}
		if( !userRole.getRealTimeAccess() ) {
			if( now.diffIn(Calendar.HOUR_OF_DAY, newTime) < GambitConstants.REAL_TIME_PERIOD_HOURS ) {
				Console.showError("User Role " + userRole.getName() 
						+ " does not include access to recent data, please upgrade");
				throw new RuntimeException("Data access level exceeds user registration role");
			}
		}
		timeOfValidity = newTime;
		rebuildChars(databaseMode);
	}

	public TimeGrid getDefaultTimeGrid() {
		return defaultTimeGrid;
	}

	public EarthGrid getDefaultEarthGrid() {
		return defaultEarthGrid;
	}

	/**
	 * Refreshes content of all components of all characteristics when target time changes
	 * Rebuild optimization is possible to exclude observations if they are not needed
	 * 
	 * @param databaseMode
	 */
	private void rebuildChars(int databaseMode) {
		for (Characteristic c : chars) {
			rebuildOneChar(c, databaseMode);
		}
	}

	/**
	 * Rebuild one OneDayCharacteristic (abstract class)
	 * 
	 * @param c
	 * @param databaseMode
	 */
	public void rebuildOneChar(Characteristic c, int databaseMode) {
		if( c == null ) {
			return;
		}

		// 1. Set climate coefficients and recompute climate expansion basis
		rebuildOneChar_Climate( c );

		// wrap up if only climate is requested
		if ((databaseMode & GambitDatabase.READ) == 0) {
			return;
		}

		// 2. Download Observations and build/download Weather Coefficients
		rebuildOneChar_ObservationsAndWeather( c, databaseMode );

		// 3. Fill out Deviation Maps
		c.deviationMap.buildDeviationMap(c.weatherMap.getMap(), c.climateMap.getMap(), timeOfValidity, null );
		c.deviationMap_percent.buildDeviationMap_percent(c.weatherMap.getMap(), c.climateMap.getMap(), timeOfValidity, null );
	}
	
	public void rebuildOneChar_Climate(Characteristic c) {
		int iTime;
		CharChoice cc = c.cc;
		switch( cc ) {
		
		case FOF2:
		case HMF2:
		case B0:
		case B1:
			c.climate.getExpansionBasis().setNewTime(timeOfValidity);
			int month = timeOfValidity.get(Calendar.MONTH) + 1;
			double sunspotNumber = SunSpots.get(timeOfValidity.get(Calendar.YEAR), month).getNumber();
			// load coefficients from an external file -- that is all we need to
			// rebuild climate
			c.climate.getCoefficients().loadFromClimateFile(month, sunspotNumber);
			iTime = c.climate.getTimeGrid().findClosestTimeStep(timeOfValidity);
			c.climateMap.setMapValues(c.climate.synthesizeValueMap(iTime), timeOfValidity);
			break;
		case NMF2:
			iTime = c.climate.getTimeGrid().findClosestTimeStep(timeOfValidity);
			c.climateMap.setMapValues(c.climate.synthesizeValueMap(iTime), timeOfValidity);
			break;
		}
	}
	
	abstract public void rebuildOneChar_ObservationsAndWeather(Characteristic c, int databaseMode);

	//
	//-----------------------------------    A C C E S S    T O    D A T A   -------------------------------------------------
	//

	public Characteristic getChar(CharChoice cc) {
		return chars[cc.getIndex()];
	}

	public Characteristic getChar(int index) {
		return chars[index];
	}

	public Characteristic[] getChars() {
		return chars;
	}

	public GeoSynthesizer getClimate(CharChoice cc) {
		return chars[cc.getIndex()].climate;
	}

	public LocalData getLocalData(CharChoice cc) {
		return chars[cc.getIndex()].localData;
	}

	public GeoSynthesizer getWeather(CharChoice cc) {
		return chars[cc.getIndex()].weather;
	}

	public Geomap getDeviationMap(CharChoice cc) {
		return chars[cc.getIndex()].deviationMap;
	}

	public Geomap getDeviationMap_percent(CharChoice cc) {
		return chars[cc.getIndex()].deviationMap_percent;
	}
	
	static public double[][] synthesizeGk(ExpansionBasis2D basis, int k, EarthGrid grid) {
		double[][] valueGrid = new double[grid.getNumLatitudeNodes()][grid.getNumLongitudeNodes()];
		for (int iLat = 0; iLat < grid.getNumLatitudeNodes(); iLat++) {
			for (int iLon = 0; iLon < grid.getNumLongitudeNodes(); iLon++) {
				valueGrid[iLat][iLon] = basis.getSpatialBasisFunctionValue(k, iLat, iLon);
			}
		}
		return valueGrid;
	}

	/**
	 * Synthesize 1D map given a custom Earth grid
	 * @param exportEarthGrid
	 * @param buildToDate
	 * @param mc
	 * @return
	 */
	
	
	public double[] synthesizeCurrentMap_1D(EarthGrid exportEarthGrid, TimeScale buildToDate, MapChoice mc ) {
		return synthesizeCurrentMap( exportEarthGrid, buildToDate, mc ).getMap_1D();
	}
	
	/**
	 * Synthesize geomap given a custom Earth grid dimensions (rather than the internal grid)
	 * 
	 * @param exportEarthGrid
	 * @param buildToDate
	 * @param mc
	 * @return
	 */
	public Geomap synthesizeCurrentMap(EarthGrid exportEarthGrid, TimeScale buildToDate, MapChoice mc ) {
		
		CharChoice cc = mc.getCC();

		Geomap outputMap;
		
		// prepare custom EarthGrid
		EarthGrid savedGridClimate = getClimate(cc).getEarthGrid();
		EarthGrid savedGridWeather = getWeather(cc).getEarthGrid();
		getClimate(cc).setEarthGrid(exportEarthGrid);
		// IMPORTANT: set basis time to get magnetic field part of the basis 
		getClimate(cc).getExpansionBasis().setNewTime(buildToDate);
		getWeather(cc).setEarthGrid(exportEarthGrid);
		getWeather(cc).getExpansionBasis().setNewTime(buildToDate);
		
		// hmmm... assuming time grids are the same throughout... Geomapht?
		int iTime = getWeather(cc).getTimeGrid().findClosestTimeStep(buildToDate);
		
		
		// Because a custom EarthGrid is needed, Geomap new Maps have to be recalculated
		Geomap climateMap = null, weatherMap = null, differenceMap = null;
		
		if( mc.isClimateMap() || mc.isDeltaMap() ) {
			climateMap = new Geomap(mc, NO_VALUE.SPECIAL);
			climateMap.setEarthGrid(exportEarthGrid);
			climateMap.setMapValues(getClimate(cc).synthesizeValueMap(iTime), buildToDate );
		}
		
		if( mc.isWeatherMap() || mc.isDeltaMap() ) {
			weatherMap = new Geomap(mc, NO_VALUE.SPECIAL);
			weatherMap.setEarthGrid(exportEarthGrid);
			weatherMap.setMapValues(getWeather(cc).synthesizeValueMap(iTime), buildToDate );
		}
		
		if( mc.isDeltaMap() ) {
			differenceMap = new Geomap(mc, NO_VALUE.SPECIAL);
			differenceMap.setEarthGrid(exportEarthGrid);
			if( mc.isPercentDeltaMap() ) {
				differenceMap.buildDeviationMap_percent(weatherMap.getMap(), climateMap.getMap(), buildToDate, null);
			} else {
				differenceMap.buildDeviationMap(weatherMap.getMap(), climateMap.getMap(), buildToDate, null);
			}
			outputMap = differenceMap;
		} else {
			outputMap = mc.isClimateMap() ? climateMap : weatherMap;
		}
			
		getClimate(cc).setEarthGrid(savedGridClimate);
		getClimate(cc).getExpansionBasis().setNewTime(buildToDate);
		getWeather(cc).setEarthGrid(savedGridWeather);
		getWeather(cc).getExpansionBasis().setNewTime(buildToDate);
			
		return outputMap;
	}

	public UserRole getUserRole() {
		return userRole;
	}

	public void setUserRole(UserRole userRole) {
		this.userRole = userRole;
	}
	
}
