/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.core;

import java.awt.Font;
import java.io.File;
import java.io.IOException;
import java.text.ParseException;

import edu.uml.lgdc.time.TimeScale;
import edu.uml.lgdc.geospace.CharChoice;
import edu.uml.lgdc.geospace.MapChoice;
import edu.uml.lgdc.project.CommandLineParam;
import edu.uml.lgdc.project.CommandLineParam.KeyType;
import static edu.uml.lgdc.project.CommandLineParam.KeyType.*;

/**
 */
public class GambitConstants {
	static public final String[] KEYS = { "sd", "rproc", "dtodproc", "avproc", "offset", "startdate", "enddate", "procdiagram", "saveframes",
		"charident", "update_db", "showsites", "grayoutsites", "excluded_assimilation", "xpos", "ypos", "controlsite", "exportcoeffs", "submit_db",
		"excluded_chars"};
	static public final KeyType[] KEY_TYPES = { VALUE_ONLY, KEY_ONLY, KEY_ONLY, KEY_ONLY, VALUE_ONLY, VALUE_ONLY, VALUE_ONLY, KEY_ONLY, 
		VALUE_ONLY, VALUE_ONLY, VALUE_ONLY, KEY_ONLY, KEY_ONLY, VALUE_ONLY,	VALUE_ONLY, VALUE_ONLY, VALUE_ONLY, KEY_ONLY, KEY_ONLY,  
		VALUE_ONLY };

	static {
		if (KEYS.length != KEY_TYPES.length)
			throw new RuntimeException("KEYS.length != KEY_TYPES.length");
	}

	static public final String VERSION = "GAMBIT v0.9.05";
	static public final String FILE_CREDENTIALS = "# GLOBAL IONOSPHERE RADIO OBSERVATORY (G.I.R.O.)\n# giro.uml.edu\n#\n# IRI-based Real-Time Assimilative Model (IRTAM)\n#\n";
	static public final String EXPORT_START_HEADER = "# START_HEADER"+System.lineSeparator();
	static public final String EXPORT_END_HEADER = "# END_HEADER"+System.lineSeparator();

	static public final String LOG_LAST_COEFFS_SUBMISSION = "coeffs_proc.log";

	static public final String DATE_FORMAT = "yyyy.MM.dd.hh.mm";
	static public final String FULL_DATE_FORMAT = "yyyy/MM/dd (DDD) HH:mm:ss.SSS";

	static public final int INDEX_MIN = 0;
	static public final int INDEX_MAX = 1;
	static public final int INDEX_AVERAGE = 2;

	public static final double ERROR_RANGE_STEP[] = { 0.1, 200, 2.5 };
	public static final int ERROR_RANGE_NUM_BINS = 61;

	// errors are approximate and are known from previous IRTAM runs
	static public final double AVERAGE_ERRORS_FOF2 = 0.38;
	static public final double AVERAGE_ERRORS_NMF2 = 1000.;
	static public final double AVERAGE_ERRORS_HMF2 = 17.0;
	static public final double[] AVERAGE_ERRORS = { AVERAGE_ERRORS_FOF2, AVERAGE_ERRORS_NMF2, AVERAGE_ERRORS_HMF2 };

	// Expansion synthesizer constants
	static public final double REFERENCE_ALTITUDE_KM = 300.0;

	static public final double HUE_BLUE = 240d / 360d;
	static public final double HUE_RED = 0d / 360d;
	// These constants are used to reduce effects of incorrect data on
	// displaying
	static public final double USED_AVARAGE_FOR_SITES = 0;
	static public final double USED_SIGMA_FOR_SITES = 2.;

	// color saturation factor: reduction of the min-max values when mapping
	// deviation maps to color.
	// smaller factor means greater saturation
	static public double COLOR_SATURATION_FACTOR = 1.0;

	public static String TYPE_IMAGES = "Images"; // type of annotation

	// Diurnal and spatial grid constants
	static public final int NUM_LONGITUDE_NODES = 45; // longitude nodes dimension
	static public final int NUM_LATITUDE_NODES = 46; // latitude nodes dimension; includes both poles
	static public final int LATITUDE_RANGE_DEGREE = 180;
	static public final int LONGITUDE_RANGE_DEGREE = 360;
	static public final int LATITUDE_STEP = LATITUDE_RANGE_DEGREE / (NUM_LATITUDE_NODES - 1); // -1 to exclude poles
	static public final int LONGITUDE_STEP = LONGITUDE_RANGE_DEGREE / NUM_LONGITUDE_NODES;
	static public final double LONDEGREE_IN_HOUR = 15.;
	
	static public final int FORECAST_DURATION_MIN = 15;//300;

	static public final String USER_FOLDER = "user" + File.separator;
	static public final String OUTPUT_FOLDER = "output" + File.separator;
	static public final String INPUT_FOLDER = "input" + File.separator;
	static public final String MOVIES_OUTPUT_FOLDER = "Animations" + File.separator;
	static public final String WEBPAGE_MOVIES_FOLDER = "RTAM" + File.separator;
	static public final String TEMP_FOLDER = OUTPUT_FOLDER + "temp" + File.separator;
	static public final String GPS_TEC_FOLDER = INPUT_FOLDER + "gps_tec" + File.separator;
	static public final String KP_AP_FOLDER = INPUT_FOLDER + "kp_ap" + File.separator;
	static public final String DST_FOLDER = INPUT_FOLDER + "dst" + File.separator;
	static public final String AVTEC_FOLDER = INPUT_FOLDER + "avtec" + File.separator;
	static public final String REPORT_IMAGE_FOLDER = OUTPUT_FOLDER + "graphs" + File.separator;

	static public final String ARGENTINA_COEFFS_J6_FOLDER = "hmf2_J6" + File.separator;
	static public final String ARGENTINA_COEFFS_J6_PREF = "uf_uh.grid.";
	static public final String ARGENTINA_COEFFS_J6_EXT = "asc";

	static private final String RTAM_FILENAME_PREF_FOF2 = "IRTAM_foF2";
	static private final String RTAMVSIRI_FILENAME_PREF_FOF2 = "IRTAMvsIRI_foF2";
	static private final String RTAMVSIRIPERCENT_FILENAME_PREF_FOF2 = "IRTAMvsIRI_foF2_per";
	static private final String RTAM_FILENAME_PREF_NMF2 = "IRTAM_NmF2";
	static private final String RTAMVSIRI_FILENAME_PREF_NMF2 = "IRTAMvsIRI_NmF2";
	static private final String RTAMVSIRIPERCENT_FILENAME_PREF_NMF2 = "IRTAMvsIRI_NmF2_per";
	static private final String RTAM_FILENAME_PREF_HMF2 = "IRTAM_hmF2";
	static private final String RTAMVSIRI_FILENAME_PREF_HMF2 = "IRTAMvsIRI_hmF2";
	static private final String RTAMVSIRIPERCENT_FILENAME_PREF_HMF2 = "IRTAMvsIRI_hmF2_per";
	// create working directories if any does not exists
	static {
		(new File(USER_FOLDER)).mkdirs();
		(new File(OUTPUT_FOLDER)).mkdirs();
		(new File(INPUT_FOLDER)).mkdirs();
		(new File(MOVIES_OUTPUT_FOLDER)).mkdirs();
		(new File(TEMP_FOLDER)).mkdirs();
		(new File(GPS_TEC_FOLDER)).mkdirs();
	}

	static public final String FILEPREF_GRID = "grid";
	static public final String FILEPREF_GRID_IONEX = "ionex";
	static public final String FILEPREF_SAMELT = "SameLT";
	static public final String FILEPREF_LONGSLICE = "LonSlice";
	static public final String FILE_GUEST_INFO = "guest_info.txt";
	static public final String FILE_GAMBIT_INFO = "gambit_info.txt";
	static public final String FILE_DISTANCES = "distances.txt";
	static public final String FILE_GIRO_INFO = "giro_stations.txt";
	static public final String FILE_MODEL_RESOLUTION = "model_resolution.txt";

	// display constants
	static public final Font LEGEND_DEFAULT_FONT = Font.decode("Arial-BOLD-12");
	static public final Font COLOR_LEGEND_DEFAULT_FONT = Font.decode("Arial-BOLD-12");
	static public final int DEFAULT_HEIGHT = 640;
	static public final int DEFAULT_WIDTH = 1060;
	static public final int DEFAULT_WIDTH_PANEL = 385;
	static public final int DEFAULT_WIDTH_DISPLAY = DEFAULT_WIDTH - DEFAULT_WIDTH_PANEL;
	static public final int DEFAULT_HEIGHT_DISPLAY = DEFAULT_HEIGHT - 5;
	static public final int VIEW_TYPE_FLAT = 0;
	static public final int VIEW_TYPE_ROUND = 1;
	static public final int VIEW_TYPE_FLAT_WEB = 2;

	static public final int WORKING_WIDTH = 500;

	static public final int LEGEND_HEIGHT_ON_SCREEN = 100;
	static public final int LEGEND_WIDTH_ON_SCREEN = WORKING_WIDTH / 2 - 80;
	static public final int TWO_LEGENDS_WIDTH_ON_SCREEN = (LEGEND_WIDTH_ON_SCREEN+80) * 2 ;
	static public final int XOFFSET_FOR_TWO_LEGENDS = WORKING_WIDTH / 4 - 5;
	static public final int YOFFSET_FOR_EXT_LEGEND = -80;
	static public final int XSHIFT_EXTRA_LINUX = 5;

	static public final int WEBVIEW_MOVIE_ROI_WIDTH = GambitConstants.WORKING_WIDTH;
	static public final int WEBVIEW_MOVIE_ROI_HEIGHT = 360;
	static public final int WEBVIEW_MOVIE_ROI_HEIGHT_WITH_LEGEND = WEBVIEW_MOVIE_ROI_HEIGHT + 150;
	static public final int WEBVIEW_MOVIE_ROI_XPOSITION = 92;
	static public final int WEBVIEW_MOVIE_ROI_YPOSITION = 75;

	static public final int LEGEND_ROUND_TO_FOR_PERCENT_MAP = 10;
	static public final int LEGEND_ROUND_TO_FOR_HMF2_DEVIATION = 10;
	static public final int LEGEND_ROUND_TO_FOR_HMF2_OBS = 50;

	static public final double MIN_FOF2_MHZ = 2.0;
	static public final double MIN_NMF2_CM3 = 50000.0;
	static public final double MIN_HMF2_KM = 200.0;
	static public final double MIN_B0_KM = 50.0;
	static public final double MIN_B1 = 1.0;
	static public final double MIN_VTEC = 5;
	static public final double MIN_TAU_KM = 140.0;
	static public final double MIN_CUSTOM = 600.0;

	static public final double[] USER_MINMAX_FOF2 = { -3., 3. };
	static public final double[] USER_MINMAX_NMF2 = { -10000., 10000. };
	static public final double[] USER_MINMAX_HMF2 = { -100., 100. };
	static public final double[] USER_MINMAX_B0 = { -100., 100. };
	static public final double[] USER_MINMAX_B1 = { -2., 2. };
	static public final double[] USER_MINMAX_VTEC = { -20., 20. };
	static public final double[] USER_MINMAX_TAU = { -100., 100. };
	static public final double[] USER_MINMAX_CUSTOM = { 600., 1200. };

	static public final int REAL_TIME_PERIOD_HOURS = 48;

	// data management
	static public final boolean OVERWRITE_EXISTING_GPS_TEC = false;

	static public String getFileNamePrefix(MapChoice choice) {
		String prefix = new String();
		switch (choice) {
		case FOF2_CLIMATE:
			prefix = "IRI_foF2";
			break;
		case FOF2_WEATHER:
			prefix = RTAM_FILENAME_PREF_FOF2;
			break;
		case FOF2_DELTA:
			prefix = RTAMVSIRI_FILENAME_PREF_FOF2;
			break;
		case FOF2_DELTA_PERCENT:
			prefix = RTAMVSIRIPERCENT_FILENAME_PREF_FOF2;
			break;
		case NMF2_CLIMATE:
			prefix = "IRI_NmF2";
			break;
		case NMF2_WEATHER:
			prefix = RTAM_FILENAME_PREF_NMF2;
			break;
		case NMF2_DELTA:
			prefix = RTAMVSIRI_FILENAME_PREF_NMF2;
			break;
		case NMF2_DELTA_PERCENT:
			prefix = RTAMVSIRIPERCENT_FILENAME_PREF_NMF2;
			break;
		case HMF2_CLIMATE:
			prefix = "Brunini_hmF2";
			break;
		case HMF2_WEATHER:
			prefix = RTAM_FILENAME_PREF_HMF2;
			break;
		case HMF2_DELTA:
			prefix = RTAMVSIRI_FILENAME_PREF_HMF2;
			break;
		case HMF2_DELTA_PERCENT:
			prefix = RTAMVSIRIPERCENT_FILENAME_PREF_HMF2;
			break;
		case GK:
			prefix = "GK";
			break;
		case B0_CLIMATE:
			prefix = "IRI_B0";
			break;
		case B0_WEATHER:
			prefix = "IRTAM_B0";
			break;
		case B0_DELTA:
			prefix = "IRTAMvsIRI_B0";
			break;
		case B0_DELTA_PERCENT:
			prefix = "IRTAMvsIRI_B0_per";
			break;
		case B1_CLIMATE:
			prefix = "IRI_B1";
			break;
		case B1_WEATHER:
			prefix = "IRTAM_B1";
			break;
		case B1_DELTA:
			prefix = "IRTAMvsIRI_B1";
			break;
		case B1_DELTA_PERCENT:
			prefix = "IRTAMvsIRI_B1_per";
			break;
		default:
			prefix = "AddName";
			break;
		}
		return prefix;
	}

	// INPUT PARAMETERS PARSING
	static public CharChoice getCharChoice(CommandLineParam commandLineParam) {
		if (commandLineParam.isKeyExists("charident")) {
			if (commandLineParam.getKeyValue("charident").contains(CharChoice.FOF2.getName()))
				return CharChoice.FOF2;
			if (commandLineParam.getKeyValue("charident").contains(CharChoice.HMF2.getName()))
				return CharChoice.HMF2;
			System.err.println("Error retrieving DIDBase ident for characteristic " + commandLineParam.getKeyValue("charident"));
		}
		System.err.println("Error retrieving DIDBase ident: characteristic name is missing in exev parameters");
		return null;
	}

	static public TimeScale[] getTimePeriod(CommandLineParam commandLineParam) throws IOException {
		TimeScale[] period = new TimeScale[2];
		if (commandLineParam.isKeyExists("startdate") && commandLineParam.isKeyExists("enddate")) {
			String startDateString = commandLineParam.getKeyValue("startdate");
			String endDateString = commandLineParam.getKeyValue("enddate");
			startDateString = startDateString.replace(':', '.');
			endDateString = endDateString.replace(':', '.');
			period[0] = new TimeScale();
			period[1] = new TimeScale();
			try {
				period[0].set(startDateString, DATE_FORMAT);
				period[1].set(endDateString, DATE_FORMAT);
			} catch (ParseException ex) {
				throw new RuntimeException("Unable to parse \"-startdate\" or -\"enddate\"; format is: " + DATE_FORMAT);
			}
		} else {
			throw new IOException("No parameters, which are required for program run: \"-startdate\" and/or \"-enddate\".");
		}
		return period;
	}

	static public TimeScale getStartTime(CommandLineParam commandLineParam) throws IOException {
		TimeScale startDate = new TimeScale();
		if (commandLineParam.isKeyExists("startdate")) {
			String startDateString = commandLineParam.getKeyValue("startdate");
			startDateString = startDateString.replace(':', '.');
			try {
				startDate.set(startDateString, DATE_FORMAT);
			} catch (ParseException ex) {
				throw new RuntimeException("Unable to parse \"-startdate\"; format is: " + DATE_FORMAT);
			}
		} else {
			return null;
		}
		return startDate;
	}

	static public boolean checkCaptureEnabled(CommandLineParam commandLineParam) {
		if (commandLineParam.isKeyExists("saveframes")) {
			String keyValue = commandLineParam.getKeyValue("saveframes");
			return Boolean.valueOf(keyValue);
		} else
			return true;
	}

	static public boolean getSubmitDB(CommandLineParam commandLineParam) throws IOException {
		return commandLineParam.isKeyExists("submit_db");
	}
	
	static public boolean getUpdateDB(CommandLineParam commandLineParam) throws IOException {
		if (commandLineParam.isKeyExists("update_db")) {
			String keyValue = commandLineParam.getKeyValue("update_db").trim().toLowerCase();
			if (keyValue.equals("yes") || keyValue.equals("1"))
				return true;
			else if (keyValue.equals("no") || keyValue.equals("0"))
				return false;
			else
				throw new IOException("Unable to parse \"-update_db\". Possible values yes/no or 1/0");
		} else
			throw new IOException("No \"-update_db\" parameter, which is required for program run. Possible values are: " + "yes/no or 1/0");
	}
	static public boolean getExportCoeffs(CommandLineParam commandLineParam) throws IOException {
		return commandLineParam.isKeyExists("exportcoeffs");
	}
}
