/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.core;

import java.io.IOException;
import java.io.InputStream;
import java.sql.SQLException;
import java.util.TreeMap;

import edu.uml.giro.gambit.coefficients.Coefficients;
import edu.uml.giro.gambit.coefficients.Coefficients2D;
import edu.uml.giro.gambit.coefficients.Coefficients_B0;
import edu.uml.giro.gambit.coefficients.Coefficients_B1;
import edu.uml.giro.gambit.coefficients.Coefficients_foF2;
import edu.uml.giro.gambit.coefficients.Coefficients_hmF2;
import edu.uml.giro.gambit.database.GambitDatabase;
import edu.uml.giro.gambit.database.GambitDatabaseReader;
import edu.uml.giro.gambit.database.StationLocation;
import edu.uml.giro.gambit.expansions.ExpansionBasis2D;
import edu.uml.giro.gambit.expansions.ExpansionBasis_JonesGallet;
import edu.uml.giro.gambit.expansions.ExpansionBasis_JonesGalletWithTrend;
import edu.uml.giro.gambit.synthesizers.GeoSynthesizer;
import edu.uml.lgdc.geospace.CharChoice;
import edu.uml.lgdc.geospace.Geomap;
import edu.uml.lgdc.geospace.MapChoice;
import edu.uml.lgdc.project.Console;

public class GambitCore extends Core {

	private GambitDatabaseReader gambitBase;
	public TreeMap<Integer, StationLocation> gambitStationLocations;
	
	public static final int TRANSFORM_MAPS_DEMAG = 1;
	public static final int TRANSFORM_MAPS_REMAG = 2;
	
	public GambitCore( int dbErrorHandlingMode ) {
		super();
		gambitBase = new GambitDatabaseReader(dbErrorHandlingMode);
		gambitStationLocations = gambitBase.getAllStationLocs();
		reconstruct( gambitBase.GAMBIT_DB_VERSIONS.get(0) ); // use default versions
	}
	
	static public GambitCore getInstance(int dbErrorHandlingMode) {
		GambitCore gambitCore = new GambitCore( dbErrorHandlingMode );
		if (gambitCore.gambitBase.isGood())
			return gambitCore;
		else
			return null;
	}
	
	public void reconstruct( AlgorithmVersions versions ) {

		// Reconstruct weather synthesizer using specific algorithm versions

		for (CharChoice cc : CharChoice.values()) {
			switch (cc) {
			case FOF2:
				// Weather synthesizer of foF2
				Coefficients2D coeffs_foF2w = new Coefficients_foF2();
				// Basis depends on the choice of algorithm
				ExpansionBasis2D basis_foF2 = null;
				if( versions.expansionBasisName.equals("JonesGallet") )
					basis_foF2 = new ExpansionBasis_JonesGallet();
				if( versions.expansionBasisName.equals("JonesGallet_LinTrend") )
					basis_foF2 = new ExpansionBasis_JonesGalletWithTrend();
				basis_foF2.setup(ExpansionBasis_JonesGallet.CCIR_DIURNAL_ORDER, ExpansionBasis_JonesGallet.SPATIAL_ORDER,
						ExpansionBasis_JonesGallet.SPATIAL_DEGREE);
				basis_foF2.setExpansionBasisName( versions.expansionBasisName);
				coeffs_foF2w.setExpansionBasis(basis_foF2);
				weather_foF2 = new GeoSynthesizer(CharChoice.FOF2, basis_foF2, coeffs_foF2w, null);
				weather_foF2.setTimeGrid(defaultTimeGrid);
				weather_foF2.setEarthGrid(defaultEarthGrid);
				weather_foF2.setAssimilationMethodName( versions.assimilationName );
				weather_foF2.setIrtamVersionIdent( versions.versionID );
				chars[CharChoice.FOF2.getIndex()].setNewWeatherSynthesizer(weather_foF2);
				break;
			case NMF2:
				// hmF2 weather is based on foF2 synthesizer
				GeoSynthesizer weather_NmF2 = new GeoSynthesizer(CharChoice.NMF2, weather_foF2.getExpansionBasis(), null, weather_foF2);
				chars[CharChoice.NMF2.getIndex()].setNewWeatherSynthesizer(weather_NmF2);
				break;
			case HMF2:
				// Weather synthesizer of hmF2
				Coefficients_hmF2 coeffs_hmF2w = new Coefficients_hmF2();
				// Basis depends on the choice of algorithm
				ExpansionBasis2D basis_hmF2 = null;
				if( versions.expansionBasisName.equals("JonesGallet") )
					basis_hmF2 = new ExpansionBasis_JonesGallet();
				if( versions.expansionBasisName.equals("JonesGallet_LinTrend") )
					basis_hmF2 = new ExpansionBasis_JonesGalletWithTrend();
				basis_hmF2.setup(ExpansionBasis_JonesGallet.CCIR_DIURNAL_ORDER, ExpansionBasis_JonesGallet.SPATIAL_ORDER,
						ExpansionBasis_JonesGallet.SPATIAL_DEGREE);
				basis_hmF2.setExpansionBasisName( versions.expansionBasisName);
				coeffs_hmF2w.setExpansionBasis(basis_hmF2);
				weather_hmF2 = new GeoSynthesizer(CharChoice.HMF2, basis_hmF2, coeffs_hmF2w, null);
				weather_hmF2.setTimeGrid(defaultTimeGrid);
				weather_hmF2.setEarthGrid(defaultEarthGrid);
				weather_hmF2.setAssimilationMethodName( versions.assimilationName );
				weather_hmF2.setIrtamVersionIdent( versions.versionID );
				chars[CharChoice.HMF2.getIndex()].setNewWeatherSynthesizer(weather_hmF2);
				break;
			case B0:
				// Weather synthesizer of B0
				Coefficients2D coeffs_B0w = new Coefficients_B0();
				// Basis depends on the choice of algorithm
				ExpansionBasis2D basis_B0 = null;
				if( versions.expansionBasisName.equals("JonesGallet") )
					basis_B0 = new ExpansionBasis_JonesGallet();
				if( versions.expansionBasisName.equals("JonesGallet_LinTrend") )
					basis_B0 = new ExpansionBasis_JonesGalletWithTrend();
				basis_B0.setup(ExpansionBasis_JonesGallet.CCIR_DIURNAL_ORDER, ExpansionBasis_JonesGallet.SPATIAL_ORDER,
						ExpansionBasis_JonesGallet.SPATIAL_DEGREE);
				basis_B0.setExpansionBasisName( versions.expansionBasisName);
				coeffs_B0w.setExpansionBasis(basis_B0);
				GeoSynthesizer weather_B0 = new GeoSynthesizer(CharChoice.B0, basis_B0, coeffs_B0w, null); // same basis as climate (for now)
				weather_B0.setTimeGrid(defaultTimeGrid);
				weather_B0.setEarthGrid(defaultEarthGrid);
				weather_B0.setAssimilationMethodName( versions.assimilationName );
				weather_B0.setIrtamVersionIdent( versions.versionID );
				chars[CharChoice.B0.getIndex()].setNewWeatherSynthesizer(weather_B0);
				break;
			case B1:
				// Weather synthesizer of B1
				Coefficients2D coeffs_B1w = new Coefficients_B1();
				// Basis depends on the choice of algorithm
				ExpansionBasis2D basis_B1 = null;
				if( versions.expansionBasisName.equals("JonesGallet") )
					basis_B1 = new ExpansionBasis_JonesGallet();
				if( versions.expansionBasisName.equals("JonesGallet_LinTrend") )
					basis_B1 = new ExpansionBasis_JonesGalletWithTrend();
				basis_B1.setup(ExpansionBasis_JonesGallet.CCIR_DIURNAL_ORDER, ExpansionBasis_JonesGallet.SPATIAL_ORDER,
						ExpansionBasis_JonesGallet.SPATIAL_DEGREE);
				basis_B1.setExpansionBasisName( versions.expansionBasisName);
				coeffs_B1w.setExpansionBasis(basis_B1);
				GeoSynthesizer weather_B1 = new GeoSynthesizer(CharChoice.B1, basis_B1, coeffs_B1w, null); // same basis as climate (for now)
				weather_B1.setTimeGrid(defaultTimeGrid);
				weather_B1.setEarthGrid(defaultEarthGrid);
				weather_B1.setAssimilationMethodName( versions.assimilationName );
				weather_B1.setIrtamVersionIdent( versions.versionID );
				chars[CharChoice.B1.getIndex()].setNewWeatherSynthesizer(weather_B1);
				break;
			default:
				throw new RuntimeException("design error: overlooked characteristic, " + cc.getName());
			}
			chars[cc.getIndex()].reset();
		}
		this.VERSIONS = versions;
	}
	

	/**
	 * Download and rebuild Observations and Weather Coefficients from GAMBIT
	 * database calls, the components know to avoid re-reading
	 * 
	 * @param c
	 * @param databaseMode
	 */
	public void rebuildOneChar_ObservationsAndWeather(Characteristic c, int databaseMode) {

		// Download and rebuild Observations and Weather Coefficients from
		// GAMBIT

		LocalData localData = c.localData;
		localData.getTimeGrid().setUTs(timeOfValidity, GambitConstants.FORECAST_DURATION_MIN);
		CharChoice cc = c.cc;

		if (cc.isDerived()) {
			// ==============================================================================
			// this is a derived characteristic from some other characteristics
			//
			localData.deriveClimateValues(c.climate);

			if ((databaseMode & GambitDatabase.READ_OBSERVATIONS) != 0) {
				// derive "observed" values of this char from other observations
				localData.deriveFrom(chars[cc.getDerivedFrom()].localData);
			}
			if ((databaseMode & GambitDatabase.READ_COEFFS) != 0) {
				int iTime = chars[cc.getDerivedFrom()].weather.getTimeGrid().getNumDiurnalSteps()-1;
				c.weatherMap.setMapValues(c.weather.synthesizeValueMap(iTime), timeOfValidity);
				// Fill internal global weather arrays in local data
				for( int i = 0; i < localData.totalLocations; i++ ) {
					localData.sites[i].deriveGlobalWeatherValues(c.weather);
					localData.sites[i].calculateO2GWdifferences();
				}
				
			}
		} else {
			// ==============================================================================
			// this is an assimilated characteristic available in GAMBIT
			// database
			//
			Coefficients<?> coeffs = c.weather.getCoefficients();

			// ==============================================================================
			// data have to be downloaded from GAMBIT database first
			// get blobs from GAMBIT database and mate them to input streams
			try {
				InputStream is[] = gambitBase.download(c.weather, localData, timeOfValidity, databaseMode);

				// observations needed?
				if ((databaseMode & GambitDatabaseReader.READ_OBSERVATIONS) != 0) {
					if (is[1] != null) {
						// load observations from input stream is[1]
						is[1].mark(is[1].available());
						localData.loadFromInputStream(timeOfValidity, is[1], gambitStationLocations, VERSIONS );
						// now the sites are initialized, we can fill local climate data
						localData.fillClimateValues(c.climate);
						// now remove poor quality values and fill gaps
						localData.processRawData(timeOfValidity);
						is[1].reset();
					} else
						System.err.println("GambitCore-ERROR: no observations found... database stream is empty");
				}
				if ((databaseMode & GambitDatabase.READ_COEFFS) != 0) {
					if (is[0] != null) {
						// load coefficients from input stream is[0]
						is[0].mark(is[0].available());
						coeffs.loadFromInputStream(is[0]);
						is[0].reset();
						// Recompute expansion basis
						c.weather.getExpansionBasis().setNewTime(timeOfValidity);
						int iTime = c.weather.getTimeGrid().findClosestTimeStep(timeOfValidity);
						c.weatherMap.setMapValues(c.weather.synthesizeValueMap(iTime), timeOfValidity);
						// Fill internal global weather arrays in local data
						for( int i = 0; i < localData.totalLocations; i++ ) {
							localData.sites[i].fillGlobalWeatherValues(c.weather);
							localData.sites[i].calculateO2GWdifferences();
						}
					} else
						System.err.println("GambitCore-ERROR: no weather coefficients found... database stream is empty");
				}
			} catch (SQLException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	public GambitDatabaseReader getGambitDatabaseReader() {
		return gambitBase;
	}
	
	public void transformOneMap( Geomap map, int option ) {
		
	}
	
	public void transformAllMaps( int option ) {
		for (Characteristic c : chars) {
			transformMap( c.climateMap, option );
			transformMap( c.weatherMap, option );
			transformMap( c.deviationMap, option );
			transformMap( c.deviationMap_percent, option );
		}
	}
	
	public void transformMap( Geomap map, int option ) {
		if( map == null )
			return;
		ExpansionBasis_JonesGallet basis = (ExpansionBasis_JonesGallet) chars[0].climate.getExpansionBasis();
		double[] buffer = new double [basis.getEarthGrid().getNumLatitudeNodes()];
		int numLats = basis.getEarthGrid().getNumLatitudeNodes();
		int iLat, iLon = 0;
		int newLatIndices[] = new int[numLats];
		if( option == TRANSFORM_MAPS_DEMAG ) {
			for( iLon = 0; iLon < basis.getEarthGrid().getNumLongitudeNodes(); iLon++ ) {
				// Compute new indices into demag array
				for( iLat = 0; iLat < numLats; iLat++ ) {
					double modip = basis.getModip( iLat, iLon);
					newLatIndices[iLat] = (int)((modip + Math.PI/2)*(numLats-1)/Math.PI);
//					System.out.println("i=" + iLat + " newI=" + newLatIndices[iLat] + " modip="+modip);
				}
				moveDemagValues( map, iLon, newLatIndices, buffer);
			}
			Geomap.compute1Dfrom2D(map.getMap_1D(), map.getMap());
		}
		if( option == TRANSFORM_MAPS_REMAG ) {
			MapChoice mapChoice = map.getMapChoice();
			CharChoice cc = mapChoice.getCC();
			Characteristic c = getChar(cc);
			rebuildOneChar(c, GambitDatabase.READ_COEFFS );
		}
	}
	
	private void moveDemagValues(Geomap map, int iLon, int[] newLatIndices, double[] buffer) {
		int iLat;
		if( map == null )
			return;
		for( iLat = 0; iLat < buffer.length; iLat++ ) 
			buffer[iLat] = map.getNoValue();
		boolean emptyMap = true;
		for( iLat = 0; iLat < buffer.length; iLat++ ) {
			if( newLatIndices[iLat] < 0 || newLatIndices[iLat] >= buffer.length)
				continue;
			double value = map.getMap()[iLat][iLon];
			buffer[newLatIndices[iLat]] = value;
			if( value != map.getNoValue())
				emptyMap = false;
		}
		if( emptyMap )
			return;
		// Fill gaps after demag operation. This algorithm assumes that demag translates poles to poles (as in modip)
		int iNext = 0;
		for( iLat = 0; iLat < buffer.length; iLat++ ) {
			if( buffer[iLat] == map.getNoValue() && iLat > 0) {
				iNext = iLat+1;
				while( iNext < buffer.length && buffer[iNext] == map.getNoValue())
					iNext++;
				// interpolate the gap
				double slope = (buffer[iNext] - buffer[iLat-1])/(iNext-iLat+1);
				for( int i = iLat; i < iNext; i++ )
					buffer[i] = buffer[iLat-1]+slope*(i-iLat+1);
			}
		}
		// update the map
		for( iLat = 0; iLat < buffer.length; iLat++ ) {
			map.getMap()[iLat][iLon] = buffer[iLat];
			map.getMap_1D()[iLon*buffer.length+iLat] = buffer[iLat];
		}
	}

		
}