/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.core;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Calendar;
import java.util.Map;
import java.util.TreeMap;

import edu.uml.giro.gambit.core.LocalDataElements.LocalDataType;
import edu.uml.giro.gambit.database.StationLocation;
import edu.uml.giro.gambit.synthesizers.GeoSynthesizer;
import edu.uml.giro.gambit.synthesizers.Synthesizer;
import edu.uml.lgdc.geospace.CharChoice;
import edu.uml.lgdc.geospace.CharChoice;
import edu.uml.lgdc.geospace.PlasmaMath;
import edu.uml.lgdc.geospace.TimeGrid;
import edu.uml.lgdc.geospace.TimeGrid;
import edu.uml.lgdc.quantities.NO_VALUE;
import edu.uml.lgdc.time.TimeScale;

public class LocalData {

	public static final int MAX_SITES = 150;

	// the number of locations with data depends on DIDBase availability
	// the memory allocation is done for MAX locations, and number of available
	// locations is smaller
	public LocalDataOneSiteAllTimes[] sites = new LocalDataOneSiteAllTimes[MAX_SITES];
	public int totalLocations; // depends on DIDBAse data availability for each time
	
	public boolean windowBoundaryMitigationEnabled = false;

	protected CharChoice cc;

	public CharChoice getCharChoice() {
		return cc;
	}

	protected TimeGrid timeGrid;

	public TimeGrid getTimeGrid() {
		return timeGrid;
	}

	public LocalData(CharChoice cc, TimeGrid timeGrid) {
		this.timeGrid = timeGrid;
		this.cc = cc;
		for (int i = 0; i < MAX_SITES; i++)
			sites[i] = new LocalDataOneSiteAllTimes(cc, timeGrid);
	}

	public void reset() {
		totalLocations = 0;
	}

	public boolean isEmpty() {
		return totalLocations == 0;
	}

	public LocalDataOneSiteAllTimes getSite(String ursi) {
		for (int iSite = 0; iSite < totalLocations; iSite++)
			if (sites[iSite].getLocation().ursi.toUpperCase().equals(ursi.toUpperCase()))
				return sites[iSite];
		return null;
	}

	public LocalDataOneSiteAllTimes[] getSites() {
		return sites;
	}

	public void fillClimateValues( Synthesizer climate ) {
		for (int iSite = 0; iSite < totalLocations; iSite++) {
			sites[iSite].fillClimateValues(climate);
		}
	}
	
	public void deriveClimateValues( GeoSynthesizer climate ) {
		for (int iSite = 0; iSite < totalLocations; iSite++) {
			sites[iSite].deriveClimateValues(climate);
		}
	}
	
	public void fillGlobalWeatherValues( Synthesizer weather ) {
		for (int iSite = 0; iSite < totalLocations; iSite++) {
			sites[iSite].fillGlobalWeatherValues(weather);
		}
	}
	
	public void deriveGlobalWeatherValues( GeoSynthesizer weather) {
		for (int iSite = 0; iSite < totalLocations; iSite++) {
			sites[iSite].deriveGlobalWeatherValues(weather);
		}
	}
	
	
	
	/*
	 * processRawData: prepare observations for assimilation -- fill gaps --
	 * smooth the day boundary to preserve latest data
	 */

	public void processRawData(TimeScale tov ) {
		for (int iSite = 0; iSite < totalLocations; iSite++) {
			if (sites[iSite].isUnused() == true)
				continue;
			// fill climate values in order to fill gaps
			if( windowBoundaryMitigationEnabled ) 
				sites[iSite].removeOldestData( 120 ); // 120 minutes
			// fill climate values in order to fill gaps
			sites[iSite].fillObservationGaps();
			// System.out.println(sites[totalLocations].location.ursi + "   " +
			// sites[totalLocations].adjustTail(toTime));
			sites[iSite].smoothDayBoundary(tov, timeGrid.timeToDiurnalPhaseIndex(tov));
			sites[iSite].calculateO2Cdifferences();
		}
	}
	
	public void deriveFrom(LocalData input) {
		int iSite;
		totalLocations = input.totalLocations;

		// first, copy all data from input
		for (iSite = 0; iSite < totalLocations; iSite++) {
			sites[iSite] = new LocalDataOneSiteAllTimes(cc, input.sites[iSite].timeGrid);
			sites[iSite].resetLocalData();
			sites[iSite].putLocation( input.sites[iSite].getLocation() );
			sites[iSite].putTov(input.sites[iSite].getTov());
		}

		// now patch the original data with derived one

		switch (cc) // this is my cc
		{
		case NMF2:
			for (iSite = 0; iSite < totalLocations; iSite++) {
				// patch input foF2 observations with NmF2 values
				double observations[] = input.sites[iSite].getObservations();
				for( int i = 0; i< observations.length; i++ ) {
					if( observations[i] != NO_VALUE.SPECIAL )
						observations[i] = PlasmaMath.freq2density(observations[i]);
				}
				this.sites[iSite].putObservations( observations );
				// raw observations -- same thing
				double rawObservations[] = input.sites[iSite].getRawObservations();
				for( int i = 0; i< rawObservations.length; i++ ) {
					if( rawObservations[i] != NO_VALUE.SPECIAL )
						rawObservations[i] = PlasmaMath.freq2density(rawObservations[i]);
				}
				this.sites[iSite].putRawObservations( rawObservations );
				// confidence scores
				int confidences[] = input.sites[iSite].getConfidenceScores();
				this.sites[iSite].putConfidences( confidences );
				// flags
				int flags[] = input.sites[iSite].getFlags();
				this.sites[iSite].putFlags( flags );
			}
			break;
		default:
			System.err.println("ERROR-Observations: no rules to derive values of " + cc.getName() + " from " + input.cc.getName());
		}
	}

	public void loadFromInputStream(TimeScale tov, InputStream obsIS, TreeMap<Integer, StationLocation> locs, AlgorithmVersions versions) {
		DataInputStream in = new DataInputStream(obsIS);
		// try { in.reset(); }
		// catch (IOException e1) { e1.printStackTrace(); }

		// IRTAM version 1 used TY buffers, next versions use ring buffers (YTF)
		boolean ringBufferArchitecture = versions.versionID > 1;
		
		try {
			totalLocations = in.readInt();
			int nTimeSteps = in.readInt();
			
			for (int iSite = 0; iSite < totalLocations; iSite++) {
				int id = (int) in.readShort();
				// System.out.print(FC.DoubleToString(id, 7, 2) + " ");
				// fill location attr
				StationLocation loc = locs.get(id);
				sites[iSite] = new LocalDataOneSiteAllTimes(cc, timeGrid);
				// no need to use smart replenishment of the ring buffers, this has all data
				sites[iSite].resetLocalData();
				sites[iSite].putLocation(loc);
				// fill the site's circular buffer
				sites[iSite].putTov( tov );
				TimeScale obsTime = sites[iSite].getMidnightToday();
				 
				for (int iTime = 0; iTime < nTimeSteps; iTime++) {
					double obs = (double) in.readFloat(); // order matters first obs then conf
					int conf = in.readShort();
					if( ringBufferArchitecture ) {
						if( iTime < sites[iSite].data.getLength()) {
							sites[iSite].data.put(iTime, LocalDataType.OBSERVED_VALUE_RAW, obs);
							sites[iSite].data.put(iTime, LocalDataType.CONFIDENCE_SCORE, conf);
						}
					}
					else {	
						// this is TY scheme (today-yesterday)
						sites[iSite].addDataValue(obsTime, obs, conf);
						obsTime.add(Calendar.MINUTE, timeGrid.getTimeStep_min() );
						if( obsTime.after(tov) )
							obsTime.add(Calendar.MINUTE, -TimeScale.MINUTES_PER_DAY);
					}
					// System.out.print(FC.DoubleToString(obs, 7, 2) + " "+
					// FC.DoubleToString(conf, 7, 2) + " ");
				}
				// System.out.println();
			}
			in.close();

		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void setSites(LocalDataOneSiteAllTimes[] sites) {
		totalLocations = 0;
		this.sites = sites;
		for (int iSite = 0; iSite < sites.length; iSite++)
			if (sites[iSite] != null) {
				// this.sites[totalLocations] =
				// sites[iSite].cloneOnlyObservations();
				// System.out.println(sites[iSite].location.ursi +
				// Misc.d2S(sites[iSite].location.lat, 7, 2) +
				// Misc.d2S(sites[iSite].location.lon, 7, 2));
				totalLocations++;
			}
	}

	public void updateSitesObservations(LocalDataOneSiteAllTimes[] sites) {

	}
	
	public ByteArrayInputStream toInputStream( StationLocation[] usedLocations, Map<String, Integer> gambitStationIDs ) {
		
		// make float and short arrays [site][time] with data and point input stream to it
		int totalTimeSteps = getSite(usedLocations[0].ursi).data.getLength();
		float[][] observations = new float[usedLocations.length][totalTimeSteps];
		short[][] confidences = new short[usedLocations.length][totalTimeSteps];

		for( int iSite = 0; iSite < usedLocations.length; iSite++ ) {
			LocalDataOneSiteAllTimes site = getSite( usedLocations[iSite].ursi);
			double[] siteObs = site.getRawObservations();
			int[] siteConf = site.getConfidenceScores();
			for( int iTime = 0; iTime < totalTimeSteps; iTime++ ) {
				observations[iSite][iTime] = (float)siteObs[iTime];
				confidences[iSite][iTime] = (short)siteConf[iTime];
			}
		}		
	    DataOutputStream out;
	    ByteArrayInputStream obsIS = null;
	    ByteArrayOutputStream byteAOS = new ByteArrayOutputStream();
	    try 
	    {
	      out = new DataOutputStream(byteAOS);
	      out.writeInt(usedLocations.length);
	      out.writeInt(totalTimeSteps);
	      for(int i = 0; i < usedLocations.length; i++)
	      {
	        //System.out.print(FC.DoubleToString(siteID[i], 7, 2) +  " ");
	        out.writeShort((short)gambitStationIDs.get(usedLocations[i].ursi).intValue());
	        for(int j = 0; j < totalTimeSteps; j++)
	        {
	          out.writeFloat(observations[i][j]);
	          out.writeShort(confidences[i][j]);
	          //System.out.print(FC.DoubleToString(obs[i][j], 7, 2) + " "+ FC.DoubleToString(conf[i][j], 7, 2) + " ");
	        }
	        //System.out.println();
	      }
	      out.flush();
	      obsIS = new ByteArrayInputStream( byteAOS.toByteArray());
	      out.close();
	    } catch (IOException e) { e.printStackTrace(); }
	    return obsIS;
	  }

	public String getAllUrsiCodes() {
		String codes = "";
		for( LocalDataOneSiteAllTimes site : sites ) {
			if( site.getLocation() == null )
				continue;
			codes += site.getLocation().ursi+" ";
		}
		return codes;
	}

}
