/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.core;

import java.util.HashMap;

import edu.uml.lgdc.quantities.NO_VALUE;

public class LocalDataElements extends HashMap<Object, Object> {

	public enum LocalDataType {
		OBSERVED_VALUE_RAW, 
		OBSERVED_VALUE_PROCESSED, // gaps filled, day boundary fixed
		TIMEGRID_MISMATCH_PERCENT, // how far from the time grid point
		CONFIDENCE_SCORE, 
		FLAG, // see IonogramScaling for explanations
		CLIMATE_VALUE, 
		LOCAL_WEATHER_VALUE, 
		GLOBAL_WEATHER_VALUE, 
		O2C_DIFFERENCE,  // observations minus climate 
		O2LW_DIFFERENCE, // observations minus local weather
		O2GW_DIFFERENCE, // observations minus global weather
		LW2C_DIFFERENCE, // local weather from climate
		GW2C_DIFFERENCE, // global weather from climate
		LW2GW_DIFFERENCE, // local weather from global weather
		LOCAL_TREND, // local weather trend over 24 hours
		GLOBAL_TREND // global weather trend over 24 hours
	}

	public double getObservation() {
		Double value = (Double)get(LocalDataType.OBSERVED_VALUE_PROCESSED);
		return value == null? NO_VALUE.SPECIAL : value.doubleValue(); 
	}

	public double getRawObservation() {
		Double value = (Double)get(LocalDataType.OBSERVED_VALUE_RAW);
		return value == null? NO_VALUE.SPECIAL : value.doubleValue(); 
	}

	public int getConfidenceScore() {
		Integer value = (Integer)get(LocalDataType.CONFIDENCE_SCORE);
		return value == null? IonogramScaling.UNKNOWN_CONFIDENCE_SCORE: value; 
	}

	public int getTimegridMismatch() {
		Integer value = (Integer)get(LocalDataType.TIMEGRID_MISMATCH_PERCENT);
		return value == null? 0: value; 
	}

	public int getFlag() {
		Integer value = (Integer)get(LocalDataType.FLAG);
		return value == null? 0: value; 
	}
	
	public double getClimate() {
		Double value = (Double)get(LocalDataType.CLIMATE_VALUE);
		return value == null? NO_VALUE.SPECIAL : value.doubleValue(); 
	}

	public double getLocalWeather() {
		Double value = (Double)get(LocalDataType.LOCAL_WEATHER_VALUE);
		return value == null? NO_VALUE.SPECIAL : value.doubleValue(); 
	}

	public double getGlobalWeather() {
		Double value = (Double)get(LocalDataType.GLOBAL_WEATHER_VALUE);
		return value == null? NO_VALUE.SPECIAL : value.doubleValue(); 
	}

	public double getO2C() {
		Double value = (Double)get(LocalDataType.O2C_DIFFERENCE);
		return value == null? NO_VALUE.SPECIAL : value.doubleValue(); 
	}
	
	public double getO2C_percent() {
		double value = NO_VALUE.SPECIAL;
		double obs = getObservation();
		double climate = getClimate();
		if( obs != NO_VALUE.SPECIAL && climate != NO_VALUE.SPECIAL )
			value = 100.* (obs-climate)/climate;
		return value;
	}
			
			
	public double getO2LW() {
		Double value = (Double)get(LocalDataType.O2LW_DIFFERENCE);
		return value == null? NO_VALUE.SPECIAL : value.doubleValue(); 
	}
	
	public double getO2LW_percent() {
		double value = NO_VALUE.SPECIAL;
		double obs = getObservation();
		double localWeather = getLocalWeather();
		if( obs != NO_VALUE.SPECIAL && localWeather != NO_VALUE.SPECIAL )
			value = 100.* (obs-localWeather)/localWeather;
		return value;
	}
	
	public double getO2GW() {
		Double value = (Double)get(LocalDataType.O2GW_DIFFERENCE);
		return value == null? NO_VALUE.SPECIAL : value.doubleValue(); 
	}
	
	public double getO2GW_percent() {
		double value = NO_VALUE.SPECIAL;
		double obs = getObservation();
		double globalWeather = getGlobalWeather();
		if( obs != NO_VALUE.SPECIAL && globalWeather != NO_VALUE.SPECIAL )
			value = 100.* (obs-globalWeather)/globalWeather;
		return value;
	}
	
	public boolean isBetter(double competitionValue, int competitionConfidence, int competitionMismatch) {
		if (IonogramScaling.isMissing(getRawObservation()))
			return false;
		if (competitionValue == NO_VALUE.SPECIAL)
			return true; // anything is better than a blank
		if (IonogramScaling.isHighQuality(getConfidenceScore()) && IonogramScaling.isHighQuality(competitionConfidence))
			// two great values, one should be closer to the grid point
			return this.getTimegridMismatch() < competitionMismatch;
		if (IonogramScaling.isHighQuality(getConfidenceScore()) && !IonogramScaling.isHighQuality(competitionConfidence))
			// more confident data
			return true;
		if (IonogramScaling.isPoorQuality(getConfidenceScore()) && !IonogramScaling.isPoorQuality(competitionConfidence))
			// cannot believe this value...
			return false;
		// no clear decision can be made.
		// ignore the value that is too far
		if (IonogramScaling.isTooFar( getTimegridMismatch()))
			return false;
		// OK. Take the one closer to grid
		return getTimegridMismatch() < competitionMismatch;
	}

}