/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.core;

import java.util.concurrent.TimeUnit;

import edu.uml.giro.gambit.core.LocalDataElements.LocalDataType;
import edu.uml.giro.gambit.database.StationLocation;
import edu.uml.giro.gambit.expansions.ExpansionBasis_JonesGalletWithTrend;
import edu.uml.giro.gambit.synthesizers.GeoSynthesizer;
import edu.uml.giro.gambit.synthesizers.Synthesizer;
import edu.uml.lgdc.datatype.RingBuffer;
import edu.uml.lgdc.geospace.CharChoice;
import edu.uml.lgdc.geospace.TimeGrid;
import edu.uml.lgdc.quantities.NO_VALUE;
import edu.uml.lgdc.time.TimeScale;

/**
 * 
 * OneDayHistory: contains all initial, intermediate, and final results of RTAM
 * for one GIRO station including
 * 
 * observed values IRI values RTAM values Differences between observed and IRI
 * values
 * 
 * @author Galkin, Vesnin
 *
 */

public class LocalDataOneSiteAllTimes {
	protected CharChoice cc;
	protected TimeGrid timeGrid;

	public RingBuffer<LocalDataElements> data;

	private StationLocation location;
	private boolean controlPoint = false; // if TRUE this station is not
											// included in assimilation
											// procedure
	private boolean unused = false;

	public static final int SUFFICIENT_DATA_POINTS = 4;

	// observation preprocessing

	public static final int TIME_STEPS_TO_AVERAGE = getTimeStepAverage(SUFFICIENT_DATA_POINTS);
	public static final int TIME_STEPS_TO_SKIP = 12;
	public static final int TIME_STEPS_SEARCH = 5;
	public static final double THRESHOLD_JUMP_OF_OBSERVATIONS_FOF2 = 2; // in
																		// orders
																		// of
																		// magnitude
																		// of
																		// derivative
	public static final double THRESHOLD_JUMP_OF_OBSERVATIONS_HMF2 = 20;

	public LocalDataOneSiteAllTimes(CharChoice cc, TimeGrid timeGrid) {
		this.cc = cc;
		this.timeGrid = timeGrid;
		data = new RingBuffer(LocalDataElements.class, TimeUnit.MINUTES, (int)TimeUnit.DAYS.toMinutes(1), 
			timeGrid.getTimeStep_min(), GambitConstants.FORECAST_DURATION_MIN);
	}

	static private int getTimeStepAverage(int sufficientPoints) {
		return (sufficientPoints / 2) > 3 ? 3 : (sufficientPoints / 2);
	}

	public void resetLocalData() {
		data.clear();
	}

	/*
	 * public void setDeltaDiurnalCoeffs(double[] diurnalCoeffs) {
	 * deltaDiurnalCoeffs = diurnalCoeffs; }
	 */

	public void setControlPoint( boolean value ) {
		this.controlPoint = value;
	}

	public boolean isControlPoint() {
		return this.controlPoint;
	}
	
	public StationLocation getLocation() {
		return location;
	}

	public void setUnused(boolean b) {
		this.unused = b;
	}

	public boolean isUnused() {
		return this.unused || location == null;
	}

	public void putTov(TimeScale tov) {
		data.setTov(tov);
	}

	public void putLocation( StationLocation location ) {
		this.location = location;
	}
	
	public void putObservations(double[] observations) {
		for (int i = 0; i < observations.length; i++) {
			data.put(i, LocalDataType.OBSERVED_VALUE_PROCESSED, observations[i]);
		}
	}

	public void putRawObservations(double[] observations) {
		for (int i = 0; i < observations.length; i++) {
			data.put(i, LocalDataType.OBSERVED_VALUE_RAW, observations[i]);
		}
	}

	public void putConfidences(int[] confidences) {
		for (int i = 0; i < confidences.length; i++) {
			data.put(i, LocalDataType.CONFIDENCE_SCORE, confidences[i]);
		}
	}

	public void putFlags(int[] flags) {
		for (int i = 0; i < flags.length; i++) {
			data.put(i, LocalDataType.FLAG, flags[i]);
		}
	}

/*
	public Double[] getDoubleArray( LocalDataType type ) {
		Object[] array = data.get(type);
		Double[] output = new Double[array.length];
		for (int i = 0; i < array.length; i++)
			output[i] = (Double)array[i];
		return output;
	}
*/
	public double[] getDoubleArray( LocalDataType type ) {
		Object[] array = data.get(type);
		double[] output = new double[array.length];
		for (int i = 0; i < array.length; i++) {
			output[i] = array[i] != null ? ((Double)array[i]).doubleValue() : NO_VALUE.SPECIAL;
		}
		return output;
	}

	public int[] intArray( LocalDataType type ) {
		Object[] array = data.get(type);
		int[] output = new int[array.length];
		for (int i = 0; i < array.length; i++) {
			output[i] = array[i] != null ? ((Integer)array[i]).intValue() : NO_VALUE.INT;
		}
		return output;
	}
	
	public double[] getObservations() {
		return getDoubleArray(LocalDataType.OBSERVED_VALUE_PROCESSED);
	}

	public double[] getRawObservations() {
		return getDoubleArray(LocalDataType.OBSERVED_VALUE_RAW);
	}

	public double[] getClimateValues() {
		return getDoubleArray(LocalDataType.CLIMATE_VALUE);
	}

	public double[] getLocalWeatherValues() {
		return getDoubleArray(LocalDataType.LOCAL_WEATHER_VALUE);
	}

	public double[] getGlobalWeatherValues() {
		return getDoubleArray(LocalDataType.GLOBAL_WEATHER_VALUE);
	}

	public double[] getObs2Climate() {
		return getDoubleArray(LocalDataType.O2C_DIFFERENCE);
	}

	public double[] getObs2GlobalWeather() {
		return getDoubleArray(LocalDataType.O2GW_DIFFERENCE);
	}

	public double[] getObs2LocalWeather() {
		return getDoubleArray(LocalDataType.O2LW_DIFFERENCE);
	}

	public double[] getTimegridRawObservations() {
		return getDoubleArray(LocalDataType.OBSERVED_VALUE_RAW);
	}

	public int[] getConfidenceScores() {
		return intArray(LocalDataType.CONFIDENCE_SCORE);
	}

	public int[] getFlags() {
		return intArray(LocalDataType.FLAG);
	}

	public int[] getTimegridMismatches() {
		return intArray(LocalDataType.TIMEGRID_MISMATCH_PERCENT);
	}

	public TimeGrid getTimeGrid() {
		return timeGrid;
	}

	public TimeScale getMidnightToday() {
		return data.getMidnightToday();
	}

	public TimeScale getTov() {
		return data.getTov();
	}

	static private double getThreshold(CharChoice cc) {
		switch (cc) {
		case FOF2:
			return THRESHOLD_JUMP_OF_OBSERVATIONS_FOF2;
		case HMF2:
			return THRESHOLD_JUMP_OF_OBSERVATIONS_HMF2;
		default:
			return Double.MAX_VALUE; // disable detection in derived characteristics
		}
	}

	/**
	 * Returns UML Characteristic Ident.<br>
	 * <br>
	 * Note:
	 * <ul>
	 * In DIDB, ident of corresponding characteristic is one more. For example,
	 * <code>CH.FOF2</code> is 0, and in DIDB it is 1.
	 * </ul>
	 * 
	 * @return
	 */
	public CharChoice getCharChoice() {
		return cc;
	}

	/**
	 * Set one data value to array <code>observation</code>, noting that
	 * OneDayHistory has particular organization first point has UT of 0, and
	 * there are TIME_STEPS_IN_DAY of points per day at resolution of TIME_STEP
	 * (see IRTAMConstants.java )
	 * 
	 * @param value
	 * @param confidenceScore
	 * @param time
	 */
	public void addDataValue(TimeScale time, double observedValue, int confidenceScore ) {
		int timeGridDeltaPercent = ((int)data.timeGridMismatch_min(time)*100)/timeGrid.getTimeStep_min();
		LocalDataElements cell = data.getCell(time);
		if (cell.isBetter(observedValue, confidenceScore, timeGridDeltaPercent)) {
			return;
		}		
		data.put(time, LocalDataType.OBSERVED_VALUE_RAW, observedValue);
		data.put(time, LocalDataType.CONFIDENCE_SCORE, confidenceScore);
		data.put(time, LocalDataType.TIMEGRID_MISMATCH_PERCENT, timeGridDeltaPercent);
	}

	/*
	 * public void addDataValue( double value, int confidenceScore, TimeScale
	 * time, int iTime) { // check time, figure out index into data array double
	 * minuteOfDay = time.getTimeSinceMidnightIn( TimeScale.MINUTE ); //int
	 * iTime = basis.timeToDiurnalPhaseIndex( time ); if( iTime ==
	 * NO_VALUE.INDEX ) return; double timeGridDelta = Math.abs( minuteOfDay -
	 * iTime*timeStep_min );
	 * 
	 * if( observation[iTime] != NO_VALUE.SPECIAL ) { if(
	 * timeGridMismatch[iTime] < timeGridDelta && confidence[iTime] >=
	 * confidenceScore ) // we already have data point at that grid return; //
	 * new candidate is not better if( confidenceScore < POOR_CONFIDENCE_SCORE )
	 * // is this data value worth even considering? return; if( timeGridDelta >
	 * 0.33*timeStep_min && // we have a potential good value for replacement
	 * isConfident( iTime ) ) // is it too far, though? return; }
	 * observation[iTime] = value; confidence[iTime] = confidenceScore;
	 * timeGridMismatch[iTime] = timeGridDelta; }
	 */

	public boolean sufficientData() {
		int counter = 0;
		double[] rawObservations = getRawObservations();
		int[] confidences = getConfidenceScores();
		for (int i = 0; i < timeGrid.getNumDiurnalSteps(); i++) {
			if (!IonogramScaling.isMissing(rawObservations[i]) && !IonogramScaling.isPoorQuality(confidences[i])) {
				counter++;
			}
		}
		return counter >= SUFFICIENT_DATA_POINTS;
	}

	public void fillClimateValues(Synthesizer climate ) {
		for (int i = 0; i < data.getLength(); i++) {
			data.put(i, LocalDataType.CLIMATE_VALUE, climate.synthesizeValue(location.lat, location.lon, i));
		}
	}

	public void deriveClimateValues(GeoSynthesizer climate ) {
		for (int i = 0; i < data.getLength(); i++) {
			data.put(i, LocalDataType.CLIMATE_VALUE, climate.synthesizeValue(location.lat, location.lon, i));
		}
	}

	public void fillGlobalWeatherValues(Synthesizer weather) {
		for (int i = 0; i < data.getLength(); i++) {
			data.put(i, LocalDataType.GLOBAL_WEATHER_VALUE, weather.synthesizeValue(location.lat, location.lon, i));
			if( weather.getExpansionBasis() instanceof ExpansionBasis_JonesGalletWithTrend ) {
				data.put(i, LocalDataType.GLOBAL_TREND, weather.synthesizeTrend(location.lat, location.lon, i));
			}
		}
	}

	public void deriveGlobalWeatherValues(GeoSynthesizer weather) {
		for (int i = 0; i < data.getLength(); i++) {
			data.put(i, LocalDataType.GLOBAL_WEATHER_VALUE, weather.synthesizeValue(location.lat, location.lon, i));
		}
	}
	
	public void removeOldestData( int period_min ) {
		int stopIndex = period_min/timeGrid.getTimeStep_min();
		for (int i = 0; i < stopIndex; i++) {
			data.put(i, LocalDataType.OBSERVED_VALUE_RAW, NO_VALUE.SPECIAL);
		}
	}

	/**
	 * For now, just fill in missing values by plain interpolation this will not
	 * work for big gaps need to involve model values. Uses observation IRI
	 * model values.
	 * 
	 * @return
	 */
	public void fillObservationGaps() {
		int i;
		int iBefore, iAfter;
		double[] rawObservations = getRawObservations();
		int length = rawObservations.length;
		double[] processedObservations = rawObservations.clone();
		int[] confidences = getConfidenceScores();
		int[] flags = new int[confidences.length];
		double[] climateValues = getClimateValues();
		for (i = 0; i < length; i++) {
			if (!IonogramScaling.isMissing(rawObservations[i]) && !IonogramScaling.isPoorQuality(confidences[i])) {
				flags[i] = IonogramScaling.ACCEPTED_ASIS;
			} else {
				// bad value... need to interpolate... find good point before
				// and after
				// 1. Look back into past for a good value
				iBefore = i - 1;
				do {
					if (iBefore < 0) {
						// no data prior to the missing value - reached the beginning of the buffer
						// wrap around into the today's data
						iBefore = timeGrid.getNumDiurnalSteps() - 1;
					}
					if (IonogramScaling.isMissing(rawObservations[iBefore]) || IonogramScaling.isPoorQuality(confidences[iBefore])) {
						iBefore--;
					} else {
						break;
					}
				} while( iBefore != i ); // full circle
				
				if( iBefore == i && rawObservations[iBefore] == NO_VALUE.SPECIAL) {
					// this is completely empty dataset
					return;
				}
				
				// found a good point for interpolation
				double sumBefore = rawObservations[iBefore];
				// keep going back, we need more points for averaging
				int countBefore = 1;
				for (int j = 1; j < TIME_STEPS_TO_AVERAGE; j++) {
					// keep going into the past
					int k = iBefore - j;
					if (k == -1) {
						// no more points can be found; this is the oldest data.
						// we do not want to wrap into today here. One point is good enough  
						break;
					} 
					if (!IonogramScaling.isMissing(rawObservations[k]) && !IonogramScaling.isPoorQuality(confidences[k])
							&& IonogramScaling.isAcceptedAsIs(flags[k])) {
						sumBefore += rawObservations[k];
						countBefore++;
					}
				}
				// the observation prior to missing point is now available
				// use its deviation from climate as the starting point for interpolation
				double delta1 = sumBefore / countBefore - climateValues[iBefore];

				// 2. Now find good point after the missing value
				// may run into the forecast mode if no data available up to the end of the buffer
				boolean forecast = true;
				iAfter = i + 1;
				if (iAfter < timeGrid.getNumDiurnalSteps()) {
					while (IonogramScaling.isMissing(rawObservations[iAfter]) || IonogramScaling.isPoorQuality(confidences[iAfter])) {
						iAfter++;
						if (iAfter == timeGrid.getNumDiurnalSteps()) {
							break;
						}
					}
					if( iAfter < timeGrid.getNumDiurnalSteps() ) {
						forecast = false;
					}
				}
				if( !forecast ) {
					// now find average "observation" to reduce dependence on
					// particular values
					double sumAfter = rawObservations[iAfter];
					int countAfter = 1;
					for (int j = 1; j < TIME_STEPS_TO_AVERAGE; j++) {
						int k = iAfter + j;
						if (k == length)
						 {
							break; // we do not want to take point from 24 hours ago
						}
						if (!IonogramScaling.isMissing(rawObservations[k]) && !IonogramScaling.isPoorQuality(confidences[k])
								&& IonogramScaling.isAcceptedAsIs(flags[k])) {
							sumAfter += rawObservations[k];
							countAfter++;
						}
					}
					// now we found "countAfter" points after the missing one. 
					// 3. Ready to interpolate
					double separation = iAfter - iBefore;
					if (separation < 0) {
						separation += timeGrid.getNumDiurnalSteps();
					}
					// double delta1 = observation[iBefore] - climatology[iBefore];
					// double delta2 = observation[iAfter] - climatology[iAfter];
					double delta2 = sumAfter / countAfter - climateValues[iAfter];
					double slope = (delta2 - delta1) / separation;
					int distanceToLeft = i - iBefore;
					if (distanceToLeft < 0) {
						distanceToLeft += timeGrid.getNumDiurnalSteps();
					}
					processedObservations[i] = climateValues[i] + (delta1 + distanceToLeft * slope);
					if( IonogramScaling.isMissing(rawObservations[i]) ) {
						flags[i] = IonogramScaling.INTERPOLATED_MISSING;
						data.put(i, LocalDataType.CONFIDENCE_SCORE, IonogramScaling.NOT_APPLICABLE_CONFIDENCE_SCORE);
					} else {
						flags[i] = IonogramScaling.INTERPOLATED_POOR_QUALITY;
					}
				}
				else {
				
				// 4. Extrapolation (forecast) is required	
					processedObservations[i] = climateValues[i] + delta1;
					flags[i] = IonogramScaling.FORECAST;
					data.put(i, LocalDataType.CONFIDENCE_SCORE, IonogramScaling.FORECAST_CONFIDENCE_SCORE);
				}				
			}
			data.put(i, LocalDataType.OBSERVED_VALUE_PROCESSED, processedObservations[i]);
			data.put(i, LocalDataType.FLAG, flags[i]);
		}
		return;
	}

	/**
	 * Adjust tail (oldest data), if there is big jump in data when go from
	 * yesterday to today.
	 * 
	 * @param toTime
	 * @return true if tail was adjusted, false otherwise
	 */
	public boolean smoothDayBoundary(TimeScale toTime, int nowIndex) {
		// don't adjust boundary for derived characteristics
		if (cc.isDerived()) {
			return false;
		}
		double threshold = getThreshold(cc);
		int pointNeeded = 0;
		int nTimes = timeGrid.getNumDiurnalSteps();
		int beforeIndex_conf = (nowIndex + nTimes) % nTimes;
		int afterIndex_conf = (nowIndex + 1) % nTimes;
		double[] observations = getObservations();
		int[] confidences = getConfidenceScores();
		int[] flags = getFlags();
		double[] climateValues = getClimateValues();
		// search for first confident point before now including now
		for (int iBefore = 0; iBefore < TIME_STEPS_SEARCH; iBefore++) {
			int beforeIndex = (nowIndex - iBefore + nTimes) % nTimes;
			if (IonogramScaling.isAcceptedAsIs(flags[beforeIndex]) && !IonogramScaling.isPoorQuality(confidences[beforeIndex])) {
				beforeIndex_conf = beforeIndex;
				pointNeeded++; // first point needed for adjustment found
				break;
			}
		}
		if (pointNeeded != 1) {
			return false;
		}
		// search for first confident point after now
		for (int iAfter = 1; iAfter < TIME_STEPS_SEARCH; iAfter++) {
			int afterIndex = (nowIndex + iAfter) % nTimes;
			if (IonogramScaling.isAcceptedAsIs(flags[afterIndex]) && !IonogramScaling.isPoorQuality(confidences[afterIndex])) {
				// TODO: what if there is no such point?
				afterIndex_conf = afterIndex;
				break;
			}
		}
		if (Math.abs(observations[afterIndex_conf] - observations[beforeIndex_conf]) > threshold) {
			int endIndex_conf = (nowIndex + TIME_STEPS_TO_SKIP) % nTimes;
			// search for first confident point around TIME_STEPS_TO_SKIP
			// time steps after latest data point
			int i = 0;
			for (int iEnd = 1; iEnd < TIME_STEPS_SEARCH * 2; iEnd++) {
				i = (iEnd / 2) * (-1) * (i != 0 ? i : 1) / Math.abs(i != 0 ? i : 1);
				// System.out.println(i);
				int endIndex = (nowIndex + TIME_STEPS_TO_SKIP + i) % nTimes;
				if (IonogramScaling.isAcceptedAsIs(flags[endIndex]) && !IonogramScaling.isPoorQuality(confidences[endIndex])) {
					endIndex_conf = endIndex;
					// second point needed for adjustmentfound
					pointNeeded++; 
					break;
				}
			}
			if (pointNeeded == 2) {
				double delta1 = observations[beforeIndex_conf] - climateValues[beforeIndex_conf];
				double delta2 = observations[endIndex_conf] - climateValues[endIndex_conf];
				int interval = (endIndex_conf - beforeIndex_conf + nTimes) % nTimes - 1;
				double slope = (delta2 - delta1) / (interval);
				for (int iStep = 0; iStep < interval; iStep++) {
					// start with point next to confident
					int index = (beforeIndex_conf + iStep + 1) % nTimes; 
					data.put(index, LocalDataType.OBSERVED_VALUE_PROCESSED, climateValues[index] + (delta1 + iStep * slope));
					data.put(index, LocalDataType.FLAG, flags[index] + IonogramScaling.DAY_BOUNDARY);
				}
			} else {
				return false;
			}
		} else {
			return false;
		}
		return true;
	}

	public void calculateO2Cdifferences() {
		double[] observations = getObservations();
		double[] climateValues = getClimateValues();
		for (int i = 0; i < observations.length; i++) {
			if( !IonogramScaling.isMissing( observations[i])) {
				data.put( i, LocalDataType.O2C_DIFFERENCE, observations[i] - climateValues[i]);
			} else {
				data.put( i, LocalDataType.O2C_DIFFERENCE, NO_VALUE.SPECIAL);
			}
		}
	}
	
	public void calculateO2LWdifferences() {
		double[] observations = getObservations();
		double[] localWeatherValues = getLocalWeatherValues();
		for (int i = 0; i < observations.length; i++) {
			if( !IonogramScaling.isMissing( observations[i])) {
				data.put( i, LocalDataType.O2LW_DIFFERENCE, observations[i] - localWeatherValues[i]);
			} else {
				data.put( i, LocalDataType.O2LW_DIFFERENCE, NO_VALUE.SPECIAL);
			}
		}
	}
	
	public void calculateO2GWdifferences() {
		double[] observations = getObservations();
		double[] globalWeatherValues = getGlobalWeatherValues();
		for (int i = 0; i < observations.length; i++) {
			if( !IonogramScaling.isMissing( observations[i])) {
				data.put( i, LocalDataType.O2GW_DIFFERENCE, observations[i] - globalWeatherValues[i]);
			} else {
				data.put( i, LocalDataType.O2GW_DIFFERENCE, NO_VALUE.SPECIAL);
			}
		}
	}
}
