/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.database;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.TreeMap;

import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.geospace.EarthGrid;
import edu.uml.lgdc.geospace.TimeGrid;
import edu.uml.lgdc.project.Console;
import edu.uml.lgdc.time.TimeScale;
import edu.uml.giro.gambit.core.GambitConstants;
import edu.uml.giro.gambit.core.IonogramScaling;

public class DataFunctions {
	static public final int NO_CORRELATION = -2;

	static public void printoutGridValues(double[] vals, int latN, int lonN) {
		for (int iLat = 0; iLat < latN; iLat++) {
			for (int iLon = 0; iLon < lonN; iLon++) {
				System.out.print(FC.DoubleToString(vals[iLon + iLat * lonN], 5, 2) + " ");
			}
			System.out.println();
		}
	}

	// ////////// COEFFS OPERATIONS
	static public void addCoeffs(double[][] base, double[][] addition) {
		if (base.length != addition.length)
			return;
		for (int i = 0; i < base.length; i++)
			for (int j = 0; j < base[i].length; j++) {
				base[i][j] += addition[i][j];
			}
	}

	// ////////

	/**
	 * Calculates correlation between two data sets (initially developed for
	 * observations), using Pearson approach. Coordinates are given to calculate
	 * the LT for each data set and shift data sets correctly.
	 * 
	 * @param data1
	 * @param data2
	 * @param coord1
	 * @param coord2
	 * @return
	 */
	static public double findCorrelation_cyclic(double[] data1, double[] data2, int[] conf1, int[] conf2, double[] coord1, double[] coord2,
			boolean offset) {
		if (data1.length != data2.length) {
			Console.showError("Correlation_cyclic call: Data arrays are not the same length.");
			return 0;
		}
		if (coord1[1] > 180.)
			coord1[1] = coord1[1] - 360.;
		if (coord2[1] > 180.)
			coord2[1] = coord2[1] - 360.;

		double mean1 = 0;
		double mean2 = 0;
		double sdo1 = 0;
		double sdo2 = 0;
		double correlation = 0;
		int num = 0;
		double timeDifference = (coord1[1] - coord2[1]) / GambitConstants.LONDEGREE_IN_HOUR * TimeScale.MINUTES_PER_HOUR;
		int shift = (int) Math.round(timeDifference / TimeGrid.DEFAULT_TIME_STEP);
		if (offset == false)
			shift = 0;
		for (int i = 0; i < data1.length; i++) {
			int j = (i + shift + data1.length) % data1.length;
			if (!IonogramScaling.isPoorQuality(conf1[i]) && !IonogramScaling.isPoorQuality(conf2[j]) ) {
				mean1 += data1[i];
				mean2 += data2[j];
				num++;
			}
		}
		if (num < data1.length / 4)
			return NO_CORRELATION;
		mean1 /= num;
		mean2 /= num;
		// System.out.println(coord1[1] + "   " + coord2[1] + "   " + shift);
		for (int i = 0; i < data1.length; i++) {
			int j = (i + shift + data1.length) % data1.length;
			if (!IonogramScaling.isPoorQuality(conf1[i]) && !IonogramScaling.isPoorQuality(conf2[j]) ) {
				correlation += (data1[i] - mean1) * (data2[j] - mean2);
				sdo1 += (data1[i] - mean1) * (data1[i] - mean1);
				sdo2 += (data2[j] - mean2) * (data2[j] - mean2);
				/*
				 * System.out.println( FC.IntegerToString(i,2) + " " +
				 * FC.IntegerToString(j, 2) + "       " +
				 * FC.DoubleToString(data1[i], 5,2) + " " +
				 * FC.DoubleToString(data2[i], 5,2) + "       " +
				 * FC.DoubleToString(data1[i], 5,2) + " " +
				 * FC.DoubleToString(data2[j], 5,2) + "       " +
				 * FC.DoubleToString(data1[i] - data2[j], 5,2));
				 */
			}
		}
		correlation /= Math.sqrt(sdo1 * sdo2);
		if (Double.isNaN(correlation)) {
			System.err.println("Check data in correlations");
			return NO_CORRELATION;
		}
		return correlation;
	}

	static public boolean isBufferUpdated(final TimeScale startBufTime, final TimeScale endBufTime, final TimeScale fromTime,
			final TimeScale endTime, int numForecastDays, boolean forecast) {
		TimeScale end = forecast == true ? new TimeScale(endBufTime.getTimeInMinutes() + numForecastDays * TimeScale.MINUTES_PER_DAY) : endBufTime;
		if (startBufTime == null || end == null)
			return false;
		else if (startBufTime.after(fromTime) || end.before(endTime))
			return false;
		return true;
	}

	static public boolean isBufferUpdated(final TimeScale startBufTime, final TimeScale endBufTime, final TimeScale fromTime, final TimeScale endTime) {
		return isBufferUpdated(startBufTime, endBufTime, fromTime, endTime, 0, false);
	}

	// /////////////

	static public TreeMap<Integer, StationLocation> getSitesInSector(TreeMap<Integer, StationLocation> locs, double latMin, double latMax,
			double lonMin, double lonMax) {
		lonMin = lonMin > 180 ? lonMin - 360 : lonMin;
		lonMax = lonMax > 180 ? lonMax - 360 : lonMax;
		TreeMap<Integer, StationLocation> locations_sector = new TreeMap<Integer, StationLocation>();
		for (Integer i : locs.keySet()) {
			double lon = locs.get(i).lon > 180 ? (locs.get(i).lon - 360) : locs.get(i).lon;
			double lat = locs.get(i).lat;
			if (lonMin < lonMax) {
				if (lon >= lonMin && lon <= lonMax && lat >= latMin && lat <= latMax)
					locations_sector.put(i, locs.get(i));
			} else if (lonMin > lonMax) {
				if ((lon >= lonMin || lon <= lonMax) && lat >= latMin && lat <= latMax)
					locations_sector.put(i, locs.get(i));
			} else {
				if (lat >= latMin && lat <= latMax)
					locations_sector.put(i, locs.get(i));
			}
		}
		return locations_sector;
	}

	/**
	 * Finds distance "all with all".
	 * 
	 * @param locations
	 * @return 2D matrix of the distances with size equals to locations.size()
	 *         and with order appears in locations
	 */
	static public double[][] getDistanceMatrix(TreeMap<Integer, StationLocation> locations) {
		int size = locations.size();
		double[][] dists = new double[size][size];
		int ii = 0;
		int jj = 0;
		for (Integer i : locations.keySet()) {
			jj = 0;
			for (Integer j : locations.keySet()) {
				dists[ii][jj] = EarthGrid.onEarthDistance_deg(locations.get(i).lat, locations.get(i).lon, locations.get(j).lat,
						locations.get(j).lon);
				jj++;
			}
			ii++;
		}
		return dists;
	}

	// /////////////DATA FILE OPERATION
	/**
	 * Reads file for URSI codes.
	 * 
	 * @param filename
	 * @return
	 */
	static public TreeMap<String, Integer> getSitesFromFile(String filename) {
		TreeMap<String, Integer> ursi = new TreeMap<String, Integer>();
		File fileCheck = new File(filename);
		if (!fileCheck.exists())
			return null;
		try {
			BufferedReader br = new BufferedReader(new FileReader(filename));
			String temp = br.readLine();
			while (temp != null) {
				ursi.put(temp.trim(), ursi.size());
				temp = br.readLine();
			}
			br.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return ursi;
	}
}
