/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.database;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.TreeMap;

import edu.uml.giro.gambit.core.AlgorithmVersions;
import edu.uml.lgdc.appuserdb.api.DatabaseCredentials;
import edu.uml.lgdc.database.ConnectItem;
import edu.uml.lgdc.database.DatabaseContents;
import edu.uml.lgdc.project.Console;

public class GambitDatabase {
	static public final String JAYBIRD_DRIVER = "org.firebirdsql.jdbc.FBDriver";
	// static protected final String URL =
	// "jdbc:firebirdsql://129.63.134.12/C:\\GAMBIT-DB\\gambit";
	//static protected final String ROLE_NAME_READER = "COMMON";

	// various database operational modes
	static public final int READ_COEFFS = 1;
	static public final int READ_OBSERVATIONS = 2;
	static public final int READ = READ_COEFFS + READ_OBSERVATIONS;
	static public final int WRITE = 4;
	static public final int USE_DIDBASE_ASSIMILATION = 8;
	
	// various modes of handling errors of database connection
	static public final int DBERROR_QUIT = 1;
	static public final int DBERROR_LOOP = 2;
	// static public final int DBERROR_ASK = 3;  TODO will be deleted
	
	private int dbErrorHandlingMode;
	private boolean goodConnection = false;
	
	public final TreeMap<Integer,AlgorithmVersions> GAMBIT_DB_VERSIONS;

	public GambitDatabase( int dbErrorHandlingMode) {
		this.dbErrorHandlingMode = dbErrorHandlingMode;
		GAMBIT_DB_VERSIONS = getAlgorithmVersions();
	}

	private Driver getDriver(ConnectItem c) throws SQLException {
		String driverClassName = c.getDriverClassName();
		if( driverClassName == null || driverClassName.length() < 2 )
			driverClassName = JAYBIRD_DRIVER;   
		try {
			
			Class.forName(driverClassName);
		} catch (ClassNotFoundException e) {
			throw new SQLException(e);
		}
		goodConnection = false;
		return DriverManager.getDriver(c.getURL());
	}
	
	protected Connection open(ConnectItem c ) throws SQLException {
		final int smallInterval_sec = 5;
		final int bigInterval_sec = 60;
		final Properties properties = new Properties();
		int wait_sec = smallInterval_sec;
		boolean noMoreWaitMessage = false;
		while (true) { 
			if (c.getUsername() != null) {
				properties.put("user", c.getUsername());
			}
			if (c.getPass() != null) {
				properties.put("password", c.getPass());
			}
			if (c.getRole() != null) {
				properties.put("roleName", c.getRole());
			}
			goodConnection = false;
			Driver driver = getDriver( c );
			try {
				Connection connection = driver.connect(c.getURL(), properties);
				goodConnection = true;
				return connection;
			}
			catch (SQLException e) {
				String text = "Connection to GAMBIT database " + c.getURL() + " in the " + c.getRole() + " role failed";
				switch( dbErrorHandlingMode ) {
				case DBERROR_QUIT:
					Console.showError(text + "... no connection possible.");
					throw e;
				case DBERROR_LOOP:
					if (!noMoreWaitMessage) {
						if (wait_sec >= bigInterval_sec) {
							wait_sec = bigInterval_sec;
							noMoreWaitMessage = true;
							Console.showError(text + "... connection attempts will continue every " + wait_sec + " sec");
						} else {
							Console.showError(text + "... will try again in " + wait_sec + " sec");
						}
					}
					try {
						Thread.sleep(wait_sec * 1000);
						if (wait_sec < bigInterval_sec) {
							wait_sec += wait_sec;
						}
					} catch (InterruptedException e1) {
						throw new RuntimeException(text + "... restoring attempts interrupted.");
					}
					break;
				/*  TODO will be deleted
				case DBERROR_ASK:
					Console.showError(text+"... going interactive...");
					// connection impossible, request new credentials interactively
					UserInfoFrame infoFrame = new UserInfoFrame(GambitConstants.userInfo);
					infoFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
					user = GambitConstants.userInfo.get(GambitConstants.USER_GAMBIT_LOGIN);
					password = GambitConstants.userInfo.get(GambitConstants.USER_GAMBIT_PASSWORD);
					break;
				*/
				default:
					throw new RuntimeException("Database connection impossible for in unknown error management mode "+dbErrorHandlingMode);
				}
			}
		}
	}

	public Connection openRead() throws SQLException {
		ConnectItem item = DatabaseCredentials.getDatabaseConnectItem(DatabaseContents.IRTAM); 
		return open(item);
	}

	public boolean isGood() {
		return goodConnection;
	}
	
	public int getExpansionID(String expansionName, int temporalLen, int spatialLen) throws SQLException {
		Connection connect = openRead();
		String querySelect = "SELECT Ident FROM Expansion WHERE Name ='" + expansionName + "'" + " AND TemporalBasisLength=" + temporalLen
			+ " AND SpatialBasisLength=" + spatialLen;
		ResultSet rs = connect.createStatement().executeQuery(querySelect);
		if (!rs.next()) {
			throw new SQLException("Expansion basis " + expansionName + "with " + temporalLen + " x " + spatialLen
				+ " coefficient set is not registered in this GAMBIT database" );
		}
		int id = rs.getInt("Ident");
		rs.close();
		connect.close();
		return id;
	}

	public int getAssimilationID(String assimilationName, int nLats, int nLons) throws SQLException {
		Connection connect = openRead();
		String querySelect = "SELECT Ident FROM Assimilation WHERE Name ='" + assimilationName + "'" + " AND EarthGridLats=" + nLats
			+ " AND EarthGridLons=" + nLons;
		ResultSet rs = connect.createStatement().executeQuery(querySelect);
		if (!rs.next()) {
			throw new SQLException("Assimilation technique " + assimilationName + " on " + nLats + " x " + nLons
				+ " grid is not registered in this GAMBIT database");
		}
		int id = rs.getInt("Ident");
		rs.close();
		connect.close();
		return id;
	}

	public int getIrtamVersionID(String irtamVersionName, int aID, int exID) throws SQLException {
		Connection connect = openRead();
		String querySelect =
			"SELECT Ident FROM Version WHERE Name ='" + irtamVersionName + "'" + " AND AssimilationID=" + aID + " AND ExpansionID=" + exID;
		ResultSet rs = connect.createStatement().executeQuery(querySelect);
		if (!rs.next()) {
			throw new SQLException("IRTAM configuration " + irtamVersionName + " with engine " + aID + " and basis " + exID
				+ " is not registered in this GAMBIT database");
		}
		int id = rs.getInt("Ident");
		rs.close();
		connect.close();
		return id;
	}

	/**
	 * Returns stations IDs as they are in database
	 * 
	 * @param connect
	 * @param stationsUsed
	 * @return
	 * @throws SQLException
	 */
	public String[] getStationUrsiCodes() throws SQLException {
		try (Connection connect = openRead()) {
			return getStationUrsiCodes(connect);
		}
	}

	private String[] getStationUrsiCodes(Connection connect) throws SQLException {
		String querySelect = "SELECT UrsiCode FROM Station ORDER BY UrsiCode";
		try (Statement statement = connect.createStatement()) {
			ResultSet rs = statement.executeQuery(querySelect);
			List<String> ursiCodes = new ArrayList<>();
			while (rs.next()) {
				ursiCodes.add(rs.getString(1).trim().toUpperCase());
			}
			return ursiCodes.toArray(new String[0]);
		}
	}

	/**
	 * Returns all records of station constants as they are registered in GAMBIT database
	 * 
	 * @return constants
	 */
	public TreeMap<Integer,StationLocation> getAllStationLocs() {
		TreeMap<Integer,StationLocation> locs = null;
		try {
			locs = getStations();
		} catch (SQLException e) {
			e.printStackTrace();
		}
		return locs;
	}

	private TreeMap<Integer,StationLocation> getStations() throws SQLException {

		Connection connect = openRead();
		TreeMap<Integer,StationLocation> stationsIdents = new TreeMap<Integer,StationLocation>();
		String querySelect = "SELECT UrsiCode, Ident, Lat, Lon, Name FROM Station";
		Statement statement = connect.createStatement();
		ResultSet rs = statement.executeQuery(querySelect);
		while (rs.next()) {
			StationLocation loc = new StationLocation(rs.getString("UrsiCode").trim().toUpperCase(), rs.getString("Name").trim().toUpperCase(),
				rs.getFloat("Lat"), rs.getFloat("Lon"), rs.getInt("Ident"));
			stationsIdents.put(rs.getInt("Ident"), loc);
		}
		rs.close();
		return stationsIdents;
	}

	public TreeMap<Integer,AlgorithmVersions> getAlgorithmVersions() {
		TreeMap<Integer,AlgorithmVersions> algorithms = new TreeMap<Integer,AlgorithmVersions>();
		try {
			Connection connect = openRead();
			String querySelect =
				"SELECT v.Ident, v.Name as VersionName, a.Name as AssimilationName, a.EarthGridLats, a.EarthGridLons, e.Name as BasisName, e.TemporalBasisLength, "
					+ "e.SpatialBasisLength, v.Description FROM Version v "
					+ "JOIN Assimilation a ON v.AssimilationID=a.Ident JOIN Expansion e ON v.ExpansionID=e.Ident " + "ORDER BY v.Ident DESC";
			Statement statement = connect.createStatement();
			ResultSet rs = statement.executeQuery(querySelect);
			int count = 0;
			while (rs.next()) {
				AlgorithmVersions av = new AlgorithmVersions();
				av.versionID = rs.getInt("Ident");
				av.name = rs.getString("VersionName").trim();
				av.assimilationName = rs.getString("AssimilationName").trim();
				av.nLats = rs.getInt("EarthGridLats");
				av.nLons = rs.getInt("EarthGridLons");
				av.expansionBasisName = rs.getString("BasisName").trim();
				av.temporalExpansionTerms = rs.getInt("TemporalBasisLength");
				av.spatialExpansionTerms = rs.getInt("SpatialBasisLength");
				algorithms.put(count++, av);
			}
			rs.close();
		} catch (SQLException e) {
			System.err.println("Could not retrieve algorithm descriptions from GAMBIT database. Assuming old database design");
			AlgorithmVersions av = new AlgorithmVersions();
			av.versionID = 1;
			av.name = "IRTAM 0.1B";
			av.assimilationName = "NECTAR";
			av.nLats = 36;
			av.nLons = 35;
			av.expansionBasisName = "JonesGallet";
			av.temporalExpansionTerms = 13;
			av.spatialExpansionTerms = 76;
			algorithms.put(0, av);
		}
		return algorithms;
	}

}
