/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.database;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Calendar;

import edu.uml.lgdc.database.FirebirdUtil;
import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.geospace.CharChoice;
import edu.uml.lgdc.project.Console;
import edu.uml.lgdc.quantities.NO_VALUE;
import edu.uml.lgdc.time.TimeScale;
import edu.uml.giro.gambit.core.LocalData;
import edu.uml.giro.gambit.synthesizers.GeoSynthesizer;

public class GambitDatabaseReader extends GambitDatabase {

	private InputStream[][] buffer_coeffs = new InputStream[CharChoice.SIZE][];
	private InputStream[][] buffer_obs = new InputStream[CharChoice.SIZE][];
	private TimeScale[][] buffer_times_coeffs = new TimeScale[CharChoice.SIZE][];
	private TimeScale[][] buffer_times_obs = new TimeScale[CharChoice.SIZE][];
	private TimeScale[] startBufTime_coeffs = new TimeScale[CharChoice.SIZE];
	private TimeScale[] endBufTime_coeffs = new TimeScale[CharChoice.SIZE];
	private TimeScale[] startBufTime_obs = new TimeScale[CharChoice.SIZE];
	private TimeScale[] endBufTime_obs = new TimeScale[CharChoice.SIZE];

	public GambitDatabaseReader( int dbErrorHandlingMode ) {
		super( dbErrorHandlingMode );
	}
	
	public InputStream[] download(GeoSynthesizer weather, LocalData obs, TimeScale refTime, int content) throws SQLException, IOException {
		CharChoice cc = obs.getCharChoice();
		TimeScale now = new TimeScale();
		TimeScale endTime = new TimeScale(refTime.getTimeInMinutes() );
		endTime = endTime.before(now) ? endTime : now;

		// First, determine if any actual download is needed to buffers
		// and if coeffs. obs, or both coeffs & obs are to be fetched

		downloadToBuffers(cc, weather, obs, refTime, endTime, content);

		// Now all data are supposedly in buffers
		// position pointer to the buffers, return input stream, if any

		int pointerCoeffs = NO_VALUE.INDEX;
		int pointerObs = NO_VALUE.INDEX;

		if ((content & READ_COEFFS) != 0) {
			pointerCoeffs = weather.getTimeGrid().findTimeNode(refTime, weather.getTimeGrid().snapToClosestGrid(startBufTime_coeffs[cc.getIndex()]));
		}
		if ((content & READ_OBSERVATIONS) != 0) {
			pointerObs = obs.getTimeGrid().findTimeNode(refTime, obs.getTimeGrid().snapToClosestGrid(startBufTime_obs[cc.getIndex()]));
		}

		return new InputStream[] { pointerCoeffs == NO_VALUE.INDEX ? null : buffer_coeffs[cc.getIndex()][pointerCoeffs],
				pointerObs == NO_VALUE.INDEX ? null : buffer_obs[cc.getIndex()][pointerObs] };
	}

	/**
	 * 
	 */
	private void downloadToBuffers(CharChoice cc, GeoSynthesizer weather, LocalData obs, final TimeScale startTime, final TimeScale endTime,
			int content) throws SQLException, IOException {
		boolean downloadCoeffs = ((content & READ_COEFFS) != 0)
				&& !DataFunctions.isBufferUpdated(startBufTime_coeffs[cc.getIndex()], endBufTime_coeffs[cc.getIndex()], startTime, startTime);
		boolean downloadObs = ((content & READ_OBSERVATIONS) != 0)
				&& !DataFunctions.isBufferUpdated(startBufTime_obs[cc.getIndex()], endBufTime_obs[cc.getIndex()], startTime, startTime);

		if (!downloadCoeffs && !downloadObs)
			return;

		TimeScale utStart = new TimeScale();
		// FIXME: this download implementation does not support different time grids of obs and weather

		TimeScale stTime = obs.getTimeGrid().lowerBoundSnapped(startTime);
		TimeScale eTime = obs.getTimeGrid().upperBoundSnapped(endTime);
		int numTimeSteps = obs.getTimeGrid().calculateNumberTimeSteps(startTime, endTime);

		String query = "SELECT TimeUT,";
		if (downloadCoeffs) {
			query += "Data";
			buffer_coeffs[cc.getIndex()] = new InputStream[numTimeSteps];
			buffer_times_coeffs[cc.getIndex()] = new TimeScale[numTimeSteps];
		}
		if (downloadCoeffs && downloadObs)
			query += ",";
		if (downloadObs) {
			query += "Observations";
			buffer_obs[cc.getIndex()] = new InputStream[numTimeSteps];
			buffer_times_obs[cc.getIndex()] = new TimeScale[numTimeSteps];
		}
		
		query += " FROM Coefficient WHERE TimeUT>='" + stTime.toTimestamp() + "' AND TimeUT<'" + eTime.toTimestamp()
				+ "' AND CharacteristicID=" + cc.getDidbaseIdent();
		
		// the original GAMBIT database version 1 does not have versionID
		if( weather.getIrtamVersionIdent() > 1)
			query += " AND VersionID=" + weather.getIrtamVersionIdent();

		if (downloadCoeffs) {
			startBufTime_coeffs[cc.getIndex()] = (TimeScale) stTime.clone();
			endBufTime_coeffs[cc.getIndex()] = (TimeScale) eTime.clone();
		}
		if (downloadObs) {
			startBufTime_obs[cc.getIndex()] = (TimeScale) stTime.clone();
			endBufTime_obs[cc.getIndex()] = (TimeScale) eTime.clone();
		}
		// for correct calculation of time grid points
		stTime = obs.getTimeGrid().snapToClosestGrid(stTime);
		eTime = obs.getTimeGrid().snapToClosestGrid(eTime);
		try (Connection connect = openRead()) {
			Statement statement = connect.createStatement();
			ResultSet rs = statement.executeQuery(query);
			int count = 0;
			while (rs.next()) {
				TimeScale time = FirebirdUtil.getTime(rs, 1);
				int iTime = obs.getTimeGrid().findTimeNode(time, stTime);
				if (downloadCoeffs) {
					buffer_coeffs[cc.getIndex()][iTime] = copyInputStream(rs.getBinaryStream("Data"));
					buffer_times_coeffs[cc.getIndex()][iTime] = (TimeScale) time.clone();
				}
				if (downloadObs) {
					buffer_obs[cc.getIndex()][iTime] = copyInputStream(rs.getBinaryStream("Observations"));
					buffer_times_obs[cc.getIndex()][iTime] = (TimeScale) time.clone();
				}
				count++;
			}
		}
	}

	private InputStream copyInputStream(InputStream is) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		byte[] buffer = new byte[1024];
		int len;
		while ((len = is.read(buffer)) > -1)
			baos.write(buffer, 0, len);
		baos.flush();
		return new ByteArrayInputStream(baos.toByteArray());
	}
	
	public TimeScale getLatestTime(CharChoice cc) {
		return getLatestTime(cc, GAMBIT_DB_VERSIONS.get(0).versionID);
	}
	
	public TimeScale getLatestTime(CharChoice cc, int versionID) {
		try (Connection gambitConnect = openRead(); Statement statement = gambitConnect.createStatement()) {
			String queryStr = 
				"SELECT Max(TimeUT) FROM Coefficient WHERE CharacteristicID=" + cc.getDidbaseIdent() + " AND VersionID=" + versionID;
			ResultSet rs = statement.executeQuery(queryStr);
			if( rs.next() )
				return FirebirdUtil.getTime(rs, 1);
		} catch (SQLException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public TimeScale getClosestTime(TimeScale tov, CharChoice cc) {
		return getClosestTime(tov, cc, GAMBIT_DB_VERSIONS.get(0).versionID);
	}
	
	public TimeScale getClosestTime(TimeScale tov, CharChoice cc, int versionID) {
		try (Connection gambitConnect = openRead(); Statement statement = gambitConnect.createStatement()) {
			StringBuilder sb = new StringBuilder(256);
			sb.append("SELECT Max(TimeUT) FROM Coefficient WHERE CharacteristicID=").append(cc.getDidbaseIdent()).append(" AND VersionID=")
				.append(versionID).append(" AND TimeUT<='").append(tov.toTimestamp()).append("' UNION ")
				.append("SELECT Min(TimeUT) FROM Coefficient WHERE CharacteristicID=").append(cc.getDidbaseIdent()) 
				.append(" AND VersionID=").append(versionID).append(" AND TimeUT>'").append(tov.toTimestamp()).append("'");
			ResultSet rs = statement.executeQuery(sb.toString());
			rs.next();
			TimeScale closestLessOrEqual = FirebirdUtil.getTime(rs, 1);
			rs.next();
			TimeScale closestGreater = FirebirdUtil.getTime(rs, 1);
			if (closestGreater != null && closestLessOrEqual != null) {
				if (closestGreater.diffIn(Calendar.SECOND, tov) <= tov.diffIn(Calendar.SECOND, closestLessOrEqual))
					return closestGreater;
				else
					return closestLessOrEqual;
			}
			return closestGreater != null ? closestGreater : closestLessOrEqual;
		} catch (SQLException e) {
			e.printStackTrace();
			return null;
		}
	}
	
	public boolean recordExists(TimeScale tov, int versionID) {
		try (Connection gambitConnect = openRead(); Statement statement = gambitConnect.createStatement()) {
			String querySelect = "SELECT TimeUT FROM Coefficient WHERE TimeUT='" + tov.toTimestamp() + "' AND VersionID=" + versionID;
			ResultSet rs = statement.executeQuery(querySelect);
			if( rs.next() )
				return true;
		} catch (SQLException e) {
			e.printStackTrace();
		}
		return false;
	}
}
