/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.expansions;

import edu.uml.lgdc.geospace.EarthGrid;
import edu.uml.lgdc.geospace.TimeGrid;

/**
 * Diurnal/spatial expansion API for empirical models
 * 
 * This class contains methods to manipulate base functions for (1) diurnal and
 * (2) spatial expansions. All coefficients are assumed to be 2D (compact form),
 * one dimension for time and one dimension for space. Compact from of a spatial
 * coefficients can be transformed into full 2D matrix and back using methods:
 * 
 * <p>
 * <code>abstract int orderDegreeToCompact(int sOrder, int sDegree, int sTerm);</code>
 * <p>
 * <code>abstract int[] compactToOrderDegree(int sCompact);   </code>
 * 
 * <p>
 * Spatial coefficients includes (2a) longitudinal (order) and (2b) latitudinal
 * (degree). Also only diurnal, i.e. 24-hour, period assumes for time variation.
 * 
 * 
 * @author Ivan
 *
 */

abstract public class ExpansionBasis2D extends ExpansionBasis {
	protected static final int INDEX_ORDER = 0;
	protected static final int INDEX_DEGREE = 1;
	protected static final int INDEX_TERM = 2;
	protected static final int MAX_TEMPORAL_ORDER = 12;

	protected int temporalOrder;
	protected int spatialOrder;
	protected int spatialDegree;

	public ExpansionBasis2D() {
	}

	public boolean setup(int temporalOrder, int spatialOrder, int spatialDegree) {
		return setTemporalOrder(temporalOrder) && setSpatialOrder(spatialOrder) && setSpatialDegree(spatialDegree);
	}

	public boolean setTemporalOrder(int diurnalOrder) {
		this.temporalOrder = diurnalOrder;
		if (diurnalOrder > MAX_TEMPORAL_ORDER) {
			System.err.println("Requested temporal order of expansion " + diurnalOrder + " is larger than maximum " + MAX_TEMPORAL_ORDER);
			return false;
		}
		return true;
	}

	abstract protected boolean setupTemporalBasisFunctions();

	abstract public boolean setSpatialOrder(int spatialOrder);

	abstract public boolean setSpatialDegree(int spatialDegree);

	abstract protected boolean setupSpatialBasisFunctions();

	/**
	 * returns number of coefficients in diurnal expansion dimension
	 * 
	 * @return
	 */
	public int getTemporalBasisLength() {
		return temporalOrder * 2 + 1;
	}

	/**
	 * Returns total compact number of coefficients in spatial expansion
	 * dimension. Single count of all non zero basis functions of various orders
	 * and degrees. Total is not necessarily equal to Order x Degree Some
	 * compact 1D representations can be used
	 * 
	 * @return int
	 */
	abstract public int getCompactSpatialBasisLength();

	/**
	 * In general depends which combinations of orders and degrees are used
	 * 
	 * @param sOrder
	 * @param sDegree
	 * @param sTerm
	 *            - can take two values 0 and 1. 0 corresponds to cos() in
	 *            longitudinal terms and 1 to sin() -- FIXME: really?
	 * 
	 * @return
	 */
	abstract public int orderDegreeTermToCompact(int sOrder, int sDegree, int sTerm);

	/**
	 * 
	 * @param sCompact
	 * @return array of three elements {order, degree, term}
	 */
	abstract public int[] compactToOrderDegreeTerm(int sCompact);

	/**
	 * access methods to diurnal order
	 * 
	 * @return diurnalOrder
	 */
	public int getTemporalOrder() {
		return temporalOrder;
	}

	protected int getSpatialOrder() {
		return spatialOrder;
	}

	protected int getSpatialDegree() {
		return spatialDegree;
	}

	/**
	 * Returns base function value for diurnal expansion
	 * 
	 * @param index
	 *            -- index of diurnal expansion functions, goes from 0 to
	 *            timeOrder*2
	 * @param iTime
	 *            -- index of the diurnal phase, from 0 to numDiurnalNodes-1
	 * @return
	 */
	abstract public double getTemporalBasisFunctionValue(int index, int iTime);

	// here k_index is a "cumulative" spatial index that is translated to
	// particular spatial order and degree
	// this function must return the value from its pre-computed buffer
	abstract public double getSpatialBasisFunctionValue(int k_index, int iLat, int iLon);

	public double getSpaceBasisFunctionValue(int k, Double latitude_deg, Double longitude_deg) {
		int iLat = earthGrid.getGridLatIndex(latitude_deg);
		int iLon = earthGrid.getGridLonIndex(longitude_deg);
		return getSpatialBasisFunctionValue(k, iLat, iLon);
	}

	public void setEarthGrid(EarthGrid earthGrid) {
		this.earthGrid = earthGrid;
		setupSpatialBasisFunctions();
	}

	public void setTimeGrid(TimeGrid timeGrid) {
		this.timeGrid = timeGrid;
		setupTemporalBasisFunctions();
	}

}
