/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.expansions;

import java.util.Calendar;

import edu.uml.giro.gambit.core.GambitConstants;
import edu.uml.lgdc.geospace.IGRF;
import edu.uml.lgdc.geospace.TimeGrid;
import edu.uml.lgdc.project.Console;
import edu.uml.lgdc.time.TimeScale;

/**
 * Diurnal/spherical expansion API for empirical models
 * 
 * Used by CCIR/URSI model of foF2 in IRI
 *
 * Uses diurnal order
 * 
 * @author Ivan
 *
 */

public class ExpansionBasis_JonesGallet extends ExpansionBasis2D {
	static public final int CCIR_DIURNAL_ORDER = 6;
	static public final int SPATIAL_ORDER = 8; // longitudinal spherical harmonics (classic), m
	static public final int SPATIAL_DEGREE = 11; // latitudinal functions (more complicated than just harmonics), n
	static public final int USED_DEGREES[] = { 11, 11, 8, 4, 1, 0, 0, 0, 0 }; // n[m]
	// how many coefficients are used for each degree
	// n[m]
	// order m=0 -- special case, no order terms, only degree terms n=0..11 (12)
	// 1, sinM, sin2M, sin3M...sin11M.
	// other orders -- (n[m]+1)*2. Multiply x2 because cos and sin of longitude
	// T
	// cosLcosT, cosLsinT, sinMcosLsinT, sinMcosLsinT, ...
	static public final int NUMBER_FUNCTIONS_PER_ORDER[] = calculateUsedFunctionsPerOrder();
	static public final int COMPACT_SPATIAL_LENGTH = calculateCompactSpatialLength(NUMBER_FUNCTIONS_PER_ORDER); // 76 total

	protected double sinT[][];
	protected double cosT[][];

	// Longitude classic base functions (order)
	protected double sinLon[][];
	protected double cosLon[][];
	// Pole contribution faders to stabilize expansion (order, power of cosLat )
	protected double cosLat[][];

	protected final IGRF igrf = new IGRF();

	// Modip can be useful
	protected double modip[][];

	// Sine of modip is used in the basis
	protected double sinModip[][][];

	public ExpansionBasis_JonesGallet() {
		super();
		setExpansionBasisName("JonesGallet");
	}

	public ExpansionBasis_JonesGallet( TimeGrid timeScale ) {
		super();
		setExpansionBasisName("JonesGallet");
		setTimeGrid( timeScale );
	}

	@Override
	public boolean setupTemporalBasisFunctions() {
		sinT = new double[temporalOrder][timeGrid.getNumDiurnalSteps()];
		cosT = new double[temporalOrder][timeGrid.getNumDiurnalSteps()];
		for (int i = 0; i < timeGrid.getNumDiurnalSteps(); i++) {
			double dayPhase_deg = (360 * ((double) i / timeGrid.getNumDiurnalSteps())) - 180; 
			double dayPhase_rad = Math.toRadians(dayPhase_deg);
			sinT[0][i] = Math.sin(dayPhase_rad);
			cosT[0][i] = Math.cos(dayPhase_rad);
			for (int j = 1; j < temporalOrder; j++) {
				sinT[j][i] = cosT[0][i] * sinT[j - 1][i] + sinT[0][i] * cosT[j - 1][i];
				cosT[j][i] = cosT[0][i] * cosT[j - 1][i] - sinT[0][i] * sinT[j - 1][i];
			}
		}

		return true;
	}

	/**
	 * Returns base function value for diurnal expansion
	 * 
	 * @param index
	 *            -- index of diurnal expansion functions, goes from 0 to
	 *            timeOrder*2
	 * @param iTime
	 *            -- index of the diurnal phase, from 0 to numDiurnalNodes-1
	 * @return
	 */

	public double getTemporalBasisFunctionValue(int index, int iTime) {
		if (index == 0)
			return 1.;
		// CosT and sinT are computed for day phase varying from -180 to +180 degrees (for UT hour going from 0 to 24)
		// Here the starting UT of the ring buffer is no longer zero UT, it can be anything
		// We need to compute offset to the correct index of the array of pre-computed cosT and sinT 
		// sine first, cosine second
		
		TimeScale startUT = timeGrid.getUTs()[0];
		double minutesAtStart = startUT.getTimeSinceMidnightIn(Calendar.MINUTE);
		int iDayPhaseOffset = (int)Math.round( minutesAtStart/timeGrid.getTimeStep_min() );
		int i = (iTime+iDayPhaseOffset)%timeGrid.getNumDiurnalSteps();
		double value = (index % 2 == 0) ? cosT[(index - 1) / 2][i] : sinT[(index - 1) / 2][i];
		return value;
	}

	@Override
	public boolean setSpatialOrder(int spatialOrder) {
		this.spatialOrder = spatialOrder;
		if (spatialOrder != SPATIAL_ORDER) {
			System.err.println("ERROR: requested Jones-Gallet spatial order " + spatialOrder + " is not supported");
			return false;
		}
		return true;
	}

	@Override
	public boolean setSpatialDegree(int spatialDegree) {
		this.spatialDegree = spatialDegree;
		if (spatialDegree != SPATIAL_DEGREE) {
			System.err.println("ERROR: requested Jones-Gallet spatial degree " + spatialDegree + " is not supported");
			return false;
		}
		return true;
	}

	@Override
	public boolean setupSpatialBasisFunctions() {
		if (earthGrid == null) {
			System.err.println("ERROR: Jones-Gallet expansion basis cannot initialize without defined grid");
			return false;
		}
		sinLon = new double[SPATIAL_ORDER][earthGrid.getNumLongitudeNodes()];
		cosLon = new double[SPATIAL_ORDER][earthGrid.getNumLongitudeNodes()];
		cosLat = new double[SPATIAL_ORDER][earthGrid.getNumLatitudeNodes()];

		int i, j;

		// 1. Setup harmonic base functions in longitude

		for (i = 0; i < earthGrid.getNumLongitudeNodes(); i++) {
			double longitude_deg = i * earthGrid.getLongitudeStep();
			double longitude_rad = Math.toRadians(longitude_deg);
			sinLon[0][i] = Math.sin(longitude_rad);
			cosLon[0][i] = Math.cos(longitude_rad);
			for (j = 1; j < SPATIAL_ORDER; j++) {
				sinLon[j][i] = cosLon[0][i] * sinLon[j - 1][i] + sinLon[0][i] * cosLon[j - 1][i];
				cosLon[j][i] = cosLon[0][i] * cosLon[j - 1][i] - sinLon[0][i] * sinLon[j - 1][i];
			}
		}

		// 2. Setup power cos functions in latitude

		for (i = 0; i < earthGrid.getNumLatitudeNodes(); i++) {
			double latitude_deg = i * earthGrid.getLatitudeStep() - 90;
			double latitude_rad = Math.toRadians(latitude_deg);
			double cosLatitude = Math.cos(latitude_rad);
			cosLat[0][i] = cosLatitude;
			for (j = 1; j < SPATIAL_ORDER; j++)
				cosLat[j][i] = cosLat[j - 1][i] * cosLatitude;
		}

		// 3. Allocate memory for modip and sin(modip)
		// Calculate when time is known

		modip = new double[earthGrid.getNumLatitudeNodes()][earthGrid.getNumLongitudeNodes()];
		sinModip = new double[SPATIAL_DEGREE][earthGrid.getNumLatitudeNodes()][earthGrid.getNumLongitudeNodes()];
		return true;
	}

	/************************** TIME-DEPENDENT METHODS **********************************************/

	/**
	 * Initialize all time-dependent data structures this is overwritten method
	 * specific to foF2 expansion
	 * 
	 * @param ut
	 */
	public void setNewTime(TimeScale ut) {
		recalculateMagneticField(ut);
		timeGrid.setUTs(ut, GambitConstants.FORECAST_DURATION_MIN);
	}

	/**
	 * Recalculates magnetic field using IGRF
	 * 
	 * @param ut
	 */
	protected void recalculateMagneticField(TimeScale ut) {
		for (int i = 0; i < earthGrid.getNumLatitudeNodes(); i++) {
			double latitude_deg = i * earthGrid.getLatitudeStep() - 90;
			double latitude_rad = Math.toRadians(latitude_deg);
			double cos_latitude = Math.cos(latitude_rad);
			for (int j = 0; j < earthGrid.getNumLongitudeNodes(); j++) {
				double longitude_deg = j * earthGrid.getLongitudeStep();
				igrf.calc(ut, latitude_deg, longitude_deg, GambitConstants.REFERENCE_ALTITUDE_KM);
				double dip_rad = Math.toRadians(igrf.dip);
				double dipdiv = dip_rad / Math.sqrt(dip_rad * dip_rad + cos_latitude);
				dipdiv = Math.min(dipdiv, 1.0);
				dipdiv = Math.max(dipdiv, -1.0);
				sinModip[0][i][j] = dipdiv;
				for (int k = 1; k < SPATIAL_DEGREE; k++)
					sinModip[k][i][j] = sinModip[k - 1][i][j] * dipdiv;
				// modip can be useful in various transformations
				modip[i][j] = Math.asin(dipdiv);
			}
		}
	}

	public double getModip(int iLat, int iLon) {
		return modip[iLat][iLon];
	}

	protected double sinModipFactor(int degree, int iLat, int iLon) {
		return degree == 0 ? 1. : sinModip[degree - 1][iLat][iLon];
	}

	public int getCompactSpatialBasisLength() {
		return COMPACT_SPATIAL_LENGTH;
	}

	static private int[] calculateUsedFunctionsPerOrder() {
		int[] usedFunctions = new int[USED_DEGREES.length];
		usedFunctions[0] = USED_DEGREES[0] + 1; // no x2 for two terms because
												// order is zero (no lon terms)
		for (int i = 1; i <= SPATIAL_ORDER; i++) {
			usedFunctions[i] = (USED_DEGREES[i] + 1) * 2;
		}
		return usedFunctions;
	}

	// calculate total number of spatial expansion coefficients/base functions
	// remember it as compactSphericalLength
	// has to come out to 76

	static private int calculateCompactSpatialLength(int[] usedCoeffs) {
		int compactSpatialLength = 0;
		for (int i = 0; i < usedCoeffs.length; i++)
			compactSpatialLength += usedCoeffs[i];
		return compactSpatialLength;
	}

	/**
	 * 1D spatial coefficients are arranged somewhat unintuitively
	 * See Jones and Gallett for the table of coefficients:
	 * 76 coefficients are split in SPATIAL_ORDER+1 groups.
	 * Number of coefficients in groups is not the same.
	 * Group 0 has (USED_DEGREES[0]+1) coefficients
	 * There are (USED_DEGREES[j]+1)*2 coefficients for groups j=1..SPATIAL_ORDER
	 * USED DEGREES as a function of j is {11, 11, 8, 4, 1, 0, 0, 0, 0}  
	 * Number of coefficients per group is 12, 24, 18, 10, 4, 2, 2, 2, 2
	 * 
	 */
	public int orderDegreeTermToCompact(int sOrder, int sDegree, int sTerm) {
		if (sTerm > 1 && sTerm < 0)
			return 0;
		if (sOrder > SPATIAL_ORDER || sDegree > USED_DEGREES[sOrder])
			return 0;
		// lowest zero order has only one set of USED_DEGREES[0] degrees, not two of them (as all others)
		if( sOrder == 0) 
			return sDegree;
		int sCompact = 0;
		for (int iOrder = 0; iOrder < sOrder; iOrder++)
			sCompact += NUMBER_FUNCTIONS_PER_ORDER[iOrder];
		sCompact += sDegree * 2 + sTerm;
		return sCompact;
	}

	public int[] compactToOrderDegreeTerm(int sCompact) {
		int[] order_degree = new int[3];
		if (sCompact >= COMPACT_SPATIAL_LENGTH)
			return null;
		int order = 0;
		int counter = NUMBER_FUNCTIONS_PER_ORDER[0] - 1;
		while (sCompact > counter && order <= SPATIAL_ORDER) {
			order++;
			counter += NUMBER_FUNCTIONS_PER_ORDER[order];
		}
		int degree = 0;
		if (order == 0)
			degree = sCompact;
		else {
			int remaining = counter - sCompact;
			int position = NUMBER_FUNCTIONS_PER_ORDER[order] - remaining - 1;
			degree = position / 2;
		}
		order_degree[INDEX_ORDER] = order;
		order_degree[INDEX_DEGREE] = degree;
		order_degree[INDEX_TERM] = sCompact % 2 == 0 ? 0 : 1;
		return order_degree;
	}

	/**
	 * Latitude expansion factor -- degree
	 * 
	 * @param degree
	 * @param iLat
	 * @return
	 */
	protected double latitudeFactor(int order, int iLat) {
		return order == 0 ? 1. : cosLat[order - 1][iLat];
	}

	protected double longitudeFactor(int order, int term, int iLon) {
		if (order == 0)
			return 1;
		// cosine first, sine second
		return term == 0 ? cosLon[order - 1][iLon] : sinLon[order - 1][iLon];
	}

	/**
	 * Get value of the geographic base function using its compact index k is
	 * wrapper for
	 * <code>public double getGeoFunction( int order, int term, int degree, int iLat, int iLon )</code>
	 * Converts compact index to order and degree of complete 2D GeoFunctions
	 * 
	 * Refers to synthesis.
	 * 
	 * @param sCompact
	 *            -- compact index that enumerates all spherical base functions
	 *            used
	 * @param iLat
	 * @param iLon
	 * @return function value at location
	 */
	public double getSpatialBasisFunctionValue(int sCompact, int iLat, int iLon) {
		if (sCompact >= COMPACT_SPATIAL_LENGTH)
			return 0;
		int[] order_degree_term = compactToOrderDegreeTerm(sCompact);
		// System.out.println(
		// "k="+k+", order="+order+", degree="+degree+", term="+term);
		return getGeoFunction(order_degree_term[INDEX_ORDER], order_degree_term[INDEX_TERM], order_degree_term[INDEX_DEGREE], iLat, iLon);
	}

	public double getGeoFunction(int order, int term, int degree, int iLat, int iLon) {
		double value = sinModipFactor(degree, iLat, iLon) * latitudeFactor(order, iLat) * longitudeFactor(order, term, iLon);
		return value;
	}
}
