/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.expansions;

import java.util.Calendar;

import edu.uml.lgdc.time.TimeScale;

	
/**
 * Diurnal/spherical expansion API for empirical models
 * 
 * Used by CCIR/URSI model of foF2 in IRI
 *
 * Uses diurnal order
 * 
 * @author Ivan
 *
 */

public class ExpansionBasis_JonesGalletWithTrend extends ExpansionBasis_JonesGallet {
	public ExpansionBasis_JonesGalletWithTrend() {
		super();
		setExpansionBasisName("JonesGallet_LinTrend");
	}

	/**
	 * returns number of coefficients in diurnal expansion dimension
	 * this now include one additional coefficient for the linear trend
	 * 
	 * @return
	 */
	@Override
	public int getTemporalBasisLength() {
		return (temporalOrder+1) * 2;
	}

	/**
	 * Returns base function value for diurnal expansion
	 * 
	 * @param index
	 *            -- index of diurnal expansion functions, goes from 0 to
	 *            timeOrder*2+1
	 * @param iTime
	 *            -- index of the diurnal phase, from 0 to numDiurnalNodes-1
	 * @return
	 */

	@Override 
	public double getTemporalBasisFunctionValue(int index, int iTime) {
        // index goes from 0 to 13
        // index #0 is a0, index #1 is b1, index #2 is a1 (sin), index #3 is b1 (cos), etc.
        // iTime goes from 0 to number of time steps; last step is TOV
		if (index == 0)
			// constant (average)
			return 1.;
		if (index == 1) {
			// slope of the linear term, shifted so that intercept is at mid-point of the analysis window
			// slope per minute
			double iTimeAverage = (timeGrid.getNumDiurnalSteps()-1)/2.;
			return timeGrid.getTimeStep_min()*(iTime - iTimeAverage);
		}
		// Harmonics (pre-computed)
		// CosT and sinT are computed for day phase varying from -180 to +180 degrees (for UT hour going from 0 to 24)
		// Here the starting UT of the ring buffer is no longer zero UT, it can be anything
		// We need to compute offset to the correct index of the array of pre-computed cosT and sinT 
		TimeScale startUT = timeGrid.getUTs()[0];
		double minutesAtStart = startUT.getTimeSinceMidnightIn(Calendar.MINUTE);
		int iDayPhaseOffset = (int)Math.round( minutesAtStart/timeGrid.getTimeStep_min() );
		int i = (iTime+iDayPhaseOffset)%timeGrid.getNumDiurnalSteps();
		double value = (index % 2 == 0) ? sinT[(index - 2) / 2][i] : cosT[(index - 2) / 2][i];
		return value;
	}
	
	
}
