/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.expansions;

import java.time.YearMonth;
import java.util.Calendar;

import edu.uml.giro.gambit.core.GambitConstants;
import edu.uml.lgdc.geospace.IGRF;
import edu.uml.lgdc.project.Console;
import edu.uml.lgdc.time.TimeScale;

/**
 * Spherical cap expansion API for empirical models
 * 
 * This is a spherical cap harmonics (SCH) expansion introduced by Haines [1988] for local modeling
 * as an alternative to the global modeling using conventional spherical harmonics
 * 
 * The SCH algorithm uses fractional Legendre polynomials of non-integral degree n, as introduced by Haines
 * 
 * This approach is used in IRI-2016 to model B0 on the global scale...
 * In order to configure SCH for global coverage, the cap's half-angle THETA is set to 180 degrees, and the cap radius is Earth radius.
 * TODO: How this compares to conventional harmonics? See also ASHA technique by De Franceschi et al., 1994. (INGV)
 *    
 * The Java implementation is a replica of the code operating in IRI, with simplifications
 * Some of the code uses variable names similar to the FORTRAN code (c) 1988. This is not per our naming conventions.  
 * 
 * 
 * Note:
 * To standardize the language of spherical expansion, we use ORDER and DEGREE instead of various Fortran indices 
 * The double sum by k = 0..Kmax and m=0..k (see the definition of the SCH) is described in terms of degree k and order m
 * Kmax is the [maximum) spatial DEGREE of the formulation
 * The Legendre polynomial is them correctly described as Pkm (degree k and order m)
 * The Longitudinal factor is a harmonic expansion of ORDER m  
 * The monthly expansion has order 4, unrelated to k or m
 *  
 * 
 * @author Ivan
 *
 */

public class ExpansionBasis_SphericalCapHarmonicsK6 extends ExpansionBasis2D {
	
	static public final int SPATIAL_DEGREE = 6; // Kmax of the Legendre polynomial
	
	// representing month-to-month variation
	// represented as Fourier series with 0 to 2-pi (TODO: check this)
	static public final int MONTHLY_ORDER = 2; 
	static private final TimeScale JANUARY = new TimeScale("2000.01"); 
	static private final TimeScale DECEMBER = new TimeScale("2000.12"); 
	static public final double TZERO = JANUARY.get(Calendar.MONTH);
	static public final double THINT = (DECEMBER.get(Calendar.MONTH)-JANUARY.get(Calendar.MONTH))/2.;
	// DELT[i] from the code	
	static public double[] DELT = new double[MONTHLY_ORDER*2+1];

	// Longitude classic base functions (order)
	// [0..SPATIAL_ORDER][lon-step]
	protected double sinLon[][];
	protected double cosLon[][];
	
	// Modip angles (for transposition along the dip lines
	// [lat-step][lon-step]
	protected double modip[][];

	// normalizing constants for Legendre polynomials
	// [0..DEGREE]0..ORDER]
	static public final double NORMALIZERS[][] = {
		{1.,0.,0.,0.,0.,0.,0.},
		{1.,1.,0.,0.,0.,0.,0.},
		{1.,1.73205,0.866025,0.,0.,0.,0.},
		{1.,2.44949,1.93649,0.7905691,0.,0.,0.},
		{1.,3.16228,3.35410,2.09165,0.739510,0.,0.},
		{1.,3.87298,5.12348,4.18330,2.21853,0.701561,0.},
		{1.,4.58258,7.24569,7.24569,4.96078,2.32681,0.671693}
	};
	
	// degrees (FN) -- FIXME why do we need this?? 
	// [0..DEGREE][0..ORDER]
	static public final double DEGREES[][] = {
		{0,0,0,0,0,0,0},
		{1,1,0,0,0,0,0},
		{2,2,2,0,0,0,0},
		{3,3,3,3,0,0,0},
		{4,4,4,4,4,0,0},
		{5,5,5,5,5,5,0},
		{6,6,6,6,6,6,6}
	};

	//[compact-index][longitude_steps]
	protected double legendre[][];
	
	private final IGRF igrf = new IGRF();
	private TimeScale tov = null;
	
	public boolean debugging = false;

	public ExpansionBasis_SphericalCapHarmonicsK6() {
		super();
//		test();
	}
	
	@Override
	public boolean setSpatialOrder(int spatialOrder) {
		this.spatialOrder = spatialOrder;
		if (spatialOrder != SPATIAL_DEGREE) {
			System.err.println("ERROR: requested K6 SCH order " + spatialOrder + " is not supported");
			return false;
		}
		return true;
	}

	@Override
	/**
	 * This computes the constants that do not change unless EarthGrid changes
	 */
	public boolean setupSpatialBasisFunctions() {
		if (earthGrid == null) {
			System.err.println("ERROR: Spherical Cap Harmonics basis cannot initialize without defined grid");
			return false;
		}
		int i, j;
		
		// 1. Longitudinal expansion
		
		sinLon = new double[SPATIAL_DEGREE+1][earthGrid.getNumLongitudeNodes()];
		cosLon = new double[SPATIAL_DEGREE+1][earthGrid.getNumLongitudeNodes()];

		for (i = 0; i < earthGrid.getNumLongitudeNodes(); i++) {
			double longitude_deg =  earthGrid.getGridLonDegree(i);
			double longitude_rad = Math.toRadians(longitude_deg);
			cosLon[0][i] = 1;
			sinLon[0][i] = 0;
			cosLon[1][i] = Math.cos(longitude_rad);
			sinLon[1][i] = Math.sin(longitude_rad);
			for (j = 2; j <= SPATIAL_DEGREE; j++) {
				cosLon[j][i] = cosLon[1][i] * cosLon[j - 1][i] - sinLon[1][i] * sinLon[j - 1][i];
				sinLon[j][i] = cosLon[1][i] * sinLon[j - 1][i] + sinLon[1][i] * cosLon[j - 1][i];
			}
		}
		
		// 2. Legendre polynomials
		
		legendre = new double[getCompactSpatialBasisLength()][earthGrid.getNumLatitudeNodes()];
		for (i = 0; i < earthGrid.getNumLatitudeNodes(); i++) {
			double latitude_deg = i * earthGrid.getLatitudeStep() - 90;
			double colat = 90 - latitude_deg;
			for( int order = 0; order <= SPATIAL_DEGREE; order++ )
				for( int degree = 0; degree <= order; degree++ ) {
					legendre[orderDegreeTermToCompact(order, degree, 0)][i] = legfun( order, degree, Math.toRadians(colat) );
					legendre[orderDegreeTermToCompact(order, degree, 1)][i] = legendre[orderDegreeTermToCompact(order, degree, 0)][i];
				}
		}
		
		// rebuild modip table if this is an update to previous earth grid
		if( tov != null && 
				(modip.length != earthGrid.getNumLatitudeNodes() ||
				modip[0].length != earthGrid.getNumLongitudeNodes() )) 
			recalculateMagneticField( tov );
		return true;
	}

	/**
	 * Ported from FORTRAN code in IRI-2016
	 * @param order -- order
	 * @param degree -- degree
	 * @param colat_rad -- colatitude
	 * @return
	 */
	private double legfun( int order, int degree, double colat_rad ) {
		double fnn = DEGREES[order][degree] * (DEGREES[order][degree]+1);
		double c = Math.cos( colat_rad );
		double x = (1-c)/2;
		double s = Math.sin( colat_rad );
		double al = NORMALIZERS[order][degree] * (degree<2? 1 : Math.pow(s, degree-1));
		double pnm = al;
		double a = 0;
		int count = 0;
		while( count++ < 60 ) {
			int jpm = count+degree;
			double b = al * ((jpm-1)-fnn/jpm);
			a = (b*x)/(double)count;
			pnm += a;
			al = a;
			if( Math.abs(a) > 1.e13 || Math.abs(b) > 1e13) {
				System.out.println("ERROR: Legendre polynomial did not converge.");
				return 0;
			}
			if( Math.abs(a) < 1.e-7 && Math.abs(b) < 1e-7)
				break;
		}
		return pnm*(degree>0?s:1);
	}
	
	
	/************************** TIME-DEPENDENT METHODS **********************************************/

	/**
	 * Initialize all time-dependent data structures 
	 * this is overwritten method specific to B0 expansion
	 * 
	 * @param ut
	 */
	public void setNewTime(TimeScale ut) {
		super.setNewTime(ut);
		tov = new TimeScale( ut.getTimeInMillis() );
		setupTemporalBasisFunctions();
		recalculateMagneticField(ut);
	}

	@Override
	protected boolean setupTemporalBasisFunctions() {
		if( tov == null )
			return false;
		// variable names and data organization are kept similar to IRI 2016
		// this is a temporal harmonic expansion over the month as the time variable
		int month = tov.get( Calendar.MONTH );
		YearMonth ym = YearMonth.of( tov.get( Calendar.YEAR), month+1);
		double dMonth = month + (tov.get(Calendar.DAY_OF_MONTH) - 1)/(double)ym.lengthOfMonth();
		
		DELT[0] = 1.;
		double del = (dMonth-TZERO)/THINT; // this goes from 0 to 2-pi

		// DELT array uses 1, sin, cos, sin2, cos2, sin3, cos3 arrangement
		// FIXME this is different from the rest of the code! 
		// Convention: cos goes first, sin goes second, and cos/sin terms are grouped separately

		double sinT = Math.sin( del );
		double cosT = Math.cos( del );
		DELT[1] = sinT;
		DELT[2] = cosT;
		for( int i = 3; i < MONTHLY_ORDER*2+1; i++ ) {
			if( i%2 != 0) 
				DELT[i] = DELT[i-2]*cosT + DELT[i-1]*sinT; // sine harmonic
			else	
				DELT[i] = DELT[i-2]*cosT - DELT[i-3]*sinT; // cosine harmonic
		}
		return true;
	}
	
	/**
	 * Recalculates modip angles using IGRF B-field model 
	 * Knowledge of all modip values on the grid is needed for latitude transposition at UT != 0     
	 * 
	 * @param ut
	 */
	protected void recalculateMagneticField(TimeScale ut) {
		modip = new double[earthGrid.getNumLatitudeNodes()][earthGrid.getNumLongitudeNodes()];
		for (int i = 0; i < earthGrid.getNumLatitudeNodes(); i++) {
			double latitude_deg = i * earthGrid.getLatitudeStep() - 90;
			double latitude_rad = Math.toRadians(latitude_deg);
			double cos_latitude = Math.cos(latitude_rad);
			double root_cos_lat = Math.sqrt(cos_latitude);
			for (int j = 0; j < earthGrid.getNumLongitudeNodes(); j++) {
				double longitude_deg = j * earthGrid.getLongitudeStep();
				igrf.calc(ut, latitude_deg, longitude_deg, GambitConstants.REFERENCE_ALTITUDE_KM);
				double dip_rad = Math.toRadians(igrf.dip);
				modip[i][j] = Math.atan2(dip_rad, root_cos_lat);
			}
		}
	}

	public int findModipIndex( double modip0, int iLon ) {
		// go from south to north until dip matches
		for( int i = 0; i< earthGrid.getNumLatitudeNodes(); i++ ) {
			double diff = modip[i][iLon] - modip0; 
			if( diff >=0 ) {
				return ( i > 0 && (modip0-modip[i-1][iLon]) < diff)? i-1 : i;
			}
		}
		System.err.println("Problems locating matching modip");
		return earthGrid.getNumLatitudeNodes()-1;
	}
	
	public double getModip( double latitude_deg, double longitude_deg ) {
		int iLat = earthGrid.getGridLatIndex(latitude_deg);
		int iLon = earthGrid.getGridLonIndex(longitude_deg);
		return modip[iLat][iLon];
	}
	

	/************************** COMPACT REPRESENTATION OF LOWER TRIANGULAR MATRIX **********************************************/

	/**
	 * The double sum of the expansion goes like this:
	 * 
	 *  sum(degree=0..DEGREE_MAX) sum(order=0..degree)
	 *  
	 *  order is faster index of summation
	 *  
	 *  So, the arrangement of data is a lower triangular matrix
	 *  
	 *  a00
	 *  a10 a11
	 *  a20 a21 a22
	 *  a30 a31 a32 a33
	 *  
	 *  and storage of coefficients, etc. in a linear compacted form is
	 *  
	 *  a00-cos, a00-sin; a10-cos, a10-sin, a11-cos, a11-sin; a20c, a20s, a21c, a21s, a22c, a22s; a30c... etc
	 *  
	 *  Here cos and sin are coefficients of the longitudinal harmonics. 
	 *  
	 *  The compact index can be converted to degree, order, and term (cos or sin) and back using these formulas 
	 *  
	 */
	@Override
	public int getCompactSpatialBasisLength() {
		return (SPATIAL_DEGREE+2)*(SPATIAL_DEGREE+1);
	}
	@Override
	public int orderDegreeTermToCompact(int order, int degree, int term) {
		return ((1+order)*order/2+degree)*2+term;
	}
	@Override
	public int[] compactToOrderDegreeTerm(int sCompact) {
		int k, m, term;
		term = sCompact%2;
		int tria = sCompact/2; 
		k = (int)(-1 + Math.sqrt(1+8*tria))/2; // order
		m = tria - (1+k)*k/2; // degree
		return new int[] {k,m,term};
	}

	
	/************************** ACCESS TO BASIS FUNCTIONS **********************************************/

	/**
	 * Latitude expansion factor -- order
	 * 
	 * @param order
	 * @param iLat
	 * @return
	 */
	protected double longitudeFactor(int order, int term, int iLon) {
		return term == 0 ? cosLon[order][iLon] : sinLon[order][iLon];
	}

	/**
	 * Get value of the geographic base function using its compact index k is
	 * wrapper for
	 * <code>public double getGeoFunction( int order, int term, int degree, int iLat, int iLon )</code>
	 * Converts compact index to order and degree of complete 2D GeoFunctions
	 * 
	 * Refers to synthesis.
	 * 
	 * @param iCompact
	 *            -- compact index that enumerates all spherical base functions
	 *            used
	 * @param iLat
	 * @param iLon
	 * @return function value at location
	 */
	public double getSpatialBasisFunctionValue(int iCompact, int iLat, int iLon) {
		int[] odt = compactToOrderDegreeTerm(iCompact);
	if( debugging ) 
			System.out.println("k="+iCompact+"("+odt[0]+" "+odt[1]+" "+odt[2]+") LongitudeTerm="
		+longitudeFactor(odt[INDEX_DEGREE], odt[INDEX_TERM], iLon)+" , LegendreFactor="+legendre[iCompact][iLat]);
		return longitudeFactor(odt[INDEX_DEGREE], odt[INDEX_TERM], iLon) * legendre[iCompact][iLat];
	}
	
	/**
	 * Spatial degree is not applicable to this basis
	 */
	@Override
	public boolean setSpatialDegree(int spatialDegree) {
		return true;
	}

	@Override
	public double getTemporalBasisFunctionValue(int index, int iTime) {
		// TODO Auto-generated method stub
		return 0;
	}

	
	public void test() {
		int spatialOrder = SPATIAL_DEGREE;
		for( int k = 0; k <= spatialOrder; k++ )
			for( int m = 0; m <= k; m++ )
				System.out.println("k"+k+" m"+m+" "+legfun(k, m, Math.toRadians(90-50)));

		for( int k = 0; k <= spatialOrder; k++ )
			for( int m = 0; m <= k; m++ ) {
				System.out.println("k"+k+" m"+m+" c="+orderDegreeTermToCompact(k,m,0));
			}
		for( int s = 0; s < getCompactSpatialBasisLength(); s++ ) {
			int km[] = compactToOrderDegreeTerm( s );
			System.out.println("s"+s+" k"+km[0]+" m="+km[1]+" t="+km[2]);
		}
		System.out.println("");
		for( int degree = 0; degree <= SPATIAL_DEGREE; degree++ )
			for( int order = 0; order <= degree; order++ ) 
				for( int t = 0; t < 2; t++ ) 
				System.out.println("degree"+degree+" order"+order+" t="+t+" c="+orderDegreeTermToCompact(degree,order,t));

//		double d;
//		for( int c = 0; c < getCompactSpatialBasisLength(); c++ )
//			d = getSpatialBasisFunctionValue(c, 5, 5);
	}
	
}
