package edu.uml.giro.gambit.synthesizers;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
import java.io.IOException;
import java.io.PrintWriter;

import edu.uml.giro.gambit.coefficients.Coefficients;
import edu.uml.giro.gambit.core.GambitConstants;
import edu.uml.giro.gambit.expansions.ExpansionBasis;
import edu.uml.giro.gambit.expansions.ExpansionBasis2D;
import edu.uml.giro.gambit.synthesizers.Synthesizer2D;
import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.geospace.CharChoice;
import edu.uml.lgdc.geospace.EarthGrid;
import edu.uml.lgdc.geospace.PlasmaMath;
import edu.uml.lgdc.geospace.TimeGrid;
import edu.uml.lgdc.time.TimeScale;

public class GeoSynthesizer extends Synthesizer2D {
	
	GeoSynthesizer supportSynthesizer;
	
	
	public GeoSynthesizer(CharChoice cc, ExpansionBasis basis, Coefficients<?> coeffs, GeoSynthesizer supportSynthesizer) {
		super(cc, basis, coeffs);
		this.supportSynthesizer = supportSynthesizer;
	}

	@Override
	public double synthesizeValue(double lat, double lon, int iTime) {
		switch (cc) // this is my cc
		{
		case FOF2:
		case HMF2:
		case B0:
		case B1:
			return super.synthesizeValue(lat, lon, iTime);
		case NMF2:
			return PlasmaMath.freq2density(supportSynthesizer.synthesizeValue(lat, lon, iTime));
		default:
			System.err.println("Unknown derivation request for " + cc.getName() + " from " + supportSynthesizer.cc.getName());
		}
		return 0;
	}

	@Override
	public double[] synthesizeValueSeries(double lat, double lon) {
		double values[];
		switch (cc) // this is my cc
		{
		case FOF2:
		case HMF2:
		case B0:
		case B1:
			return super.synthesizeValueSeries(lat, lon);
		case NMF2:
			values = supportSynthesizer.synthesizeValueSeries(lat, lon);
			for (int i = 0; i < values.length; i++)
				values[i] = PlasmaMath.freq2density(values[i]);
			return values;
		default:
			System.err.println("Unknown derivation request for " + cc.getName() + " from " + supportSynthesizer.cc.getName());
		}
		return null;
	}

	@Override
	public double[][] synthesizeValueMap(int iTime) {
		double values[][];
		switch (cc) // this is my cc
		{
		case FOF2:
		case HMF2:
		case B0:
		case B1:
			return super.synthesizeValueMap(iTime);
		case NMF2:
			values = supportSynthesizer.synthesizeValueMap(iTime);
			for (int i = 0; i < values.length; i++)
				for (int j = 0; j < values[0].length; j++)
					values[i][j] = PlasmaMath.freq2density(values[i][j]);
			return values;
		default:
			System.err.println("Unknown derivation request for " + cc.getName() + " from " + supportSynthesizer.cc.getName());
		}
		return null;
	}

	public void writeToFile_oneLocation(String filename, double lat, double lon) {
		try {
			PrintWriter writer = new PrintWriter(filename);
			double values[] = synthesizeValueSeries(lat, lon);
			TimeGrid timeGrid = ((ExpansionBasis2D) basis).getTimeGrid();
			TimeScale times[] = timeGrid.getUTs();
			writer.write("#Global Ionosphere Radio Observatory\n#\n#GAMBIT " + GambitConstants.VERSION + "\n#Time Series Export of " + cc.getName()
					+ "\n#GEO latitude=" + lat + "N, longitude=" + lon + "E\n#\n");
			for (int iTime = 0; iTime < timeGrid.getNumDiurnalSteps(); iTime++) {
				writer.write(times[iTime].toHumanUT() + " " + FC.DoubleToString(values[iTime], 8, 3) + "\n");
			}
			writer.close();
		} catch (IOException ex) {
			System.out.println("SingleLocationExport: can't write to file " + filename);
		}
	}

	@Override
	public void setEarthGrid(EarthGrid earthGrid) {
		basis.setEarthGrid(earthGrid);
		if( supportSynthesizer != null )
			supportSynthesizer.setEarthGrid(earthGrid);
	}
	
	
}
