/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.synthesizers;

import edu.uml.giro.gambit.coefficients.Coefficients;
import edu.uml.giro.gambit.expansions.ExpansionBasis;
import edu.uml.giro.gambit.expansions.ExpansionBasis2D;
import edu.uml.lgdc.geospace.CharChoice;

import edu.uml.lgdc.geospace.CharChoice;
public class Synthesizer2D extends Synthesizer {
	public Synthesizer2D(CharChoice cc, ExpansionBasis basis, Coefficients<?> coeffs) {
		super(cc, basis, coeffs);
	}

	/**
	 * Synthesized value at particular coordinates and time
	 * 
	 * @param iLat
	 * @param iLon
	 * @param iTime
	 * @return
	 */

	/**
	 *********************** METHODS THAT BUILD COEFFICIENTS[] FROM COEFFICIENTS[][] ************
	 */

	/**
	 * Returns array of diurnal expansion coefficients for given location point
	 * 
	 * @param iLat
	 *            - latitude index to grid point
	 * @param iLon
	 *            - longitude index to grid point
	 * @param coefs
	 *            [compactSpatialBasisLength][diurnalBasisLength] - coefficients
	 *            of the model
	 * @return array of diurnalOrder*2+1 coefficients
	 */
	public double[] getDiurnalCoefficients(int iLat, int iLon) {
		int diurnalBasisLength = ((ExpansionBasis2D) basis).getTemporalBasisLength();
		int compactSpatialBasisLength = ((ExpansionBasis2D) basis).getCompactSpatialBasisLength();
		double[] c = new double[diurnalBasisLength];
		int i, k;
		for (i = 0; i < diurnalBasisLength; i++) {
			c[i] = 0.; // start with zero degree zero order constant
			for (k = 0; k < compactSpatialBasisLength; k++)
				c[i] += ((double[][]) coeffs.get())[k][i] * ((ExpansionBasis2D) basis).getSpatialBasisFunctionValue(k, iLat, iLon);
		}
		return c;
	}

	/**
	 * Returns array of diurnal expansion coefficients for given location point
	 * 
	 * @param latitude
	 *            - input location
	 * @param longitude
	 * @param coefs
	 *            [compactSpatialBasisLength][diurnalBasisLength] - coefficients
	 *            of the model
	 * @return array of diurnalOrder*2+1 coefficients
	 */
	public double[] getDiurnalCoefficients(double latitude_deg, double longitude_deg) {
		int iLat = basis.getEarthGrid().getGridLatIndex(latitude_deg);
		int iLon = basis.getEarthGrid().getGridLonIndex(longitude_deg);
		return getDiurnalCoefficients(iLat, iLon);
	}

	/**
	 * Returns array of spatial expansion coefficients for given time
	 * 
	 * @param iTime
	 *            - index into timeGrid
	 * @return array of compactSphericalLength coefficients
	 */
	public double[] getCompactSpatialCoefficients(int iTime) {
		double[] c = new double[((ExpansionBasis2D) basis).getCompactSpatialBasisLength()];
		double sum;
		int i, k;

		for (k = 0; k < ((ExpansionBasis2D) basis).getCompactSpatialBasisLength(); k++) {
			// sum = coefs[k][0]; // start with zero time degree constant //
			// TODO Use zero degree two times?
			sum = 0;
			for (i = 0; i < ((ExpansionBasis2D) basis).getTemporalBasisLength(); i++) {
				sum += ((double[][]) coeffs.get())[k][i] * ((ExpansionBasis2D) basis).getTemporalBasisFunctionValue(i, iTime);
			}
			c[k] = sum;
		}
		return c;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.uml.giro.gambit.expansions.Synthesizer#synthesizeValue(double, double, int)
	 * 
	 * For [][] synthesizer, this is the main un-optimized method of computing one value
	 */

	@Override
	public double synthesizeValue(double lat, double lon, int iTime) {
		double value = 0;
		double timeHarmonics[] = getDiurnalCoefficients(lat, lon);
		for (int iOrder = 0; iOrder < ((ExpansionBasis2D) basis).getTemporalBasisLength(); iOrder++)
			value += timeHarmonics[iOrder] * ((ExpansionBasis2D) basis).getTemporalBasisFunctionValue(iOrder, iTime);
		return value;
	}

	public double synthesizeTrend(double lat, double lon, int iTime) {
		double value = 0;
		double timeHarmonics[] = getDiurnalCoefficients(lat, lon);
		for (int iOrder = 0; iOrder < 2; iOrder++)
			value += timeHarmonics[iOrder] * ((ExpansionBasis2D) basis).getTemporalBasisFunctionValue(iOrder, iTime);
		return value;
	}

	/**
	 * Calculates time series of values at given location Location is given as 2
	 * indices to the earthGrid (i,j)
	 * 
	 * @param iLat
	 * @param iLon
	 */
	@Override
	public double[] synthesizeValueSeries(int iLat, int iLon) {
		double[] valueSeries = new double[basis.getTimeGrid().getNumDiurnalSteps()];

		// First compute the coefficients from space basis integration (once)
		ExpansionBasis2D basis2d = (ExpansionBasis2D) basis;
		double[] diurnalCoeffs = new double[basis2d.getTemporalBasisLength()];
		for (int iOrder = 0; iOrder < basis2d.getTemporalBasisLength(); iOrder++) {
			diurnalCoeffs[iOrder] = 0;
			for (int k = 0; k < basis2d.getCompactSpatialBasisLength(); k++) {
				diurnalCoeffs[iOrder] += ((double[][]) coeffs.get())[k][iOrder] * basis2d.getSpatialBasisFunctionValue(k, iLat, iLon);
			}
		}
		// Now run integrations for each time node
		for (int iTime = 0; iTime < basis.getTimeGrid().getNumDiurnalSteps(); iTime++) {
			valueSeries[iTime] = 0;
			for (int iOrder = 0; iOrder < basis2d.getTemporalBasisLength(); iOrder++) {
				valueSeries[iTime] += diurnalCoeffs[iOrder] * basis2d.getTemporalBasisFunctionValue(iOrder, iTime);
			}
		}
		return valueSeries;
	}

	public double[] synthesizeValueSeries( double[] diurnalCoeffs ) {
		double[] valueSeries = new double[basis.getTimeGrid().getNumDiurnalSteps()];

		for (int iTime = 0; iTime < basis.getTimeGrid().getNumDiurnalSteps(); iTime++) {
			valueSeries[iTime] = 0;
			for (int iOrder = 0; iOrder < ((ExpansionBasis2D)basis).getTemporalBasisLength(); iOrder++) {
				valueSeries[iTime] += diurnalCoeffs[iOrder] * ((ExpansionBasis2D)basis).getTemporalBasisFunctionValue(iOrder, iTime);
			}
		}
		return valueSeries;
	}

	/**
	 * Synthesizes 2D space grid of values
	 * 
	 * Latitude goes from -90 to +90, longitude goes from 0 to 360.
	 *  
	 * @param iTime
	 * @return
	 */
	@Override
	public double[][] synthesizeValueMap(int iTime) {
		double[][] valueGrid = new double[basis.getEarthGrid().getNumLatitudeNodes()][basis.getEarthGrid().getNumLongitudeNodes()];
		if( coeffs.get() == null )
			return valueGrid;
		// First compute the coefficients from space basis integration (once)
		ExpansionBasis2D basis2d = (ExpansionBasis2D) basis;
		double[] spatialCoeffs = new double[basis2d.getCompactSpatialBasisLength()];

		for (int k = 0; k < basis2d.getCompactSpatialBasisLength(); k++) {
			spatialCoeffs[k] = 0;
			for (int iOrder = 0; iOrder < basis2d.getTemporalBasisLength(); iOrder++) {
				spatialCoeffs[k] += ((double[][]) coeffs.get())[k][iOrder] * basis2d.getTemporalBasisFunctionValue(iOrder, iTime);
			}
		}

		for (int iLat = 0; iLat < basis.getEarthGrid().getNumLatitudeNodes(); iLat++) {
			for (int iLon = 0; iLon < basis.getEarthGrid().getNumLongitudeNodes(); iLon++) {
				valueGrid[iLat][iLon] = 0;
				for (int k = 0; k < basis2d.getCompactSpatialBasisLength(); k++) {
					valueGrid[iLat][iLon] += spatialCoeffs[k] * basis2d.getSpatialBasisFunctionValue(k, iLat, iLon);
				}
			}
		}
		return valueGrid;
	}
}
