/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.giro.gambit.synthesizers;

import java.util.Calendar;

import edu.uml.giro.gambit.coefficients.Coefficients;
import edu.uml.giro.gambit.expansions.ExpansionBasis;
import edu.uml.giro.gambit.expansions.ExpansionBasis_SphericalCapHarmonicsK6;
import edu.uml.lgdc.geospace.CharChoice;
import edu.uml.lgdc.time.TimeScale;

public class SynthesizerSCH extends GeoSynthesizer {
	public SynthesizerSCH(CharChoice cc, ExpansionBasis basis, Coefficients<?> coeffs) {
		super(cc, basis, coeffs, null);
	}

	/**
	 * Synthesized value at particular coordinates and time
	 * 
	 * @param iLat
	 * @param iLon
	 * @param iTime
	 * @return
	 */

	@Override
	public double synthesizeValue(double latitude_deg, double longitude_deg, int iTime) {

		ExpansionBasis_SphericalCapHarmonicsK6 basisSCH = (ExpansionBasis_SphericalCapHarmonicsK6) basis;

		// account for different time than default 00z by rotating
		
		// 1. Rotate Earth by determining adjusted longitude for time != 0
		TimeScale ut = basis.getTimeGrid().getUTs()[iTime];
		double hoursSince00z = ut.getTimeSinceMidnightIn( Calendar.HOUR_OF_DAY );
		double timeAdjustedLongitude = (int)(longitude_deg + 15 * hoursSince00z);
		if( timeAdjustedLongitude > 360 ) 
			timeAdjustedLongitude -= 360;
		int iLon = basisSCH.getEarthGrid().getGridLonIndex( timeAdjustedLongitude );

		// 2. For new longitude, find the latitude at which modip is the same as the original one 
		double referenceModip = basisSCH.getModip( latitude_deg, longitude_deg );
		int iLat = basisSCH.findModipIndex( referenceModip, iLon );

		// compute the values
		double value = 0;
		for (int k = 0; k < basisSCH.getCompactSpatialBasisLength(); k++) {
			double c = ((double[][])coeffs.get())[k/2][k%2];
			int[] odt = basisSCH.compactToOrderDegreeTerm( k );
//			System.out.println("k="+k+"("+odt[0]+" "+odt[1]+" "+odt[2]+") coef = "+ c);
			double basis = basisSCH.getSpatialBasisFunctionValue(k, iLat, iLon); 
			value += c * basis; 
		}
		return value;
	}

	/**
	 * Calculates time series of values at given location Location is given as 2
	 * indices to the earthGrid
	 * 
	 * @param iLat
	 * @param iLon
	 */
	@Override
	public double[] synthesizeValueSeries(int iLat, int iLon) {
		double[] valueSeries = new double[basis.getTimeGrid().getNumDiurnalSteps()];
		double latitude_deg = basis.getEarthGrid().getGridLatDegree(iLat);
		double longitude_deg = basis.getEarthGrid().getGridLonDegree(iLon);
		for (int iTime = 0; iTime < basis.getTimeGrid().getNumDiurnalSteps(); iTime++)
				valueSeries[iTime] = synthesizeValue(latitude_deg, longitude_deg, iTime);
		return valueSeries;
	}

	/**
	 * Synthesizes 2D space grid of values
	 * 
	 * @param iTime
	 * @return
	 */
	@Override
	public double[][] synthesizeValueMap(int iTime) {
		double[][] valueGrid = new double[basis.getEarthGrid().getNumLatitudeNodes()][basis.getEarthGrid().getNumLongitudeNodes()];
		for (int iLat = 0; iLat < basis.getEarthGrid().getNumLatitudeNodes(); iLat++) {
			double latitude_deg = basis.getEarthGrid().getGridLatDegree(iLat);
			for (int iLon = 0; iLon < basis.getEarthGrid().getNumLongitudeNodes(); iLon++) {
				double longitude_deg = basis.getEarthGrid().getGridLonDegree(iLon);
				valueGrid[iLat][iLon] += synthesizeValue( latitude_deg, longitude_deg, iTime );
				
			}
		}
		return valueGrid;
	}
}
