/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.appuserdb.api;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashMap;
import java.util.Map;

import edu.uml.lgdc.appuserdb.constants.Application;
import edu.uml.lgdc.database.ConnectItem;
import edu.uml.lgdc.database.ConnectList;
import edu.uml.lgdc.database.DatabaseContents;
import edu.uml.lgdc.project.Console;

/**
 * @author Alex Kozlov, Ivan Galkin
 * 
 * Provides authentication of registered users or guests 
 * Retrieves connection credentials to LGDC database (only one? TBR) and user role as registered 
 * 
 */
public class DatabaseCredentials {
	
	static public final String USER_ROLE_AS_REGISTERED = "UserRoleAsRegistered";
	static public final String DB_LOGIN = "DbLogin";
	static public final String DB_PASSWORD = "DbPassword";
	
	static public final String URL_USER_DATABASE_SERVLET;
	static private final boolean deployment = true;
	static {
		if (deployment) {
			URL_USER_DATABASE_SERVLET = "https://lgdc.uml.edu/rix/user-log";
		} else {
			URL_USER_DATABASE_SERVLET = "http://localhost:8080/rix/user-log";
		}
	}
	// uncomment this for testing only

	static private UserRole userRoleAsRegistered;
	static private final Map<DatabaseContents, ConnectList> dbCredentials = new HashMap<DatabaseContents, ConnectList>();
	
	static public void read( Application application, Map<String, String> userCredentials  ) throws FileNotFoundException {

		dbCredentials.clear();
		DatabaseCredentials.acquireDatabaseCredentials( application, userCredentials, dbCredentials);
		if( dbCredentials.isEmpty() ) {
			throw new RuntimeException("No access credentials could be obtained for "+application.getName());
		}
	}

	static private void acquireDatabaseCredentials(Application application, Map<String,String> userCreds,Map<DatabaseContents, ConnectList> dbCreds ) {

		boolean guestMode = !userCreds.containsKey(UserCredentials.USER_LOGIN);
		
		if( !guestMode ) {
			
			// 1. Attempt retrieval of database credentials from userdb for this user
			String userdbAddress = URL_USER_DATABASE_SERVLET  
					+ "?app="+application.getName() 
					+ "&n="+userCreds.get(UserCredentials.USER_LOGIN)
					+ "&w="+userCreds.get(UserCredentials.USER_PASSWORD);
				
			retrieveCredentialsFromUserdb( userdbAddress, userCreds, dbCreds );
			
			if( dbCredentials.isEmpty() ) {
				guestMode = true;
			}
		}
		if (dbCredentials.isEmpty() 
					&& userCreds.containsKey(UserCredentials.GUEST_EMAIL) && userCreds.containsKey(UserCredentials.GUEST_AFF)) {
				
			// 2. Attempt retrieval of database credentials from userdb as a guest
			String userdbAddress = URL_USER_DATABASE_SERVLET 
					+ "?app="+application.getName() 
					+ "&e="+userCreds.get(UserCredentials.GUEST_EMAIL)
					+ "&a="+userCreds.get(UserCredentials.GUEST_AFF);
				
			retrieveCredentialsFromUserdb( userdbAddress, userCreds, dbCreds );
		}
	}
	
	private static void retrieveCredentialsFromUserdb(String userdbAddress, 
			Map<String, String> userCreds, Map<DatabaseContents, ConnectList> dbCreds) {
		
		try {
			//String encodedURL = URLEncoder.encode(userdbAddress,java.nio.charset.StandardCharsets.UTF_8.toString());
			String encodedURL = userdbAddress.replaceAll(" ", "+");
			URL link2userdb = new URL( encodedURL );
			URLConnection conn = link2userdb.openConnection();
			BufferedReader br = new BufferedReader(new InputStreamReader(conn.getInputStream()));
			String inputLine;
			boolean firstLineParsed = false;
			while ((inputLine = br.readLine()) != null) {
				if( inputLine.startsWith("#") || inputLine.startsWith("<")) {
					continue;
				}
				String[] tokens = inputLine.split("\\s*=\\s*");
				if (tokens.length != 2) {
					Console.showError("illegal line encountered while retrieving credentials");
					break;
				}
				if (!firstLineParsed) {
					if (!tokens[0].equalsIgnoreCase(USER_ROLE_AS_REGISTERED)) {
						Console.showError("first line is not " + USER_ROLE_AS_REGISTERED + " while retrieving credentials");
						break;
					}
					userRoleAsRegistered = UserRole.get(tokens[1]);
					if (userRoleAsRegistered == null) {
						Console.showError("illegal User Role encountered while retrieving credentials, " + tokens[1]);
						break;
					}
					firstLineParsed = true;
				} else {
					DatabaseContents content  = DatabaseContents.get(tokens[0]);
					if (content == null) {
						Console.showError("illegal line encountered while retrieving credentials");
						break;
					}
					ConnectList connectList = ConnectList.parse(tokens[1]);
					dbCreds.put(content, connectList);
				}
			}
		} catch (MalformedURLException e) {
			Console.showError("Malformed URL to LGDC user authentication service");
			return;
		} catch (IOException e) {
			String errorMessage = e.getMessage();
			String pattern = "response code: ";
			int codePos = errorMessage.indexOf( pattern ); 
			if( codePos > 0) {
				int code = Integer.valueOf(errorMessage.substring(codePos+pattern.length(),codePos+pattern.length()+3).trim());
				Console.showError("LGDC user authentication service reported internal error "+code);
			} else {
				Console.showError("LGDC user authentication service is inaccessible...\n"+errorMessage);
			}
			return;
		}
		
	}

	static public ConnectItem getDatabaseConnectItem( DatabaseContents content ) {
		ConnectList list = dbCredentials.get( content );
		if (list != null) {
			return list.getLastConnectItem();
		} else {
			// all needed contents have to be checked at the user login process
			throw new RuntimeException("Design error: requested content, " + content.getName() + ", is absent");
		}
	}

	public static void put( Map<DatabaseContents, ConnectList> dbCredentials2 ) {
		for(Map.Entry<DatabaseContents, ConnectList> entry : dbCredentials2.entrySet()) {
			dbCredentials.put(entry.getKey(), entry.getValue());
		}
	}
	public static void set(Map<DatabaseContents, ConnectList> dbCredentials) {
		DatabaseCredentials.dbCredentials.clear();
		put(dbCredentials);
	}
	public static UserRole getUserRoleAsRegistered() {
		return userRoleAsRegistered;
	}
	public static void setUserRoleAsRegistered( UserRole userRole) {
		userRoleAsRegistered = userRole;
	}

}
