/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.appuserdb.api;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;

import edu.uml.lgdc.appuserdb.constants.Application;
import edu.uml.lgdc.project.Console;

/**
 * @author Alex Kozlov, Ivan Galkin
 * 
 * Provides authentication of registered users or guests 
 * Retrieves connection credentials to LGDC database (only one? TBR) and user role as registered 
 * 
 */
public class UserCredentials {
	
	static public final String USER_LOGIN = "UserLogin";
	static public final String USER_PASSWORD = "UserPassword";
	static public final String GUEST_EMAIL = "GuestEmail";
	static public final String GUEST_AFF = "GuestAffiliation";

	static private final Map<String, String> userCredentials = new HashMap<String, String>();

	static public void read( Application application, File guestFile, File userFile ) throws FileNotFoundException {
		FileInputStream guestFis = null;
		FileInputStream userFis = null;
		
		if( guestFile != null && guestFile.isFile() ) {
			guestFis = new FileInputStream(guestFile);
		}
		if( userFile.isFile() ) {
			userFis = new FileInputStream(userFile);
		}

		if( userFis == null && guestFis == null ) {
			if (guestFile != null) {
				throw new RuntimeException("No user/guest credentials are provided for "+application.getName());
			} else {
				throw new RuntimeException("No user credentials are provided for "+application.getName());
			}
		}
		UserCredentials.read( application, guestFis, userFis );
	}
	
	static public void read( Application application, InputStream isGuest, InputStream isUser ) {

		userCredentials.clear();
		
		// Try getting User login and password settings
		try {
			read(isUser, userCredentials, USER_LOGIN, USER_PASSWORD);
		} catch (IOException e) {
			Console.showMsg("User credentials are not available");
		}
		// Try getting email and organization settings for Guest login
		try {
			read(isGuest, userCredentials, GUEST_EMAIL, GUEST_AFF);
		} catch (IOException e) {
			Console.showMsg("Guest credentials are not available");
		}

		if( userCredentials.isEmpty() ) {
			throw new RuntimeException("No local user/guest credentials could be found for "+application.getName());
		}

	}

	static public Map<String, String> getCredentials() {
		return userCredentials;
	}

	static public String getUserLogin() {
		return userCredentials.get(USER_LOGIN);
	}
	static public String getUserPassword() {
		return userCredentials.get(USER_PASSWORD);
	}
	static public String getGuestEmail() {
		return userCredentials.get(GUEST_EMAIL);
	}
	static public String getGuestAffiliation() {
		return userCredentials.get(GUEST_AFF);
	}
	static public void setUserCredentials(String username, String password) {
		userCredentials.put(USER_LOGIN, username );
		userCredentials.put(USER_PASSWORD, password );
	}

	static private void read(InputStream is, Map<String,String> info, String firstLineKey, String secondLineKey) throws IOException {
		if( is == null ) {
			return;
		}
		String line;
		boolean firstItemDone = false;
		try (BufferedReader reader = new BufferedReader(new InputStreamReader(is))) {
			while ((line = reader.readLine()) != null) {
				line = line.trim();
				if( line.startsWith("#") || line.isEmpty() ) {
					continue;
				}
				if( firstItemDone ) {
					info.put(secondLineKey, line);
					break;
				}
				info.put(firstLineKey, line);
				firstItemDone = true;
			}
		}
	}

}
