/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.database;

import java.text.ParseException;
import java.util.Objects;

public class ConnectItem {
	static public final String NONE_STR = "NONE";
	
	private final String title;
	private final String protocol;
	private final String host;
	private final String filename;
	private final String username;
	private final String pass;
	private final String role;
	private final String driverClassName;
	
	public ConnectItem(String title, String protocol, String host, String filename, String username, String pass, String role, 
					String driverClassName) {
		this.title = Objects.requireNonNull(title);
		this.protocol = Objects.requireNonNull(protocol).trim();
		this.host = Objects.requireNonNull(host).trim();
		this.filename = Objects.requireNonNull(filename).trim();
		this.username = Objects.requireNonNull(username).trim();
		this.pass = Objects.requireNonNull(pass);
		this.role = role;
		this.driverClassName = Objects.requireNonNull(driverClassName).trim();
	}
	public ConnectItem(String title, String url, String username, String pass, String role,	String driverClassName) 
					throws ParseException {
		this.title = Objects.requireNonNull(title);
		String[] parts = parseURL(Objects.requireNonNull(url));
		this.protocol = parts[0];
		this.host = parts[1];
		this.filename = parts[2];
		this.username = Objects.requireNonNull(username).trim();
		this.pass = Objects.requireNonNull(pass);
		this.role = role;
		this.driverClassName = Objects.requireNonNull(driverClassName).trim();
	}
	
	static public ConnectItem createConnectItem(String title, String url, String username, String pass, String role, String driverClassName) {
		try {
			return new ConnectItem(title, url, username, pass, role, driverClassName);
		} catch (ParseException ex) {
			// TODO Auto-generated catch block
			throw new RuntimeException(ex);
		}
	}
	public String getTitle() {
		return title;
	}
	public String getProtocol() {
		return protocol;
	}
	public String getHost() {
		return host;
	}
	public String getFilename() {
		return filename;
	}
	public String getUsername() {
		return username;
	}
	public String getPass() {
		return pass;
	}
	public String getRole() {
		return role;
	}
	public String getDriverClassName() {
		return driverClassName;
	}
	public String getURL() {
		return protocol + "://" + host + "/" + filename;
	}
	static private String[] parseURL(String url) throws ParseException {
		url = url.trim();
		String[] parts = new String[3];
		int pos = url.indexOf("://");
		if (pos < 0) {
			throw new ParseException("protocol not found in url, " + url, 0);
		}
		parts[0] = url.substring(0, pos);
		if (parts[0].length() == 0 || parts[0].contains(" ")) {
			throw new ParseException("illegal protocol in url, " + url, 0);
		}
		url = url.substring(pos + 3);
		pos = url.indexOf('/');
		if (pos < 0) {
			throw new ParseException("host not found in url, " + url, 0);
		}
		parts[1] = url.substring(0, pos);
		if (parts[1].length() == 0 || parts[1].contains(" ")) {
			throw new ParseException("illegal host in url, " + url, 0);
		}
		parts[2] = url.substring(pos + 1);
		if (parts[2].length() == 0 || parts[2].contains(" ")) {
			throw new ParseException("illegal filename in url, " + url, 0);
		}
		return parts;
	}
	public boolean isEmpty() {
		return host.isEmpty();
	}
	static public ConnectItem createEmpty() {
		return new ConnectItem(NONE_STR, "", "", "", "", "", "", "");
	}
	static public ConnectItem parse(String connectItemAsStr) {
		assert connectItemAsStr != null;
		connectItemAsStr = connectItemAsStr.trim();
		final String[] parts = connectItemAsStr.split(",");
		if (parts.length < 3) {
			return createEmpty();
		}
		final String title = parts[0].trim();
		final String protocol;
		final String host;
		final String filename;
		final String user;
		final String pass;
		final String role;
		final String driver;
		if (parts.length <= 6) {
			String[] urlParts;
			try {
				urlParts = parseURL(parts[1]);
			} catch (ParseException ex) {
				protocol = parts[1].trim();
				host = parts[2].trim();
				filename = parts[3].trim();
				user = parts[4].trim();
				pass = parts.length < 6 ? "" : parts[5].trim();
				role = "";
				driver = "";
				return new ConnectItem(title, protocol, host, filename, user, pass, role, driver);
			}
			protocol = urlParts[0];
			host = urlParts[1];
			filename = urlParts[2];
			user = parts[2].trim();
			pass = parts.length < 4 ? "" : parts[3].trim();
			role = parts.length < 5 ? "" : parts[4].trim();
			driver = parts.length < 6 ? "" : parts[5].trim();
		} else {
			protocol = parts[1].trim();
			host = parts[2].trim();
			filename = parts[3].trim();
			user = parts[4].trim();
			pass = parts.length < 6 ? "" : parts[5].trim();
			role = parts.length < 7 ? "" : parts[6].trim();
			driver = parts.length < 8 ? "" : parts[7].trim();
		}
		return new ConnectItem(title, protocol, host, filename, user, pass, role, driver);
	}
	public StringBuilder append(StringBuilder sb) {
		return sb.append(title).append(',').append(protocol).append(',').append(host).append(',').append(filename).append(',')
						.append(username).append(',').append(pass).append(',').append(role).append(',').append(driverClassName);
	}
	@Override public String toString() {
		return append(new StringBuilder(128)).toString();
	}
}
