	package edu.uml.lgdc.database;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Properties;

import edu.uml.lgdc.project.Console;


/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
public class DBConnect implements AutoCloseable {
	static private final String URL_SQL_ROLE_NAME = "sql_role_name";
	static public final String JAYBIRD_DRIVER = "org.firebirdsql.jdbc.FBDriver";
    static public final String ROLE_NAME_COMMON = "COMMON";

	/**
	 * Next objects are created during construction and do not altered later. So they do not diminish thread safety degree of container object.
	 */
	protected String driverName;
	protected String databaseURL;
	protected String user;
	protected String password;
	protected String roleName;
	protected DatabaseMetaData dbMetaData;
	protected Map<String, String> tables;
	protected Map<String,StoredProcedure> procedures;

	protected Connection connect;

	/**
	 * Creates instance that is not connected. So, to connect, you need to use method <code>open</code>
	 */
	public DBConnect() {	}
	/**
	 * Constructor
	 */
	public DBConnect(String user, String password, String driverName, String databaseURL) throws SQLException {
		this(user, password, driverName, databaseURL, null);
	}
	/**
	 * Constructor
	 */
	public DBConnect(String user, String password, String driverName, String databaseURL, String roleName) throws SQLException {
		open(driverName, databaseURL, user, password, roleName);
	}
	/**
	 * Returns <b>true</b> if this instance is connected to database, otherwise returns <b>false</b>.
	 */
	public boolean isConnected() {
		boolean connected = false;
		if (connect != null) {
			try {
				connected = !connect.isClosed();
			} catch (SQLException e) {
				System.out.println(e.toString());
			}
		}
		return connected;
	}
	/**
	 * Closes connection through member variable <code>connect</code> and then nullifies this variable.
	 */
	@Override public void close() {
		if (connect != null) {
			try {
				if (!connect.isClosed()) {
					connect.rollback();
					connect.close();
				}
			} catch (SQLException e) {
				Console.printThreadStackTrace(e);
			}
			connect = null;
			dbMetaData = null;
			driverName = null;
			databaseURL = null;
		}
	}
	/**
	 * Sends commit command to database
	 */
	public void commit() throws SQLException {
		if (connect != null) {
			connect.commit();
		}
	}
	/**
	 * Sends rollback command to database
	 */
	public void rollback() throws SQLException {
		if (connect != null) {
			connect.rollback();
		}
	}
	public Statement createStatement() throws SQLException {
		if (connect != null) {
			return connect.createStatement();
		}
		return null;
	}

	public PreparedStatement prepareStatement(String sqlString) throws SQLException {
		if (connect != null) {
			return connect.prepareStatement(sqlString);
		}
		return null;
	}

	/**
	 * Runs SQL-script from given file, <code>filename</code>.
	 */
	public void runSQLScript(String filename) throws SQLException, IOException {
		Statement st = null;
		BufferedReader in = null;
		String sep = ";";
		int sepLen = sep.length();
		String line;
		boolean endOfSQLStr = false;
		boolean eof = false;
		try {
			st = connect.createStatement();
			in = new BufferedReader(new FileReader(filename), 4000);
			while (in.ready()) {
				line = in.readLine();
				if (line == null) {
					eof = true;
					endOfSQLStr = true;
				} else {
					line = line.trim();
					if (line.substring(line.length() - sepLen, line.length()).equals(sep)) {
						line = line.substring(0, line.length() - sepLen);
						line = line.trim();
						endOfSQLStr = true;
					}
				}
				if (endOfSQLStr) {
					if (!line.equals("")) {
						if (line.toUpperCase().startsWith("SET ") && line.substring(4).toUpperCase().startsWith("TERM ")) {
							sep = line.substring(4).trim();
							sep = sep.substring(5).trim();
							sepLen = sep.length();
						} else {
							if (line.toUpperCase().startsWith("SELECT ")) {
								st.executeQuery(line);
							} else if (line.toUpperCase().startsWith("UPDATE ") || line.toUpperCase().startsWith("INSERT ")
								|| line.toUpperCase().startsWith("DELETE ")) {
								st.executeUpdate(line);
							} else if (line.trim().toUpperCase().equals("COMMIT")) {
								connect.commit();
							} else if (line.trim().toUpperCase().equals("ROLLBACK")) {
								connect.rollback();
							} else {
								st.execute(line);
							}
						}
					}
					endOfSQLStr = false;
				}
				if (eof) {
					break;
				}
			}
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (IOException ex) {
					Console.printThreadStackTrace(ex);
				}
			}
			close(st);
		}
	}
	/**
	 * Returns <b>true</b> if given <code>tableName</code> exists, otherwise returns <b>false</b>.<br><br>
	 *
	 * Note:
	 * <ul>
	 *   Comparison is case sensitive
	 * </ul>
	 */
	public boolean isTableExist(String tableName) {
		initTableNames();
		return (tables == null) ? false : tables.get(tableName.toUpperCase()) != null;
	}
	/**
	 * Returns <b>true</b> if given <code>procedureName</code> exists, otherwise returns <b>false</b>.<br>
	 * Note:
	 * <ul>
	 *   Comparison is case insensitive
	 * </ul>
	 */
	public boolean isProcedureExist(String procedureName) {
		initProcedureNames();
		return (procedures == null) ? false : (procedures.get(procedureName.toUpperCase()) != null);
	}
	public StoredProcedure getProcedure(String procedureName) {
		initProcedureNames();
		return (procedures == null) ? null : procedures.get(procedureName.toUpperCase());
	}
	/**
	 * Returns name of unique index for given <code>table</code> first column of which is equal to given <code>column</code>.
	 * If no such an unique index exists than returns <b>null</b>.
	 */
	public String getUniIndName(String table, String column) {
		return getUniIndName(table, new String[] {column});
	}
	/**
	 * Returns name of unique index for given <code>table</code>, first columns of which are equaled to those in array <code>columns</code>.<br>
	 * If no such an unique index exists than returns <b>null</b>.
	 */
	public String getUniIndName(String table, String[] columns) {
		return getIndName(table, columns, true);
	}
	/**
	 * Returns name of index for given <code>table</code> first column of which is equal to <code>column</code>.<br>
	 * If no such an index exists than returns <b>null</b>.
	 */
	public String getIndName(String table, String column) {
		return getIndName(table, new String[] {column});
	}
	/**
	 * Returns name of index for given <codetable</code>, first columns of which are equal to those in array <code>columns</code>.<br>
	 * If no such an index exists than returns <b>null</b>.
	 */
	public String getIndName(String table, String[] columns) {
		return getIndName(table, columns, false);
	}
	/**
	 * Establishes connection to given <code>databaseURL</code>, using given <code>user</code>, <code>password</code> and <code>roleName</code>.<br><br>
	 *
	 * Note:
	 * <ul>
	 *   SQL Role name can be given either by parameter <code>roleName</code> (which will be saved in {@link Properties} instance under
	 *   name <i>roleName</i>), or as embedded parameter of URL (sql_role_name=value)<br>
	 *   Firebird SQL driver, {@link org.firebirdsql.jdbc.FBDriver}, of version 1.5 did not recognize parameter <i>roleName</i> from
	 *   {@link Properties} instance. Starting from version 2.0, it recognizes parameter <i>roleName</i> from {@link Properties} instance,
	 *   but does not accept that role name is given in both, {@link Properties} instance and as as embedded parameter of URL. So, to make
	 *   applications to be working for all versions of Firebird SQL drivers, the following check is implemented:
	 *   <ul>
	 *     if role name given in both, by parameter <code>roleName</code> and  as embedded parameter of URL, then, if names are different
	 *     then error message is issued and connection is aborted, otherwise <code>roleName</code> will NOT be saved in {@link Properties} instance.
	 *   </ul>
	 * </ul>
	 */
	protected void open(String driverName, String databaseURL, String user, String password, String roleName) throws SQLException {
		Objects.requireNonNull(user);
		if (user.isEmpty()) {
			throw new IllegalArgumentException("user is empty string");
		} else {
			this.user = user;
		}
		Objects.requireNonNull(databaseURL);
		if (databaseURL.isEmpty()) {
			throw new IllegalArgumentException("databaseURL is empty string");
		} else {
			this.databaseURL = databaseURL;
		}
		Properties properties = new Properties();
		properties.put("user", user);
		if (password != null && !password.isEmpty()) {
			properties.put("password", password);
		} else {
			password = null;
		}
		if (roleName != null && !roleName.isEmpty()) {
			this.roleName = roleName;
			String roleNameFromURL = getSQLRole(databaseURL);
			if (roleNameFromURL == null) {
				properties.put("roleName", roleName);
			} else if (!roleName.equals(roleNameFromURL)) {
				Console.showError("DBConnect.open(): SQL role name supplied in both, as a parameter roleName\n"
					+ "and as sql_role_name in databaseURL, and they have different values,  " + roleName + " and " + roleNameFromURL);
				return;
			}
		} else {
			String roleNameFromURL = getSQLRole(databaseURL);
			if (roleNameFromURL == null) {
				this.roleName = ""; 
			} else {
				this.roleName = roleNameFromURL; 
			}
		}
		if (driverName != null) {
			open(driverName, databaseURL, properties);
		} else {
			open(databaseURL, properties);
		}
	}
	/**
	 * Establishes connection to given <code>databaseURL</code> taking user, password, and roleName from given <code>properties</code> and
	 * using driver with given class name, <code>driverName</code>.<br>
	 */
	private void open(String driverName, String databaseURL, Properties properties) throws SQLException {
		Driver driver = loadDriver(driverName);
		if (driver == null) {
			throw new IllegalArgumentException("Can not find driver " + driverName);
		} else {
			open(driver, databaseURL, properties);
		}
	}
	/**
	 * Establishes connection to given <code>databaseURL</code> taking user and password from given <code>properties</code>.<br>
	 *
	 * Note:
	 * <ul>
	 *   This method assumes that appropriate driver was already loaded
	 * </ul>
	 */
	private void open(String databaseURL, Properties properties) throws SQLException {
		Driver driver = DriverManager.getDriver(databaseURL);
		if (driver == null) {
			throw new IllegalArgumentException("Can't associate string '" + databaseURL + "' with SQL driver");
		} else {
			open(driver, databaseURL, properties);
		}
	}
	/**
	 * Establishes connection to given <code>databaseURL</code>, taking user and password from given <code>properties</code>, and using
	 * given SQL <code>driver</code>.
	 */
	private void open(Driver driver, String databaseURL, Properties properties) throws SQLException {
		Connection conn = driver.connect(databaseURL, properties);
		if (conn != null) {
			connect = conn;
			connect.setAutoCommit(false);
			this.driverName = driver.getClass().getName();
			readMetaData();
		} else {
			throw new IllegalArgumentException("wrong driver, " + driver + ", to connect to database " + databaseURL);
		}
	}
	/**
	 * Extracts SQL Role name from Database URL and returns it. If no SQL Role name is embedded, then returns <b>null</b>.
	 */
	static public String getSQLRole(String databaseURL) {
		int pos = -1;
		String mask = "?" + URL_SQL_ROLE_NAME + "=";
		pos = databaseURL.lastIndexOf(mask);
		if (pos < 0) {
			mask = "&" + URL_SQL_ROLE_NAME + "=";
			pos = databaseURL.lastIndexOf(mask);
		}
		if (pos >= 0) {
			int startOfValue = pos + mask.length();
			int endOfValue = databaseURL.indexOf('&', startOfValue + 1);
			if (endOfValue < 0) {
				return databaseURL.substring(startOfValue);
			} else {
				return databaseURL.substring(startOfValue, endOfValue);
			}
		} else {
			return null;
		}
	}
	static protected Connection getConnection(String driverName, String databaseURL, String user, String password) {
		Connection tmpConnect = null;
		Properties properties = null;
		Driver driver = loadDriver(driverName);
		if (driver != null) {
			try {
				properties = new Properties();
				properties.put("user", user);
				properties.put("password", password);
				int oldLoginTimeout = DriverManager.getLoginTimeout();
				DriverManager.setLoginTimeout(30);
				tmpConnect = DriverManager.getConnection(databaseURL, properties);
				DriverManager.setLoginTimeout(oldLoginTimeout);
			} catch (SQLException ex) {
				System.out.println("Database is unreachable or shut down.");
				System.out.println(ex.toString());
			}
		}
		return tmpConnect;
	}
	/**
	 * Loads driver, given by its class name, <code>driverName</code>.<br><br>
	 *
	 * Note:
	 * <ul>
	 *   Each SQL-driver should have static initialization piece of code that contains creation of instance of driver and registrating it
	 *   within DriverManager through the call to static method registerDriver.
	 * </ul>
	 *
	 * Returns driver corresponding to given class name, <code>driverName</code>, if load was successful or driver was already loaded,
	 * otherwise returns <b>null</b>.
	 */
	synchronized static private Driver loadDriver(String driverName) {
		Driver driver = getDriver(driverName);
		if (driver != null) {
			return driver;
		}
		try {
			Class.forName(driverName);
			driver = getDriver(driverName);
		} catch (java.lang.ClassNotFoundException e) {
			// A call to Class.forName() forces us to consider this exception
			System.out.println("Class " + driverName + " not found in class path");
			System.out.println(e.toString());
		}
		return driver;
	}
	/**
	 * Returns driver from DriverManager by its name.<br>
	 * Returns <b>null</b> if driver was not loaded.
	 */
	synchronized static private Driver getDriver(String driverName) {
		Enumeration<Driver> enumer = DriverManager.getDrivers();
		Driver driver;
		while (enumer.hasMoreElements()) {
			driver = enumer.nextElement();
			if (driver.getClass().getName().equals(driverName)) {
				return driver;
			}
		}
		return null;
	}
	/**
	 * Reads Metadata
	 */
	private void readMetaData() {
		if (connect != null) {
			try {
				dbMetaData = connect.getMetaData();
			} catch (SQLException e) {
				Console.printThreadStackTrace(e);
			}
		}
	}
	/**
	 * Reads names of non-system tables into array <code>tables</code>, if it did not yet read.
	 */
	private void initTableNames() {
		if (tables == null) {
			readTableNames();
		}
	}
	/**
	 * Reads names of procedures into the list <code>procedures</code>, if it did not yet read.
	 */
	private void initProcedureNames() {
		if (procedures == null) {
			readProcedureNames();
		}
	}
	/**
	 * Reads names of non-system tables in internal array <code>tables</code>.
	 */
	private void readTableNames() {
		tables = new HashMap<>();
		try (ResultSet rs = dbMetaData.getTables(null, null, null, new String[] {"TABLE"})) {
			while (rs.next()) {
				String name = rs.getString("TABLE_NAME");
				tables.put(name, name);
			}
		} catch (SQLException e) {
			metaDataError(e);
		}
	}
	/**
	 * Reads names of procedures into array <code>procedures</code>.
	 */
	private void readProcedureNames() {
		try (ResultSet rs = dbMetaData.getProcedureColumns(null, null, null, null)) {
			procedures = new HashMap<>();
			while (rs.next()) {
				String procName = rs.getString("PROCEDURE_NAME");
				String colName = rs.getString("COLUMN_NAME");
				int type = rs.getShort("COLUMN_TYPE");
				int sqlType = rs.getInt("DATA_TYPE");
				SPColumn column = new SPColumn(colName, SPColumnType.get(type), sqlType);
				StoredProcedure sp = procedures.get(procName);
				if (sp == null) {
					List<SPColumn> list = new ArrayList<>();
					list.add(column);
					procedures.put(procName.toUpperCase(), new StoredProcedure(procName, list));
				} else {
					sp.addColumn(column);
				}
			}
			/*for (String name : procedures.keySet())
				System.out.println(name);*/
		} catch (SQLException e) {
			metaDataError(e);
		}
	}
	private void metaDataError(SQLException e) {
		System.out.println("Error during reading metadata");
		e.printStackTrace();
	}
	/**
	 * Returns name of index for given <code>table</code>, first columns of which are equal to those in array <code>columns</code>.<br>
	 * if <code>unique</code> is <b>true</b>, than searching is done among the unique indexes, otherwise - regardless of index uniqueness.<br>
	 * If no such an index exists than returns <b>null</b>.
	 */
	private String getIndName(String table, String[] columns, boolean unique) {
		String indexName = null;
		if (dbMetaData == null || connect == null) {
			return indexName;
		}
		String index = null;
		String column;
		int ordPos = -1;
		boolean check = false;
		ResultSet rs = null;
		try {
			rs = dbMetaData.getIndexInfo(null, null, table, unique, false);
			while (rs.next()) {
				ordPos = rs.getShort("ORDINAL_POSITION");
				if (ordPos <= 1) {
					check = false;
					index = rs.getString("INDEX_NAME");
					if (ordPos == 1) {
						check = true;
					}
				}
				if (check) {
					column = rs.getString("COLUMN_NAME").toUpperCase();
					check = column.equals(columns[ordPos - 1].toUpperCase());
					if (check && ordPos == columns.length) {
						indexName = index;
						break;
					}
				}
			}
		} catch (SQLException e) {
			metaDataError(e);
		} finally {
			close(rs);
		}
		return indexName;
	}
	/**
	 * Shortcut for closing Statement, <code>st</code>.
	 */
	static public void close(Statement st) {
		if (st != null) {
			try {
				st.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
	}
	/**
	 * Shortcut for closing ResultSet, <code>rs</code>.
	 */
	static public void close(ResultSet rs) {
		if (rs != null) {
			try {
				rs.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
	}
	/**
	 * Cleanup processing. It should be called by JVM just before garbage collection process will start for this object.
	 */
	@Override protected void finalize() throws Throwable {
		close();
		super.finalize();
	}
}