/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.database;

import java.util.HashMap;
import java.util.Map;

public enum DatabaseContents {
	NONE		(0x0000, "NONE", "No contents"),
	IONO_VI		(0x0001, "IONO_VI", "Raw vertical incidence ionogram data"),
	SCALED_VI	(0x0002, "SCALED_VI", "VI ionogram-derived scaled record, complete content"),
	FASTCHAR_VI	(0x0004, "FASTCHAR_VI", "VI ionogram-derived scaled record, only a subset of characteristics"),
	DOP_VI		(0x0008, "DOP_VI", "Raw 4-channel VI spectral data in UMS format"),
	DFT_VI		(0x0010, "DFT_VI", "Raw 4-channel VI spectral data in legady DFT format"),
	SKY_VI		(0x0020, "SKY_VI", "VI skymaps derived from raw VI spectral data"),
	DVL_VI		(0x0040, "DVL_VI", "Overhead bulk velocity of plasma drift from VI skymaps"),
	TLT_VI		(0x0080, "TLT_VI", "Tilt of overhead ionosphere from VI skymaps"),
	DOP_OI		(0x0100, "DOP_OI", "Oblique incidence raw 4-channel spectral data in UMS format"),
	SKY_OI		(0x0200, "SKY_OI", "OI skymaps derived from raw OI spectral data"),
	TID_OI		(0x0400, "TID_OI", "TRaveling ionospheric disturbance parameters derived from DOP_OI and signal tracks"),
	IONO_OI		(0x1000, "IONO_OI", "Raw oblique incidence ionogram data"),
	SCALED_OI	(0x2000, "SCALED_OI", "VI ionogram-derived scaled record, complete content"),
	TAV			(0x4000, "TAV", "Time-domain averaged RF samples"),
	IRTAM		(0x8000, "IRTAM", "IRTAM coefficients and 24-hour time series from SCALED_VI"),
	RPI			(0x10000, "RPI", "NASA Radio Plasma Imager"),
	TNT			(0x100000, "TNT", "AF TNT instrument");

	static private final Map<Long,DatabaseContents> MAP_BY_ID = new HashMap<>();
	static private final Map<String,DatabaseContents> MAP_BY_NAME = new HashMap<>();
	static {
		for (DatabaseContents contents : values()) {
			if (MAP_BY_ID.put(contents.getId(), contents) != null) {
				throw new RuntimeException("design error: duplication of id, " + contents.getId());
			}
			if (MAP_BY_NAME.put(contents.getName(), contents) != null) {
				throw new RuntimeException("design error: duplication of name, " + contents.getName());
			}
		}
	}
	private final long id;
	private final String name;
	private final String description;
	
	private DatabaseContents(long id, String name, String description) {
		this.id = id;
		this.name = name;
		this.description = description;
	}
	public long getId() {
		return id;
	}
	public String getName() {
		return name;
	}
	public String getDescription() {
		return description;
	}
	static public DatabaseContents get(long id) {
		return MAP_BY_ID.get(id);
	}
	static public DatabaseContents get(String name) {
		return MAP_BY_NAME.get(name.toUpperCase());
	}
}
