package edu.uml.lgdc.datatype;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import edu.uml.lgdc.multithread.Semaphore;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
abstract public class PoolOfObjects<E> {
	/**
	 * Default size of pool
	 */
	static private final int DEFAULT_SIZE = 10;

	private int size;
	private final List<E> OBJECTS;
	private final List<PoolObjectDesc<E>> OBJECT_DESC;
	private final Stack<Integer> freeObjectIndexes = new Stack<Integer>();

	private Semaphore semaphore;

	/**
	 * Constructor. Creates pool of default size, {@value #DEFAULT_SIZE}.
	 */
	public PoolOfObjects() {
		this(DEFAULT_SIZE);
	}
	/**
	 * Constructor. Creates pool of given size, <code>size</code>.
	 */
	public PoolOfObjects(int size) {
		if (size <= 0)
			throw new IllegalArgumentException("size <= 0");
		semaphore = new Semaphore(size);
		this.size = size;
		OBJECTS = new ArrayList<>(size);
		OBJECT_DESC = new ArrayList<>(size);
		init();
	}
	private void init() {
		for (int i = 0; i < size; i++) {
			OBJECT_DESC.add(null);
			OBJECTS.add(null);
			freeObjectIndexes.push(size - 1 - i);
		}
	}
	/**
	 * Creates and returns the instance of type <code>E</code>.
	 */
	abstract protected E createInstanceOfObject();
	/**
	 * Removes object from the pool and returns object descriptor to the user. If pool is empty, then it will wait until the pool is not
	 * empty.<br>
	 * This method calls protected method {@link #initObject(Object)} just before returning object. So, user can add some functionality,
	 * related with method <code>get</code>, by extended method {@link #initObject(Object)}, which does nothing by default.
	 */
	public PoolObjectDesc<E> get() {
		try {
			semaphore.pend();
		} catch (InterruptedException ex) {}
		synchronized (freeObjectIndexes) {
			int index = freeObjectIndexes.pop();
			if (OBJECTS.get(index) == null)
				OBJECTS.set(index, createInstanceOfObject());
			initObject(OBJECTS.get(index));
			if (OBJECT_DESC.get(index) != null) {
				OBJECT_DESC.get(index).index = index;
				OBJECT_DESC.get(index).object = OBJECTS.get(index);
			} else {
				OBJECT_DESC.set(index, new PoolObjectDesc<E>(index, OBJECTS.get(index)));
			}
			return new PoolObjectDesc<E>(index, OBJECTS.get(index));
		}
	}
	/**
	 * Returns object, given by descriptor <code>desc</code>, back to the pool.<br>
	 * This method calls protected method {@link #finalizeObject(Object)} just before returning object to the pool. So, user can add some
	 * functionality, related with method <code>free</code>, by extended method {@link #finalizeObject(Object)}, which does nothing by
	 * default.
	 */
	public void free(PoolObjectDesc<E> desc) {
		if (desc.index < 0 || desc.index >= size)
			throw new IllegalArgumentException("illegal desc.index");
		if (desc.object == null)
			throw new IllegalArgumentException("desc.object is null");
		if (OBJECTS.get(desc.index) != desc.object)
			throw new IllegalArgumentException("illegal descriptor");
		synchronized (freeObjectIndexes) {
			if (freeObjectIndexes.indexOf(desc.index) < 0) {
				finalizeObject(desc.object);
				freeObjectIndexes.push(desc.index);
				semaphore.post();
			} else
				throw new RuntimeException("object with index " + desc.index + " allready freed");
		}
	}
	/**
	 * This method is empty. User can redefine this method to add some functionality to method {@link #get()}.
	 */
	protected void initObject(E object) {}
	/**
	 * This method is empty. User can redefine this method to add some functionality to method {@link #free(PoolObjectDesc)}.
	 */
	protected void finalizeObject(E object) {}
}