/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.datatype;

import java.lang.reflect.Array;
import java.util.Calendar;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

import edu.uml.lgdc.quantities.NO_VALUE;
import edu.uml.lgdc.time.TimeScale;

public class RingBuffer<T extends Map<Object,Object>> {
	public static final int FLAG_UNPROCESSED = 0;
	public static final int FLAG_PROCESSED = 1;
	
	/**
	 * Internal backstage-array which is the storage of <i>Ring Buffer</i> elements.<br>
	 * Users never have a deal with this array directly, they are working with <i>Abstract Array</i> that contains the same number of elements 
	 * circularly mapped to this internal array.<br> 
	 * The following is <b>true</b>,
	 * <ul>
	 *   <li>both, <i>Abstract Array</i> and internal array have the same length</li>
	 *   <li>internal array has special index, denote it <b>S</b>, which can be changed under some circumstances</li>
	 *   <li>let denote <b>L</b> as the length of these array then the following 
	 *       <i>Abstract_Array_Elements</i> --> <i>Internal_Array_Elements</i> map is one-to-one correspondence between elements of these
	 *          arrays, 
	 *     <ul>
	 *        <code>AbstractIndex</code> ---> <code>(AbstractIndex + <b>S</b>) % <b>L</b></code>  
	 *     </ul>
	 *   </li>    
	 * </ul>
	 */
	private final T[] buffer;
	private final int length;
	/**
	 * Time units that used for {@link #bufferDuration_units}, {@link #step_units}, and {@link #forecast_units}.<br>
	 * Allowed Time Units are,
	 * <ul>
	 *   <li>{@link TimeUnit#SECONDS} - seconds</li>
	 *   <li>{@link TimeUnit#MINUTES} - minutes</li>
	 *   <li>{@link TimeUnit#HOURS} - hours</li>
	 *   <li>{@link TimeUnit#DAYS} - days</li>
	 * </ul>
	 */
	private final TimeUnit timeUnit;
	/**
	 * Length of non-forecast part of buffer, given in Time Units
	 */
	private final int bufferDuration_units;
	/**
	 * Time step, given in Time Units, between cells of this RingBuffer
	 */
	private final int step_units;
	/**
	 * Length of forecast part of buffer, given in Time Units. It can be considered as future time zone.<br>
	 * Note that this zone can absent, i.e. {@code forcast_units} can be 0.
	 * 
	 */
	private final int forecast_units;
	/**
	 * Units index from class {@link Calendar}, that corresponds to Time Unit. For example {@link Calendar#SECOND} will correspond to
	 * {@link TimeUnit#SECONDS}, and so on.
	 */
	private final int calendarUnitIndex;
	/**
	 * Number of milliseconds in one Time Unit.
	 */
	private final int millisecondsPerUnit;
	private int startIndex;
	/**
	 * Time of validity. It is kind of "now time", so that greater (later) times are going into forecast zone, if forecast zone presents in
	 * this buffer, otherwise it is also the latest time in buffer.
	 */
	private TimeScale tov;
	private TimeScale midnightToday = new TimeScale();
	/**
	 * The earliest time this ring buffer contains.
	 */
	private TimeScale firstTime = new TimeScale();
	/**
	 * The latest time this ring buffer contains.<br>
	 * If forecast part of buffer is not zero-length, then {@code lastTime} is greater (later) than {@link #tov}, otherwise it equals {@link #tov}. 
	 */
	private TimeScale lastTime = new TimeScale();
	private TimeScale firstTime_lowerBound = new TimeScale(); // for index computations
	private TimeScale lastTime_upperBound = new TimeScale(); // for index computations
	private int processingFlag;

	@SuppressWarnings("unchecked")
	public RingBuffer(Class<T> clazz, TimeUnit timeUnit, int bufferDuration_units, int step_units, int forecast_units) {
		Objects.requireNonNull(clazz);
		this.timeUnit = Objects.requireNonNull(timeUnit);
		assert timeUnit.ordinal() >= TimeUnit.SECONDS.ordinal();
		this.bufferDuration_units = step_units * (bufferDuration_units / step_units);
		this.step_units = step_units;
		this.forecast_units = forecast_units;
		calendarUnitIndex = getCalendarUnitIndex(timeUnit);
		millisecondsPerUnit = getMillisecondsPerUnit(timeUnit);
		length = (bufferDuration_units + this.forecast_units) / this.step_units;
		buffer = (T[]) Array.newInstance(clazz, length);
		try	{
			for (int i = 0; i < length; i++) {
				this.buffer[i] = clazz.newInstance();
			}
		} catch (InstantiationException | IllegalAccessException ex) {
			throw new RuntimeException("design error");
		}
	}

	public T getCell( TimeScale time ) {
		int i = timeToInternalIndex(time);
		checkIndex(i);
		return buffer[i];
	}
	
	// the input index is relative to 0 (oldest time)
	public T getCell( int iRelative) {
		int i = userIndextoInternalIndex(iRelative);
		checkIndex(i);
		return buffer[i];
	}
	
	public Object get(int userIndex, Object type) {
		checkIndex(userIndex);
		return buffer[userIndextoInternalIndex(userIndex)].get(type);
	}

	public Object get(TimeScale time, Object type) {
		int i = timeToInternalIndex(time);
		return i != NO_VALUE.INDEX ? buffer[i].get(type) : null;
	}

	public Object[] get(Object type) {
		return get(startIndex, startIndex + length - 1, type);
	}

	public Object[] getTodayValues(Object type) {
		int iStart = timeToInternalIndex(midnightToday);
		int iStop = timeToInternalIndex(tov);
		if (iStart == NO_VALUE.INDEX || iStop == NO_VALUE.INDEX) {
			return null;
		}
		return get(iStart, iStop, type);
	}

	public Object[] getPastValues(Object type) {
		int iStart = timeToInternalIndex(firstTime);
		int iStop = timeToInternalIndex(tov);
		if (iStart == NO_VALUE.INDEX || iStop == NO_VALUE.INDEX) {
			return null;
		}
		return get(iStart, iStop, type);
	}

	/** internal method */
	private Object[] get(int internalIndex1, int internalIndex2, Object type) {
		if( internalIndex2 < internalIndex1 ) {
			internalIndex2 += length;
		}
		Object[] output = new Object[internalIndex2 - internalIndex1 + 1];
		for (int i = internalIndex1, k = 0; i <= internalIndex2; i++, k++) {
			output[k] = buffer[i % length].get(type);
		}
		return output;
	}

	public TimeScale getMidnightToday() {
		return midnightToday;
	}
	
	public TimeScale getTov() {
		return tov;
	}
	
	/**
	 * Returns first time, the earliest time this ring buffer contains.
	 */
	public TimeScale getFirstTime() {
		return firstTime;
	}
	
	/**
	 * Returns last time, the latest time this ring buffer contains.<br>
	 * If forecast part of buffer is not zero-length, then {@code lastTime} is greater (later) than {@link #tov}, otherwise it equals {@link #tov}. 
	 */
	public TimeScale getLastTime() {
		return lastTime;
	}
	
	public int getLength() {
		return length;
	}
	
	public TimeScale getOldestObjectTime() {
		int iStart = timeToInternalIndex(firstTime);
		int iStop = timeToInternalIndex(tov);
		if (iStart == NO_VALUE.INDEX || iStop == NO_VALUE.INDEX) {
			throw new RuntimeException("iStart and iStop has to be >= 0, but iStart is " + iStart + ", iStop is " + iStop);
		}
		if( iStop < iStart ) {
			iStop += length;
		}
		for (int i = iStart; i <= iStop; i++) {
			int k = i % length;
			if( !buffer[k].isEmpty() ) {
				return relativeIndexToTime(internalndexToUserIndex(k));
			}
		}
		return null;
	}
	
	public TimeScale getRecentObjectTime() {
		int iStart = timeToInternalIndex(firstTime);
		int iStop = timeToInternalIndex(tov);
		if (iStart == NO_VALUE.INDEX || iStop == NO_VALUE.INDEX) {
			throw new RuntimeException("iStart and iStop has to be >= 0, but iStart is " + iStart + ", iStop is " + iStop);
		}
		if (iStop < iStart) {
			iStop += length;
		}
		for (int i = iStop; i >= iStart; i--) {
			int k = i % length;
			if (!buffer[k].isEmpty()) {
				return relativeIndexToTime(internalndexToUserIndex(k));
			}
		}
		return null;
	}

	public boolean isEmpty() {
		for (T map : buffer) {
			if( !map.isEmpty() ) {
				return false;
			}
		}
		return true;
	}

	public int filledCellCount() {
		int count = 0;
		for( int i = 0; i < length; i++ ) {
			if( !buffer[i].isEmpty() ) {
				count++;
			}
		}
		return count;
	}
	
	public boolean put(TimeScale time, Object type, Object object) {
		int i = timeToInternalIndex(time);
		if (i == NO_VALUE.INDEX) {
			return false;
		}
		buffer[i].put(type, object);
		return true;
	}

	public boolean put(int userIndex, Object type, Object object) {
		buffer[userIndextoInternalIndex(userIndex)].put(type, object);
		return true;
	}

	/**
	 * Internal method, must not be exposed.<br>
	 * It returns internal index, i.e. index in array {@link #buffer}, which user has not be aware of.
	 */
	private int timeToInternalIndex(TimeScale time) {
		if( !isValidTime(time) ) {
			return NO_VALUE.INDEX;
		}
		int index = (int)Math.round(time.diffIn(calendarUnitIndex, firstTime) / step_units);
		// check and correct it because of marginal time values gives indexes -1 and *length* 
		if (index == -1) {
			index = 0;
		} else if (index == length) {
			index = length - 1;
		}
		return (startIndex + index) % length;
	}
	
	/**
	 * Simply check if the input time is within the buffer 
	 */
	public boolean isValidTime(TimeScale time) {
		return (!time.after(lastTime_upperBound) && !time.before(firstTime_lowerBound));
	}

	public TimeScale relativeIndexToTime(int i) {
		checkIndex(i);
		TimeScale time = new TimeScale(firstTime.getTimeInMillis());
		time.add(calendarUnitIndex, i * step_units);
		return time;
	}
	
	private int userIndextoInternalIndex(int index) {
		checkIndex(index);
		return (startIndex + index) % length;
	}
	
	private int internalndexToUserIndex(int index) {
		checkIndex(index);
		return index - startIndex + ((startIndex > index) ? length : 0);
	}
	
	private void checkIndex(int i) {
		if (i < 0 || i >= length) {
			throw new IllegalArgumentException("i < 0 || i >= length, i=" + i + ", length=" + length);
		}
	}

	public double timeGridMismatch_min(TimeScale time) {
		return convertTo(timeGridMismatch_units(time), TimeUnit.MINUTES);
	}

	public double timeGridMismatch_sec(TimeScale time) {
		return convertTo(timeGridMismatch_units(time), TimeUnit.SECONDS);
	}

	public double timeGridMismatch_units(TimeScale time) {
		double diffInUnits = time.diffIn(calendarUnitIndex, firstTime);
		return Math.abs(diffInUnits - (int)(diffInUnits / step_units) * step_units);
	}
	

	public void setTov(TimeScale newTov) {
		// compute the new first time
		TimeScale newFirstTime = new TimeScale(newTov.getTimeInMillis() + millisecondsPerUnit * (step_units - bufferDuration_units));
		TimeScale newLastTime = new TimeScale(newTov.getTimeInMillis() + millisecondsPerUnit * forecast_units);
		if (tov != null) {
			// to avoid confusion with expired data, force clear it
			clearExpired(firstTime, newFirstTime, newLastTime);
			startIndex += (int) (newFirstTime.diffIn(calendarUnitIndex, firstTime) / step_units);
			startIndex %= length;
			if (startIndex < 0) {
				startIndex += length;
			}
		} else {
			// first ever TOV
			startIndex = 0;
			tov = new TimeScale();
		}
		firstTime.setTimeInMillis(newFirstTime.getTimeInMillis());
		firstTime_lowerBound.setTimeInMillis(firstTime.getTimeInMillis() - step_units * millisecondsPerUnit / 2);
		tov.setTimeInMillis(newTov.getTimeInMillis());
		lastTime.setTimeInMillis(newLastTime.getTimeInMillis());
		lastTime_upperBound.setTimeInMillis(lastTime.getTimeInMillis() + step_units * millisecondsPerUnit / 2);
		midnightToday.setTimeInMillis(tov.getTimeInMillis());
		midnightToday.add(Calendar.SECOND, -(int) tov.getTimeSinceMidnightIn(Calendar.SECOND));
	}

	/**
	 * Clears expired data from the buffer.<br>
	 * Checks the buffer dynamics: cell at the larger time of the two is not cleared
	 */ 
	private void clearExpired(TimeScale firstTime, TimeScale newFirstTime, TimeScale newLastTime) {
		int iFirst = timeToInternalIndex(firstTime);
		if (newFirstTime.after(firstTime)) {
			// newFirstTime time still within the ring
			int iNewFirst = timeToInternalIndex(newFirstTime);
			if (iNewFirst != NO_VALUE.INDEX) {
				if (iFirst != iNewFirst) {
					if (iNewFirst < iFirst) {
						iNewFirst += length;
					}
					for (int i = iFirst; i < iNewFirst; i++) {
						buffer[i % length].clear();
					}
				}
			} else {
				// clear whole buffer
				clearBuffer();
			}
		} else {	
			int iNewLast = timeToInternalIndex(newLastTime);
			if (iNewLast != NO_VALUE.INDEX) {
				if (iFirst <= iNewLast) {
					iFirst += length;
				}
				for (int i = iNewLast + 1; i < iFirst; i++) {
					buffer[i % length].clear();
				}
			} else {
				// clear whole buffer
				clearBuffer();
			}
		}
	}
	
	private void clearBuffer() {
		for (int i = 0; i < length; i++) {
			buffer[i].clear();
		}
	}
	
	public void clear() {
		clearBuffer();
		tov = null;
	}

	public int getProcessingFlag() {
		return processingFlag;
	}

	public void setProcessingFlag(int processingFlag) {
		this.processingFlag = processingFlag;
	}
	
	public double getStep_min() {
		return convertTo(step_units, TimeUnit.MINUTES);
	}

	public int getStep_sec() {
		return (int)convertTo(step_units, TimeUnit.SECONDS);
	}

	public long getStep_ms() {
		return 1000 * getStep_sec();
	}
	
	public int getStep_units() {
		return step_units;
	}

	public TimeUnit getTimeUnit() {
		return timeUnit;
	}

	public int getCalendarUnitIndex() {
		return calendarUnitIndex;
	}
	
	static private int getCalendarUnitIndex(TimeUnit timeUnit) {
		switch (timeUnit) {
		case SECONDS:
			return Calendar.SECOND;
		case MINUTES:
			return Calendar.MINUTE;
		case HOURS:
			return Calendar.HOUR_OF_DAY;
		case DAYS:
			return Calendar.DAY_OF_MONTH;
		default:
			throw new RuntimeException("illegal TimeUnit, " + timeUnit); 
		}
	}
	
	static private int getMillisecondsPerUnit(TimeUnit timeUnit) {
		switch (timeUnit) {
		case SECONDS:
			return 1000;
		case MINUTES:
			return 60000;
		case HOURS:
			return 3600000;
		case DAYS:
			return 24 * 3600000;
		default:
			throw new RuntimeException("illegal TimeUnit, " + timeUnit); 
		}
	}
	
	static private double convertToSeconds(double value_units, TimeUnit timeUnit) {
		switch (timeUnit) {
		case SECONDS:
			return value_units;
		case MINUTES:
			return 60 * value_units;
		case HOURS:
			return 3600 * value_units;
		case DAYS:
			return 24 * 3600 * value_units;
		default:
			throw new RuntimeException("illegal TimeUnit, " + timeUnit); 
		}
	}
	
	private double convertTo(double value_units, TimeUnit otherTimeUnit) {
		if (timeUnit != otherTimeUnit) {
			switch (otherTimeUnit) {
			case SECONDS:
				return convertToSeconds(value_units, timeUnit);
			case MINUTES:
				return convertToSeconds(value_units, timeUnit) / 60.0;
			case HOURS:
				return convertToSeconds(value_units, timeUnit) / 3600.0;
			case DAYS:
				return convertToSeconds(value_units, timeUnit) / (24 * 3600.0);
			default:
				throw new RuntimeException("illegal TimeUnit for ring buffer duration, " + otherTimeUnit); 
			}
		} else {
			return value_units;
		}
	}
}
