package edu.uml.lgdc.datatype;

import edu.uml.lgdc.project.CommonConst;
import edu.uml.lgdc.time.TimeMarked;
import edu.uml.lgdc.time.TimeScale;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
abstract public class TimeOrderedMetric implements OrderedMetricable<TimeOrderedMetric>, TimeMarked {
	static private final String FILE_TIME_FORMAT = CommonConst.getFileTimeFormat();

	public TimeScale getTime() {
		double time = getTimeInMinutes();
		if (!Double.isNaN(time))
			return new TimeScale(time);
		else
			return null;
	}
	@Override public int compareTo(TimeOrderedMetric timeOrderedMetric) {
		if (timeOrderedMetric != null) {
			double time1 = getTimeInMinutes();
			double time2 = timeOrderedMetric.getTimeInMinutes();
			if (!Double.isNaN(time1) && !Double.isNaN(time2))
				return (int)Math.signum(time1 - time2);
		}
		return 1;
	}
	/**
	 * Returns distance as time difference, in minutes.
	 */
	@Override public double dist(TimeOrderedMetric timeOrderedMetric) {
		double dist = Double.NaN;
		if (timeOrderedMetric != null) {
			double time1 = getTimeInMinutes();
			double time2 = timeOrderedMetric.getTimeInMinutes();
			if (!Double.isNaN(time1) && !Double.isNaN(time2))
				dist = Math.abs(time1 - time2);
		}
		return dist;
	}
	/**
	 * Returns File Name (=special string) representation of this object as formatted time string using default formatting.<br>
	 * You can redefine or enhance this method in subclass.
	 */
	public String createFileName() {
		return getTimeAsString();
	}
	/**
	 * Returns File Name (=special string) representation of this object as formatted time string using given formatting string
	 * <code>format</code>.<br>
	 * You can redefine or enhance this method in subclass.
	 */
	public String createFileName(String format) {
		return getTimeAsString(format);
	}
	/**
	 * Returns Temporary File Name (=special string) representation of this object as formatted time string using default formatting.<br>
	 * You can redefine or enhance this method in subclass.
	 */
	public String createTempFileName() {
		return getTimeAsString();
	}
	/**
	 * Returns Temporary File Name (=special string) representation of this object as formatted time string using given formatting string
	 * <code>format</code>.<br>
	 * You can redefine or enhance this method in subclass.
	 */
	public String createTempFileName(String format) {
		return getTimeAsString(format);
	}
	/**
	 * Returns Temporary Name (=special string) of this object as formatted time string using default formatting.<br>
	 * You can redefine or enhance this method in subclass.
	 */
	public String createTempName() {
		return getTimeAsString(FILE_TIME_FORMAT);
	}
	/**
	 * Returns Temporary Name (=special string) of this object as formatted time string using given formatting string <code>format</code>.<br>
	 * You can redefine or enhance this method in subclass.
	 */
	public String createTempName(String format) {
		return getTimeAsString(format);
	}
	/**
	 * Returns Name (=special string) of this object as formatted time string using default formatting.<br>
	 * You can redefine or enhance this method in subclass.
	 */
	public String createName() {
		return getTimeAsString();
	}
	/**
	 * Returns Name (=special string) of this object as formatted time string using given formatting string <code>format</code>. You can
	 * redefine or enhance this method in subclass.
	 */
	public String createName(String format) {
		return getTimeAsString(format);
	}
	/**
	 * Returns Formatted Time String using default formatting.<br>
	 * You can't redefine this method in subclass.
	 */
	public final String getTimeAsString() {
		return getTimeAsString(FILE_TIME_FORMAT);
	}
	/**
	 * Returns Formatted Time String using specified time format string <code>format</code>.<br>
	 * You can't redefine this method in subclass.
	 */
	public final String getTimeAsString(String format) {
		return getTime().toFormatUT(format);
	}
}
