/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.fileio;

// file: FileRW.java
// started: January 3, 1999

// (c) Copyright 1999
// University of Massachusetts Lowell
import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;

public class FileRW extends RandomAccessFile implements Closeable {
	private String filePathName;
	private File file;
	private String fileName;
	public String filePath;

	public static final int BIG_ENDIAN = 1;
	public static final int LITTLE_ENDIAN = 2;

	private byte w[];

	private int endian;

	public FileRW(String inPathName) throws IOException {
		this(inPathName, "rw");
	}
	public FileRW(String inPathName, String mode) throws IOException {
		super(inPathName, mode);
		init(inPathName);
	}
	private void init(String inPathName) {
		filePathName = inPathName;
		file = new File(filePathName);
		fileName = file.getName();
		filePath = file.getParent();
		if (filePath == null) {
			filePath = ".";
		}
		w = new byte[8];
		endian = BIG_ENDIAN;
	}
	public int getEndian() {
		return endian;
	}
	public void setEndian(int in) {
		endian = in;
	}
	public String getFileName() {
		return fileName;
	}
	public String getPathName() {
		return filePath;
	}
	public String getFullFileName() {
		return filePathName;
	}
	public boolean ready() throws IOException {
		return (length() - getFilePointer()) > 0;
	}
	public void newLine() throws IOException {
		writeBytes("\r\n");
	}
	public void write(String str) throws IOException {
		writeBytes(str);
	}
	public short readShortE() throws IOException {
		if (endian == BIG_ENDIAN) {
			return readShort();
		}
		readFully(w, 0, 2);
		return (short)((w[1] & 0xff) << 8 | (w[0] & 0xff));
	}
	public int readIntE() throws IOException {
		if (endian == BIG_ENDIAN) {
			return readInt();
		}
		readFully(w, 0, 4);
		return (w[3]) << 24 | (w[2] & 0xff) << 16 | (w[1] & 0xff) << 8 | (w[0] & 0xff);
	}
	public void copyFromFile(String fileName) throws IOException {
		byte b[] = new byte[32000];
		int len;
		FileRW in = null;
		try {
			in = new FileRW(fileName, "r");
			while (in.ready()) {
				len = in.read(b);
				write(b, 0, len);
			}
		} finally {
			if (in != null) {
				in.close();
			}
		}
	}
	/**
	 * Copies portion of this instance beginning from current position and up to the end into newly created file, <code>filename</code>.
	 */
	public void copyToFile(String filename) throws IOException {
		byte b[] = new byte[32000];
		int len;
		FileRW to = null;
		long savedPos = getFilePointer();
		try {
			new File(filename).delete();
			to = new FileRW(filename, "rw");
			while (ready()) {
				len = read(b);
				to.write(b, 0, len);
			}
		} finally {
			if (to != null) {
				to.close();
			}
			seek(savedPos);
		}
	}
	public void copyFromStream(InputStream inputStream) throws IOException {
		byte b[] = new byte[32000];
		int len;
		while (inputStream.available() > 0) {
			len = inputStream.read(b);
			write(b, 0, len);
		}
	}
	/**
	 * This function was primarily written for Mac platform where line ended with the only Carriage Return.<br>
	 * Reads the next line of text from this file. This method successively reads bytes from the file until it reaches the end of a line of
	 * text.<br>
	 * A line of text is terminated one of the following,
	 * <ul>
	 *   <li>carriage-return character</li>
	 *   <li>newline character</li>
	 *   <li>carriage-return character</li>
	 * </ul>  
	 * immediately followed by a newline character; 4) the end of the input stream.<br>
	 * The line-terminating character(s), if any, are not included into returned string.<br>
	 * return the next line of text from this file or null if EOF encountered.
	 */
	public final String smartReadLine() throws IOException {
		StringBuffer input = new StringBuffer(200);
		int c, b;
		while (((c = read()) != -1) && (c != '\r') && (c != '\n')) {
			input.append((char)c);
		}
		if (c == '\r') {
			long offset = getFilePointer();
			if (((b = read()) != -1) && (b != '\n')) {
				this.seek(offset);
			}
		}
		if ((c == -1) && (input.length() == 0)) {
			return null;
		}
		return input.toString();
	}
	/**
	 * Returns extension of file, i.e. all what is followed last dot symbol ("."), excluding this dot symbol.<br>
	 * If no dot exists or nothing follow after last dot, then empty string ( "" ) is returned.
	 */
	static public String getExt(String fileName) {
		String ext = "";
		int pos = fileName.lastIndexOf('.');
		if (pos >= 0 && pos < fileName.length() - 1) {
			ext = fileName.substring(pos + 1);
		}
		return ext;
	}
	/**
	 * Takes given file, <b>fileName</b>, possibly with extension, and returns this file name without extension and dot itself. For example,
	 * <ul>
	 *   <code>foo.txt ---> foo</code>
	 * </ul>
	 */
	static public String getNameButExt(String fileName) {
		int pos = fileName.lastIndexOf('.');
		if (pos >= 0) {
			fileName = fileName.substring(0, pos);
		}
		return fileName;
	}
	/**
	 * Returns file name and extension (if any) of full file, <code>fullFileName</code>. So any path preceded will be removed.
	 */
	static public String getFileName(String fullFileName) {
		return new File(fullFileName).getName();
	}
	/**
	 * Returns file name only of file, given as full filename, <code>fullFileName</code>. So any path preceded and extension (including 
	 * last dot) will be removed.
	 */
	public String getFileNameOnly() {
		return getNameButExt(fileName);
	}
	/**
	 * Returns file name only of given full filename, <code>fullFileName</code>. So any path preceded and extension (including last dot) 
	 * will be discarded.
	 */
	static public String getFileNameOnly(String fullFileName) {
		return getNameButExt(new File(fullFileName).getName());
	}
	/**
	 * Returns file path with separator (if present) at the end.<br>
	 * Note:
	 * <ul>
	 *   Separator is not omitted because if in Unix file located in root directory (though it is bad style) you have no possibility 
	 *   to show where it is located but in "/". This is because in Unix symbol "/" overloaded,
	 *   <ul> 
	 *     at the beginning of file name it means "root directory" and in the middle just serves as separator between directories
	 *   </ul>
	 * </ul>  
	 */
	static public String getPath(String fullFileName) {
		String parent = new File(fullFileName).getParent();
		if (parent != null) {
			return parent;
		} else {
			return ".";
		}
	}
	/**
	 * Returns length of file with given name, <code>fileName</code>, as long number. If this file doesn't exist returns -1, so you can 
	 * differentiate this case from zero-length existing file.
	 */
	static public long getLength(String fileName) {
		if (fileName == null) {
			throw new IllegalArgumentException("fileName == null");
		}
		return get_Length(null, fileName);
	}
	/**
	 * Returns length of file with given path, <code>path</code>, and name, <code>fileName</code> as long number.<br>
	 * If this file doesn't exist returns -1, so you can differentiate this case from zero-length existing file.
	 */
	static public long getLength(String path, String fileName) {
		if (path == null && fileName == null) {
			throw new IllegalArgumentException("path == null and fileName == null");
		}
		return get_Length(path, fileName);
	}
	/**
	 * Returns length of file with given <code>path</code>> and <code>fileName</code> as long number.<br>
	 * If this file doesn't exist returns -1, so you can differentiate this case from zero-length existing file.<br>
	 * Note:
	 * <ul>
	 *   it is acceptable <code>fileName</code>> equaled to <b>null</b>, - in this case <code>path</code> will be considered as file name
	 * </ul>
	 */
	static private long get_Length(String path, String fileName) {
		File file;
		if (fileName == null) {
			file = new File(path);
		} else {
			file = new File(path, fileName);
		}
		if (file.isFile()) {
			return file.length();
		}
		return -1;
	}
	/**
	 * Deletes specified file
	 */
	static public boolean deleteFile(String fileName) {
		if (fileName == null) {
			throw new IllegalArgumentException("fileName == null");
		}
		return delete_File(null, fileName);
	}
	/**
	 * Deletes specified file
	 */
	static public boolean deleteFile(String path, String fileName) {
		if (path == null && fileName == null) {
			throw new IllegalArgumentException("path == null and fileName == null");
		}
		return delete_File(path, fileName);
	}

	/**
	 * Deletes specified file.<br>
	 * Note:
	 * <ul>
	 *   it is acceptable <code>fileName</code> or <code>path</code> is <b>null</b>, but not both
	 * </ul>
	 */
	static private boolean delete_File(String path, String fileName) {
		File file;
		if (fileName != null) {
			file = new File(path, fileName);
		} else {
			file = new File(path);
		}
		return file.delete();
	}
	/**
	 * Copies file <code>from</code> onto file <code>to</code>. Before start copying file <code>to</code> is deleting.
	 */
	static public void fileCopy(String from, String to) throws IOException {
		File target = new File(to);
		if (target.exists()) {
			if (!target.isFile()) {
				throw new IOException(to + " is not a file ");
			}
			if (!target.delete()) {
				throw new IOException("Cannot delete file " + to);
			}
		}
		FileRW out = null;
		try {
			out = new FileRW(to, "rw");
			out.copyFromFile(from);
		} finally {
			if (out != null) {
				out.close();
			}
		}
	}
	/**
	 * Copies file <code>from</code> onto file <code>to</code>. Before copying, file <code>to</code> is deleting.<br>
	 * Note:
	 * <ul>
	 *   copying uses <code>.tmp</code> extension while in progress
	 * </ul>
	 */
	static public void safeFileCopy(String from, String to) throws IOException {
		File source = new File(from);
		if (source.exists() && !source.isFile()) {
			throw new IOException(from + " is not a file");
		}
		File f2 = new File(to);
		if (f2.exists() && !f2.isFile()) {
			throw new IOException(to + " is not a file");
		}
		if (f2.equals(source)) {
			throw new IOException("Try to copy file onto itself, " + from);
		}
		String filePathNamePart = to + ".tmp";
		File f2tmp = new File(filePathNamePart);
		if (f2tmp.exists() && !f2tmp.delete()) {
			throw new IOException("Cannot delete file " + filePathNamePart);
		}
		FileRW out = null;
		boolean ok = false;
		try {
			out = new FileRW(filePathNamePart, "rw");
			out.copyFromFile(from);
			ok = true;
		} finally {
			if (out != null) {
				try {out.close();} catch (IOException ex) {}
				if (!ok) {
					f2tmp.delete();
				}
			}
		}
		File f1 = new File(filePathNamePart);
		if (f2.exists() && !f2.delete()) {
			throw new IOException("Cannot delete file " + to);
		}
		if (!f1.renameTo(f2)) {
			throw new IOException("Cannot rename file " + filePathNamePart + " to " + to);
		}
	}
}