/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.format;

import java.text.*;
import java.util.*;

import edu.uml.lgdc.time.TimeScale;

public class FC
{
  static public final int ORDER_NORMAL   = 0;
  static public final int ORDER_INVERTED = 1;

  static public final int LOW_BYTE = 0x000000FF;

  static private short a_short;

  static public final String HEX_DIGITS = "0123456789ABCDEF";
  static public final String STR_BLANKS = "                                                              "
		  + "                                                              ";
  static public final String STR_ZEROES = "00000000000000000000000000000000000000000000000000000000000000"
		  + "00000000000000000000000000000000000000000000000000000000000000";

  
  static public byte high_nibble( byte in ) 
  { 
    return (byte)( ( in & 0xF0 ) >> 4 ); 
  }
  
  static public byte low_nibble( byte in ) 
  { 
    return (byte)( in & 0xF ); 
  }

  static public final double getDelta( int precision )
  {
    return 5 * Math.pow( 10.0, -( precision + 1 ) );
  }

  static public final double round( double value, int precision )
  {
    double factor = Math.pow( 10, precision );
    long l = Math.round( value * factor );
    double d = (double)l / factor;
    return d;
  }

  static public final String padLeft( String str, int width )
  {
    return padLeft( str, width, ' ' );
  }

  static public final String padLeft( String str, int width, char ch )
  {
    int length = str.length();
    if( length == width )
      return str;
    if( length > width )
      return str.substring( length - width, length );
    int padLen = width - length;
    char[] chArr = new char[width];
    for (int i = 0; i < padLen; i++)
    	chArr[i] = ch;
    for (int i = padLen, j = 0; i < width; i++, j++)
    	chArr[i] = str.charAt(j);
    return new String(chArr);
  }

  static public final String padRight( String str, int width )
  {
    return padRight( str, width, ' ' );
  }

  static public final String padRight( String str, int width, char ch )
  {
    int length = str.length();
    if( length == width )
      return str;
    if( length > width )
      return str.substring( 0, width );
    char[] chArr = new char[width];
    for (int i = 0; i < length; i++)
    	chArr[i] = str.charAt(i);
    for (int i = length; i < width; i++)
    	chArr[i] = ch;
    return new String(chArr);
  }

  /**
   * Returns sequence of nibbles as integer putting these nibbles into
   * integer memory container so that the last nibble adjust to the right
   * edge (LSB edge) of container.
   */
  static public int nibbles2Int( byte in[], int nibbleOffset, int qty )
  {
    return nibbles2Int( in, nibbleOffset, qty, false );
  }

  /**
   * Returns sequence of nibbles as integer putting these nibbles into
   * integer memory container so that the last nibble adjust to the right
   * edge (LSB edge) of container.
   */
  static public int nibbles2Int( byte in[], int nibbleOffset, int qty,
                                 boolean inverted )
  {
    int value = 0;
    boolean upper = ( ( nibbleOffset + qty - 1 ) % 2 ) == 0;
    int nibblePosition;
    byte nibble;

    for( int i = 0; i < qty; i++ )
    {
      if( !inverted )
        nibblePosition = ( nibbleOffset + qty - 1 - i ) / 2;
      else
        nibblePosition = ( nibbleOffset + i ) / 2;
      if( upper )
        nibble = (byte)( in[nibblePosition] >> 4 );
      else
        nibble = in[nibblePosition];
      nibble &= 0x0f;
      value |= nibble << ( i * 4 );
      upper = !upper;
      nibbleOffset++;
    }

    return value;
  }

  /**
   * Converts BCD (binary coded decimal - each decimal digit is contained in
   * in half of byte - 4 bits - which is called "nibble" - so there are
   * two decimal digits per byte ) sequence to integer and return this integer.
   * If non-decimal encountered in any nibble then returns Integer.MAX_VALUE.
   * No check is made against overflowing.
   */
  static public int BCDtoInt( byte in[], int nibbleOffset, int qty )
  {
    int value = 0;
    boolean upper = ( nibbleOffset % 2 ) == 0;
    byte digit;

    for( int i = 0; i < qty; i++ )
    {
      if( upper )
        digit = (byte)( in[nibbleOffset/2] >> 4 );
      else
        digit = in[nibbleOffset/2];
      digit &= 0x0f;
      if( digit > 9 )    // should be revisited, e.g. throw IllegalArgumentException
      {
        value = Integer.MAX_VALUE;
        break;
      }
      value = 10 * value + digit;
      upper = !upper;
      nibbleOffset++;
    }

    return value;
  }

  /**
   * Converts BCD (binary coded decimal - each byte of array <in> is decimal
   * digit - i.e. value of it >= 0 and <= 9) decimal digit sequence to integer
   * and return this integer.
   * If non-decimal encountered in any byte then returns Integer.MAX_VALUE.
   * No check is made against overflowing.
   */
  static public int BCDBytes2Int( byte in[], int start, int length )
  {
    return BCDBytes2Int( in, start, length, false );
  }

  /**
   * Converts BCD (binary coded decimal - each byte of array <in> is decimal
   * digit - i.e. value of it >= 0 and <= 9) decimal digit sequence to integer
   * and return this integer.
   * If non-decimal encountered in any byte then returns Integer.MAX_VALUE.
   * No check is made against overflowing.
   */
  static public int BCDBytes2Int( byte in[], int start, int length,
                                  boolean inverted )
  {
    int value = 0;
    byte digit;
    for( int i = 0; i < length; i++ )
    {
      if( !inverted )
        digit = in[start + i];
      else
        digit = in[start + length - 1 - i];
      if( digit > 9  ||  digit < 0 )
      {
        value = Integer.MAX_VALUE;
        break;
      }
      value = 10 * value + digit;
    }
    return value;
  }

  /**
   * Converts integer to BCD (binary coded decimal - each byte of array <in>
   * is decimal digit - i.e. value of it >= 0 and <= 9) decimal digit sequence.
   * No check is made against overflowing.
   */
  static public void int2BCDBytes( int value, byte in[], int start,
                                   int length )
  {
    int2BCDBytes( value, in, start, length, false );
  }

  /**
   * Converts integer to BCD (binary coded decimal - each byte of array <in>
   * is decimal digit - i.e. value of it >= 0 and <= 9) decimal digit sequence.
   * No check is made against overflowing.
   */
  static public void int2BCDBytes( int value, byte in[], int start, int length,
                                    boolean inverted )
  {
    int index;
    if( !inverted )
      index = start + length - 1;
    else
      index = start;
    for( int i = 0; i < length; i++ )
    {
      if( !inverted )
        in[index--] = (byte)( value % 10 );
      else
        in[index++] = (byte)( value % 10 );
      value /= 10;
    }
  }

  /**
   * Converts BCHD (binary coded hexa-decimal - each byte of array <in> is
   * hexadecimal digit - i.e. value of it >= 0 and <= F) hexadecimal digit
   * sequence to integer and return this integer.
   * If non-hexadecimal encountered in any byte then returns
   * Integer.MAX_VALUE.
   * No check is made against overflowing.
   */
  static public int BCHDBytes2Int( byte in[], int start, int length )
  {
    return BCHDBytes2Int( in, start, length, false );
  }

  /**
   * Converts BCHD (binary coded hexa-decimal - each byte of array <in> is
   * hexadecimal digit - i.e. value of it >= 0 and <= F) hexadecimal digit
   * sequence to integer and return this integer.
   * If non-hexadecimal encountered in any byte then returns
   * Integer.MAX_VALUE.
   * No check is made against overflowing.
   */
  static public int BCHDBytes2Int( byte in[], int start, int length,
                                   boolean inverted )
  {
    int value = 0;
    byte digit;
    for( int i = 0; i < length; i++ )
    {
      if( !inverted )
        digit = in[start + i];
      else
        digit = in[start + length - 1 - i];
      if( digit > 15  ||  digit < 0 )
      {
        value = Integer.MAX_VALUE;
        break;
      }
      value = 16 * value + digit;
    }
    return value;
  }

  /**
   * Converts integer value to BCHD (binary coded hexa-decimal - each byte
   * of array <in> is hexadecimal digit - i.e. value of it >= 0 and <= F)
   * hexadecimal digit sequence to integer and return this integer.
   * No check is made against overflowing.
   */
  static public void int2BCHDBytes( int value, byte in[], int start,
                                    int length )
  {
    int2BCHDBytes( value, in, start, length, false );
  }

  /**
   * Converts integer value to BCHD (binary coded hexa-decimal - each byte
   * of array <in> is hexadecimal digit - i.e. value of it >= 0 and <= F)
   * hexadecimal digit sequence to integer and return this integer.
   * No check is made against overflowing.
   */
  static public void int2BCHDBytes( int value, byte in[], int start,
                                    int length, boolean inverted )
  {
    int index;
    if( !inverted )
      index = start + length - 1;
    else
      index = start;
    for( int i = 0; i < length; i++ )
    {
      if( !inverted )
        in[index--] = (byte)( value % 16 );
      else
        in[index++] = (byte)( value % 16 );
      value /= 16;
    }
  }

  /**
   * Converts sequence of bytes containing binary representation of signed
   * integer (in two's compliment format) to integer variable and returns
   * this variable.
   *
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   *
   * @return
   *    Integer value represented by given bytes.
   *
   * @exceptions
   * <b>IllegalArgumentException</b>
   *      when <b>start</b> is less than 0 or start + length is greater
   *      than length of array <b>in</b><br>
   *      when <b>length</b> is less or equal than 0 or greater than 4
   */
  static public int bytes2Int( byte in[], int start, int length )
  {
    return bytes2Int( in, start, length, false );
  }

  /**
   * Converts sequence of bytes containing binary representation of signed
   * integer (in two's compliment format) to integer variable and returns
   * this variable.
   *
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   * @param <b>inverted</b> if <b>true</b> then first byte represent the least
   *                        significant byte
   *
   * @return
   *    Integer value represented by given bytes.
   *
   * @exceptions
   * <b>IllegalArgumentException</b>
   *      when <b>start</b> is less than 0 or <b>start + length</b> is
   *      greater than length of array <b>in</b><br>
   *      when <b>length</b> is less or equal than 0 or greater than 4
   */
  static public int bytes2Int( byte in[], int start, int length,
                               boolean inverted )
  {
    checkBytes2Int( in, start, length );
    return convertBytes2Int( in, start, length, inverted, false );
  }

  /**
   * Checks parameters are correct <b>bytes2Int</b> and <b>unsignedBytes2Int</b> methods.<br>
   * Throws {@link IllegalArgumentException} when,
   * <ul>
   *    <li><b>start</b> is less than 0 or <b>start + length</b> is</li>
   *    <li>greater than length of array <b>in</b></li>
   *    <li>when <b>length</b> is less or equal than 0 or greater than 4</li>
   * </ul>
   *
   * @param in       array of bytes
   * @param start    offset of start of first byte of sequence
   * @param length   number of bytes
   */
	static private void checkBytes2Int(byte in[], int start, int length) {
		if (length <= 0 || length > 4)
			throw new IllegalArgumentException("Length should be > 0 and less or equal than 4");
		if (start < 0)
			throw new IllegalArgumentException("Start should not be less than 0");
		if (start + length > in.length)
			throw new IllegalArgumentException("Start + length should not be greater than in.lenght()");
	}

	/**
	 * Converts sequence of bytes containing binary representation of signed (parameter <code>asUnsigned</code> is <b>false</b>), 
	 * in two's compliment format, or unsigned (parameter <code>asUnsigned</code> is <b>true<?b>) integer variable and returns this variable.
	 *
	 * @param <b>in</b> array of bytes
	 * @param <b>start</b> offset of start of first byte of sequence
	 * @param <b>length</b> number of bytes
	 * @param <b>inverted</b> if <b>true</b> then first byte represent the least significant byte
	 *
	 * @return Integer value represented by given bytes.
	 */
	static private int convertBytes2Int(byte in[], int start, int length, boolean inverted, boolean asUnsigned) {
		final int inc;
		int value = 0;
		int ind;
		boolean negative = false;
		if (!inverted) {
			ind = start;
			inc = 1;
			negative = !asUnsigned && (in[start] < 0);
		} else {
			ind = start + length - 1;
			inc = -1;
			negative = !asUnsigned && (in[start + length - 1] < 0);
		}
		for (int i = 0; i < length; i++) {
			value = (value << 8) | (in[ind] & LOW_BYTE);
			ind += inc;
		}
		if (negative && length < 4) {
			int shift = (4 - length) * 8;
			value <<= shift;
			value >>= shift;
		}
		return value;
	}

  /**
   * Checks parameters are correct <b>bytes2Long</b> and
   * <b>unsignedBytes2Long</b> methods.
   *
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   *
   * @return
   *    None
   *
   * @exceptions
   * <b>IllegalArgumentException</b>
   *      when <b>start</b> is less than 0 or <b>start + length</b> is
   *      greater than length of array <b>in</b><br>
   *      when <b>length</b> is less or equal than 0 or greater than 8
   */
  static private void checkBytes2Long( byte in[], int start, int length )
  {
    if( length <= 0 || length > 8 )
      throw new IllegalArgumentException(
                    "Length should be > 0 and less or equal than 8" );
    if( start < 0 )
      throw new IllegalArgumentException( "Start should not be less than 0" );
    if( start + length > in.length )
      throw new IllegalArgumentException(
                    "Start + length should be greater than length in *in*" );
  }

  /**
   * Converts sequence of bytes containing binary representation of signed
   * in two's compliment format.
   *
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   * @param <b>inverted</b> if <b>true</b> then first byte represent the least
   *                        significant byte
   *
   * @return
   *    Integer value represented by given bytes.
   */
  static public long bytes2Long( byte in[], int start, int length,
                                 boolean inverted )
  {
    checkBytes2Long( in, start, length );
    return convertBytes2Long( in, start, length, inverted, false );
  }

  /**
   * Converts sequence of bytes containing binary representation of signed
   * (parameter *asUnsigned* is *false*), in two's compliment format, or
   * unsigned (parameter *asUnsigned* is *true*)  variable and returns
   * this variable as long.
   *
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   * @param <b>inverted</b> if <b>true</b> then first byte represent the least
   *                        significant byte
   *
   * @return
   *    Integer value represented by given bytes.
   */
  static private long convertBytes2Long( byte in[], int start, int length,
                                         boolean inverted, boolean asUnsigned )
  {
    long value = 0;
    int ind;
    int inc;
    boolean negative = false;
    if( !inverted )
    {
      ind = start;
      inc = 1;
      negative = !asUnsigned && ( in[start] < 0 );
    }
    else
    {
      ind = start + length - 1;
      inc = -1;
      negative = !asUnsigned && ( in[start + length - 1] < 0 );
    }
    for( int i = 0; i < length; i++ )
    {
      value = ( value << 8 ) | ( in[ind] & LOW_BYTE );
      ind += inc;
    }
    if( negative && length < 8 )
    {
      int shift = ( 8 - length ) * 8;
      value <<= shift;
      value >>= shift;
    }
    return value;
  }

  /**
   * Converts sequence of 8 bytes containing binary representation of
   * double value to this double value and returns it.
   *
   * @param <b>data</b>     array of bytes
   * @param <b>offset</b>   offset of first byte of sequence
   *
   * @return
   *    Double value represented by given bytes.
   */
  static public double bytes2Double( byte data[], int offset )
  {
    return bytes2Double( data, offset, false );
  }

  /**
   * Converts sequence of 8 bytes containing binary representation of
   * double value to this double value and returns it.
   *
   * @param <b>data</b>     array of bytes
   * @param <b>offset</b>   offset of first byte of sequence
   * @param <b>inverted</b> if <b>true</b> then first byte represent the least
   *                        significant byte
   *
   * @return
   *    Double value represented by given bytes.
   */
  static public double bytes2Double( byte data[], int offset, boolean inverted )
  {
    return Double.longBitsToDouble( bytes2Long( data, offset, 8, inverted ) );
  }

  /**
   * Converts sequence of 4 bytes containing binary representation of
   * float value to this float value and returns it.
   *
   * @param <b>data</b>     array of bytes
   * @param <b>offset</b>   offset of first byte of sequence
   *
   * @return
   *    Float value represented by given bytes.
   */
  static public float bytes2Float( byte data[], int offset )
  {
    return bytes2Float( data, offset, false );
  }

  /**
   * Converts sequence of 4 bytes containing binary representation of
   * float value to this float value and returns it.
   *
   * @param <b>data</b>     array of bytes
   * @param <b>offset</b>   offset of first byte of sequence
   * @param <b>inverted</b> if <b>true</b> then first byte represent the least
   *                        significant byte
   *
   * @return
   *    Float value represented by given bytes.
   */
  static public float bytes2Float( byte data[], int offset, boolean inverted )
  {
    return Float.intBitsToFloat( bytes2Int( data, offset, 4, inverted ) );
  }

  /**
   * Converts sequence of bytes containing binary representation of unsigned
   * integer to integer variable and returns this variable.
   *
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   *
   * @return
   *    Non-negative integer value represented by given bytes.
   *
   * @exceptions
   * <b>IllegalArgumentException</b>
   *      when <b>start</b> is less than 0 or <b>start + length</b> is
   *      greater than length of array <b>in</b><br>
   *      when <b>length</b> is less or equal than 0 or greater than 4<br>
   *      when <b>length</b> is 4 and high order bit of high order byte is set
   */
  static public int unsignedBytes2Int( byte in[], int start, int length )
  {
    return unsignedBytes2Int( in, start, length, false );
  }

  /**
   * Converts sequence of bytes containing binary representation of unsigned
   * integer to integer variable and returns this variable.
   *
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   * @param <b>inverted</b> if <b>true</b> then first byte represent the least
   *                        significant byte
   *
   * @return
   *    Non-negative integer value represented by given bytes.
   *
   * @exceptions
   * <b>IllegalArgumentException</b>
   *      when <b>start</b> is less than 0 or <b>start + length</b> is
   *      greater than length of array <b>in</b><br>
   *      when <b>length</b> is less or equal than 0 or greater than 4<br>
   *      when <b>length</b> is 4 and high order bit of high order byte is set
   */
  static public int unsignedBytes2Int( byte in[], int start, int length,
                                       boolean inverted )
  {
    checkBytes2Int( in, start, length );
    int ind = start;
    if( inverted )
      ind = start + length - 1;
    if( length == 4 && in[ind] < 0 )
      throw new IllegalArgumentException( "Value too big to fit in integer" );
    return convertBytes2Int( in, start, length, inverted, true );
  }

  /**
   * Converts sequence of bytes containing binary representation of unsigned
   * integer to long variable and returns this variable.
   *
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   *
   * @return
   *    Non-negative long value represented by given bytes.
   *
   * @exceptions
   * <b>IllegalArgumentException</b>
   *      when <b>start</b> is less than 0 or <b>start + length</b> is
   *      greater than length of array <b>in</b><br>
   *      when <b>length</b> is less or equal than 0 or greater than 8<br>
   *      when <b>length</b> is 8 and high order bit of high order byte is set
   */
  static public long unsignedBytes2Long( byte in[], int start, int length )
  {
    return unsignedBytes2Long( in, start, length, false );
  }

  /**
   * Converts sequence of bytes containing binary representation of unsigned
   * integer to long variable and returns this variable.
   *
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   * @param <b>inverted</b> if <b>true</b> then first byte represent the least
   *                        significant byte
   *
   * @return
   *    Non-negative integer value represented by given bytes.
   *
   * @exceptions
   * <b>IllegalArgumentException</b>
   *      when <b>start</b> is less than 0 or <b>start + length</b> is
   *      greater than length of array <b>in</b><br>
   *      when <b>length</b> is less or equal than 0 or greater than 8<br>
   *      when <b>length</b> is 8 and high order bit of high order byte is set
   */
  static public long unsignedBytes2Long( byte in[], int start, int length,
                                         boolean inverted )
  {
    checkBytes2Long( in, start, length );
    int ind = start;
    if( inverted )
      ind = start + length - 1;
    if( length == 8 && in[ind] < 0 )
      throw new IllegalArgumentException( "Value too big to fit in long" );
    return convertBytes2Long( in, start, length, inverted, true );
  }

  /**
   * Packs boolean array into the long number and returns this long number.
   * Each boolean element occupies one bit starting from the most significant
   * in the long number. The rest are zeroed.<br>
   * If length of array greater than 64 then IllegalArgumentException will be
   * thrown.
   *
   * @param switches boolean[]
   * @return long
   */
  static public long packBooleanArrayToLong( boolean[] switches )
  {
    if( switches.length > 64 )
      throw new IllegalArgumentException( "length of switches > 64, (" +
                                          switches.length + ")" );
    byte[] array = FC.booleanArray2Bytes_left2right( switches );
    return FC.unsignedBytes2Long( array, 0, array.length );
  }

  /**
   * Unpack boolean array of given <code>length</code> from the given long
   * number <code>value</code>. First element of array is taken from the
   * most significant bit of long number and so on.<br>
   * If length of array greater than 64 then IllegalArgumentException will be
   * thrown.
   *
   * @param value long
   * @param length int
   * @return boolean[]
   */
  static public boolean[] getBooleanArrayFromLong( long value, int length )
  {
    if( length > 64 )
      throw new IllegalArgumentException( "length > 64, (" + length + ")" );
    int nBytes = ( length + 7 ) / 8;
    byte[] array = new byte[nBytes];
    FC.convertLong2Bytes( value, array, 0, array.length );
    return FC.bytes2BooleanArray_left2right( array, length );
  }

  /**
   * Copy integer on to sequence of bytes in such a manner that if the number
   * of bytes less than 4 then high order byte(s) of integer will be discarded.
   * If the number of bytes is greater than 4 then extraneous bytes
   * (high order bytes of byte sequence) will be fill with all 0-bit value or
   * all 1-bit value depending of given integer is >= 0 or < 0.
   *
   * @param <b>value</b>    integer value to copy into sequence of bytes
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   *
   * @return
   *    None
   */
  static public void convertInt2Bytes( int value, byte in[], int start,
                                       int length )
  {
    convertInt2Bytes( value, in, start, length, false );
  }

  /**
   * Copy integer on to sequence of bytes in such a manner that if the number
   * of bytes less than 4 then high order byte(s) of integer will be discarded.
   * If the number of bytes is greater than 4 then extraneous bytes
   * (high order bytes of byte sequence) will be fill with all 0-bit value or
   * all 1-bit value depending of given integer is >= 0 or < 0.
   * 
   * @param <b>value</b>    integer value to copy into sequence of bytes
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   * @param <b>inverted</b> if <b>true</b> then first byte represent the least
   *                        significant byte
   *
   * @return
   *    None
   */
  static public void convertInt2Bytes( int value, byte in[], int start,
                                       int length, boolean inverted )
  {
    int ind;
    int inc;
    byte extraneousByteValue = ( value >= 0 ) ? (byte)0 : (byte)0xff;
    if( !inverted )
    {
      ind = start + length - 1;
      inc = -1;
    }
    else
    {
      ind = start;
      inc = 1;
    }
    for( int i = 0; i < length; i++ )
    {
      if( i < 4 )
      {
        in[ind] = (byte)( value & LOW_BYTE );
        value >>= 8;
      }
      else
        in[ind] = extraneousByteValue;
      ind += inc;
    }
  }

  /**
   * Copy long on to sequence of bytes in such a manner that if the number
   * of bytes less than 8 then high order byte(s) of integer will be discarded.
   * If the number of bytes is greater than 8 then extraneous bytes
   * (high order bytes of byte sequence) will be fill with all 0-bit value or
   * all 1-bit value depending of given integer is >= 0 or < 0.
   *
   * @param <b>value</b>    integer value to copy into sequence of bytes
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   *
   * @return
   *    None
   */
  static public void convertLong2Bytes( long value, byte in[], int start,
                                        int length )
  {
    convertLong2Bytes( value, in, start, length, false );
  }

  /**
   * Copy long on to sequence of bytes in such a manner that if the number
   * of bytes less than 8 then high order byte(s) of integer will be discarded.
   * If the number of bytes is greater than 8 then extraneous bytes
   * (high order bytes of byte sequence) will be fill with all 0-bit value or
   * all 1-bit value depending of given integer is >= 0 or < 0.
   *
   * @param <b>value</b>    long value to copy into sequence of bytes
   * @param <b>in</b>       array of bytes
   * @param <b>start</b>    offset of start of first byte of sequence
   * @param <b>length</b>   number of bytes
   * @param <b>inverted</b> if <b>true</b> then first byte represent the least
   *                        significant byte
   *
   * @return
   *    None
   */
  static public void convertLong2Bytes( long value, byte in[], int start,
                                        int length, boolean inverted )
  {
    int ind;
    int inc;
    byte extraneousByteValue = ( value >= 0 ) ? (byte)0 : (byte)0xff;
    if( !inverted )
    {
      ind = start + length - 1;
      inc = -1;
    }
    else
    {
      ind = start;
      inc = 1;
    }
    for( int i = 0; i < length; i++ )
    {
      if( i < 8 )
      {
        in[ind] = (byte)( value & LOW_BYTE );
        value >>= 8;
      }
      else
        in[ind] = extraneousByteValue;
      ind += inc;
    }
  }

  /**
   * Copy double on to sequence of bytes.
   *
   * @param <b>value</b>    double value to copy into sequence of bytes
   * @param <b>data</b>     array of bytes
   * @param <b>offset</b>   offset of first byte for writing to
   *
   * @return
   *    None
   */
  static public void convertDouble2Bytes( double value, byte data[], int offset )
  {
    convertDouble2Bytes( value, data, offset, false );
  }

  /**
   * Copy double on to sequence of bytes.
   *
   * @param <b>value</b>    double value to copy into sequence of bytes
   * @param <b>data</b>     array of bytes
   * @param <b>offset</b>   offset of first byte for writing to
   * @param <b>inverted</b> if <b>true</b> then first byte represent the least
   *                        significant byte
   *
   * @return
   *    None
   */
  static public void convertDouble2Bytes( double value, byte data[], int offset,
                                          boolean inverted )
  {
    FC.convertLong2Bytes( Double.doubleToLongBits( value ), data, offset,
                          8, inverted );
  }

  /**
   * Copy float on to sequence of bytes.
   *
   * @param <b>value</b>    float value to copy into sequence of bytes
   * @param <b>data</b>     array of bytes
   * @param <b>offset</b>   offset of first byte for writing to
   *
   * @return
   *    None
   */
  static public void convertFloat2Bytes( float value, byte data[], int offset )
  {
    convertFloat2Bytes( value, data, offset, false );
  }

  /**
   * Copy float on to sequence of bytes.
   *
   * @param <b>value</b>    float value to copy into sequence of bytes
   * @param <b>data</b>     array of bytes
   * @param <b>offset</b>   offset of first byte for writing to
   * @param <b>inverted</b> if <b>true</b> then first byte represent the least
   *                        significant byte
   *
   * @return
   *    None
   */
  static public void convertFloat2Bytes( float value, byte data[], int offset,
                                         boolean inverted )
  {
    FC.convertInt2Bytes( Float.floatToIntBits( value ), data, offset, 4,
                         inverted );
  }

  /**
   * Convert byte array of not more than 8 element to one byte
   * using standard positional notation with lsb being rightmost
   *
   * @param values boolean[]
   * @return byte
   */
  static public byte booleanArray2Byte( boolean[] values )
  {
    byte result = 0;
    int len = Math.min( values.length, 8 );
    for( int i = 0; i < len; i++ )
    {
      if( values[i] )
        result |= ( 1 << i );
    }
    return result;
  }

  /**
   * Represent an 8 bit value as array of booleans
   * using standard positional notation with lsb being the rightmost
   *
   * @param value byte
   * @return boolean[]
   */
  static public boolean[] byte2BooleanArray( byte value )
  {
    boolean[] result = new boolean[8];
    for( int i = 0; i < 8; i++ )
    {
      if( (value&1) != 0 )
        result[i] = true;
      value >>= 1;
    }
    return result;
  }

  /**
   * Create byte array of minimum length that can fit values.length bits and,
   * considering it as bit array, set individual bits according with boolean
   * values in array *values*. Set bits starting from left.
   *
   * @param values boolean[]
   * @return byte[]
   */
  static public byte[] booleanArray2Bytes_left2right( boolean[] values )
  {
    int bitQty = values.length;
    int byteQty = ( bitQty + 7 ) / 8;
    byte[] array = new byte[byteQty];
    int byteIndex = -1;
    int bitIndex = 7;

    for( int i = 0; i < bitQty; i++ )
    {
      bitIndex++;
      if( bitIndex >= 8 )
      {
        bitIndex = 0;
        byteIndex++;
        array[byteIndex] = 0;
      }
      if( values[i] )
        array[byteIndex] |= ( 1 << bitIndex );
    }
    return array;
  }

  /**
   * Create boolean array of given length *length* and, considering byte array
   * as bit array, set values of boolean array with according to individual
   * bit values in bit array *array*. Bits starting from left.
   *
   * @param array byte[]
   * @param length
   *
   * @return boolean[]
   */
  static public boolean[] bytes2BooleanArray_left2right( byte[] array, int length )
  {
    if( length > array.length * 8 )
      throw new IllegalArgumentException( "FC.bytes2BooleanArray: length is " +
          "larger than number of bits in given array" );
    boolean[] values = new boolean[length];
    int byteIndex = -1;
    int bitIndex = 7;
    byte arrElement = 0;

    for( int i = 0; i < length; i++ )
    {
      bitIndex++;
      if( bitIndex >= 8 )
      {
        bitIndex = 0;
        byteIndex++;
        arrElement = array[byteIndex];
      }
      else
        arrElement >>= 1;
      if( ( arrElement & 0x01 ) != 0 )
        values[i] = true;
    }
    return values;
  }

  static public byte asciiToNibble( byte in )
  {
    return (byte)( in > '9' ? ( in - 'A' + 10 ) : ( in -'0' ) );
  }

  static public byte BCD2toByte( byte in[], int offset )
  {
    return (byte)( high_nibble( in[offset] ) * 10 + low_nibble( in[offset] ) );
  }

  static public byte byteToBCD2( byte in )
  {
    byte result = (byte)(in%10);
    byte tens = (byte)((in/10)%10);
    result += (tens<<4);
    return result;
  }

  static public short BCD3stoShort(byte in[], int offset)
  {
    a_short = (short)
              (low_nibble( in[offset] )*100 +
               high_nibble( in[offset+1] )*10 +
               low_nibble( in[offset+1] ));
    if( high_nibble( in[offset] )==9 )
      return (short)-a_short;
    else
      return a_short;
  }

  static public short BCD4toShort(byte in[], int offset)
  {
    return ( (short)(
             high_nibble( in[offset] )*1000 +
             low_nibble( in[offset] )*100 +
             high_nibble( in[offset+1] )*10 +
             low_nibble( in[offset+1] ) ));
  }

  static public short shortToBCD4( short in )
  {
    short result = (short)(in%10);
    byte tens = (byte)((in/10)%10);
    byte hundreds = (byte)((in/100)%10);
    byte thousands = (byte)((in/1000)%10);
    result += (short)((tens<<4)+(hundreds<<8)+(thousands<<12));
    return result;
  }

  static public void shortToBCD4( short number, byte out[], int offset )
  {
    // preface style ( high digit first )
    int d = 100;
    for( int i = 0; i< 2; i++ )
    {
      out[offset+i] = byteToBCD2( (byte)((number / d )%100));
      d /= 100;
    }
  }

  static public int BCD6toInt(byte in[], int offset)
  {
    return ( high_nibble( in[offset] )*100000 +
             low_nibble( in[offset] )*10000 +
             high_nibble( in[offset+1] )*1000 +
             low_nibble( in[offset+1] )*100 +
             high_nibble( in[offset+2] )*10 +
             low_nibble( in[offset+2] ) );
  }

  static public void intToBCD6(int number, byte out[], int offset)
  {
    // preface style ( high digit first )
    int d = 10000;
    for( int i = 0; i< 3; i++ )
    {
      out[offset+i] = byteToBCD2( (byte)((number / d )%100));
      d /= 100;
    }
  }

  // takes two ascii chars using 0..9,A..F and makes one BCD2 byte
  // used for conversion of SAO preface "FF" to DataPassport.preface[]

  static public byte asciiToBCD2( byte inUpperChar, byte inLowerChar )
  {
    return byteToBCD2( (byte)(asciiToNibble(inUpperChar)*10 + asciiToNibble(inLowerChar)) );
  }

  static public double AWStoDouble(byte in[], int offset)
  {
    int value, exp10, shift;

    value = (int)BCD4toShort( in, offset );
    exp10 = (int)low_nibble( in[offset+2] );
    shift = (int)high_nibble( in[offset+2] );

    if( shift!= 0 )
    {
        if( shift<9 )
            value = -value;
        if( shift==7 || shift==9)
            exp10 = -exp10;
    }
    exp10 -= 3;
    return (Math.pow( 10., (double)exp10 ) * value);
  }

  /**
   * Translates string value, <code>str</code> to long value and returns this long value.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns 0.
   */
  static public long StringToLong( String str )
  {
    return StringToLong( str, 0 );
  }

  /**
   * Translates string value, <code>str</code>, to long value and returns this long value.<br>
   * if <code>width</code> &gt; 0 and <code>width</code> &lt; <i>length_of_<code>str</code></i>,
   * then only 1st <code>width</code> characters of string <code>str</code> will be 
   * translated, otherwise the whole string <code>str</code> will be translated.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns 0.
   */
  static public long StringToLong( String str, int width )
  {
    return StringToLong( str, width, 0 );
  }

  /**
   * Translates string value, <code>str</code>, to long value and returns this long value.<br>
   * if <code>width</code> &gt; 0 and <code>width</code> &lt; <i>length_of_<code>str</code><i>,
   * then only 1st <code>width</code> characters of string <code>str</code> will be
   * translated, otherwise the whole string <code>str</code> will be translated.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns <code>defaultValue>.
   */
  static public long StringToLong( String str, int width, long defaultValue )
  {
    long returnVal;
    if( width > 0  && width < str.length() )
      str = str.substring( 0, width );
    str = str.trim();
    try
    {
      returnVal = Long.parseLong( str );
    }
    catch( NumberFormatException e )
    {
      returnVal = defaultValue;
    }
    return returnVal;
  }

  /**
   * Translates string value, <code>str</code>, to long value and returns this long value.<br>
   * It differs from StringToLong in accepting float/scientific notification
   * and then casting given double number to long.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns 0.
   */
  static public long LenientStringToLong( String str )
  {
    return LenientStringToLong( str, 0 );
  }

  /**
   * Translates string value, <code>str</code>, to long value and returns this long value.<br>
   * if <code>width</code> &gt; 0 and <code>width</code> &lt; <i>length_of_<code>str</code></i>,
   * then only 1st <code>width</code> characters of string <code>str</code> will be translated,
   * otherwise the whole string <code>str</code> will be translated.<br>
   * It differs from StringToLong in accepting float/scientific notification
   * and then casting given double number to long.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns 0.
   */
  static public long LenientStringToLong( String str, int width )
  {
    return LenientStringToLong( str, width, 0 );
  }

  /**
   * Translates string value, <code>str</code>, to long value and returns this long value.<br>
   * if <code>width</code> &gt; 0 and <code>width</code> &lt; <i>length_of_<code>str</code></i>,
   * then only 1st <code>width</code> characters of string <code>str</code> will be translated,
   * otherwise the whole string <code>str</code> will be translated.<br>
   * It differs from StringToLong in accepting float/scientific notification
   * and then casting given double number to long.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns <code>defaultValue</code>.
   */
  static public long LenientStringToLong( String str, int width, long defaultValue )
  {
    return (long)StringToDouble( str, width, defaultValue );
  }

  /**
   * Translates hexadecimal string, <code>str</code>, to integer value and returns 
   * this integer value.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns -1.
   */
  static public int hexStr2Int( String str )
  {
    return hexStr2Int( str, str.length() );
  }

  /**
   * Translates hexadecimal string, <code>str</code>, to integer value and returns 
   * this integer value.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns -1.
   */
  static public int hexStr2Int( String str, int length )
  {
    return hexStr2Int( str, length, ORDER_NORMAL );
  }

  /**
   * Translates given hexadecimal string, <code>str</code>, to integer value and 
   * returns this integer value.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns -1.
   */
  static public int hexStr2Int( String str, int length, int order )
  {
    int value = 0;
    char symbol;
    int digit;
    int pos = 0;
    int inc = 1;
    str = str.toUpperCase();
    // order == ORDER_NORMAL   : ABCD -> ABCD
    // order == ORDER_INVERTED : ABCD -> DCBA
    if( order == ORDER_INVERTED )
    {
      pos = length - 1;
      inc = -1;
    }

    if( length > str.length() )
      value = -1;
    else if( length > 8 ) // number is larger than maximum integer
      value = -1;
    else if ( length == 8 )
    {
      symbol = str.charAt( pos );
      digit = HEX_DIGITS.indexOf( symbol );
      if( digit < 0 ) // illegal symbol
        value = -1;
      else if( digit > 7 ) // negative
        value = -1;
      else
      {
        value = digit;
        pos += inc;
        length--;
      }
    }
    if( value >= 0 )
    {
      for( int k = 0; k < length; k++ )
      {
        symbol = str.charAt( pos );
        digit = HEX_DIGITS.indexOf( symbol );
        if( digit < 0 )
        {
          value = -1;
          break;
        }
        value = 16 * value + digit;
        pos += inc;
      }
    }
    return value;
  }

  /**
   * Translates hexadecimal string, <code>str</code>, to long value and returns 
   * this long value.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns -1.
   */
  static public long hexStr2Long( String str )
  {
    return hexStr2Long( str, str.length() );
  }

  /**
   * Translates hexadecimal string, <code>str</code>, to long value and returns 
   * this long value.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns -1.
   */
  static public long hexStr2Long( String str, int length )
  {
    return hexStr2Long( str, length, ORDER_NORMAL );
  }

  /**
   * Translates given hexadecimal string, <code>str</code>, to long value and 
   * returns this long value.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns -1.
   */
  static public long hexStr2Long( String str, int length, int order )
  {
    long value = 0;
    char symbol;
    int digit;
    int pos = 0;
    int inc = 1;
    str = str.toUpperCase();
    // order == ORDER_NORMAL   : ABCD -> ABCD
    // order == ORDER_INVERTED : ABCD -> DCBA
    if( order == ORDER_INVERTED )
    {
      pos = length - 1;
      inc = -1;
    }
    if( length > str.length() )
      value = -1L;
    else if( length > 16 ) // number is larger than maximum integer
      value = -1L;
    else if ( length == 16 )
    {
      symbol = str.charAt( pos );
      digit = HEX_DIGITS.indexOf( symbol );
      if( digit < 0 ) // illegal symbol
        value = -1L;
      else if( digit > 7 ) // negative
        value = -1L;
      else
      {
        value = digit;
        pos += inc;
        length--;
      }
    }
    if( value >= 0 )
    {
      for( int k = 0; k < length; k++ )
      {
        symbol = str.charAt( pos );
        digit = HEX_DIGITS.indexOf( symbol );
        if( digit < 0 )
        {
          value = -1L;
          break;
        }
        value = 16 * value + digit;
        pos += inc;
      }
    }
    return value;
  }

  /**
   * Translates string value, <code>str</code>, to integer value and returns this 
   * integer value.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns 0.
   */
  static public int StringToInteger( String str )
  {
    return (int)StringToLong( str );
  }

  /**
   * Translates string value, <code>str</coder>, to integer value and returns this 
   * integer value.<br>
   * if <code>width</code> &gt; 0 and <code>width</code> &lt; <i>length_of_<code>str</code></i>,
   * then only 1st <code>width</code> characters of string <code>str</code> will be 
   * translated, otherwise the whole string <code>str</code> will be translated.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns 0.
   */
  static public int StringToInteger( String str, int width )
  {
    return (int)StringToLong( str, width );
  }

  /**
   * Translates string value, <code>str</code>, to integer value and returns this 
   * integer value.<br>
   * if <code>width</code> &gt; 0 and <code>width</code> &lt; <i>length_of_<code>str</code></i>, 
   * then only 1st <code>width</code> characters of string <code>str</code> will be 
   * translated, otherwise the whole string <code>str</code> will be translated.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns <code>defaultValue</code>.
   */
  static public int StringToInteger( String str, int width, int defaultValue )
  {
    return (int)StringToLong( str, width, defaultValue );
  }

  /**
   * Translates string value, <code>str</code>, to integer value and returns this
   * integer value.<br>
   * It differs from {@link #StringToInteger(String)} in accepting float/scientific notification
   * and then casting given double number to integer.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns 0.
   */
  static public int LenientStringToInteger( String str )
  {
    return (int)LenientStringToLong( str );
  }

  /**
   * Translates string value, <code>str</code>, to integer value and returns this
   * integer value.<br>
   * if <code>width</code> &gt; 0 and <code>width</code> &lt; <i>length_of_<code>str</code></i>,
   * then only 1st <code>width</code> characters of string <code>str</code> will be translated,
   * otherwise the whole string <code>str</code> will be translated.<br>
   * It differs from {@link #StringToInteger(String, int)} in accepting float/scientific
   * notification and then casting given double number to integer.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns 0.
   */
  static public int LenientStringToInteger( String str, int width )
  {
    return (int)LenientStringToLong( str, width );
  }

  /**
   * Translates string value, <code>str</code>, to integer value and returns this
   * integer value.<br>
   * if <code>width</code> &gt; 0 and <code>width</code> &lt; <i>length_of_<code>str</code></i>,
   * then only 1st <code>width</code> characters of string <code>str</code> will be translated,
   * otherwise the whole string <code>str</code> will be translated.<br>
   * It differs from {@link #StringToInteger(String, int, int)} in accepting float/scientific
   * notification and then casting given double number to integer.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns <code>defaultValue</code>.
   */
  static public int LenientStringToInteger( String str, int width, int defaultValue )
  {
    return (int)LenientStringToLong( str, width, defaultValue );
  }

  /**
   * Translates string value, <code>str</code>, to double value and returns this 
   * double value.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns 0.
   */
  static public double StringToDouble( String str )
  {
    return StringToDouble( str, 0 );
  }

  /**
   * Translates string value, <code>str</code>, to double value and returns this
   * double value.<br>
   * if <code>width</code> &gt; 0 and <code>width</code> &lt; <i>length_of_<code>str</code></i>,
   * then only 1st <code>width</code> characters of string <code>str</code> will be translated,
   * otherwise the whole string <code>str</code> will be translated.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns 0.
   */
  static public double StringToDouble( String str, int width )
  {
    return StringToDouble( str, width, 0 );
  }

  /**
   * Translates string value, <code>str</code>, to double value and returns this
   * double value.<br>
   * if <code>width</code> &gt; 0 and <code>width</code> &lt; <i>length_of_<code>str</code></i>,
   * then only 1st <code>width</code> characters of string <code>str</code> will be translated,
   * otherwise the whole string <code>str</code> will be translated.<br>
   * If some error occurred (for example, illegal characters in string)
   * then it returns <code>defaultValue</code>.
   */
  static public double  StringToDouble( String str, int width, double defaultValue )
  {
    double returnVal;
    if( width > 0  &&  width < str.length() )
      str = str.substring( 0, width );
    str = str.trim();
    try
    {
      returnVal = Double.valueOf( str ).doubleValue();
    }
    catch( NumberFormatException e )
    {
      returnVal = defaultValue;
    }
    return returnVal;
  }

  static public String IntegerToString( int value )
  {
    return Integer.toString( value );
  }

  static public String IntegerToString( int value, int width )
  {
    return IntegerToString( value, width, ' ' );
  }

  static public boolean IntegerToString( int value, int width,
                                         StringBuilder strBuf, int offset )
  {
    return IntegerToString( value, width, ' ', strBuf, offset );
  }

  static public String IntegerToString( int value, int width, char fill )
  {
    StringBuilder strBuf = new StringBuilder( width );
    strBuf.setLength( width );
    IntegerToString( value, width, fill, strBuf, 0 );
    return strBuf.toString();
  }

  /**
   * This method encodes integer given <code>value</code> into given 
   * <code>StringBuilder</code> buffer, <code>strBuf</code>.<br>
   * This value will be placed into <code>strBuf</code> region that starts 
   * from given <code>offset</code> and has given <code>width</code>. Encoded 
   * value will be right justified in that region and unused characters of this 
   * region will be replaced by given fill character, <code>fill</code>.<br>
   * If <code>value</code> does not fit into given <code>width</code>, then 
   * region will be filled up with symbol * (asterisk) and <b>false</b> will 
   * be returned, otherwise <b>true</b> will be returned.
   */
  static public boolean IntegerToString( int value, int width, char fill,
                                         StringBuilder strBuf, int offset )
  {
    boolean overflow = false;
    if( width <= 0 )
      throw new IllegalArgumentException( "width is not positive, " + width );
    if( strBuf.length() < offset + width )
      throw new IllegalArgumentException( "strBuf size is small, " +
          strBuf.length() + ", offset = " + offset + ", width = " + width );
    boolean isNegative = false;
    int numberOfFreeDigitPositions = width;
    if( value < 0 )
    {
      value = -value;
      isNegative = true;
      numberOfFreeDigitPositions--;
    }
    int numberOfSymbolsInWholePartSoFar = 0;
    while( value > 0  &&  numberOfFreeDigitPositions-- > 0 )
    {
      numberOfSymbolsInWholePartSoFar++;
      strBuf.setCharAt( offset + width - numberOfSymbolsInWholePartSoFar,
                       (char)( '0' + value % 10 ) );
      value /= 10;
    }
    if( value > 0 )
    {
      for( int i = offset; i < offset + width; i++ )
        strBuf.setCharAt( i, '*' );
      overflow = true;
    }
    else
    {
      if( numberOfSymbolsInWholePartSoFar == 0 )
      {
        numberOfSymbolsInWholePartSoFar++;
        strBuf.setCharAt(offset + width - numberOfSymbolsInWholePartSoFar, '0');
      }
      if( isNegative )
      {
        numberOfSymbolsInWholePartSoFar++;
        strBuf.setCharAt(offset + width - numberOfSymbolsInWholePartSoFar, '-');
      }
      while( numberOfSymbolsInWholePartSoFar < width )
      {
        numberOfSymbolsInWholePartSoFar++;
        strBuf.setCharAt(offset + width - numberOfSymbolsInWholePartSoFar, fill);
      }
    }
    return !overflow;
  }

  static public String int2HexStr( int value, int width )
  {
	return String.format("%0" + width + "X", value);
    /*String str = Integer.toHexString( value ).toUpperCase();
    int len = str.length();
    if( len > width )
      return str.substring( len - width );
    if( len == width )
      return str;
    return C.STR_ZEROES.substring( 0, width - len ) + str;*/
  }

  static public String long2HexStr( long value, int width )
  {
    String str = Long.toHexString( value ).toUpperCase();
    int len = str.length();
    if( len > width )
      return str.substring( len - width );
    if( len == width )
      return str;
    return STR_ZEROES.substring( 0, width - len ) + str;
  }

  static public String bytes2HexStr( byte[] bytes )
  {
    if( bytes.length == 0 )
      throw new IllegalArgumentException( "array bytes is zero length" );
    return bytes2HexStr( bytes, 0, bytes.length );
  }

  static public String bytes2HexStr( byte[] bytes, int start )
  {
    if( start < 0 || start >= bytes.length )
      throw new IllegalArgumentException( "start is illegal, " + start );
    return bytes2HexStr( bytes, start, bytes.length - start );
  }

  static public String bytes2HexStr( byte[] bytes, int start, int length )
  {
    if( start < 0 || length <= 0 || start + length > bytes.length )
      throw new IllegalArgumentException( "start, length are illegal or " +
                                          "inconsistent with length of bytes" );
    int end = start + length;
    StringBuilder str = new StringBuilder( 2 * length );
    byte c;
    char hi;
    char low;
    for( int i = start; i < end; i++ )
    {
      c = bytes[i];
      low = HEX_DIGITS.charAt( c & 0x0f );
      hi = HEX_DIGITS.charAt( ( c >> 4 ) & 0x0f );
      str.append( hi );
      str.append( low );
    }
    return str.toString();
  }

  static public String DoubleToString( double value, int width, int precision )
  {
    StringBuilder strBuf = new StringBuilder( width );
    strBuf.setLength( width );
    DoubleToString( value, width, precision, strBuf, 0 );
    return strBuf.toString();
  }

  /**
   * This method encodes given double <code>value</code> into StringBuilder 
   * buffer, <code>strBuf</code>.<br>
   * This value will be placed into <code>strBuf</code> region that starts 
   * from given <code>offset</code> and has given <code>width</code>. Encoded 
   * value will be right justified in this region and unused characters of 
   * this region will be replaced by given fill character, <code>fill</code>.<br>
   * If <code>value</code> does not fit into given <code>width</code>, then 
   * region will be filled up with symbol *, asterisk (symbol . will not be 
   * replaced with *), and <b>false</b> will be returned, otherwise 
   * <b>true</b> will be returned.
   */
  static public boolean DoubleToString( double value, int width, int precision,
                                        StringBuilder strBuf, int offset )
  {
    boolean overflow = false;
    if( width <= 0 )
      throw new IllegalArgumentException( "width is not positive, " + width );
    if( precision > width - 1 )
      throw new IllegalArgumentException(
        "precision > width - 1, width = " + width + ", precision = " + precision );
    if( strBuf.length() < offset + width )
      throw new IllegalArgumentException( "strBuf size is small, " +
          strBuf.length() + ", *offset* = " + offset + ", *width* = " + width );
    int wholePartWidth;
    int fractionPartOffset;
    int fractionPartWidth;
    if( precision < 0 )
    {
      precision = 0;
      wholePartWidth = width;
      fractionPartOffset = offset + width;
      fractionPartWidth = 0;
    }
    else
    {
      wholePartWidth = width - precision - 1;
      strBuf.setCharAt( offset + wholePartWidth, '.' );
      fractionPartOffset = offset + wholePartWidth + 1;
      fractionPartWidth = precision;
    }

    long wholePart, fractionPart;
    int leftToPoint = wholePartWidth;
    boolean isNegative = false;
    if( value < 0 )
    {
      value = -value;
      isNegative = true;
      leftToPoint--;
    }
    wholePart    = (long)Math.floor( value );
    fractionPart = Math.round(Math.pow(10.0, precision) * (value - wholePart) );
    while( precision-- > 0 )
    {
      if( fractionPart > 0 )
      {
        strBuf.setCharAt( fractionPartOffset + precision,
                          (char)( '0' + fractionPart % 10 ) );
        fractionPart /= 10;
      }
      else
        strBuf.setCharAt( fractionPartOffset + precision, '0' );
    }
    wholePart += fractionPart;   // usually fractionPart is 0 here, but sometimes is 1 :)

    int numberOfSymbolsInWholePartSoFar = 0;

    while( wholePart > 0  &&  leftToPoint-- > 0 )
    {
      numberOfSymbolsInWholePartSoFar++;
      strBuf.setCharAt(
          offset + wholePartWidth - numberOfSymbolsInWholePartSoFar,
          (char)( '0' + wholePart % 10 ) );
      wholePart /= 10;
    }
    if( wholePart > 0 )
    {
      for( int i = offset; i < offset + wholePartWidth; i++ )
        strBuf.setCharAt( i, '*' );
      for( int i = fractionPartOffset;
           i < fractionPartOffset + fractionPartWidth; i++ )
        strBuf.setCharAt( i, '*' );
      overflow = true;
    }
    else
    {
      if( numberOfSymbolsInWholePartSoFar == 0 && leftToPoint > 0 )
      {
        numberOfSymbolsInWholePartSoFar++;
        strBuf.setCharAt( offset + wholePartWidth -
                          numberOfSymbolsInWholePartSoFar, '0' );
      }
      if( isNegative )
      {
        numberOfSymbolsInWholePartSoFar++;
        strBuf.setCharAt( offset + wholePartWidth -
                          numberOfSymbolsInWholePartSoFar, '-' );
      }
      while( numberOfSymbolsInWholePartSoFar < wholePartWidth )
      {
        numberOfSymbolsInWholePartSoFar++;
        strBuf.setCharAt( offset + wholePartWidth -
                          numberOfSymbolsInWholePartSoFar, ' ' );
      }
    }
    return !overflow;
  }

  /**
   * Converts given double <code>value</code> to decimal string rounded to
   * not more than given <code>maxPrecision</code> decimals positions after dot.<br>
   * For example, if <code>maxPrecision</code> is 2, then
   * <ul>
   *    <code>5.756</code> will be printed as <code>5.76</code>
   * </ul>   
   */
  static public String doubleToString( double value, int maxPrecision )
  {
    return doubleToString( value, maxPrecision, false );
  }

  /**
   * Converts given double <code>value</code> to decimal string rounded to
   * not more than given <code>maxPrecision</code> decimals positions after dot.<br>
   * For example, if <code>maxPrecision</code> is 2, then
   * <ul>
   *    <code>5.756</code> will be printed as <code>5.76</code>
   * </ul>   
   * Additionally, if <code>discardTrailingZeroes</code> is <b>true</b> then
   * <ul>
   *   <li><code>5.2</code> will be printed as <code>5.2</code> (but not as <code>5.20</code>)</li>
   *   <li><code>5</code> will be printed as <code>5</code> (but not as <code>5.00</code>)</li>
   * </ul>   
   */
  static public String doubleToString( double value, int maxPrecision,
                                       boolean discardTrailingZeroes )
  {
    int mult = (int)Math.pow( 10, maxPrecision );
    value = ( (double)Math.round( mult * value ) ) / mult;
    String str = "" + value;
    str = str.toUpperCase();
    int ePos = str.indexOf( 'E' );
    int pointPos;
    if( ePos > 0 )
    {
      int shift = StringToInteger( str.substring( ePos + 1 ) );
      str = str.substring( 0, ePos );
      pointPos = str.indexOf( '.' );
      if( pointPos >= 0 )
      {
        if( shift > 0 )
        {
          if( str.length() - pointPos - 1 > shift )
            str = str.substring( 0, pointPos ) +
                  str.substring( pointPos + 1, pointPos + 1 + shift ) + "." +
                  str.substring( pointPos + 1 + shift );
          else if( str.length() - pointPos - 1 < shift )
            str = str.substring( 0, pointPos ) + str.substring( pointPos + 1 ) +
                  STR_ZEROES.substring( 0,
                      shift - ( str.length() - pointPos - 1 ) );
          else
            str = str.substring( 0, pointPos ) + str.substring( pointPos + 1 );
        }
        else if( shift < 0 )
        {
          if( -shift < pointPos )
            str = str.substring( 0, pointPos + shift ) + "." +
                  str.substring( pointPos + shift, pointPos ) +
                  str.substring( pointPos + 1 );
          else
            str = "0." + STR_ZEROES.substring( 0, -shift - pointPos ) +
                  str.substring( 0, pointPos ) +
                  str.substring( pointPos + 1 );
        }
      }
      else
      {
        if( shift > 0 )
        {
          str = str + STR_ZEROES.substring( 0, shift );
        }
        else if( shift < 0 )
        {
          if( -shift < str.length() )
            str = str.substring( 0, str.length() + shift ) + "." +
                  str.substring( str.length() + shift );
          else
            str = "0." + STR_ZEROES.substring( 0, -shift - str.length() ) + str;
        }
      }
    }
    pointPos = str.indexOf( '.' );
    if( pointPos >= 0 )
    {
      if( maxPrecision <= 0 )
        str = str.substring( 0, pointPos );
      else if( str.length() - 1 - pointPos > maxPrecision )
        str = str.substring( 0, pointPos + 1 + maxPrecision );
      else if( !discardTrailingZeroes &&
               str.length() - 1 - pointPos < maxPrecision )
        str = padRight( str, pointPos + 1 + maxPrecision, '0' );
      if( maxPrecision > 0 && discardTrailingZeroes )
      {
        int trZerPos = -1;
        for( int i = str.length() - 1; i > pointPos; i-- )
        {
          if( str.charAt( i ) == '0' )
            trZerPos = i;
          else
            break;
        }
        if( trZerPos > 0 )
          str = str.substring( 0, trZerPos );
        if( str.length() > 0 )
        {
          if( str.charAt( str.length() - 1 ) == '.' )
            str = str.substring( 0, str.length() - 1 );
        }
        if( str.length() == 0 )
          str = "0";
      }
    }
    else if( maxPrecision > 0 )
      str += "." + StrUtil.repeat( '0', maxPrecision );
    return str;
  }

  static public String ScientificToString( double value, int width,
                                           int precision, int expWidth )
  {
    int power = 0;
    double val = Math.abs( value );
    while( val >= 1.0 )
    {
      val = val / 10;
      power++;
    }
    while( val < 0.1 && val != 0.0 )
    {
      val = val * 10;
      power--;
    }

    if( value < 0 )
      val = -val;

    String str = DoubleToString( val, width - expWidth - 2, precision );

    str += "E";

    if( power < 0 )
      str += "-";
    else
      str += "+";

    str += IntegerToString( Math.abs( power ), expWidth ).replace( ' ', '0' );

    return str;
  }

  static public String nowToString()
  {
    SimpleDateFormat formatter = new SimpleDateFormat( "yyyy-MMM-dd HH.mm.ss " );
    Date now = new Date();
    String str = formatter.format( now );
    return str;
  }

  static public String secToTimestamp( int tsec )
  {
    int hour = tsec/3600;
    int min = (tsec%3600)/60;
    int sec = tsec%60;
    return hour + ":" + IntegerToString(min,2).replace(' ', '0' ) + "." +
                        IntegerToString(sec,2).replace(' ', '0' );
  }

  static public int swapBytes( int oldValue )
  {
    int newValue = (oldValue&0xff)<<24;
    newValue += (oldValue&0xff00)<<8;
    newValue += (oldValue&0xff0000)>>8;
    newValue += ((oldValue&0xff000000)>>24)&0xFF;
    return newValue;
  }

  static public short swapBytes( short oldValue )
  {
    short newValue = (short)((oldValue&0xff)<<8);
    newValue += ((oldValue&0xff00)>>8)&0xFF;
    return newValue;
  }

  static public float floatLE( int oldValue )
  {
    float newValue = Float.intBitsToFloat( swapBytes(oldValue) );
    return newValue;
  }

  static public short swapBytes( byte byte1, byte byte2 )
  {
    return (short) ((byte2<<8)|byte1);
  }

  static public long intToLong( int hi, int lo )
  {
    return ( ((long)hi) << 32 ) | ( ((long)lo) & 0xffffffffL );
  }

  static public int longToHi( long val )
  {
    return (int)( val >>> 32 );
  }

  static public int longToLo( long val )
  {
    return (int)val;
  }

  static public String doubleArrayToString( double[] array, int width, int fraction )
  {
    return doubleArrayToString( array, array.length, width, fraction );
  }

  static public String doubleArrayToString( double[] array, int maxIndex, int width, int fraction )
  {
    String str = "";
    for( int i = 0; i < maxIndex; i++ )
    {
      str += DoubleToString( array[i], width, fraction ).trim();
      if( i < maxIndex - 1 )
        str += ", ";
    }
    return str;
  }

  static public String byteArrayToString( byte[] array, int width )
  {
    return byteArrayToString( array, array.length, width );
  }

  static public String byteArrayToString( byte[] array, int maxIndex, int width )
  {
    String str = "";
    for( int i = 0; i < maxIndex; i++ )
    {
      str += IntegerToString( (int)array[i], width ).trim();
      if( i < maxIndex - 1 )
        str += ", ";
    }
    return str;
  }

  static public String intArrayToString( int[] array, int width )
  {
    return intArrayToString( array, array.length, width );
  }

  static public String intArrayToString( int[] array, int maxIndex, int width )
  {
    String str = "";
    for( int i = 0; i < maxIndex; i++ )
    {
      str += IntegerToString( array[i], width ).trim();
      if( i < maxIndex - 1 )
        str += ", ";
    }
    return str;
  }

  static public double [] stringToDoubleArray( String str )
  {
    return stringToDoubleArray( str, ',' );
  }

  static public double [] stringToDoubleArray( String str, char separator )
  {
    int pos, width;
    double tmpValue = Double.MIN_VALUE;
    int startPos = 0;
    int count = 0;
    while( startPos < str.length() )
    {
      pos = str.indexOf( separator, startPos );
      if( pos == -1 )
        pos = str.length();
      width = pos - startPos;
      if( width > 0 )
        tmpValue = StringToDouble( str.substring( startPos ),
          width, Double.MIN_VALUE );
      else
        tmpValue = Double.MIN_VALUE;

      if( tmpValue == Double.MIN_VALUE )
      {
        count = 0;
        break;
      }

      startPos = pos + 1;
      count++;
    }
    if( count == 0 )
      return null;

    return stringToDoubleArray( str, separator, count );
  }

  static public double [] stringToDoubleArray( String str, char separator, int count )
  {
    double [] array = new double[count];
    if( stringToDoubleArray( str, array, separator, count ) )
      return array;
    else
      return null;
  }

  static public boolean stringToDoubleArray( String str, double [] array, char separator, int count )
  {
    int pos, width;
    double tmpValue = Double.MIN_VALUE;
    int startPos = 0;
    startPos = 0;
    int tmpCount = 0;
    while( startPos < str.length() )
    {
      pos = str.indexOf( separator, startPos );
      if( pos == -1 )
        pos = str.length();
      width = pos - startPos;
      if( width > 0 )
        tmpValue = StringToDouble( str.substring( startPos ),
          width, Double.MIN_VALUE );
      startPos = pos + 1;
      array[tmpCount] = tmpValue;
      tmpCount++;
    }
    return count == tmpCount;
  }

  static public int [] stringToIntArray( String str )
  {
    return stringToIntArray( str, ',' );
  }

  static public int [] stringToIntArray( String str, char separator )
  {
    int pos, width;
    int tmpValue = Integer.MIN_VALUE;
    int startPos = 0;
    int count = 0;
    while( startPos < str.length() )
    {
      pos = str.indexOf( separator, startPos );
      if( pos == -1 )
        pos = str.length();
      width = pos - startPos;
      if( width > 0 )
        tmpValue = StringToInteger( str.substring( startPos ),
          width, Integer.MIN_VALUE );
      else
        tmpValue = Integer.MIN_VALUE;

      if( tmpValue == Integer.MIN_VALUE )
      {
        count = 0;
        break;
      }

      startPos = pos + 1;
      count++;
    }
    if( count == 0 )
      return null;

    return stringToIntArray( str, separator, count );
  }

  static public int [] stringToIntArray( String str, char separator, int count )
  {
    int [] array = new int[count];
    if( stringToIntArray( str, array, separator, count ) )
      return array;
    else
      return null;
  }

  static public boolean stringToIntArray( String str, int[] array, char separator, int count )
  {
    int pos, width;
    int tmpValue = Integer.MIN_VALUE;
    int startPos = 0;
    startPos = 0;
    int tmpCount = 0;
    while( startPos < str.length() )
    {
      pos = str.indexOf( separator, startPos );
      if( pos == -1 )
        pos = str.length();
      width = pos - startPos;
      if( width > 0 )
        tmpValue = StringToInteger( str.substring( startPos ),
          width, Integer.MIN_VALUE );
      startPos = pos + 1;
      array[tmpCount] = tmpValue;
      tmpCount++;
    }
    return count == tmpCount;
  }

  static public byte [] stringToByteArray( String str )
  {
    return stringToByteArray( str, ',' );
  }

  static public byte[] stringToByteArray( String str, char separator )
  {
    int pos, width;
    byte tmpValue = Byte.MIN_VALUE;
    int startPos = 0;
    int count = 0;
    while( startPos < str.length() )
    {
      pos = str.indexOf( separator, startPos );
      if( pos == -1 )
        pos = str.length();
      width = pos - startPos;
      if( width > 0 )
        tmpValue = (byte) StringToInteger( str.substring( startPos ),
          width, Byte.MIN_VALUE );
      else
        tmpValue = Byte.MIN_VALUE;

      if( tmpValue == Byte.MIN_VALUE )
      {
        count = 0;
        break;
      }

      startPos = pos + 1;
      count++;
    }
    if( count == 0 )
      return null;

    return stringToByteArray( str, separator, count );
  }

  static public byte[] stringToByteArray( String str, char separator, int count )
  {
    byte [] array = new byte[count];
    if( stringToByteArray( str, array, separator, count ) )
      return array;
    else
      return null;
  }

  static public boolean stringToByteArray( String str, byte[] array, char separator, int count )
  {
    int pos, width;
    byte tmpValue = Byte.MIN_VALUE;
    int startPos = 0;
    startPos = 0;
    int tmpCount = 0;
    while( startPos < str.length() )
    {
      pos = str.indexOf( separator, startPos );
      if( pos == -1 )
        pos = str.length();
      width = pos - startPos;
      if( width > 0 )
        tmpValue = (byte) StringToInteger( str.substring( startPos ),
          width, Byte.MIN_VALUE );
      startPos = pos + 1;
      array[tmpCount] = tmpValue;
      tmpCount++;
    }
    return count == tmpCount;
  }

  /**
   * Converts given string <code>text</code> to arrays of strings (lines), 
   * building this array by extracting lines (substrings that ended with either 
   * native End Of Line Symbol(s) or with Carriage Return + Line Feed).<br> 
   * Symbols that end lines are not included.
   */
  static public String[] str2LinesArray( String text )
  {
    String[] lines = null;
    Vector<String> vLines = new Vector<String>( 50, 100 );
    String altSep = "\n";
    if( System.lineSeparator().equals( "\n" ) )
      altSep = "\r\n";
    int pos1 = -1;
    int pos2 = -1;

    while( text.length() != 0 )
    {
      pos1 = text.indexOf( System.lineSeparator() );
      pos2 = text.indexOf( altSep );
      if( pos1 >= 0 )
      {
        if( pos2 < 0 || pos2 > pos1 )
        {
          vLines.addElement( text.substring( 0, pos1 ) );
          text = text.substring( pos1 + System.lineSeparator().length() );
        }
        else
        {
          vLines.addElement( text.substring( 0, pos2 ) );
          text = text.substring( pos2 + altSep.length() );
        }
      }
      else if( pos2 >= 0 )
      {
        vLines.addElement( text.substring( 0, pos2 ) );
        text = text.substring( pos2 + altSep.length() );
      }
      else
      {
        vLines.addElement( text );
        break;
      }
    }
    lines = new String[vLines.size()];
    vLines.copyInto( lines );
    return lines;
  }

  static public String timeIntervalToString( long intervalInMicrosec )
  {
    if( intervalInMicrosec > 0 )
    {
      String s = "";
      int val = ( int ) ( intervalInMicrosec / ( 60 * 60 * 1000000L ) );
      if( val > 0 )
      {
        s += val + " h ";
      }
      intervalInMicrosec %= ( 60 * 60 * 1000000L );
      val = ( int ) ( intervalInMicrosec / ( 60 * 1000000L ) );
      if( val > 0 )
      {
        s += val + " m ";
      }
      intervalInMicrosec %= ( 60 * 1000000L );
      val = ( int ) ( intervalInMicrosec / 1000000 );
      intervalInMicrosec %= 1000000;
      if( val != 0 || ( s.length() > 0 && intervalInMicrosec > 0 ) )
      {
        s += val + " s ";
      }
      val = ( int ) ( intervalInMicrosec / 1000 );
      intervalInMicrosec %= 1000;
      if( val != 0 || ( s.length() > 0 && intervalInMicrosec > 0 ) )
      {
        s += val + " ms ";
      }
      if( intervalInMicrosec > 0 )
      {
        s += intervalInMicrosec + " us ";
      }
      return s;
    }
    else if( intervalInMicrosec == 0 )
    {
      return "0 s";
    }
    else
    {
      return "NEG";
    }
  }

  static public String hex2String( int value, int width )
  {
    return padLeft( Integer.toHexString( value ), width, '0' );
  }
  
  static public String millis2period_ISO8601( long input_millis )
  {
    if( input_millis == 0)
      return "PT0S";
    long input = input_millis;
    int years, months, days, hours, minutes, seconds, millis;
    millis = (int)(input%TimeScale.MILLISECONDS_PER_SECOND);
    input /= TimeScale.MILLISECONDS_PER_SECOND;
    seconds = (int)(input%TimeScale.SECONDS_PER_MINUTE);
    input /= TimeScale.SECONDS_PER_MINUTE;
    minutes = (int)(input%TimeScale.MINUTES_PER_HOUR);
    input /= TimeScale.MINUTES_PER_HOUR;
    hours = (int)(input%TimeScale.HOURS_PER_DAY);
    input /= TimeScale.HOURS_PER_DAY;
    days = (int)(input%TimeScale.DAYS_PER_YEAR);
    String output = "P";
    if( days > 0)
      output += ""+days+"D";
    if( hours == 0 && minutes == 0 && seconds == 0 && millis == 0 )
      return output;
    output += "T";
    if ( hours > 0 || days > 0)
      output += ""+hours+"H";
    if( minutes == 0 && seconds == 0 && millis == 0 )
      return output;
    if( minutes > 0 || hours > 0 )
      output += ""+minutes+"M";
    if( seconds == 0 && millis == 0 )
      return output;
    if( seconds > 0 || millis > 0 )
      output += ""+seconds;
    if( millis >0 )
    {
      // remove trailing zeros per ISO (e.g.,"0.5S")
      if( millis % 10 == 0 ) millis /= 10;
      if( millis % 10 == 0 ) millis /= 10;
      output += "."+millis;
    }
    output += "S";
    return output;
  }

public static String nC2S(char c, int n) {
	String s = "";
	for (int i = 0; i < n; i++)
		s += c;
	return s;
}
  
}