/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.format;

import java.util.Calendar;

import edu.uml.lgdc.geospace.EarthGrid;
import edu.uml.lgdc.project.Console;
import edu.uml.lgdc.time.TimeScale;

public class IonexGrid {
	
	// latitude, longitude, altitude grids per IONEX specs... 
	// element 0 is start, element 1 is stop, element 2 is step
	// [bad way to store this as array of doubles]... legacy code
	private double[] latGrid = new double[3];
	private int latNodes;
	private double[] lonGrid = new double[3];
	private int lonNodes;
	private double[] altGrid = new double[3];
	private int altNodes;

	// timestamps of maps
	private TimeScale[] times = null;
	private int cadence_sec = -1;
	private int totalTimes = 0;

	private boolean isGood = false;
	
	public IonexGrid() {
		
	}
	
	// At this time, only partial setup is implemented for GAMBIT geomaps (fixed altitude reference)
	
	public boolean set( EarthGrid earthGrid, double referenceAltitude_km, TimeScale start, TimeScale stop, int cadence_sec ) {
		
		// fill the IONEX legacy triplets

		latGrid[0] = 90.-earthGrid.getLatitudeStep();
		latGrid[1] = -90.+earthGrid.getLatitudeStep();
		// note that latitude step is negative because IONEX output uses 90 to -90 ordering of latitudes
		latGrid[2] = -earthGrid.getLatitudeStep();
		lonGrid[0] = -180;
		lonGrid[1] = 180;
		lonGrid[2] = earthGrid.getLongitudeStep();
		
		// only one reference altitude for these 2D IONEX records
		altGrid[0] = referenceAltitude_km;
		altGrid[1] = referenceAltitude_km;
		altGrid[2] = 0;
		
		latNodes = (int) Math.abs((latGrid[1] - latGrid[0]) / latGrid[2]) + 1;
		lonNodes = (int) Math.abs((lonGrid[1] - lonGrid[0]) / lonGrid[2]) + 1;
		altNodes = 1;
		
		isGood = set( start, stop, cadence_sec ) && checkGridStepping();
		
		return isGood;
	}
		
	private boolean set( TimeScale start, TimeScale stop, int cadence_sec ) {
		this.cadence_sec = cadence_sec;
		totalTimes = cadence_sec == 0? 1 : ((int)(stop.getTimeInSeconds() - start.getTimeInSeconds()))/cadence_sec + 1;
		times = new TimeScale[totalTimes];
		long startTime = start.getTimeInMillis();
		long cadence_ms = cadence_sec * 1000;
		for( int i = 0; i < totalTimes; i++ ) {
			times[i] = new TimeScale( startTime+cadence_ms*i );
		}
		return checkCadence();
	}

	public boolean set( double[] latGrid, double[] lonGrid, double referenceAltitude_km, TimeScale start, TimeScale stop, int cadence_sec ) {
	
		this.latGrid = latGrid;
		this.lonGrid = lonGrid;
		// only one reference altitude for these 2D IONEX records
		altGrid[0] = referenceAltitude_km;
		altGrid[1] = referenceAltitude_km;
		altGrid[2] = 0;
		latNodes = (int) Math.abs((latGrid[1] - latGrid[0]) / latGrid[2]) + 1;
		lonNodes = (int) Math.abs((lonGrid[1] - lonGrid[0]) / lonGrid[2]) + 1;
		altNodes = 1;

		isGood = set( start, stop, cadence_sec ) && checkGridStepping();
		return isGood;
	}
	
	
	public boolean setTimes( TimeScale[] timestamps ) {
		this.times = timestamps;
		cadence_sec = (int) (times[0].getTimeInSeconds() - times[1].getTimeInSeconds());
		
		return checkCadence();
	}
	
	private boolean checkGridStepping() {

		if( latGrid[0]+(latNodes-1)*latGrid[2] != latGrid[1] ) {
			Console.showError("IONEX Formatter: Latitude grid is not even: "+latGrid[0]+ " + " + latNodes + " steps of "+ latGrid[2] 
					+ " does not match expected "+ latGrid[1]);
			return false;
		}
		
		if( lonGrid[0]+(lonNodes-1)*lonGrid[2] != lonGrid[1] ) {
			Console.showError("IONEX Formatter: Longitude grid is not even: "+lonGrid[0]+ " + " + lonNodes + " steps of "+ lonGrid[2] 
					+ " does not match expected "+ lonGrid[1]);
			return false;
		}

		if( altGrid[0]+(altNodes-1)*altGrid[2] != altGrid[1] ) {
			Console.showError("IONEX Formatter: Altitude grid is not even: "+altGrid[0]+ " + " + altNodes + " steps of "+ altGrid[2] 
					+ " does not match expected "+ altGrid[1]);
			return false;
		}
		
		return true;
	}
	
	
	public boolean checkStandardSpatialGrid() {
		if ((latGrid[0] == 90.+latGrid[2] && latGrid[1] == -90.-latGrid[2])) {
			if ((lonGrid[0] == -180. && lonGrid[1] == 180.) || (lonGrid[0] == 0. && lonGrid[1] == 360.))
				return true;
			else {
				Console.showError("IONEX Formatter: Longitude grids are non-standard: "+lonGrid[0]+".."+lonGrid[1]);
				return false;
			}
		} else {
			Console.showError("IONEX Formatter: Latitude grids are non-standard: "+latGrid[0]+".."+latGrid[1]);
			return false;
		}
	}
	
	private boolean checkCadence() {
		if (times == null ) {
			Console.showError("IONEX Formatter: times are not initialized for use");
			return false;
		}
		if( times.length == 0 ) {
			System.out.println("IONEX Formatter: array of timestamps is empty");
			return false;
		}
		if (times.length > 1) {
			if (!TimeScale.checkTimesUniformity(times, cadence_sec))
				System.out.println("IONEX Formatter: Map timestamps are unsorted or not uniform");
			return false;
		}
		return true;
	}

	public boolean checkCadence( double[][] maps ) {
		if (maps == null ) {
			Console.showError("IONEX Formatter: maps are not initialized for use");
			return false;
		}
		if( times.length != maps.length ) {
			System.out.println("IONEX Formatter: Number of timestamps does not match number of maps");
			return false;
		}
		return checkCadence();
	}

	public double[] getLatitudeTriplet() {
		return latGrid;
	}

	public double[] getLongitudeTriplet() {
		return lonGrid;
	}
	
	public double[] getAltitudeTriplet() {
		return altGrid;
	}
	
	public int getTotalLatitudes() {
		return latNodes;
	}
	
	public int getTotalLongitudes() {
		return lonNodes;
	}
	
	public int getTotalAltitudes() {
		return altNodes;
	}
	
	public TimeScale[] getTimes() {
		return times;
	}
	
	public TimeScale getTime( int index ) {
		return times[index];
	}

	public int getIndex(TimeScale time) {
		double secondsSinceStart = time.diffIn(Calendar.SECOND, times[0] );
		return (int) Math.round(secondsSinceStart / cadence_sec);
	}
	
	public int getTotalTimes() {
		return totalTimes;
	}
	
	public TimeScale getStartTime() {
		return getTime(0);
	}

	public TimeScale getStopTime() {
		return getTime(getTotalTimes()-1);
	}
	
	public int getCadence_sec() {
		return cadence_sec;
	}

	public boolean isGood() {
		return isGood;
	}
	
	public void reset() {
		isGood = false;
		times = null;
	}
	
}
