package edu.uml.lgdc.format;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
public final class StrUtil {
	static private final char CR = '\r';
	static private final char LF = '\n';
	static public final String[] INFO_VOLUME_UNITS_NAMES = {"B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"};

	/**
	 * Returns position of first encountered digit in the string <code>str</code> starting from the <code>startPos</code>.<br>
	 * If no digit was found then it returns length of string.
	 */
	static public int getDigitalPos(String str, int startPos) {
		int pos = str.length();
		for (int i = startPos; i < str.length(); i++) {
			if (Character.isDigit(str.charAt(i))) {
				pos = i;
				break;
			}
		}
		return pos;
	}

	/**
	 * Returns array of three {@link String} elements,
	 * <ul>
	 *   <li>first element is first line in given {@link String} instance <code>text</code>. Note that line is returned without terminator.</li>
	 *   <li>second element is the rest of text, i.e. part that immediately follows after the first line terminator.</li>
	 *   <li>line terminator, which is either "\n", "\r\n", or "\r". Also it will be empty string, "", if no terminator was found.</li>
	 * </ul>
	 * Notes, that because of disagreement between three software giants, the line terminator can be any of the following,
	 * <ul>
	 *   <li><b>Line feed</b>, hex value 0x10<br>
	 *       this line terminator is used in Linux/Unix and Mac OS X</li>
	 *   <li><b>Carriage Return</b>, hex value 0x13<br>
	 *       this line terminator is used in Classic Mac<li>
	 *   <li><b>Carriage Return</b> + <b>Line feed</b>, hex value 0x1310<br>
	 *       this line terminator is used in Windows</li>
	 * </ul>
	 * Throws {@link IllegalArgumentException} if parameter <code>text</code> is <b>null</b>
	 */
	static public String[] cutLine(String text) {
		Objects.requireNonNull(text);
		int length = text.length();
		int terminatorStartPos = -1;
		int terminatorEndPos = -1;
		int pos = -1;
		String terminator = "";
		while (++pos < length) {
			char ch = text.charAt(pos);
			if (ch != LF && ch != CR) {
				if (terminatorStartPos >= 0) {
					break;
				}
			} else if (terminatorStartPos >= 0) {
				if (ch == LF) {
					terminatorEndPos = pos;
					terminator += ch; 
				}
				break;
			} else {
				terminatorStartPos = pos;
				terminator += ch; 
				if (ch == LF) {
					break;
				}
			}
		}
		if (terminatorStartPos >= 0) {
			if (terminatorEndPos < 0) {
				terminatorEndPos = terminatorStartPos;
			}
			return new String[] {text.substring(0, terminatorStartPos), text.substring(terminatorEndPos + 1), terminator};
		} else {
			return new String[] {text, "", terminator};
		}
	}
	
	/**
	 * Converts given {@link String} instance <code>text</code> into array of lines. Lines terminators are not included in lines.<br>
	 * Note, that line terminator can be any of the following,
	 * <ul>
	 *   <li><b>Line feed</b>, hex value 0x10, this line terminator is used in Linux/Unix</li>
	 *   <li><b>Carriage Return</b>, hex value 0x13, this line terminator is used in Classic Mac</li>
	 *   <li><b>Carriage Return</b> + <b>Line feed</b>, hex value 0x1310, this line terminator is used in Windows</li>
	 * </ul>
	 * Note,
	 * <ul>
	 *   if given <code>text</code> is not empty and does not have line terminator at the end then it has <i>last line without terminator</i>,
	 *   which is also returned.
	 * </ul>
	 */
	static public String[] textToLines(String text) {
		Objects.requireNonNull(text);
		List<String> lines = new ArrayList<>(128);
		while (!text.isEmpty()) {
			String[] result = cutLine(text);
			lines.add(result[0]);
			text = result[1];
		}
		return lines.toArray(new String[lines.size()]);
	}

	/**
	 * Returns array of tokens extracted from given string <code>strList</code>. Token are supposed to be separated by given character
	 * <code>delimiter</code>.<br>
	 * <br>
	 * 
	 * Notes,
	 * <ul>
	 *   <li>algorithm silently skips adjacent delimiters, so empty tokens will be never produced</li>
	 *   <li>algorithm removes leading blanks from every token</li>
	 *   <li>if delimiter is space character, then there are no trailing blanks for any token</li>
	 * </ul>
	 */
	public static String[] listToArray(String strList, char delimiter) {
		return listToArray(strList, new String(new char[] {delimiter}));
	}

	/**
	 * Returns array of tokens extracted from given string <code>strList</code>. Token are supposed to be separated by given string
	 * <code>delimiter</code>.<br>
	 * <br>
	 * 
	 * Notes,
	 * <ul>
	 *   <li>algorithm silently skips adjacent delimiters, so empty tokens will be never produced</li>
	 *   <li>algorithm removes leading blanks from every token</li>
	 *   <li>if delimiter is space character, then there are no trailing blanks for any token</li>
	 * </ul>
	 */
	public static String[] listToArray(String strList, String delimiter) {
		return listToArray(strList, delimiter, false);
	}

	/**
	 * Returns array of tokens extracted from given string <code>strList</code>. Token are supposed to be separated with either,
	 * <ul>
	 *   <li>given string <code>delimiter</code>, if <code>anyCharAsDelimiter</code> is <b>false</b></li>
	 *   <li>any character from given string <code>delimiter</code>, if <code>anyCharAsDelimiter</code> is <b>true</b></li>
	 * </ul>
	 * 
	 * Notes,
	 * <ul>
	 *   <li>algorithm silently skips adjacent delimiters, so empty tokens will be never produced</li>
	 *   <li>algorithm removes leading blanks from every token</li>
	 *   <li>if delimiter is space character, then there are no trailing blanks for any token</li>
	 * </ul>
	 */
	public static String[] listToArray(String strList, String delimiter, boolean anyCharAsDelimiter) {
		List<String> list = new ArrayList<>(64);
		int prevPos = 0, nextPos;
		int delLength = delimiter.length();
		String element = null;

		if (delLength == 1) {
			anyCharAsDelimiter = false;
		}

		if (anyCharAsDelimiter) {
			delLength = 1;
		}

		if (delimiter.equals(" ")) {
			strList = strList.trim();
		}

		while (prevPos < strList.length()) {
			nextPos = indexOfDel(strList, delimiter, prevPos, anyCharAsDelimiter);
			if (nextPos == -1) {
				nextPos = strList.length();
			}
			element = strList.substring(prevPos, nextPos).trim();
			if (element.length() > 0) {
				list.add(element);
			}
			prevPos = nextPos + delLength;
			// skip delimiters and spaces
			while (prevPos < strList.length()) {
				if (strList.charAt(prevPos) == ' ') {
					prevPos++;
				} else if (startWith(strList, delimiter, prevPos, anyCharAsDelimiter)) {
					prevPos += delLength;
				} else {
					break;
				}
			}
		}
		return list.toArray(new String[list.size()]);
	}

	/**
	 * Returns position in string <code>str</code> of next character from string <code>chars</code>. Search starts from given position
	 * <code>from</code> in string <code>str</code.<br>
	 * If not found then returns -1.
	 */
	static public int sniff(String str, String chars, int from) {
		for (int i = from; i < str.length(); i++) {
			if (chars.indexOf(str.charAt(i)) >= 0) {
				return i;
			}
		}
		return -1;
	}

	static public int[] sniffStr(String str, String[] strs, int from) {
		int j;
		for (int i = from; i < str.length(); i++) {
			for (j = 0; j < strs.length; j++) {
				if (str.substring(i).startsWith(strs[j])) {
					return new int[] {i, i + strs[j].length()};
				}
			}
		}
		return new int[] {str.length(), str.length()};
	}

	static public int sniffOthers(String str, String chars, int from) {
		for (int i = from; i < str.length(); i++) {
			if (chars.indexOf(str.charAt(i)) < 0) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Returns position of next delimiter (which can be either the whole string <code>delimiters</code> or any symbol from string
	 * <code>delimiters</code>) searching them in string <code>str</code> beginning from position <code>from</code>.<br>
	 * If not found then returns -1.<br>
	 * If <code>anyCharAsDelimiter</code> is <b>false</b> then the whole string <code>delimiters</code> is considered as delimiter, otherwise any
	 * character from string <code>delimiters</code> is considered as delimiter.
	 */
	static private int indexOfDel(String str, String delimiters, int from, boolean anyCharAsDelimiter) {
		if (!anyCharAsDelimiter) {
			return str.indexOf(delimiters, from);
		} else {
			return sniff(str, delimiters, from);
		}
	}

	/**
	 * Returns <b>true</b> if string <code>str</code> starting from position <code>from</code> coincides with string <code>compare</code>. If not
	 * found then returns -1.<br>
	 * If <code>anyCharFromCompare</code> is <b>true</b> then returns <b>true</b> if character in string <code>str</code> at position
	 * <code>from</code> is equal to some character in string <code>compare</code>, otherwise returns <b>false</b>.
	 */
	static private boolean startWith(String str, String compare, int from, boolean anyCharFromCompare) {
		if (!anyCharFromCompare) {
			return str.startsWith(compare, from);
		} else {
			return compare.indexOf(str.charAt(from)) >= 0;
		}
	}

	/**
	 * This function count number of occurrence of character <code>c</code>
	 */
	public static int countChar(String strList, char c) {
		int count = 0;
		int prevPos = 0, nextPos = 0;
		while (nextPos >= 0) {
			nextPos = strList.indexOf(c, prevPos);
			if (nextPos >= 0) {
				count++;
			}
			prevPos = nextPos + 1;
		}
		return count;
	}

	/**
	 * Trims spaces at the start of given string <code>str</code> and returns this left-trimmed string.
	 */
	static public String trimLeft(String str) {
		return trimLeft(str, ' ');
	}

	/**
	 * Trims given character <code>trimChar</code> at the start of given string <code>str</code> and returns this left-trimmed string.
	 */
	static public String trimLeft(String str, char trimChar) {
		String leftTrimmed = "";
		int pos = -1;
		for (int i = 0; i < str.length(); i++) {
			if (str.charAt(i) != trimChar) {
				pos = i;
				break;
			}
		}
		if (pos > 0) {
			leftTrimmed = str.substring(pos);
		} else if (pos == 0) {
			leftTrimmed = str;
		}
		return leftTrimmed;
	}

	/**
	 * Trims trailing spaces of String <code>str</code> and returns this right-trimmed string.
	 */
	static public String trimRight(String str) {
		return trimRight(str, ' ');
	}

	/**
	 * Trims trailing given character <code>trimChar</code> of given string <code>str</code> and returns this right-trimmed string.
	 */
	static public String trimRight(String str, char trimChar) {
		String rightTrimmed = "";
		int pos = -1;
		for (int i = str.length() - 1; i >= 0; i--) {
			if (str.charAt(i) != trimChar) {
				pos = i;
				break;
			}
		}
		if (pos >= 0) {
			if (pos < str.length() - 1) {
				rightTrimmed = str.substring(0, pos + 1);
			} else {
				rightTrimmed = str;
			}
		}
		return rightTrimmed;
	}

	/**
	 * Returns <b>true</b> if and only if string <code>str</code> is built entirely of digits (symbols '0' through '9'), otherwise returns
	 * <b>false</b>
	 */
	static public boolean isOnlyDigits(String str) {
		return isOnlyDigits(str, false);
	}

	/**
	 * Returns <b>true</b> if and only if string <code>str</code> is built entirely of digits (except, may be, leading blanks), otherwise returns
	 * <b>false</b><br>
	 * If parameter <code>ignoreLeadingBlanks</code> is <b>true</b> then leading blanks will be ignored.
	 */
	static public boolean isOnlyDigits(String str, boolean ignoreLeadingBlanks) {
		if (ignoreLeadingBlanks) {
			str = trimLeft(str);
		}
		for (int i = 0; i < str.length(); i++) {
			if (!Character.isDigit(str.charAt(i))) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Returns <b>true</b> if and only if string <code>str</code> is built entirely of letters, otherwise returns <b>false</b>
	 */
	static public boolean isOnlyLetters(String str) {
		return isOnlyLetters(str, false);
	}

	/**
	 * Returns <b>true</b> if and only if string <code>str</code> is built entirely of letters (except, may be, leading blanks), otherwise
	 * returns <b>false</b><br>
	 * If parameter <code>ignoreLeadingBlanks</code> is <b>true</b> then leading blanks will be ignored.
	 * @return boolean
	 */
	static public boolean isOnlyLetters(String str, boolean ignoreLeadingBlanks) {
		if (ignoreLeadingBlanks) {
			str = trimLeft(str);
		}
		for (int i = 0; i < str.length(); i++) {
			if (!Character.isLetter(str.charAt(i))) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Returns <b>true</b> if and only if string <code>str</code> is built entirely of letters and digits, otherwise returns <b>false</b>
	 */
	static public boolean isOnlyLettersOrDigits(String str) {
		return isOnlyLettersOrDigits(str, false);
	}

	/**
	 * Returns <b>true</b> if and only if string <code>str</code> is built entirely of letters and digits (except, may be, leading blanks), 
	 * otherwise returns <b>false</b><br>
	 * If parameter <code>ignoreLeadingBlanks</code> is <b>true</b> then leading blanks will be ignored.
	 */
	static public boolean isOnlyLettersOrDigits(String str, boolean ignoreLeadingBlanks) {
		if (ignoreLeadingBlanks) {
			str = trimLeft(str);
		}
		for (int i = 0; i < str.length(); i++) {
			if (!Character.isLetterOrDigit(str.charAt(i))) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Returns <b>true</b> if, and only if, string <code>str</code> represents integer number, i.e. built entirely of digits (symbols '0' through
	 * '9'), except 1st symbol which might be - (minus sign) or + (plus sign).
	 */
	static public boolean isIntegerNumber(String str) {
		return isIntegerNumber(str, false);
	}

	/**
	 * Returns <b>true</b> if, and only if, string <code>str</code> represents integer number, i.e. built entirely of digits (symbols '0' through
	 * '9'), except 1st symbol, which might be - (minus sign) or + (plus sign), and contain leading blanks in case of parameter
	 * <code>leadingBlanksAdmitted</code> is <b>true</b>.<br>
	 * At least one digits should be present.
	 */
	static public boolean isIntegerNumber(String str, boolean leadingBlanksAdmitted) {
		boolean ok = true;
		boolean digitEncountered = false;
		boolean signEncountered = false;
		char c;
		for (int i = 0; i < str.length(); i++) {
			c = str.charAt(i);
			if (!Character.isDigit(c)) {
				if (c == ' ') // blank symbols might be onle accepted as leading blanks
				{
					if (!leadingBlanksAdmitted || signEncountered || digitEncountered) {
						ok = false;
					}
				} else if (c == '-' || c == '+') // sign accepted only if no other
													// symbols, except may ne blanks,
													// have been encountered yet
				{
					if (!signEncountered && !digitEncountered) {
						signEncountered = true;
					} else {
						ok = false;
					}
				} else {
					ok = false;
				}
			} else {
				digitEncountered = true;
			}
			if (!ok) {
				break;
			}
		}

		// at least one digit should be present
		if (!digitEncountered) {
			ok = false;
		}

		return ok;
	}

	/**
	 * Returns <b>true</b> if, and only if, string <code>str</code> represents floating point number, i.e. built entirely of digits (symbols '0'
	 * through '9'), except,
	 * <ul>
	 *   <li>1st symbol might be - (minus sign) or + (plus sign)</li>
	 *   <li>decimal point might be present;</li>
	 * </ul>
	 * At least one digits should be present.
	 */
	static public boolean isFloatingPointNumber(String str) {
		return isFloatingPointNumber(str, false);
	}

	/**
	 * Returns <b>true</b> if, and only if, string <code>str</code> represents floating point number, i.e. built entirely of digits (symbols '0'
	 * through '9') except,
	 * <ul>
	 *   <li>1st symbol might be - (minus sign) or + (plus sign)</li>
	 *   <li>decimal point might be present</li>
	 *   <li>leading blanks in case parameter <code>leadingBlanksAdmitted</code> is <b>true</b></li>
	 * </ul>
	 * Note:
	 * <ul>
	 *   At least one digits should be present
	 * </ul>
	 */
	static public boolean isFloatingPointNumber(String str, boolean leadingBlanksAdmitted) {
		boolean ok = true;
		boolean digitEncountered = false;
		boolean signEncountered = false;
		boolean pointEncountered = false;
		char c;
		for (int i = 0; i < str.length(); i++) {
			c = str.charAt(i);
			if (!Character.isDigit(c)) {
				if (c == ' ') // blank symbols might be only accepted as leading blanks
				{
					if (!leadingBlanksAdmitted || signEncountered || pointEncountered || digitEncountered) {
						ok = false;
					}
				} else if (c == '-' || c == '+') // sign accepted only if no other symbols, except may be blanks, have been encountered yet
				{
					if (!signEncountered && !digitEncountered && !pointEncountered) {
						signEncountered = true;
					} else {
						ok = false;
					}
				} else if (c == '.') // only one point can be accepted
				{
					if (!pointEncountered) {
						pointEncountered = true;
					} else {
						ok = false;
					}
				} else {
					ok = false;
				}
			} else {
				digitEncountered = true;
			}

			if (!ok) {
				break;
			}
		}
		// at least one digit should be present
		if (!digitEncountered) {
			ok = false;
		}
		return ok;
	}

	/**
	 * Returns <b>true</b> if given string <code>str</code> can be converted to double using Java method <code>Double.valueOf( str )</code>,
	 * otherwise returns <b>false</b>.
	 */
	static public boolean isFloating(String str) {
		try {
			Double.valueOf(str).doubleValue();
			return true;
		} catch (NumberFormatException e) {
			return false;
		}
	}

	/**
	 * Returns <b>true</b> if all bytes of given <code>array</code>, beginning from <code>start</code> up to byte <code>start + length - 1</code>
	 * contains digits, i.e. values between 0 and 9 inclusively. Otherwise returns <b>false</b>.
	 */
	static public boolean isOnlyDigits(byte[] array, int start, int length) {
		boolean ok = true;
		for (int i = start; i < start + length; i++) {
			if (array[i] < 0 || array[i] > 9) {
				ok = false;
				break;
			}
		}
		return ok;
	}

	/**
	 * Returns <b>true</b> if all nibbles (4-bit halves of bytes) of given <code>array</code>, beginning from nibble
	 * <code>firstNibleOffset</code> up to nibble <code>firstNibleOffset + qtyOfNimbles - 1</code>, contains numbers less than 10. Otherwise
	 * returns <b>false<?b>.
	 */
	static public boolean isOnlyBCD(byte[] array, int firstNibleOffset, int qtyOfNimbles) {
		boolean ok = true;
		boolean upper = (firstNibleOffset % 2) == 0;
		byte val;
		for (int i = 0; i < qtyOfNimbles; i++) {
			if (upper) {
				val = (byte)(array[firstNibleOffset / 2] >> 4);
			} else {
				val = array[firstNibleOffset / 2];
			}
			val &= 0x0f;
			if (val > 9) {
				ok = false;
				break;
			}
			upper = !upper;
			firstNibleOffset++;
		}
		return ok;
	}

	/**
	 * Returns <b>true</b> if, and only if, string <code>str</code> is built entirely of hexadecimal digits (symbols '0' through '9' and 'A',
	 * 'a', 'B', 'b', 'C', 'c', 'D', 'd', 'E', 'e', 'F', 'f' ).
	 */
	static public boolean isOnlyHexDigits(String str) {
		return consistOf(str.toUpperCase(), FC.HEX_DIGITS);
	}

	/**
	 * Returns <b>true</b> if all bytes of given <code>array</code>, beginning from <code>start</code> up to byte <code>start + length - 1</code>
	 * contain hex digits, i.e. values between 0 and 15 inclusively. Otherwise returns <b>false</b>.
	 */
	static public boolean isOnlyHexDigits(byte[] array, int start, int length) {
		boolean ok = true;
		for (int i = start; i < start + length; i++) {
			if (array[i] < 0 || array[i] > 15) {
				ok = false;
				break;
			}
		}
		return ok;
	}

	/**
	 * Returns <b>true</b> if string <code>str</code> consists of legal symbols (i.e. of symbols which string <code>legalSymbols</code> consists
	 * of). Otherwise returns <b>false</b>.
	 */
	static public boolean consistOf(String str, String legalSymbols) {
		boolean result = true;
		for (int i = 0; i < str.length(); i++) {
			if (legalSymbols.indexOf(str.charAt(i)) < 0) {
				result = false;
				break;
			}
		}
		return result;
	}

	/**
	 * Returns string that is <code>count</code> times duplication of given character <code>c</code>.
	 */
	static public String repeat(char c, int count) {
		StringBuilder sb = new StringBuilder(count);
		for (int i = 0; i < count; i++) {
			sb.append(c);
		}
		return sb.toString();
	}

	/**
	 * Converts given long integer number to string so that triplets are separated with comma.<br>
	 * <br>
	 * 
	 * Example:
	 * <ul>
	 *   <code>6459877  ->   6,459,877</code>
	 * </ul>
	 */
	static public String toTriplets(long value) {
		long posValue = Math.abs(value);
		if (posValue > 0) {
			String resultStr = "" + (posValue % 1000);
			posValue /= 1000;
			int length = 3;
			while (posValue > 0) {
				if (resultStr.length() < length) {
					resultStr = StrUtil.repeat('0', length - resultStr.length()) + resultStr;
				}
				resultStr = "" + (posValue % 1000) + "," + resultStr;
				posValue /= 1000;
				length += 4;
			}
			if (value < 0) {
				resultStr = "-" + resultStr;
			}
			return resultStr;
		} else {
			return "0";
		}
	}

	/**
	 * Quotes given string and returns it.
	 */
	static public String quote(String str) {
		return "\"" + str + "\"";
	}

	static public String secondsToStr(int seconds, boolean fullUnitsName) {
		int min = seconds / 60;
		int sec = seconds % 60;
		String str = "";
		if (fullUnitsName && !(min > 0 && sec > 0)) {
			if (min == 0) {
				if (sec > 1) {
					str = sec + " seconds";
				} else {
					str = sec + " second";
				}
			} else {
				if (min > 1) {
					str = min + " minutes";
				} else {
					str = min + " minute";
				}
				if (sec > 0) {
					if (sec > 1) {
						str += " " + sec + " seconds";
					} else {
						str += " " + sec + " second";
					}
				}
			}
		} else {
			String minName = fullUnitsName ? " min" : "m";
			String secName = fullUnitsName ? " sec" : "s";
			if (min > 0) {
				str = min + minName;
				if (sec > 0) {
					str += " " + sec + secName;
				}
			} else {
				str = sec + secName;
			}
		}
		return str;
	}

	/**
	 * Taking size in bytes, returns size as a string either in bytes (B), kilobytes(KB), megabytes(MB), gigabytes(GB), terabytes(TB), 
	 * and so on.<br>
	 * Returned string will not represent exact size (unless it is returned in bytes), but precision of at least 0.1 percent is guaranteed.<br>
	 * Corresponding suffix of units name is appended.
	 */
	static public String getInfoVolSizeAsStr(long numberOfBytes) {
		long volume = numberOfBytes;
		long volumeInNextUnits;
		int i;
		for (i = 0; i < INFO_VOLUME_UNITS_NAMES.length; i++) {
			volumeInNextUnits = volume / 1024;
			if (volumeInNextUnits <= 999) {
				break;
			}
			volume = volumeInNextUnits;
		}
		if (volume == 0) {
			volume = 0;
		}
		return StrUtil.toTriplets(volume) + " " + INFO_VOLUME_UNITS_NAMES[i];
	}
	
	/**
	 * Pads given string, <code>str</code>, from the left with HTML non-breakable space, {@literal &nbsp;}
	 */
	static public String padLeftNBSP(String str, int width) {
		int strLen = str.length();
		if (strLen >= width) {
			return str.substring(strLen - width);
		}
		String spaces = "";
		for (int i = 0; i < width - strLen; i++) {
			spaces += "&nbsp;";
		}
		return spaces + str;
	}

	
}