/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.geospace;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import edu.uml.lgdc.quantities.NO_VALUE;

public enum CharChoice {
	FOF2("foF2", "foF2", "Critical frequency", 1, "MHz", NO_VALUE.INDEX, 0), 
	NMF2("NmF2", "NmF2", "Peak density", -1, "cm-3", 0, 1), 
	HMF2("hmF2", "hmF2", "Peak height", 32, "km", NO_VALUE.INDEX, 2),
	B0("B0", "B0in", "B0 Shape Parameter", 41, "km", NO_VALUE.INDEX, 3),
	B1("B1", "B1in", "B1 Shape Parameter", 42, "", NO_VALUE.INDEX, 4);

	static final public int SIZE = CharChoice.values().length;
	
	static private final Map<Integer, CharChoice> mapByDatabaseIdent = new HashMap<>();
	static private final Map<Integer, CharChoice> mapByIndex = new HashMap<>();
	static private final Map<String, CharChoice> mapByName = new HashMap<>();
	static private final Set<CharChoice> regularChars = new HashSet<>();

	private final String name;
	private final String name4char;
	private final String description;
	private final String description_tooltip;
	private final String units;
	private final int didbaseIdent;
	private final int derivedFrom; // index of characteristic that this one is derived from
	private final int index;
	
	static {
		for (CharChoice cc : CharChoice.values()) {
			if (mapByDatabaseIdent.put(cc.getDidbaseIdent(), cc) != null) {
				throw new RuntimeException("CharChoice design error: DIDB ident duplication: " + cc.getDidbaseIdent());
			}
			if (mapByIndex.put(cc.getIndex(), cc) != null) {
				throw new RuntimeException("CharChoice design error: index duplication: " + cc.getIndex());
			}
			if (mapByName.put(cc.getName(), cc) != null) {
				throw new RuntimeException("CharChoice design error: name duplication: " + cc.getName());
			}
			regularChars.add(cc);
		}
	}

	private CharChoice(String name, String name4char, String description, int didbaseIdent, String units, int derivedFrom, int index) {
		this.name = name;
		this.name4char = name4char;
		this.description = description;
		this.description_tooltip = "<html>" + description.replaceAll("\n", "<br>") + "</html>";
		this.didbaseIdent = didbaseIdent;
		this.units = units;
		this.derivedFrom = derivedFrom;
		this.index = index;
	}

	public String getName() {
		return name;
	}
	public String getName4char() {
		return name4char;
	}
	public String getDescription() {
		return description;
	}
	public String getDescription_tooltip() {
		return description_tooltip;
	}
	public String getUnits() {
		return units;
	}
	public int getDidbaseIdent() {
		return didbaseIdent;
	}
	public int getDerivedFrom() {
		return derivedFrom;
	}
	public boolean isDerived() {
		return derivedFrom != NO_VALUE.INDEX;
	}
	public int getIndex() {
		return index;
	}

	static public CharChoice getByIdent(int charIdent) {
		return mapByDatabaseIdent.get(charIdent);
	}
	static public CharChoice getByName(String name) {
		return mapByName.get(name);
	}
	static public CharChoice getByIndex(int index) {
		return mapByIndex.get(index);
	}
	static public boolean isRegularChar(CharChoice cc) {
		return regularChars.contains(cc);
	}
	static public boolean isRegularChar(String name) {
		CharChoice cc = getByName(name);
		if (cc != null) {
			return regularChars.contains(cc);
		} else {
			return false;
		}
	}
	static public String getRegularCharsList() {
		return regularChars.toString();
	}
}
