package edu.uml.lgdc.geospace;


import edu.uml.lgdc.geospace.Geopack;
import edu.uml.lgdc.time.TimeScale;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/

public class EarthGrid {
	// Space grid at which source data are given
	// Latitude goes from -90 to +90, longitude goes from 0 to 360-lonStep.
	// Please note that -90 and +90 latitudes are included!
	// So, if 2D array is built using this EarthGrid as nLats x nLons, it will have nLon values stored for lat of 90 and -90. 
	protected static final int MAX_LATITUDE_NODES = 180;
	protected static final int MAX_LONGITUDE_NODES = 360;

	static private final double PI_HALF = Math.PI / 2;
	static private final double RADIANS_TO_DEGREES = 180. / Math.PI;

	static public final double EARTH_CIRCUMFERENCE_EQUATOR = 40075017.; // in
																		// meters
	static public final double EARTH_CIRCUMFERENCE_MERIDIAN = 40007860.; // in
																			// meters
	static public final double EARTH_MERIDIANAL_DEGREE_METERS = EARTH_CIRCUMFERENCE_MERIDIAN / 360.;
	static public final double EARTH_EQUATORIAL_DEGREE_METERS = EARTH_CIRCUMFERENCE_EQUATOR / 360.;
	static public final double RE = Geopack.RE;

	static public final int LAT_IND = 0;
	static public final int LON_IND = 1;

	static public final int NEIGHBORS_DATA_FIELDS = 3;

	private double neighborRadius;

	private int nLats;
	private int nLons;
	private double gridLatStep_rad;
	private double gridLonStep_rad;
	private double lonStepDegree;
	private double latStepDegree;

	private int nLats_2;
	private int nLons_2;
	private int nLatsM1;
	private double[][][] dists = null;
	private double[][][] neighbors = null;

	/**
	 * 
	 * Careful with these nLats and nLons
	 * nLats includes both poles.
	 * nLons do not include the same meridian twice.
	 * 
	 * @param nLats
	 * @param nLons
	 */
	public EarthGrid(int nLats, int nLons) {
		checkLatLon(nLats, nLons);
		// commented out it so far, because a lot of this messages issued by gambit-coeffs servlet 
		//System.out.println("Initialize grid for lat = " + nLats + "; lon = " + nLons + " nodes");
		this.nLats = nLats;
		this.nLons = nLons;
		nLatsM1 = nLats - 1;
		nLats_2 = nLats / 2;
		nLons_2 = nLons / 2;
		gridLatStep_rad = Math.PI / (nLatsM1);
		gridLonStep_rad = 2 * Math.PI / nLons;
		latStepDegree = 180. / (nLatsM1);
		lonStepDegree = 360. / nLons;
		checkSteps(latStepDegree, lonStepDegree);
	}

	public EarthGrid(int nLats, int nLons, double neighborRadius) {
		this(nLats, nLons);
		this.neighborRadius = neighborRadius;
		calcNeighbors(neighborRadius);
	}

	static public void checkLatLon(int nLats, int nLons) {
		if (nLats < 3) {
			throw new IllegalArgumentException("nLats < 3, is " + nLats);
		}
		// if( nLats % 2 == 0 )
		// throw new IllegalArgumentException( "nLats should be odd number, is "
		// + nLats );
		if (nLons < 2)
		 {
			throw new IllegalArgumentException("nLons < 2, is " + nLons);
		// if( nLons % 2 == 1 )
		// throw new IllegalArgumentException(
		// "nLons should be even number, is " + nLons );
		}
	}

	static public double[] getLatLonStep_deg(int nLats, int nLons) {
		double[] steps = new double[2];
		steps[0] = 180. / (nLats - 1);
		steps[1] = 360. / nLons;
		return steps;
	}

	static public double[] getLatLonStep_rad(int nLats, int nLons) {
		double[] steps = new double[2];
		steps[0] = Math.PI / (nLats - 1);
		steps[1] = 2. * Math.PI / nLons;
		return steps;
	}

	static public void checkSteps(double latStep_deg, double lonStep_deg) {
		if (lonStep_deg != (int) lonStep_deg || latStep_deg*10 != (int) (latStep_deg*10)) {
			throw new IllegalArgumentException("Suspicious stepping: " + "Lat. step = " + latStep_deg
					+ " lon. step = " + lonStep_deg);
		}
	}

	public double getDist(int lat1, int lon1, int lat2, int lon2) {

		if( dists == null) {
			// compute distances for the analyzers
			System.out.println((new TimeScale()).toHumanUT() + ". Start calculating distances.");
			calcDists();
			System.out.println((new TimeScale()).toHumanUT() + ". Finish calculating distances.");
		}		
		int lon = lon2 - lon1;
		if (lon < 0) {
			lon = -lon;
		}
		if (lon > nLons_2) {
			lon = nLons - lon;
		}
		if (lat1 <= nLats_2) {
			return dists[lat1][lat2][lon];
		} else if (lat2 <= nLats_2) {
			return dists[lat2][lat1][lon];
		} else {
			return dists[nLatsM1 - lat1][nLatsM1 - lat2][lon];
		}
	}

	public double[][] getNeighbors(int iLatCenter, int iLonCenter) {
		if (neighbors == null) {
			System.out.println("Precalculate neighbors!");
		}
		double[][] neighborsLocal = new double[neighbors[iLatCenter].length][NEIGHBORS_DATA_FIELDS];
		for (int iN = 0; iN < neighborsLocal.length; iN++) {
			neighborsLocal[iN][LAT_IND] = neighbors[iLatCenter][iN][LAT_IND];
			neighborsLocal[iN][LON_IND] = neighbors[iLatCenter][iN][LON_IND] + iLonCenter;
			neighborsLocal[iN][LON_IND] = neighborsLocal[iN][LON_IND] < nLons ? neighborsLocal[iN][LON_IND] : neighborsLocal[iN][LON_IND]
					- nLons;
			neighborsLocal[iN][2] = neighbors[iLatCenter][iN][2];
		}
		return neighborsLocal;
	}

	private void calcNeighbors(double radius_km) {
		System.out.println("Start searching for neighbors.");
		neighbors = new double[nLats][][];
		for (int iLat = 0; iLat < nLats; iLat++) {
			neighbors[iLat] = findNeighbors(iLat, 0, radius_km);
		}
	}

	/**
	 * Find all neighbor grid nodes on sphere, using distance between nodes as a
	 * criteria.
	 * 
	 * @param iLatCenter
	 * @param iLonCenter
	 * @param radius_km
	 * @return output[i][0] latitude indexes of i-th neighbor grid node<br>
	 *         output[i][1] longitude indexes of i-th neighbor grid node<br>
	 *         output[i][2] weight of i-th neighbor, as it would contribute to
	 *         center
	 */
	public double[][] findNeighbors(int iLatCenter, int iLonCenter, double radius_km) {
		int latOffset = 0;
		int lonOffset = 0;
		int neighborsGain = 0, neighborsTotal = 0;
		int iLat = 0, iLon = 0;
		double weightNorm = 0;
		double latStep_km = EARTH_CIRCUMFERENCE_MERIDIAN / (2 * 1000 * nLatsM1);
		latOffset = -((int) (radius_km / latStep_km) + 1); // start always from
															// north
		double[][] temp = new double[nLats * nLons][NEIGHBORS_DATA_FIELDS];
		do {
			neighborsGain = 0;
			latOffset++;
			iLat = iLatCenter + latOffset;
			lonOffset = latOffset == 0 ? 1 : 0; // we don't want to include
												// center into neighbors
			do {
				iLon = iLonCenter - lonOffset;
				// since for iLat == (0, 90) and any iLon grid converges to
				// single point(pole)
				// we process them in special order
				if ((iLat == 0 || iLat == nLats - 1) && iLon != iLonCenter) {
					lonOffset++;
					continue;
				}
				int[] latlonP = getProperIndexes(iLat, iLon);
				double dist = getDist(iLatCenter, iLonCenter, latlonP[LAT_IND], latlonP[LON_IND]);
				// System.out.println("Distance = " + dist + "; lat = " +
				// latlonP[0] + "lon = " + latlonP[1]);

				if (dist < radius_km) {
					neighborsGain++;
					temp[neighborsTotal][LAT_IND] = latlonP[LAT_IND];
					temp[neighborsTotal][LON_IND] = latlonP[LON_IND];
					temp[neighborsTotal][2] = (1 - dist / radius_km);
					weightNorm += temp[neighborsTotal][2];
					neighborsTotal++;
					// since grid is symmetric another point (lonCenter +
					// lonOffset) is also included
					if (lonOffset != 0 && lonOffset != nLons_2) {
						latlonP = getProperIndexes(iLat, iLonCenter + lonOffset);
						temp[neighborsTotal][LAT_IND] = latlonP[LAT_IND];
						temp[neighborsTotal][LON_IND] = latlonP[LON_IND];
						temp[neighborsTotal][2] = (1 - dist / radius_km);
						weightNorm += temp[neighborsTotal][2];
						neighborsTotal++;
					}
				} else {
					break;
				}
				lonOffset++;
			} while (lonOffset <= nLons_2);
		} while (neighborsGain > 0);
		double[][] neighbors = new double[neighborsTotal][NEIGHBORS_DATA_FIELDS];
		for (int i = 0; i < neighborsTotal; i++) {
			neighbors[i][LAT_IND] = temp[i][LAT_IND];
			neighbors[i][LON_IND] = temp[i][LON_IND];
			neighbors[i][2] = temp[i][2] / weightNorm;
		}
		return neighbors;
	}

	/**
	 * Returns proper indexes in sense that we never go over bounds
	 * 
	 * @param iLat
	 * @param iLon
	 * @return
	 */
	int[] getProperIndexes(int iLat, int iLon) {
		int[] latlon = new int[2];
		// match within grid
		int iLatP = iLat;
		int iLonP = iLon;
		if (iLon < 0) {
			iLonP = nLons + iLon;
		}
		if (iLon >= nLons) {
			iLonP = iLon - nLons;
		}
		// if we pass through any pole to another (east/west) hemisphere
		if (iLat < 0 || iLat >= nLats) {
			if (iLat < 0) {
				iLatP = -iLatP;
			}
			if (iLat > nLatsM1) {
				iLatP = 2 * nLatsM1 - iLat;
			}

			iLonP = iLonP >= nLons_2 ? iLonP - nLons_2 : iLonP + nLons_2;
		}
		latlon[LAT_IND] = iLatP;
		latlon[LON_IND] = iLonP;
		return latlon;
	}

	public int getNumLatitudeNodes() {
		return nLats;
	}

	public int getNumLongitudeNodes() {
		return nLons;
	}

	public double getLatitudeStep() {
		return latStepDegree;
	}

	public double getLongitudeStep() {
		return lonStepDegree;
	}

	public double getNeighborRadius() {
		return neighborRadius;
	}

	/**
	 * Returns latitude and longitude of grid point with given latitude and
	 * longitude indexes, <code>latIndex</code> and <code>lonIndex</code>.
	 */
	private double[] getCoord(int latIndex, int lonIndex) {
		return new double[] { latIndex * gridLatStep_rad - PI_HALF, lonIndex * gridLonStep_rad };
	}

	public double[] getCoordGeo_radians(int latIndex, int lonIndex) {
		return new double[] { PI_HALF - latIndex * gridLatStep_rad, -Math.PI + lonIndex * gridLonStep_rad };
	}

	public double[] getCoordGeo_degrees(int latIndex, int lonIndex) {
		double[] latlon = getCoordGeo_radians(latIndex, lonIndex);
		latlon[LAT_IND] = latlon[LAT_IND] * RADIANS_TO_DEGREES;
		latlon[LON_IND] = latlon[LON_IND] * RADIANS_TO_DEGREES;
		return latlon;
	}

	public int[] getNode_radians(double lat, double lon) {
		int iLat = (int) Math.round((lat + PI_HALF) / gridLonStep_rad);
		int iLon = (int) Math.round(lon / gridLonStep_rad);
		return new int[] { iLat, iLon };
	}

	/**
	 * For node given by <code>iLat</code>-<code>iLon</code> in
	 * <code>grid2</code> finds the nearest lat-lon node in current EarthGrid
	 * object. node in grid given as a parameter.
	 * 
	 * @param grid
	 * @param iLat
	 * @param iLon
	 * @return
	 */
	public int[] matchNode(EarthGrid grid2, int iLat, int iLon) {
		return new int[] { Math.round(iLat * this.nLatsM1 / (grid2.getNumLatitudeNodes() - 1)),
				Math.round(iLon * this.nLons / grid2.getNumLongitudeNodes()) };
	}

	/**
	 * calculate distances between grid points.
	 */
	private void calcDists() {
		double[] p1;
		double[] p2;
		int i, j, k;
		dists = new double[nLats_2 + 1][nLats][nLons_2 + 1];
		for (i = 0; i <= nLats_2; i++) {
			p1 = getCoord(i, 0);
			for (j = 0; j < nLats; j++) {
				for (k = 0; k <= nLons_2; k++) {
					p2 = getCoord(j, k);
					dists[i][j][k] = onEarthDistance_rad(p1[LAT_IND], p1[LON_IND], p2[LAT_IND], p2[LON_IND]);
				}
			}
		}
	}

	static public double onEarthDistance_rad(double lat1_rad, double lon1_rad, double lat2_rad, double lon2_rad) {

		// Convert points p1 = (1, lon1, 90 - lat1) and p2 = (1, lon2, 90 -
		// lat2),
		// given in spherical coordinates, to Cartesian coordinates:
		//
		// p1 = ( cos(lat1)*cos(lon1), cos(lat1)*sin(lon1), sin(lat1) ),
		// p2 = ( cos(lat2)*cos(lon2), cos(lat2)*sin(lon2), sin(lat2) )
		//
		// The cosine of the angle, name it D, between corresponding vectors
		// equals to its dot-product:
		//
		// cos(lat1)*cos(lon1)*cos(lat2)*cos(lon2) +
		// cos(lat1)*sin(lon1)*cos(lat2)*sin(lon2) +
		// sin(lat1)*sin(lat2)
		//
		// which is reduced to
		//
		// con(lon2-lon1)*cos(lat1)*cos(lat2) + sin(lat1)*sin(lat2)
		//
		double cosD = Math.cos(lon2_rad - lon1_rad) * Math.cos(lat1_rad) * Math.cos(lat2_rad) + Math.sin(lat1_rad) * Math.sin(lat2_rad);
		return Geopack.RE * Math.acos(cosD);
	}

	static public double onEarthDistance_deg(double lat1_deg, double lon1_deg, double lat2_deg, double lon2_deg) {
		return onEarthDistance_rad(Math.toRadians(lat1_deg), Math.toRadians(lon1_deg), Math.toRadians(lat2_deg), Math.toRadians(lon2_deg));
	}

	protected boolean checkSpaceGrid() {
		return getNumLatitudeNodes() < MAX_LATITUDE_NODES && getNumLongitudeNodes() < MAX_LONGITUDE_NODES;
	}

	/**
	 * Get index of latitude node in the grid. Lat = -90. has index = 0 and Lat
	 * = +90. has maximal index.
	 * 
	 * @param latitude_deg
	 * @return
	 */
	public int getGridLatIndex(double latitude_deg) {
		return ((int) Math.round((latitude_deg + 90) / getLatitudeStep())) % getNumLatitudeNodes();
	}

	/**
	 * Get index of longitude node in the grid. Lon = 0. has index = 0 and Lon =
	 * (360. - lonStep) has maximal index. Negative longitudes are handled
	 * correctly.
	 * 
	 * @param latitude_deg
	 * @return
	 */
	public int getGridLonIndex(double longitude_deg) {
		int i = ((int) Math.round(longitude_deg / getLongitudeStep())) % getNumLongitudeNodes();
		if (i < 0) {
			i += getNumLongitudeNodes();
		}
		return i;
	}

	/**
	 * Inverse to public int getGridLatIndex( double latitude_deg )
	 * 
	 * @param iLat
	 * @return
	 */
	public double getGridLatDegree(int iLat) {
		return iLat * getLatitudeStep() - 90.;
	}

	/**
	 * Inverse to public int getGridLonIndex( double longitude_deg )
	 * 
	 * @param iLon
	 * @return
	 */
	public double getGridLonDegree(int iLon) {
		return iLon * getLongitudeStep();
	}

	/*
	 * static public final int LATITUDE_STEP = LATITUDE_RANGE_DEGREE /
	 * (NUM_LATITUDE_NODES - 1); // -1 to exclude poles static public final int
	 * LONGITUDE_STEP = LONGITUDE_RANGE_DEGREE / NUM_LONGITUDE_NODES; static {
	 * if(checkTimeStep(LATITUDE_RANGE_DEGREE, (NUM_LATITUDE_NODES - 1)) ==
	 * false || checkTimeStep(LONGITUDE_RANGE_DEGREE, NUM_LONGITUDE_NODES) ==
	 * false) throw new RuntimeException(
	 * "Check Lat-Lon grid dimensions in RTAM.java" ); }
	 */

}