/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.geospace;

import edu.uml.lgdc.format.IonexGrid;
import edu.uml.lgdc.geospace.Geomap;
import edu.uml.lgdc.math.SimpleMetrics;
import edu.uml.lgdc.time.TimeScale;

public class Geomap {
	protected EarthGrid earthGrid;
	protected double[][] mapValues; // latitude -90..+90, east longitude 0..360;
	protected double[] mapValues_1D; // 1D packed using WorldWind convention:
	// latitude +90..-90, longitude -180..+180
	protected SimpleMetrics metrics = new SimpleMetrics();
	protected TimeScale refTime;
	protected TimeScale refTime_baseline;
	protected boolean needRecalculateGrid = true;
	protected double noValue;

	MapChoice mapChoice;
	
	public Geomap(MapChoice mapChoice, double noValue) {
		this.mapChoice = mapChoice;
		this.noValue = noValue;
	}

	public void setEarthGrid(EarthGrid earthGrid) {
		this.earthGrid = earthGrid;
		mapValues = new double[earthGrid.getNumLatitudeNodes()][earthGrid.getNumLongitudeNodes()];
		mapValues_1D = new double[earthGrid.getNumLatitudeNodes() * earthGrid.getNumLongitudeNodes()];
	}

	public TimeScale getRefTime() {
		return (new TimeScale(refTime.getTimeInMillis()));
	}

	public TimeScale getBaselineTime() {
		return (new TimeScale(refTime_baseline.getTimeInMillis()));
	}

	public boolean getNeedRecalcGrid() {
		return needRecalculateGrid;
	}

	public double[] getMap_1D() {
		return mapValues_1D;
	}

	public double[][] getMap() {
		return mapValues;
	}

	public SimpleMetrics getMetrics() {
		return metrics;
	}

	public double getMinValue() {
		return metrics.min;
	}
	public double getMaxValue() {
		return metrics.max;
	}
	public double getAverageValue() {
		return metrics.average;
	}

	public double getSDO() {
		return metrics.standardDeviation;
	}

	public EarthGrid getEarthGrid() {
		return earthGrid;
	}

	public MapChoice getMapChoice() {
		return mapChoice;
	}

	public double getNoValue() {
		return noValue;
	}

	public void setNeedRecalcGrid( boolean b) {
		needRecalculateGrid = b;
	}

	public boolean setMapValues(double values[][], TimeScale refTime) {
		if (values.length != earthGrid.getNumLatitudeNodes() || values[0].length != earthGrid.getNumLongitudeNodes()) {
			System.err.println("Map ERROR: grid does not match value array");
			return false;
		}
		mapValues = values;
		this.refTime = refTime;
		compute1Dfrom2D(mapValues_1D, mapValues);
		metrics.compute( mapValues_1D, mapValues_1D.length, noValue );
		return true;
	}

	public boolean setMapValues(double values[], TimeScale refTime) {
		if (values.length != earthGrid.getNumLatitudeNodes() * earthGrid.getNumLongitudeNodes()) {
			System.err.println("Map ERROR: grid does not match value array");
			return false;
		}
		mapValues_1D = values;
		this.refTime = refTime;
		compute2Dfrom1D(mapValues, mapValues_1D);
		metrics.compute( mapValues_1D, mapValues_1D.length, noValue );
		return true;
	}

	static public boolean compute2Dfrom1D(double[][] m2, double[] m1) {
		if (m1 == null || m2 == null) {
			System.err.println("Map ERROR: null value arrays in 2D->1D computation");
			return false;
		}
		if (m1.length != m2.length * m2[0].length) {
			System.err.println("Map ERROR: dimensions do not match in 2D->1D computation");
			return false;
		}
		// fill 2D values
		int totalLatitudes = m2.length;
		int totalLongitudes = m2[0].length;
		int halfLongitudes = totalLongitudes / 2;
		for (int i = 0; i < m1.length; i++) {
			int wLat = i / totalLongitudes;
			int iLat = totalLatitudes - wLat - 1;
			int wLon = i - wLat * totalLongitudes;
			int iLon = (wLon + halfLongitudes) % totalLongitudes;
			m2[iLat][iLon] = m1[i];
		}
		return true;
	}

	/**
	 * 
	 * Takes 2D map m2 [lats][lons] with lats -90..90 and lons 0..360
	 * Produces 1D map array [lats x lons]: lons is the faster index, lats from 90 to -90, lons from -180 to 180. 
	 * 
	 * @param m1
	 * @param m2
	 * @return
	 */
	static public boolean compute1Dfrom2D(double[] m1, double[][] m2) {
		if (m1 == null || m2 == null) {
			System.err.println("Map ERROR: null value arrays in 1D->2D computation");
			return false;
		}
		if (m1.length != m2.length * m2[0].length) {
			System.err.println("Map ERROR: dimensions do not match in 1D->2D computation");
			return false;
		}
		// fill 1D values
		int totalLatitudes = m2.length;
		int totalLongitudes = m2[0].length;
		int halfLongitudes = totalLongitudes / 2;
		for (int wLat = 0; wLat < totalLatitudes; wLat++) {
			int j = totalLongitudes * wLat;
			int iLat = totalLatitudes - 1 - wLat;
			for (int wLon = 0; wLon < totalLongitudes; wLon++) {
				int iLon = (wLon + halfLongitudes) % totalLongitudes;
				m1[j + wLon] = m2[iLat][iLon];
			}
		}
		return true;
	}

	public boolean setIonexMapValues( double[] ionexValues, IonexGrid ionexGrid ) {

		if( ionexValues == null ) {
			System.err.println("Ionex map is empty");
			return false;
		}
		
		int nLons = earthGrid.getNumLongitudeNodes();
		int nLats = earthGrid.getNumLatitudeNodes();

		if (Math.abs(ionexGrid.getLatitudeTriplet()[2]) != earthGrid.getLatitudeStep()) {
			System.err.println("Ionex Import ERROR: latitude step of the grids do not match: ionex "
					+ionexGrid.getLatitudeTriplet()[2]+" vs WorldWind " + earthGrid.getLatitudeStep());
			return false;
		}
		
		if (Math.abs(ionexGrid.getLongitudeTriplet()[2]) != earthGrid.getLongitudeStep()) {
			System.err.println("Ionex Import ERROR: longitude step of the grids do not match: ionex "
					+ionexGrid.getLongitudeTriplet()[2]+" vs WorldWind " + earthGrid.getLongitudeStep());
			return false;
		}
		
		int latCount = 0;
		int i;
		// load existing information from Ionex map, skip lon duplicates 
		// index i runs through output array without poles
		// i+latCount skips the lon duplicates in input array
		for( i = 0; i < mapValues_1D.length-2*nLons; i++ ) {
			// duplicate the longitude value at the boundary
			if( i>0 && (i%nLons) == 0 )
				latCount++;
			mapValues_1D[i+nLons] = ionexValues[i+latCount];
		}

		// The GAMBIT synthesizer produces an IRTAM map whose grid includes both poles
		// IONEX format does not include poles in the output
		// compute averages at last-but-one latitude for the poles
		double average = 0;
		for( i = 0; i < nLons; i++ )
			average += ionexValues[i];
		average /= nLons;
		for( i = 0; i < nLons; i++ )
			mapValues_1D[i] = average;
		
		// the other pole
		average = 0;
		for( i = ionexValues.length-1; i >= ionexValues.length-1-nLons; i-- )
			average += ionexValues[i];
		average /= nLons;
		for( i = mapValues_1D.length-1; i >= mapValues_1D.length-nLons-1; i-- )
			mapValues_1D[i] = average;
		
		// 2D map
		compute2Dfrom1D(mapValues, mapValues_1D);
		metrics.compute( mapValues_1D, mapValues_1D.length, noValue );
		return true;
	}
	
	

	
	public Geomap clone() {
		Geomap m = new Geomap(mapChoice, this.noValue);
		m.setEarthGrid(earthGrid);
		m.refTime = refTime;
		m.refTime_baseline = refTime_baseline;
		for( int i = 0; i< mapValues_1D.length; i++)
			m.mapValues_1D[i] = mapValues_1D[i];
		for( int j = 0; j< mapValues.length; j++)
			for( int k = 0; k< mapValues[0].length; k++)
				m.mapValues[j][k] = mapValues[j][k];
		m.metrics.compute( m.mapValues_1D, m.mapValues_1D.length, noValue );
		return m;
	}
	
	public void reset()
	{
		for (int iLat = 0; iLat < earthGrid.getNumLatitudeNodes(); iLat++)
			for (int iLon = 0; iLon < earthGrid.getNumLongitudeNodes(); iLon++)
				mapValues[iLat][iLon] = noValue;
		for (int i = 0; i < mapValues_1D.length; i++)
			mapValues_1D[i] = noValue;
	}

	public boolean buildDeviationMap(double values1[][], double values2[][], TimeScale refTime, TimeScale baselineTime) {
		if (values1.length != earthGrid.getNumLatitudeNodes() || values1[0].length != earthGrid.getNumLongitudeNodes()
			||	values2.length != earthGrid.getNumLatitudeNodes() || values2[0].length != earthGrid.getNumLongitudeNodes() ) {
			System.err.println("Map ERROR: grid does not match value array");
			return false;
		}
		this.refTime = refTime;
		this.refTime_baseline = baselineTime;
		for (int iLat = 0; iLat < values1.length; iLat++) {
			for (int iLon = 0; iLon < values1[0].length; iLon++) {
				double val1 = values1[iLat][iLon];
				double val2 = values2[iLat][iLon];
				if (val1 == noValue || val2 == noValue) {
					mapValues[iLat][iLon] = noValue;
					continue;
				}
				mapValues[iLat][iLon] = val1 - val2;
			}
		}
		compute1Dfrom2D(mapValues_1D, mapValues);
		metrics.compute( mapValues_1D, mapValues_1D.length, noValue );
		return true;
	}

	public boolean buildDeviationMap(double values1[], double values2[], TimeScale refTime, TimeScale baselineTime) {
		if (values1.length != earthGrid.getNumLatitudeNodes() * earthGrid.getNumLongitudeNodes()
			||	values2.length != earthGrid.getNumLatitudeNodes() * earthGrid.getNumLongitudeNodes() ) {
			System.err.println("Map ERROR: grid does not match value array");
			return false;
		}
		this.refTime = refTime;
		this.refTime_baseline = baselineTime;
		for (int i = 0; i < values1.length; i++) {
			double val1 = values1[i];
			double val2 = values2[i];
			if (val1 == noValue || val2 == noValue) {
				mapValues_1D[i] = noValue;
				continue;
			}
			mapValues_1D[i] = val1 - val2;
		}
		compute2Dfrom1D(mapValues, mapValues_1D);
		metrics.compute( mapValues_1D, mapValues_1D.length, noValue );
		return true;
	}

	public boolean buildDeviationMap_percent(double values1[][], double values2[][], TimeScale refTime, TimeScale baselineTime) {
		buildDeviationMap( values1, values2, refTime, baselineTime);
		
		for (int iLat = 0; iLat < values1.length; iLat++) 
			for (int iLon = 0; iLon < values1[0].length; iLon++)
				if( mapValues[iLat][iLon] != noValue )
					mapValues[iLat][iLon] /= 0.01*values2[iLat][iLon];
		compute1Dfrom2D(mapValues_1D, mapValues);
		metrics.compute( mapValues_1D, mapValues_1D.length, noValue );
		return true;
	}

}
