/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
//Title:        GEOPACK
//Version:      1.0
//Copyright:    Copyright (c) 2000 - 2005
//Author:       Nikolai Tsyganenko, plus multiple co-authors, see comments
//Company:      UMLCAR
//Description:  Ported from April 16, 1996 version of Geopack in Fortran,
//              with coefficients up to 2005
//  Jan 2005    added possibility to read geodipole moment coefficients from
//  Kozlov A.   data file geopackdata.txt. See format of this file in abstract
//              of method <readGeodipoleMomentCoeff>.

package edu.uml.lgdc.geospace;

import java.awt.Polygon;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Calendar;
import java.util.Vector;

import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.format.StrUtil;
import edu.uml.lgdc.project.CommonConst;
import edu.uml.lgdc.project.Console;
import edu.uml.lgdc.time.TimeScale;

public class Geopack {

	// ------------------------- c o n s t a n t s -----------------------------
	// Geopack data file name
	static public final String GEOPACKDATA = "geopackdata.txt";
	// Radius of the Earth in km
	static public final double RE = 6371.2;

	public static final int X_AXIS = 0;
	public static final int Y_AXIS = 1;
	public static final int Z_AXIS = 2;

	// !!! Very important.
	// Must be satisfied the following equations:
	// POLAR_ANGLE == LATITUDE
	// AZIMUTH == LONGITUDE
	// RADIUS == ALTITUDE
	// Don't violate this - otherwise some things will go wrong !!!

	// Spherical coordinate system named coordinate's indices
	static public final int POLAR_ANGLE = 0; // for polar angle
	static public final int AZIMUTH = 1; // for azimuthal angle
	static public final int RADIUS = 2; // for radius

	// Geographical coordinate system named coordinate's indices
	static public final int LATITUDE = 0; // for latitude (in deg, -90 to 90)
	static public final int LONGITUDE = 1; // for longitude (in deg, 0 to 360)
	static public final int ALTITUDE = 2; // for altitude

	static private final String DATAFILE_NAME = new File(CommonConst.getShareResourcesDir(), GEOPACKDATA).getPath();
	static private final String FILE_SIGNATURE = "GeopackData ";
	// historically, version 2.0 and 3.0 are identical format
	static private final double DATA_FILE_VERSION = 3.0;

	static public final int DATAFILE_ERROR_OK = 0;
	static public final int DATAFILE_ERROR_ZEROLENGTH = 1;
	static public final int DATAFILE_ERROR_SIGNATURE = 2;
	static public final int DATAFILE_ERROR_ELEMENT_QTY = 3;
	static public final int DATAFILE_ERROR_TIME_FORMAT = 4;
	static public final int DATAFILE_ERROR_NUMBER_FORMAT = 5;
	static public final int DATAFILE_ERROR_TIME_DECREASE = 6;
	static public final int DATAFILE_ERROR_NO_ENTRIES = 7;
	static public final int DATAFILE_ERROR_LACK_OF_ENTRIES = 8;
	static public final int DATAFILE_ERROR_IO = 9;
	static public final int DATAFILE_ERROR_OBSOLETE = 10;

	static private TimeScale[] TIMES = { new TimeScale("1965.01.01"), new TimeScale("1970.01.01"), new TimeScale("1975.01.01"),
			new TimeScale("1980.01.01"), new TimeScale("1985.01.01"), new TimeScale("1990.01.01"), new TimeScale("1995.01.01"),
			new TimeScale("2000.01.01"), new TimeScale("2005.01.01"), new TimeScale("2010.01.01"), new TimeScale("2016.01.01") };
	static private double[] G10 = { 30334, 30220, 30100, 29992, 29873, 29775, 29692, 29619.4, 29554.63, 29496.50, -11.4 };
	static private double[] G11 = { -2119, -2068, -2013, -1956, -1905, -1848, -1784, -1728.2, -1669.05, -1585.9, 16.7 };
	static private double[] H11 = { 5776, 5737, 5675, 5604, 5500, 5406, 5306, 5186.1, 5077.99, 4945.1, -28.8 };

	static private boolean datafileInUse = false;
	static private int datafileErrorCode = DATAFILE_ERROR_OK;
	static private boolean softwareObsolete = false;

	// Time range where this model is working
	static public TimeScale MIN_DATE = TIMES[0];
	static public TimeScale MAX_DATE = new TimeScale(TIMES[TIMES.length - 1].getTimeInMinutes());
	static {
		MAX_DATE.add(Calendar.MILLISECOND, -1);
	}
	static public double MIN_DATE_IN_MINUTES = MIN_DATE.getTimeInMinutes();
	static public double MAX_DATE_IN_MINUTES = MAX_DATE.getTimeInMinutes();

	public static final double RAD = 180. / Math.PI;
	public static double L_SHELL = 4.;

	// ---------------------- p u b l i c v a r i a b l e s -------------------

	//
	// S U N P O S I T I O N D A T A
	//
	// GST - GREENWICH MEAN SIDEREAL TIME,
	// CGST - Cosine of GST
	// SGST - Sine of GST
	// SLONG - LONGITUDE ALONG ECLIPTIC
	// SRASN - RIGHT ASCENSION,
	// SDEC - DECLINATION OF THE SUN (RADIANS)
	//
	public double GST = 0, SLONG = 0, SRASN = 0, SDEC = 0;
	public double CGST = 0, SGST = 0;

	// TILT ANGLE PSI

	public double PSI = 0;
	double CPS = 0, SPS = 0; // cosine and sine of PSI

	// --------------------- coordinate transformation variables
	// ----------------

	// THE ELEMENTS OF THE MATRIX GEO TO GSM:

	double A11 = 0, A21 = 0, A31 = 0;
	double A12 = 0, A22 = 0, A32 = 0;
	double A13 = 0, A23 = 0, A33 = 0;

	// THE ELEMENTS OF THE MATRIX GSE TO GSM:

	double CHI = 0, SHI = 0; // a.k.a. EM22 and EM23
	double HI = 0;

	// THE ELEMENTS OF THE MATRIX MAG TO SM:

	double CFI = 0, SFI = 0;

	// THE ELEMENTS OF THE MATRIX GEI TO GSE:

	double S1 = 0, S2 = 0, S3 = 0;
	double DZ1, DZ2, DZ3, DY1, DY2, DY3;

	// THE ELEMENTS OF THE MATRIX GE0 TO MAG:

	double ST0 = 0, CT0 = 0, SL0 = 0, CL0 = 0;
	double STCL = 0, STSL = 0, CTCL = 0, CTSL = 0;

	// cash for compatible mode
	private int prevYear = -1;
	private int prevDayOfYear = -1;
	// cash for advanced mode
	private TimeScale prevTimeUT = null;

	private TimeScale timeUT = null;

	static private boolean extrapolationInaccurateMes = false;

	static private boolean compatibleMode;
	static {
		readGeodipoleMomentCoeff();
		compatibleMode = true;
		for (int i = 0; i < TIMES.length - 1; i++) {
			if (!TIMES[i].equals(TIMES[i].startOfYear())) {
				compatibleMode = false;
				break;
			}
		}
	}

	public Geopack() {
	}

	//
	// ============================== r e c a l c ( )
	// ===========================
	//
	private void recalc(int year, int dayOfYear, int hour, int min, int sec)

	{
		boolean onlyForGeoMagTransform = false;
		TimeScale ut = new TimeScale();
		ut.set(Calendar.YEAR, year);
		ut.set(Calendar.DAY_OF_YEAR, dayOfYear);
		if (hour >= 24) {
			hour -= 24;
			onlyForGeoMagTransform = true;
		}
		ut.set(Calendar.HOUR_OF_DAY, hour);
		ut.set(Calendar.MINUTE, min);
		ut.set(Calendar.SECOND, sec);
		ut.set(Calendar.MILLISECOND, 0);
		recalc(ut, onlyForGeoMagTransform);
	}

	public void recalc(TimeScale ut) {
		recalc(ut, false);
	}

	private void recalc(TimeScale ut, boolean onlyForGeoMagTransform) {
		// THIS IS A MODIFIED VERSION OF THE SUBROUTINE RECOMP WRITTEN BY
		// N.A. TSYGANENKO. SINCE I WANT TO USE IT IN PLACE OF SUBROUTINE
		// RECALC I HAVE RENAMED THIS ROUTINE RECALC AND ELIMINATED THE
		// ORIGINAL RECALC FROM THIS VERSION OF THE GEOPACK.F PACKET. THIS
		// WAY ALL ORIGINAL CALLS TO RECALC WILL CONTINUE TO WORK WITHOUT
		// HAVING TO CHANGE THEM TO CALLS TO RECOMP.
		//
		//
		// AN ALTERNATIVE VERSION OF THE SUBROUTINE RECALC FROM THE GEOPACK
		// PACKAGE
		// BASED ON A DIFFERENT APPROACH TO DERIVATION OF ROTATION MATRIX
		// ELEMENTS
		//
		// THIS SUBROUTINE WORKS BY 20% FASTER THAN RECALC AND IS EASIER TO
		// UNDERSTAND
		//
		// ################################################
		// # WRITTEN BY N.A. TSYGANENKO ON DEC.1, 1991 #
		// ################################################
		//
		//
		// Modified by: Mauricio Peredo and Nikolai Tsyganenko
		// Hughes STX at NASA/GSFC Code 695
		// September 1992, March 4, April 16, 1996
		//
		// Modified (1) to accept dates up to year 2000 and updated IGRF
		// coefficients for 1985-1995, and
		// (2) to provide matrix elements for the GEI-GEO transformation
		//

		//
		// WE ARE RESTRICTED BY THE INTERVAL <MIN_YEAR> - <MAX_YEAR>
		// (now is 1965-2005), THE IGRF COEFFICIENTS ARE KNOWN FOR THE
		// INTERVAL 1965-2000;
		// THE IGRF COEFFICIENTS FOR 2001-2005 ARE INTERPOLATED
		// IF IYR IS OUTSIDE <MIN_YEAR> - <MAX_YEAR>, THE SUBROUTINE
		// THROWS EXCEPTION
		//

		int year = ut.get(Calendar.YEAR);
		int dayOfYear = ut.get(Calendar.DAY_OF_YEAR);

		timeUT = new TimeScale(ut.getTimeInMinutes());

		if (ut.before(MIN_DATE))
			throw new RuntimeException("Geopack: date is below " + MIN_DATE + ", - no coefficients exist");
		if (MAX_DATE.before(ut)) {
			String msg = "Geopack: date is above " + MAX_DATE + ", extrapolated coefficients may be inaccurate";
			if (!extrapolationInaccurateMes) {
				if (!datafileInUse) {
					if (datafileErrorCode == DATAFILE_ERROR_OK)
						msg += System.lineSeparator() + "  you do not use Geopack data file";
					else if (!softwareObsolete && datafileErrorCode != DATAFILE_ERROR_OBSOLETE)
						msg += System.lineSeparator() + "  your Geopack data file is corrupted";
				}
				msg += System.lineSeparator() + "  try to get up-to-date Geopack data file";
				extrapolationInaccurateMes = true;
			}
			Console.showWarn(msg);
		}

		double g10 = 0, g11 = 0, h11 = 0;
		double DJ, T, OBLIQ;

		double SQ = 0, SQQ = 0, SQR = 0;

		double f2, f1, dt;

		if (compatibleMode && (year != prevYear || dayOfYear != prevDayOfYear) || !compatibleMode && !ut.equals(prevTimeUT)) {
			prevYear = year;
			prevDayOfYear = prevYear;
			prevTimeUT = ut;

			if (compatibleMode) {
				if (year >= TIMES[TIMES.length - 2].get(Calendar.YEAR)) {
					dt = year + dayOfYear / 366.0 - TIMES[TIMES.length - 2].get(Calendar.YEAR);
					g10 = G10[TIMES.length - 2] + G10[TIMES.length - 1] * dt;
					g11 = G11[TIMES.length - 2] + G11[TIMES.length - 1] * dt;
					h11 = H11[TIMES.length - 2] + H11[TIMES.length - 1] * dt;
				} else {
					for (int i = 0; i < TIMES.length - 2; i++) {
						if (year < TIMES[i + 1].get(Calendar.YEAR)) {
							f2 = (year + dayOfYear / 365.0 - TIMES[i].get(Calendar.YEAR)) / 5;
							f1 = 1 - f2;
							g10 = f1 * G10[i] + f2 * G10[i + 1];
							g11 = f1 * G11[i] + f2 * G11[i + 1];
							h11 = f1 * H11[i] + f2 * H11[i + 1];
							break;
						}
					}
				}
			} else {
				if (!ut.before(TIMES[TIMES.length - 2])) {
					dt = ut.diffIn(Calendar.YEAR, TIMES[TIMES.length - 2]);
					g10 = G10[TIMES.length - 2] + G10[TIMES.length - 1] * dt;
					g11 = G11[TIMES.length - 2] + G11[TIMES.length - 1] * dt;
					h11 = H11[TIMES.length - 2] + H11[TIMES.length - 1] * dt;
				} else {
					for (int i = 0; i < TIMES.length - 2; i++) {
						if (ut.before(TIMES[i + 1])) {
							f2 = ut.diffIn(Calendar.MINUTE, TIMES[i]) / TIMES[i + 1].diffIn(Calendar.MINUTE, TIMES[i]);
							f1 = 1 - f2;
							g10 = f1 * G10[i] + f2 * G10[i + 1];
							g11 = f1 * G11[i] + f2 * G11[i + 1];
							h11 = f1 * H11[i] + f2 * H11[i + 1];
							break;
						}
					}
				}
			}

			// NOW CALCULATE THE COMPONENTS OF THE UNIT VECTOR EzMAG IN GEO
			// COORD.SYSTEM:
			// SIN(TETA0)*COS(LAMBDA0), SIN(TETA0)*SIN(LAMBDA0), AND COS(TETA0)
			// ST0 * CL0 ST0 * SL0 CT0

			SQ = g11 * g11 + h11 * h11;
			SQQ = Math.sqrt(SQ);
			SQR = Math.sqrt(g10 * g10 + SQ);
			SL0 = -h11 / SQQ;
			CL0 = -g11 / SQQ;
			ST0 = SQQ / SQR;
			CT0 = g10 / SQR;
			STCL = ST0 * CL0;
			STSL = ST0 * SL0;
			CTSL = CT0 * SL0;
			CTCL = CT0 * CL0;
		}

		// THE CALCULATIONS ARE TERMINATED IF ONLY GEO-MAG TRANSFORMATION
		// IS TO BE DONE (IHOUR>24 IS THE AGREED CONDITION FOR THIS CASE):

		if (onlyForGeoMagTransform)
			return;

		sunPosition(ut);

		// S1, S2, AND S3 ARE THE COMPONENTS OF THE UNIT VECTOR EXGSM=EXGSE IN
		// THE
		// SYSTEM GEI POINTING FROM THE EARTH'S CENTER TO THE SUN:

		S1 = Math.cos(SRASN) * Math.cos(SDEC);
		S2 = Math.sin(SRASN) * Math.cos(SDEC);
		S3 = Math.sin(SDEC);
		CGST = Math.cos(GST);
		SGST = Math.sin(GST);

		// DIP1, DIP2, AND DIP3 ARE THE COMPONENTS OF THE UNIT VECTOR EZSM=EZMAG
		// IN THE SYSTEM GEI:

		double DIP1 = STCL * CGST - STSL * SGST;
		double DIP2 = STCL * SGST + STSL * CGST;
		double DIP3 = CT0;

		// NOW CALCULATE THE COMPONENTS OF THE UNIT VECTOR EYGSM IN THE SYSTEM
		// GEI
		// BY TAKING THE VECTOR PRODUCT D x S AND NORMALIZING IT TO UNIT LENGTH:

		double Y1 = 0, Y2 = 0, Y3 = 0;
		double Z1 = 0, Z2 = 0, Z3 = 0;
		Y1 = DIP2 * S3 - DIP3 * S2;
		Y2 = DIP3 * S1 - DIP1 * S3;
		Y3 = DIP1 * S2 - DIP2 * S1;
		double Y = Math.sqrt(Y1 * Y1 + Y2 * Y2 + Y3 * Y3);
		Y1 = Y1 / Y;
		Y2 = Y2 / Y;
		Y3 = Y3 / Y;

		// THEN IN THE GEI SYSTEM THE UNIT VECTOR Z = EZGSM = EXGSM x EYGSM = S
		// x Y
		// HAS THE COMPONENTS:

		Z1 = S2 * Y3 - S3 * Y2;
		Z2 = S3 * Y1 - S1 * Y3;
		Z3 = S1 * Y2 - S2 * Y1;

		// THE VECTOR EZGSE (HERE DZ) IN GEI HAS THE COMPONENTS (0,-SIN(DELTA),
		// COS(DELTA)) = (0.,-0.397823,0.917462); HERE DELTA = 23.44214 DEG FOR
		// THE EPOCH 1978 (SEE THE BOOK BY GUREVICH OR OTHER ASTRONOMICAL
		// HANDBOOKS).
		// HERE THE MOST ACCURATE TIME-DEPENDENT FORMULA IS USED:

		// AL: it looks very close to ( ut - 1900.01.01 ) in days but not
		// exactly
		// and with some cranks that looks like error
		// May be the following would be more clear:
		// DJ = ut.diffIn( Calendar.DAY_OF_MONTH, new TimeScale( "1900.01.01" )
		// );
		// then this:
		// DJ = 365 * ( year - 1900 ) + (int)( ( year - 1901 ) / 4 ) +
		// dayOfYear - 0.5 + ut.get( Calendar.SECOND ) / 86400.;
		if (compatibleMode)
			DJ = 365 * (year - 1900) + (int) ((year - 1901) / 4) + dayOfYear - 0.5 + ut.get(Calendar.SECOND) / 86400.;
		else
			DJ = ut.diffIn(Calendar.DAY_OF_MONTH, new TimeScale("1900.01.01"));
		T = DJ / 36525.;
		OBLIQ = (23.45229 - 0.0130125 * T) / 57.2957795;
		DZ1 = 0.;
		DZ2 = -Math.sin(OBLIQ);
		DZ3 = Math.cos(OBLIQ);

		// THEN THE UNIT VECTOR EYGSE IN GEI SYSTEM IS THE VECTOR PRODUCT DZ x S
		// :

		DY1 = DZ2 * S3 - DZ3 * S2;
		DY2 = DZ3 * S1 - DZ1 * S3;
		DY3 = DZ1 * S2 - DZ2 * S1;

		// THE ELEMENTS OF THE MATRIX GSE TO GSM ARE THE SCALAR PRODUCTS:
		// CHI=EM22=(EYGSM,EYGSE), SHI=EM23=(EYGSM,EZGSE),
		// EM32=(EZGSM,EYGSE)=-EM23,
		// AND EM33=(EZGSM,EZGSE)=EM22

		CHI = Y1 * DY1 + Y2 * DY2 + Y3 * DY3;
		SHI = Y1 * DZ1 + Y2 * DZ2 + Y3 * DZ3;
		HI = Math.asin(SHI);

		// TILT ANGLE: PSI = ARCSIN(DIP,EXGSM)

		SPS = DIP1 * S1 + DIP2 * S2 + DIP3 * S3;
		CPS = Math.sqrt(1. - SPS * SPS);
		PSI = Math.asin(SPS);

		// THE ELEMENTS OF THE MATRIX MAG TO SM ARE THE SCALAR PRODUCTS:
		// CFI=GM22=(EYSM,EYMAG), SFI=GM23=(EYSM,EXMAG); THEY CAN BE DERIVED AS
		// FOLLOWS:

		// IN GEO THE VECTORS EXMAG AND EYMAG HAVE THE COMPONENTS
		// (CT0*CL0,CT0*SL0,-ST0)
		// AND (-SL0,CL0,0), RESPECTIVELY. HENCE, IN GEI THE COMPONENTS ARE:
		// EXMAG: CT0*CL0*COS(GST)-CT0*SL0*SIN(GST)
		// CT0*CL0*SIN(GST)+CT0*SL0*COS(GST)
		// -ST0
		// EYMAG: -SL0*COS(GST)-CL0*SIN(GST)
		// -SL0*SIN(GST)+CL0*COS(GST)
		// 0
		// THE COMPONENTS OF EYSM IN GEI WERE FOUND ABOVE AS Y1, Y2, AND Y3;
		// NOW WE ONLY HAVE TO COMBINE THE QUANTITIES INTO SCALAR PRODUCTS:

		double EXMAGX = CT0 * (CL0 * CGST - SL0 * SGST);
		double EXMAGY = CT0 * (CL0 * SGST + SL0 * CGST);
		double EXMAGZ = -ST0;
		double EYMAGX = -(SL0 * CGST + CL0 * SGST);
		double EYMAGY = -(SL0 * SGST - CL0 * CGST);
		CFI = Y1 * EYMAGX + Y2 * EYMAGY;
		SFI = Y1 * EXMAGX + Y2 * EXMAGY + Y3 * EXMAGZ;

		// THE ELEMENTS OF THE MATRIX GEO TO GSM ARE THE SCALAR PRODUCTS:
		// A11=(EXGEO,EXGSM), A12=(EYGEO,EXGSM), A13=(EZGEO,EXGSM),
		// A21=(EXGEO,EYGSM), A22=(EYGEO,EYGSM), A23=(EZGEO,EYGSM),
		// A31=(EXGEO,EZGSM), A32=(EYGEO,EZGSM), A33=(EZGEO,EZGSM),

		// ALL THE UNIT VECTORS IN BRACKETS ARE ALREADY DEFINED IN GEI:

		// EXGEO=(CGST,SGST,0), EYGEO=(-SGST,CGST,0), EZGEO=(0,0,1)
		// EXGSM=(S1,S2,S3), EYGSM=(Y1,Y2,Y3), EZGSM=(Z1,Z2,Z3)
		// AND THEREFORE:

		A11 = S1 * CGST + S2 * SGST;
		A12 = -S1 * SGST + S2 * CGST;
		A13 = S3;
		A21 = Y1 * CGST + Y2 * SGST;
		A22 = -Y1 * SGST + Y2 * CGST;
		A23 = Y3;
		A31 = Z1 * CGST + Z2 * SGST;
		A32 = -Z1 * SGST + Z2 * CGST;
		A33 = Z3;

		/*
		 * System.out.println( "DT    = " + DT ); System.out.println( "G10   = "
		 * + g10 ); System.out.println( "G11   = " + g11 ); System.out.println(
		 * "H11   = " + h11 ); System.out.println( "SQ    = " + SQ );
		 * System.out.println( "SQQ   = " + SQQ ); System.out.println(
		 * "SQR   = " + SQR ); System.out.println( "SL0   = " + SL0 );
		 * System.out.println( "CL0   = " + CL0 ); System.out.println(
		 * "ST0   = " + ST0 ); System.out.println( "CT0   = " + CT0 );
		 * System.out.println( "STCL  = " + STCL ); System.out.println(
		 * "STSL  = " + STSL ); System.out.println( "CTCL  = " + CTCL );
		 * System.out.println( "CTSL  = " + CTSL ); System.out.println(
		 * "SRASN = " + SRASN ); System.out.println( "SDEC  = " + SDEC );
		 * System.out.println( "GST   = " + GST ); System.out.println(
		 * "CGST  = " + CGST ); System.out.println( "SGST  = " + SGST );
		 * System.out.println( "S1    = " + S1 ); System.out.println( "S2    = "
		 * + S2 ); System.out.println( "S3    = " + S3 ); System.out.println(
		 * "DIP1  = " + DIP1 ); System.out.println( "DIP2  = " + DIP2 );
		 * System.out.println( "DIP3  = " + DIP3 ); System.out.println(
		 * "PSI   = " + PSI ); System.out.println( "HI    = " + HI );
		 */

	}

	//
	// ========================= s u n P o s i t i o n ( )
	// ======================
	//

	private void sunPosition(TimeScale ut) {
		sunPosition(ut.get(Calendar.YEAR), ut.get(Calendar.DAY_OF_YEAR), ut.get(Calendar.HOUR_OF_DAY), ut.get(Calendar.MINUTE),
				ut.get(Calendar.SECOND));
	}

	private void sunPosition(int IYR, int IDAY, int IHOUR, int MIN, int ISEC) {
		// CALCULATES FOUR QUANTITIES NECESSARY FOR COORDINATE TRANSFORMATIONS
		// WHICH DEPEND ON SUN POSITION (AND, HENCE, ON UNIVERSAL TIME AND
		// SEASON)
		//
		// ------- INPUT PARAMETERS:
		// IYR,IDAY,IHOUR,MIN,ISEC - YEAR, DAY, AND UNIVERSAL TIME IN HOURS,
		// MINUTES,
		// AND SECONDS (IDAY=1 CORRESPONDS TO JANUARY 1).
		//
		// ------- OUTPUT PARAMETERS:
		// GST - GREENWICH MEAN SIDEREAL TIME, SLONG - LONGITUDE ALONG ECLIPTIC
		// SRASN - RIGHT ASCENSION, SDEC - DECLINATION OF THE SUN (RADIANS)
		// THIS SUBROUTINE HAS BEEN COMPILED FROM: RUSSELL C.T., COSM.ELECTRO-
		// DYN., 1971, V.2,PP.184-196.
		//
		//
		// AUTHOR: Gilbert D. Mead
		//
		//
		// ; IMPLICIT NONE

		double T, VL, G, OBLIQ, SOB, SLP, SIND, COSD, SC;
		double DJ, FDAY;

		if ((IYR < 1901) || (IYR > 2099))
			return;
		FDAY = (IHOUR * 3600. + MIN * 60. + ISEC) / 86400.;
		DJ = 365 * (IYR - 1900) + (int) ((IYR - 1901) / 4) + IDAY - 0.5 + FDAY;
		T = DJ / 36525.;
		VL = (279.696678 + 0.9856473354 * DJ) % 360.;
		GST = ((279.690983 + .9856473354 * DJ + 360. * FDAY + 180.) % 360.) / RAD;
		G = ((358.475845 + 0.985600267 * DJ) % 360.) / RAD;
		SLONG = (VL + (1.91946 - 0.004789 * T) * Math.sin(G) + 0.020094 * Math.sin(2. * G)) / RAD;
		if (SLONG > 6.2831853)
			SLONG = SLONG - 6.2831853;
		if (SLONG < 0.0)
			SLONG = SLONG + 6.2831853;
		OBLIQ = (23.45229 - 0.0130125 * T) / RAD;
		SOB = Math.sin(OBLIQ);
		SLP = SLONG - 9.924E-5;
		//
		// THE LAST CONSTANT IS A CORRECTION FOR THE ANGULAR ABERRATION DUE TO
		// THE ORBITAL MOTION OF THE EARTH
		//
		SIND = SOB * Math.sin(SLP);
		COSD = Math.sqrt(1. - SIND * SIND);
		SC = SIND / COSD;
		SDEC = Math.atan(SC);
		SRASN = 3.141592654 - Math.atan2(Math.cos(OBLIQ) / SOB * SC, -Math.cos(SLP) / COSD);
		return;
	}

	// -------------------------- s u n G E I ( ) ------------------------------

	// Returns unit vector in GEI that points sunward

	public void sunGEI(double[] out) {
		out[0] = S1;
		out[1] = S2;
		out[2] = S3;
	}

	//
	// ============= c o o r d i n a t e t r a n s f o r m a t i o n s ========
	//

	/**
	 * Converts Geographic (GEO) to Dipole (MAG) coordinates.<br>
	 * <br>
	 * 
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void GeoToMag(double[] geo, double[] mag) {
		recalc(timeUT.get(Calendar.YEAR), 0, 25, 0, 0);
		// XMAG = XGEO*CTCL + YGEO*CTSL - ZGEO*ST0
		mag[X_AXIS] = geo[X_AXIS] * CTCL + geo[Y_AXIS] * CTSL - geo[Z_AXIS] * ST0;
		// YMAG = YGEO*CL0 - XGEO*SL0
		mag[Y_AXIS] = geo[Y_AXIS] * CL0 - geo[X_AXIS] * SL0;
		// ZMAG = XGEO*STCL + YGEO*STSL + ZGEO*CT0
		mag[Z_AXIS] = geo[X_AXIS] * STCL + geo[Y_AXIS] * STSL + geo[Z_AXIS] * CT0;
	}

	/**
	 * Converts Dipole (MAG) to Geographic (GEO) coordinates.<br>
	 * <br>
	 * 
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void MagToGeo(double[] mag, double[] geo) {
		recalc(timeUT.get(Calendar.YEAR), 0, 25, 0, 0);
		// XGEO = XMAG*CTCL - YMAG*SL0 + ZMAG*STCL
		geo[X_AXIS] = mag[X_AXIS] * CTCL - mag[Y_AXIS] * SL0 + mag[Z_AXIS] * STCL;
		// YGEO = XMAG*CTSL + YMAG*CL0 + ZMAG*STSL
		geo[Y_AXIS] = mag[X_AXIS] * CTSL + mag[Y_AXIS] * CL0 + mag[Z_AXIS] * STSL;
		// ZGEO = ZMAG*CT0 - XMAG*ST0
		geo[Z_AXIS] = mag[Z_AXIS] * CT0 - mag[X_AXIS] * ST0;
	}

	/**
	 * Converts Solar Magnetospheric (GSM) to Solar Ecliptical (GSE)
	 * coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void GsmToGse(double[] gsm, double[] gse) {
		// XGSE = XGSM
		gse[X_AXIS] = gsm[X_AXIS];
		// YGSE = YGSM*CHI - ZGSM*SHI
		gse[Y_AXIS] = gsm[Y_AXIS] * CHI - gsm[Z_AXIS] * SHI;
		// ZGSE = YGSM*SHI + ZGSM*CHI
		gse[Z_AXIS] = gsm[Y_AXIS] * SHI + gsm[Z_AXIS] * CHI;
	}

	/**
	 * Converts Solar Ecliptical (GSE) to Solar Magnetospheric (GSM)
	 * coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void GseToGsm(double[] gse, double[] gsm) {
		// XGSM = XGSE
		gsm[X_AXIS] = gse[X_AXIS];
		// YGSM = YGSE*CHI + ZGSE*SHI
		gsm[Y_AXIS] = gse[Y_AXIS] * CHI + gse[Z_AXIS] * SHI;
		// ZGSM = ZGSE*CHI - YGSE*SHI
		gsm[Z_AXIS] = gse[Z_AXIS] * CHI - gse[Y_AXIS] * SHI;
	}

	/**
	 * Converts Geographic (GEO) to Solar Magnetospheric (GSM) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void GeoToGsm(double[] geo, double[] gsm) {
		// XGSM = A11*XGEO + A12*YGEO + A13*ZGEO
		gsm[X_AXIS] = A11 * geo[X_AXIS] + A12 * geo[Y_AXIS] + A13 * geo[Z_AXIS];
		// YGSM = A21*XGEO + A22*YGEO + A23*ZGEO
		gsm[Y_AXIS] = A21 * geo[X_AXIS] + A22 * geo[Y_AXIS] + A23 * geo[Z_AXIS];
		// ZGSM = A31*XGEO + A32*YGEO + A33*ZGEO
		gsm[Z_AXIS] = A31 * geo[X_AXIS] + A32 * geo[Y_AXIS] + A33 * geo[Z_AXIS];
	}

	/**
	 * Converts Solar Magnetospheric (GSM) to Geographic (GEO) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void GsmToGeo(double[] gsm, double[] geo) {
		// XGEO = A11*XGSM + A21*YGSM + A31*ZGSM
		geo[X_AXIS] = A11 * gsm[X_AXIS] + A21 * gsm[Y_AXIS] + A31 * gsm[Z_AXIS];
		// YGEO = A12*XGSM + A22*YGSM + A32*ZGSM
		geo[Y_AXIS] = A12 * gsm[X_AXIS] + A22 * gsm[Y_AXIS] + A32 * gsm[Z_AXIS];
		// ZGEO = A13*XGSM + A23*YGSM + A33*ZGSM
		geo[Z_AXIS] = A13 * gsm[X_AXIS] + A23 * gsm[Y_AXIS] + A33 * gsm[Z_AXIS];
	}

	/**
	 * Converts Solar Magnetic (SM) to Solar Magnetospheric (GSM) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void SmToGsm(double[] sm, double[] gsm) {
		// XGSM = XSM*CPS + ZSM*SPS
		gsm[X_AXIS] = sm[X_AXIS] * CPS + sm[Z_AXIS] * SPS;
		// YGSM = YSM
		gsm[Y_AXIS] = sm[Y_AXIS];
		// ZGSM = - XSM*SPS + ZSM*CPS
		gsm[Z_AXIS] = -sm[X_AXIS] * SPS + sm[Z_AXIS] * CPS;
	}

	/**
	 * Converts Solar Magnetospheric (GSM) to Solar Magnetic (SM) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void GsmToSm(double[] gsm, double[] sm) {
		// XSM = XGSM*CPS-ZGSM*SPS
		sm[X_AXIS] = gsm[X_AXIS] * CPS - gsm[Z_AXIS] * SPS;
		// YSM = YGSM
		sm[Y_AXIS] = gsm[Y_AXIS];
		// ZSM = XGSM*SPS + ZGSM*CPS
		sm[Z_AXIS] = gsm[X_AXIS] * SPS + gsm[Z_AXIS] * CPS;
	}

	/**
	 * Get MLT by xSM, ySM - magnetic local time in hours from Solar Magnetic
	 * coordinate<br>
	 * <br>
	 */
	public static double getMLT(double xSM, double ySM) {
		return (12. + Math.atan2(ySM, xSM) * 12 / Math.PI) % 24.;
	}

	/**
	 * Converts Equatorial Inertial (GEI) to Solar Ecliptical (GSE) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void GeiToGse(double[] gei, double[] gse) {
		// xgse = S1*xgei + S2*ygei + S3*zgei
		gse[X_AXIS] = S1 * gei[X_AXIS] + S2 * gei[Y_AXIS] + S3 * gei[Z_AXIS];
		// ygse = Y1*xgei + Y2*ygei + Y3*zgei
		gse[Y_AXIS] = DY1 * gei[X_AXIS] + DY2 * gei[Y_AXIS] + DY3 * gei[Z_AXIS];
		// zgse = Z1*xgei + Z2*ygei + Z3*zgei
		gse[Z_AXIS] = DZ1 * gei[X_AXIS] + DZ2 * gei[Y_AXIS] + DZ3 * gei[Z_AXIS];
	}

	/**
	 * Converts Solar Ecliptical (GSE) to Equatorial Inertial (GEI) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void GseToGei(double[] gse, double[] gei) {
		// xgei = S1*xgse + Y1*ygse + Z1*zgse
		gei[X_AXIS] = S1 * gse[X_AXIS] + DY1 * gse[Y_AXIS] + DZ1 * gse[Z_AXIS];
		// ygei = S2*xgse + Y2*ygse + Z2*zgse
		gei[Y_AXIS] = S2 * gse[X_AXIS] + DY2 * gse[Y_AXIS] + DZ2 * gse[Z_AXIS];
		// zgei = S3*xgse + Y3*ygse + z3*zgse
		gei[Z_AXIS] = S3 * gse[X_AXIS] + DY3 * gse[Y_AXIS] + DZ3 * gse[Z_AXIS];
	}

	/**
	 * Converts Equatorial Inertial (GEI) to Geographic (GEO) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void GeiToGeo(double[] gei, double[] geo) {
		// xgeo = CGST*xgei + SGST*ygei
		geo[X_AXIS] = CGST * gei[X_AXIS] + SGST * gei[Y_AXIS];
		// ygeo = -SGST*xgei + CGST*ygei
		geo[Y_AXIS] = -SGST * gei[X_AXIS] + CGST * gei[Y_AXIS];
		// zgeo = zgei
		geo[Z_AXIS] = gei[Z_AXIS];
	}

	/**
	 * Converts Geographic (GEO) to Equatorial Inertial (GEI) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 *
	 * @author Nikolai A. Tsyganenko, Institute of Physics, St. Petersburg State
	 *         University, Stary Petergof 1989/04, St. Petersburg, Russia
	 */
	public void GeoToGei(double[] geo, double[] gei) {
		// xgei = CGST*xgeo - SGST*ygeo
		gei[X_AXIS] = CGST * geo[X_AXIS] - SGST * geo[Y_AXIS];
		// ygei = SGST*xgeo + CGST*ygeo
		gei[Y_AXIS] = SGST * geo[X_AXIS] + CGST * geo[Y_AXIS];
		// zgei = zgeo
		gei[Z_AXIS] = geo[Z_AXIS];
	}

	/**
	 * Converts Equatorial Inertial (GEI) to Dipole Magnetic (MAG) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 */
	public void GeiToMag(double[] gei, double[] mag) {
		double[] tmp = new double[3];
		GeiToGeo(gei, tmp);
		GeoToMag(tmp, mag);
	}

	/**
	 * Converts Dipole Magnetic (MAG) to Equatorial Inertial (GEI) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 */
	public void MagToGei(double[] mag, double[] gei) {
		double[] tmp = new double[3];
		MagToGeo(mag, tmp);
		GeoToGei(tmp, gei);
	}

	/**
	 * Converts Geographic (GEO) to Solar Magnetic (SM) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 */
	public void GeoToSm(double[] geo, double[] sm) {
		double gsm[] = new double[3];
		GeoToGsm(geo, gsm);
		GsmToSm(gsm, sm);
	}

	/**
	 * Converts Solar Magnetic (SM) to Geographic (GEO) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 */
	public void SmToGeo(double[] sm, double[] geo) {
		double gsm[] = new double[3];
		SmToGsm(sm, gsm);
		GsmToGeo(gsm, geo);
	}

	/**
	 * Converts Equatorial Inertial (GEI) to Solar Magnetospheric (GSM)
	 * coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 */
	public void GeiToGsm(double[] gei, double[] gsm) {
		double geo[] = new double[3];
		GeiToGeo(gei, geo);
		GeoToGsm(geo, gsm);
	}

	/**
	 * Converts Solar Magnetospheric (GSM) to Equatorial Inertial (GEI)
	 * coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion is independent of measurement units
	 * </ul>
	 */
	public void GsmToGei(double[] gsm, double[] gei) {
		double geo[] = new double[3];
		GsmToGeo(gsm, geo);
		GeoToGei(geo, gei);
	}

	/**
	 * Converts Equatorial Inertial (GEI) to Geographic Map (GEOMAP)
	 * coordinates:
	 * <ul>
	 * <code>(xGEI,yGEI,zGEI) --&gt; (latitude, longitude, altitude)</code>
	 * </ul>
	 * Latitude and longitude are given in degrees. Altitude is given in
	 * kilometers above sea level.<br>
	 * Latitude is changed from -90 degrees at South Pole through 0 degrees at
	 * equator and to 90 degrees at North Pole.<br>
	 * Longitude is changed from 0 degrees to eastward direction up to 360
	 * degrees (360 is interpreted as 0 again), - so called, east longitude
	 * convention.<br>
	 * Altitude is changed from 0 to any positive number. Negative altitude is
	 * also works, up to -RE, and can be interpreted as depth below sea level.
	 * Really, Geographic Map coordinate is Geographic (GEO) coordinate system
	 * transferred into "abstract geographical" coordinate with:
	 * <ul>
	 * <li><code>latitude</code> = <i>polar_angle_in_degrees</i> - 90deg</li>
	 * <li><code>longitude</code> = <i>azimuthal_angle_in_degrees</i></li>
	 * <li><code>altitude</code> = <code>radius</code> -
	 * <i>Earth_radius_in_km</i></li>
	 * </ul>
	 *
	 * Note:
	 * <ul>
	 * This conversion works correctly only if measurement units is kilometers.
	 * Though if you use any other units all will be correct except of altitude.
	 * </ul>
	 */
	public void GeiToGeomap(double[] gei, double[] geomap) {
		double geo[] = new double[3];
		GeiToGeo(gei, geo);
		cartToSpher(geo, geomap);
		spherToGeomap(geomap);
	}

	/**
	 * Converts Geographic Map (GEOMAP) to Equatorial Inertial (GEI)
	 * coordinates:
	 * <ul>
	 * <code>(latitude, longitude, altitude) --&gt; (xGEI,yGEI,zGEI)</code>
	 * </ul>
	 *
	 * Note:
	 * <ul>
	 * This conversion works correctly only if measurement units is kilometers.
	 * </ul>
	 */
	public void GeomapToGei(double[] geomap, double[] gei) {
		double geo[] = new double[3];
		geomapToSpher(geomap, geo);
		spherToCart(geo);
		GeoToGei(geo, gei);
	}

	/**
	 * Converts Equatorial Inertial (GEI) to Magnetic Map (MAGMAP) coordinates:
	 * <ul>
	 * <code>(xGEI,yGEI,zGEI) --&gt; (magLatitude, magLongitude, altitude)</code>
	 * </ul>
	 * Magnetic latitude and longitude were calculated in degrees. Altitude is
	 * given in kilometers above sea level.<br>
	 * Magnetic latitude is changed from -90 degrees at South Magnetic Pole
	 * through 0 degrees at magnetic equator and to 90 degrees at North Magnetic
	 * Pole.<br>
	 * Magnetic longitude is changed from 0 degrees to eastward direction up to
	 * 360 degrees (360 is interpreted as 0 again), - so called, east longitude
	 * convention.<br>
	 * Altitude is changed from 0 to any positive number. Negative altitude is
	 * also works, up to -RE, and can be interpreted as depth below sea level.
	 * Really, Magnetic Map coordinate is Dipole Magnetic (MAG) coordinate
	 * system transferred into "abstract geographical" coordinate with:
	 * <ul>
	 * <li><i>magnetic_latitude</i> = <i>magnetic_polar_angle_in_degrees</i> -
	 * 90deg</li>
	 * <li><i>magnetic_longitude</i> =
	 * <i>magnetic_azimuthal_angle_in_degrees</i></li>
	 * <li><i>altitude</i> = <i>radius</i> - <i>Earth_radius</i></li>
	 * </ul>
	 *
	 * Note:
	 * <ul>
	 * This conversion works correctly only if measurement units is kilometers.
	 * Though if you use any other units all will be correct except of altitude.
	 * </ul>
	 */
	public void GeiToMagmap(double[] gei, double[] magmap) {
		double mag[] = new double[3];
		GeiToMag(gei, mag);
		cartToSpher(mag, magmap);
		spherToGeomap(magmap);
	}

	/**
	 * Converts Magnetic Map (MAGMAP) to Equatorial Inertial (GEI) coordinates:
	 * <ul>
	 * <code>(magLatitude, magLongitude, altitude) --&gt; (xGEI,yGEI,zGEI)</code>
	 * </ul>
	 *
	 * Note:
	 * <ul>
	 * This conversion works correctly only if measurement units is kilometers.
	 * </ul>
	 */
	public void magmapToGei(double[] magmap, double[] gei) {
		double mag[] = new double[3];
		geomapToSpher(magmap, mag);
		spherToCart(mag);
		MagToGei(mag, gei);
	}

	/**
	 * Converts Solar Magnetic (SM) to Geographic Map (GEOMAP) coordinates:
	 * <ul>
	 * <code>(xGEO,yGEO,zGEO) --&gt; (latitude, longitude, altitude)</code>
	 * </ul>
	 *
	 * Note:
	 * <ul>
	 * This conversion works correctly only if measurement units is kilometers.
	 * Though if you use any other units, all will be correct except of
	 * altitude.
	 * </ul>
	 */
	public void smToGeomap(double[] sm, double[] geomap) {
		double geo[] = new double[3];
		SmToGeo(sm, geo);
		cartToSpher(geo);
		spherToGeomap(geo, geomap);
	}

	/**
	 * Converts Geographic Map (GEOMAP) to Solar Magnetic (SM) coordinates:
	 * <ul>
	 * <code>(latitude, longitude, altitude) --&gt; (xGEO,yGEO,zGEO)</code>
	 * </ul>
	 *
	 * Note:
	 * <ul>
	 * This conversion works correctly only if measurement units is kilometers.
	 * </ul>
	 */
	public void geomapToSm(double[] geomap, double[] sm) {
		double geo[] = new double[3];
		geomapToSpher(geomap, geo);
		spherToCart(geo);
		GeoToSm(geo, sm);
	}

	/**
	 * Converts Geographic Map (GEOMAP) to Magnetic Map (MAGMAP) coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * This conversion works correctly only if altitude given in km.
	 * </ul>
	 */
	public void geomapToMagmap(double[] geomap, double[] magmap) {
		double tmp[] = new double[3];
		geomapToSpher(geomap, tmp);
		spherToCart(tmp);
		GeoToMag(tmp, magmap);
		cartToSpher(magmap);
		spherToGeomap(magmap);
	}

	/**
	 * Converts Magnetic Map (MAGMAP) Geographic Map (GEOMAP) to coordinates.<br>
	 * <br>
	 *
	 * Note:<br>
	 * This conversion works correctly only if altitude is given in km.
	 */
	public void magmapToGeomap(double[] magmap, double[] geomap) {
		double tmp[] = new double[3];
		geomapToSpher(magmap, tmp);
		spherToCart(tmp);
		MagToGeo(tmp, geomap);
		cartToSpher(geomap);
		spherToGeomap(geomap);
	}

	/**
	 * Converts Spherical coordinates to Abstract Geographic Map coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * <li>Geographic Map coordinates will be written into the same array,
	 * <code>coord</code></li>
	 * <li>this conversion works correctly only if radius is given in km</li>
	 * </ul>
	 */
	static public void spherToGeomap(double[] coord) {
		spherToGeomap(coord, coord);
	}

	/**
	 * Converts Spherical coordinates to Abstract Geographic Map coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * <li>array <code>geomap</code> can be the same as <code>spher</code></li>
	 * <li>this conversion works correctly only if radius is given in km</li>
	 * </ul>
	 */
	static public void spherToGeomap(double[] spher, double[] geomap) {
		// !!! constants LONGITUDE and AZIMUTH are equaled
		geomap[LONGITUDE] = Math.toDegrees(spher[AZIMUTH]);
		// !!! constants LATITUDE and POLAR_ANGLE are equaled
		geomap[LATITUDE] = 90 - Math.toDegrees(spher[POLAR_ANGLE]);
		// !!! constants ALTITUDE and RADIUS are equaled
		geomap[ALTITUDE] = spher[RADIUS] - RE;
	}

	/**
	 * Converts Abstract Geographic Map coordinates to Spherical coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * <li>spherical coordinates will be written into the same array,
	 * <code>coord</code></li>
	 * <li>this conversion works correctly only if radius given in km, though if
	 * you use any other units, - all will be correct except of altitude</li>
	 * </ul>
	 */
	static public void geomapToSpher(double[] coord) {
		geomapToSpher(coord, coord);
	}

	/**
	 * Converts Abstract Geographic Map coordinates to Spherical coordinates.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * <li>array <code>spher</code> can be the same as <code>geomap</code></li>
	 * <li>this conversion works correctly only if radius is given in km, though
	 * if you use any other units, - all will be correct except of altitude</li>
	 * </ul>
	 */
	static public void geomapToSpher(double[] geomap, double[] spher) {
		// !!! constants LONGITUDE and AZIMUTH are equaled
		spher[AZIMUTH] = Math.toRadians(geomap[LONGITUDE]);
		// !!! constants LATITUDE and POLAR_ANGLE are equaled
		spher[POLAR_ANGLE] = Math.toRadians(90 - geomap[LATITUDE]);
		// !!! constants ALTITUDE and RADIUS are equaled
		spher[RADIUS] = geomap[ALTITUDE] + RE;
	}

	/**
	 * Converts cartesian coordinate to spherical:
	 * <ul>
	 * <code>(x,y,z) --&gt; (polar angle, azimuthal angle, radius)</code>
	 * </ul>
	 * All angles in radians.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * <li>spherical coordinates will be written into the same array,
	 * <code>coord</code>></li>
	 * <li>this conversion is independent of measurement units</li>
	 * </ul>
	 */
	static public void cartToSpher(double[] coord) {
		cartToSpher(coord, coord);
	}

	/**
	 * Converts cartesian coordinate to spherical:
	 * <ul>
	 * <code>(x,y,z) --&gt; (polar angle, azimuthal angle, radius)</code>
	 * </ul>
	 * All angles in radians.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * <li>array <code>spher</code> can be the same as <code>cart</code></li>
	 * <li>this conversion is independent of measurement units for length</li>
	 * </ul>
	 */
	static public void cartToSpher(double[] cart, double[] spher) {
		double azimuth, polarAngle, radius;
		// to accept case spher == cart put values in temporary variable first
		if (cart[X_AXIS] != 0.0 || cart[Y_AXIS] != 0.0)
			azimuth = Math.atan2(cart[Y_AXIS], cart[X_AXIS]);
		else
			azimuth = 0;
		if (azimuth < 0)
			azimuth = 2 * Math.PI + azimuth;
		radius = Math.sqrt(cart[X_AXIS] * cart[X_AXIS] + cart[Y_AXIS] * cart[Y_AXIS] + cart[Z_AXIS] * cart[Z_AXIS]);
		if (radius > 0)
			polarAngle = Math.acos(cart[Z_AXIS] / radius);
		else
			polarAngle = 0;
		// and copy them to destination
		spher[AZIMUTH] = azimuth;
		spher[POLAR_ANGLE] = polarAngle;
		spher[RADIUS] = radius;
	}

	/**
	 * Converts spherical coordinate to cartesian:
	 * <ul>
	 * <code>(polar angle, azimuthal angle, radius) --&gt; (x,y,z)</code>
	 * </ul>
	 * All angles in radians.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * <li>cartesian coordinates will be written into the same array,
	 * <code>coord</code></li>
	 * <li>this conversion is independent of measurement units for length</li>
	 * </ul>
	 */
	static public void spherToCart(double[] coord) {
		spherToCart(coord, coord);
	}

	/**
	 * Converts spherical coordinate to cartesian:
	 * <ul>
	 * <code>(polar angle, azimuthal angle, radius) --&gt; (x,y,z)</code>
	 * </ul>
	 * All angles in radians.<br>
	 * <br>
	 *
	 * Note:
	 * <ul>
	 * <li>array <code>cart</code> can be the same as <code>spher</code></li>
	 * <li>this conversion is independent of measurement units for length</li>
	 * </ul>
	 */
	static public void spherToCart(double[] spher, double[] cart) {
		double x, y, z;
		// to accept case spher == cart put values in temporary variable first
		x = spher[RADIUS] * Math.sin(spher[POLAR_ANGLE]) * Math.cos(spher[AZIMUTH]);
		y = spher[RADIUS] * Math.sin(spher[POLAR_ANGLE]) * Math.sin(spher[AZIMUTH]);
		z = spher[RADIUS] * Math.cos(spher[POLAR_ANGLE]);
		// and the copy them to destination
		cart[X_AXIS] = x;
		cart[Y_AXIS] = y;
		cart[Z_AXIS] = z;
	}

	/**
	 * Converts from local cartesian to Geographic cartesian:
	 * <ul>
	 * <code>(south, east, zenith) at (Long, Lat) --&gt; (x,y,z)</code>
	 * </ul>
	 * All angles in radians.
	 */
	public static void localToGeoWithGeomapInRad(double[] local, double[] geomap, double[] geo) {
		localToGeoWithGeomapInRad(local, Math.PI / 2. - geomap[LATITUDE], geomap[LONGITUDE], geo);
	}

	public static void localToGeoWithGeomapInRad(double[] local, double polarAngleInRad, double lonInRad, double[] geo) {
		geo[0] = (local[0] * Math.cos(polarAngleInRad) + local[2] * Math.sin(polarAngleInRad)) * Math.cos(lonInRad) - local[1]
				* Math.sin(lonInRad);

		geo[1] = (local[0] * Math.cos(polarAngleInRad) + local[2] * Math.sin(polarAngleInRad)) * Math.sin(lonInRad) + local[1]
				* Math.cos(lonInRad);

		geo[2] = -local[0] * Math.sin(polarAngleInRad) + local[2] * Math.cos(polarAngleInRad);
	}

	/**
	 * Converts from Geographic cartesian to local cartesian:
	 * <ul>
	 * <code>(x,y,z) --&gt; (south, east, zenith) at (Long, Lat)</code>
	 * </ul>
	 * All angles in radians.
	 */
	public void GeoToLocalWithGeomapInRad(double[] geo, double[] geomap, double[] local) {
		double polarAngleInRad = Math.PI / 2. - geomap[LATITUDE];
		local[0] = (geo[0] * Math.cos(geomap[LONGITUDE]) + geo[1] * Math.sin(geomap[LONGITUDE])) * Math.cos(polarAngleInRad) - geo[2]
				* Math.sin(polarAngleInRad);

		local[1] = -geo[0] * Math.sin(geomap[LONGITUDE]) + geo[1] * Math.cos(geomap[LONGITUDE]);

		local[2] = (geo[0] * Math.cos(geomap[LONGITUDE]) + geo[1] * Math.sin(geomap[LONGITUDE])) * Math.sin(polarAngleInRad) + geo[2]
				* Math.cos(polarAngleInRad);
	}

	//
	// ========= M A G N E T I C F I E L D D I R E C T I O N ================
	//

	public void getMagneticField_GSM(double[] pos_GSM, double[] b_GSM) {
		// position in a Cartesian system
		double pos_C[] = new double[3];
		// magnetic field B
		double b_C[] = new double[3];
		double b_C1[] = new double[3];
		// spatial derivative matrix of B
		// don't know where to use those yet
		double db_C[][] = new double[3][3];
		double db_C1[][] = new double[3][3];
		double db_GSM[][] = new double[3][3];
		// magnetic moment, K, [Earth magnetic moment]
		double K;
		double xoff; // offset of the dipoles

		// Earth dipole contribution at the position
		GsmToSm(pos_GSM, pos_C);
		getMagnetDipoleCart(pos_C, b_C, db_C, 1.);
		SmToGsm(b_C, b_GSM);
		SmToGsm1(db_C, db_GSM, CPS, SPS);

		// First mirror dipole contribution, dipole in ZGSM direction

		K = 80.;
		xoff = 10. * (1. + Math.pow(K, 1. / 3.));
		pos_C[X_AXIS] = pos_GSM[X_AXIS] - xoff;
		pos_C[Y_AXIS] = pos_GSM[Y_AXIS];
		pos_C[Z_AXIS] = pos_GSM[Z_AXIS];
		getMagnetDipoleCart(pos_C, b_C, db_C, K);
		scaleVector(b_C, CPS); // * cos Psi
		scaleMatrix(db_C, CPS); // * cos Psi
		addVectors(b_GSM, b_C, b_GSM);
		addMatrices3x3(db_GSM, db_C, db_GSM);

		// Second mirror dipole contribution, antiparallel to XGSM

		K = 25.;
		xoff = 11.3 * (1. + Math.pow(K, 1. / 3.));
		pos_C[X_AXIS] = pos_GSM[Z_AXIS];
		pos_C[Y_AXIS] = pos_GSM[Y_AXIS];
		pos_C[Z_AXIS] = -(pos_GSM[X_AXIS] - xoff);
		getMagnetDipoleCart(pos_C, b_C, db_C, K);
		scaleVector(b_C, SPS); // * sin Psi
		scaleMatrix(db_C, SPS); // * sin Psi
		b_C1[X_AXIS] = -b_C[Z_AXIS];
		b_C1[Y_AXIS] = b_C[Y_AXIS];
		b_C1[Z_AXIS] = b_C[X_AXIS];
		addVectors(b_GSM, b_C1, b_GSM);
		SmToGsm1(db_C, db_C1, 0., -1.);
		addMatrices3x3(db_GSM, db_C1, db_GSM);
	}

	public void getMagnetDipoleCart(double[] P, double[] B, double[][] DB, double K) {
		// ALL VARIABLE IN CARTESIAN COORDINATES
		// P(3) = (X,Y,Z) IN SM IN UNITS OF RE
		// B(3) = (BX,BY,BZ) MAGNETIC FIELD IN SM IN UNITS OF NT
		// DB(3,3) = DBXI/DXJ MAG FIELD SPATIAL DERIVATIVES IN UNITS OF NT/RE
		// K = MAGNETIC MOMENT IN UNITS OF THE EARTH'S MAGNETIC MOMENT

		double M = -31.1e3 * K;
		double X = P[X_AXIS];
		double Y = P[Y_AXIS];
		double Z = P[Z_AXIS];
		double R = Math.sqrt(X * X + Y * Y + Z * Z);
		double R5 = Math.pow(R, 5.);
		double BX = 3. * M / R5 * (Z * X);
		double BY = 3. * M / R5 * (Z * Y);
		double BZ = M / R5 * (3. * Z * Z - R * R);
		B[X_AXIS] = BX;
		B[Y_AXIS] = BY;
		B[Z_AXIS] = BZ;

		// COMPUTE SPATIAL DERIVATIVES

		DB[X_AXIS][X_AXIS] = 3. * M * Z / R5 * (1. - 5 * (X / R) * (X / R));
		DB[X_AXIS][Y_AXIS] = BX * (-5 * Y / (R * R));
		DB[X_AXIS][Z_AXIS] = 3. * M * X / R5 * (1. - 5 * (Z / R) * (Z / R));

		DB[Y_AXIS][Y_AXIS] = 3. * M * Z / R5 * (1. - 5 * (Y / R) * (Y / R));
		DB[Y_AXIS][X_AXIS] = DB[X_AXIS][Y_AXIS];
		DB[Y_AXIS][Z_AXIS] = 3. * M * Y / R5 * (1. - 5 * (Z / R) * (Z / R));

		DB[Z_AXIS][X_AXIS] = DB[X_AXIS][Z_AXIS];
		DB[Z_AXIS][Y_AXIS] = DB[Y_AXIS][Z_AXIS];
		DB[Z_AXIS][Z_AXIS] = M * Z / R5 * (9. - 15 * (Z / R) * (Z / R));
	}

	// tensor conversion for SM to GSM
	// looks kinda suspicious

	private void SmToGsm1(double[][] P, double[][] P1, double COST, double SINT) {
		// CONVERT TENSOR IN SM TO TENSOR IN GSM
		double[][] A = new double[3][3];
		A[0][0] = COST;
		A[0][2] = SINT;
		A[1][1] = 1.;
		A[2][0] = -SINT;
		A[2][2] = COST;
		for (int i = 0; i < 3; i++)
			for (int j = 0; j < 3; j++) {
				P1[i][j] = 0.;
				for (int k = 0; k < 3; k++)
					for (int l = 0; l < 3; l++)
						P1[i][j] += A[i][k] * A[j][l] * P[k][l];
			}
	}

	//
	// ========= P L A S M A P A U S E A N D M A G N E T O P A U S E =======
	//

	public static double ha_0(double mx, double kx, double a) {
		return mx * Math.cos(a) + kx * Math.sin(a);
	}

	public void ppOrbitPlaneIntersect(double[] nn, double[] x, double[] y, double[] z) {
		orbitPlaneIntersect(nn, x, y, z, 0);
	}

	public void mpOrbitPlaneIntersect(double[] nn, double[] x, double[] y, double[] z) {
		orbitPlaneIntersect(nn, x, y, z, 1);
	}

	private void orbitPlaneIntersect(double[] nn, double[] x, double[] y, double[] z, int type) {
		// nn -unit normal in GEI
		// type = 0 to plasmapause intersection, 1 to magnetopause intersection
		// x,y,z solution in RE in cartesian GEI coordinates
		double[] n = new double[3];
		double[] u = new double[3];
		double[] v = new double[3];
		// ;transform orbit normal n from GEI to either SM or GSM
		if (type == 0) {
			GeiToGse(nn, u);
			GseToGsm(u, v);
			GsmToSm(v, n);
		} else {
			GeiToGse(nn, u);
			GseToGsm(u, n);
		}

		// form right handed orthogonal basis (n,m,k)
		double[] m = new double[3];
		double[] k = new double[3];
		con_basis_2(n, m, k);

		int np = x.length;
		double[] r = new double[np];
		double[] a = new double[np];

		// ;compute intersection of orbit plane with surface
		if (type == 0) {
			pp_0(n, m, k, r, a);
		} else {
			mp_0(n, m, k, r, a);
		}

		for (int i = 0; i < np; i++) {
			x[i] = r[i] * (m[0] * Math.cos(a[i]) + k[0] * Math.sin(a[i]));
			y[i] = r[i] * (m[1] * Math.cos(a[i]) + k[1] * Math.sin(a[i]));
			z[i] = r[i] * (m[2] * Math.cos(a[i]) + k[2] * Math.sin(a[i]));

			// convert coordinates back to GEI
			v[0] = x[i];
			v[1] = y[i];
			v[2] = z[i];
			if (type == 0) {
				SmToGsm(v, u);
				GsmToGse(u, v);
				GseToGei(v, u);
			} else {
				GsmToGse(v, n);
				GseToGei(n, u);
			}
			x[i] = u[0];
			y[i] = u[1];
			z[i] = u[2];
		}
	}

	// compute intersection of orbit plane given by orbit normal 'n'
	// with the plasmapause
	// n,m,k is a right handed orthonormal basis in SM whose
	// r = l*(rho/r)^2 -> r^3 = l*rho^2 ;l-lshell value in RE
	private void pp_0(double[] n, double[] m, double[] k, double[] r, double[] a) {
		double l = L_SHELL; // (l-shell of 4)
		int np = r.length;
		for (int i = 0; i < np; i++) {
			a[i] = (double) i / (double) np * 2 * Math.PI;
			r[i] = l * (Math.pow(ha_0(m[0], k[0], a[i]), 2) + Math.pow(ha_0(m[1], k[1], a[i]), 2));
		}
	}

	// compute intersection of orbit plane given by orbit normal 'n'
	// with the magnetopause, using roelof sibeck model
	// n,m,k is a right handed orthonormal basis in GSM or GSE whose
	// a0*x^2 + b0*x + (y^2+z^2) + c = 0
	private void mp_0(double[] n, double[] m, double[] k, double[] r, double[] a) {
		double aa, bb, cc;
		double[] coef = new double[3];
		double press = 2.5;
		double bz0 = 0.;
		mp_coef_sib(coef, press, bz0);
		double a0 = coef[0];
		double b0 = coef[1];
		double c0 = coef[2];
		int np = r.length;
		for (int i = 0; i < np; i++) {
			a[i] = (double) i / (double) (np - 1) * 2 * Math.PI;
			aa = a0 * Math.pow(ha_0(m[0], k[0], a[i]), 2) + Math.pow(ha_0(m[1], k[1], a[i]), 2) + Math.pow(ha_0(m[2], k[2], a[i]), 2);
			bb = b0 * ha_0(m[0], k[0], a[i]);
			cc = c0;
			r[i] = quadMax(aa, bb, cc); // solve quadratic equation
										// aa*r^2+bb*r+cc=0
		}
	}

	// press - solar wind pressure in nPa
	// bz - IMF Bz component in nT
	// note: I clip bz if |bz| > 6.5 nT
	// also for low pressure (i.e. .1 nT) the subsolar point moves earthward
	// instead of anti-earthward
	private void mp_coef_sib(double[] coef, double press, double bz0) {
		double bz = 0;
		double x, y, z;
		if (Math.abs(bz0) > 6.5) {
			bz = bz0 / Math.abs(bz0) * 6.5;
		} else {
			bz = bz0;
		}
		y = bz + 0.1635;
		x = press / 2.088;
		z = Math.log(x);
		coef[0] = 0.171 * Math.pow(x, -0.474 - 0.616 * z + 0.023 * y) * Math.exp(-0.043 * y + 0.0391 * y * y);
		coef[1] = 18.80 * Math.pow(x, -0.120 - 0.030 * z + 0.036 * y) * Math.exp(-0.037 * y + 0.0002 * y * y);
		coef[2] = -220.8 * Math.pow(x, -0.290 - 0.110 * z + 0.018 * y) * Math.exp(-0.012 * y + 0.0017 * y * y);
	}

	//
	// =============== t e r m i n a t o r =====================================
	//

	// Scott Boardsen,2/9/2000
	// alpha in radians angle between observer and sun relative to Earth's
	// center
	// zeta in radians the angle from the X plot axis to the sun direction
	// projected onto the plot.
	// Function fill in Polygon pol in km units
	static public void terminatorPoligon(double alpha, double zeta, Polygon pol) {
		double beta = Math.PI / 2. - alpha;
		double gamma;
		double x, y;
		int len = pol.xpoints.length;
		for (int i = 0; i < len; i++) {
			gamma = (double) ((double) i * 2. / (double) (len - 2.)) * Math.PI;
			if (i < len / 2) {
				x = -Math.sin(beta) * Math.sin(gamma);
				y = Math.cos(gamma);
			} else {
				gamma = Math.PI - gamma;
				x = -Math.sin(gamma);
				y = Math.cos(gamma);
			}
			pol.xpoints[i] = (int) Math.round(RE * (x * Math.cos(zeta) - y * Math.sin(zeta)));
			pol.ypoints[i] = (int) Math.round(RE * (x * Math.sin(zeta) + y * Math.cos(zeta)));
		}
	}

	//
	// ====================== S U P P O R T M A T H ==========================
	//

	/**
	 * Calculate cross (or vector) product of two 3-dimensional vectors, *v1*
	 * and *v2*, and put result into another vector, *out*
	 *
	 * @param v1
	 *            double[], first vector
	 * @param v2
	 *            double[], second vector
	 *
	 * @param out
	 *            double[], cross/vector product will be put into this
	 *            3-dimensional vector
	 */
	public static void crossp(double[] v1, double[] v2, double[] out) {
		out[X_AXIS] = v1[Y_AXIS] * v2[Z_AXIS] - v2[Y_AXIS] * v1[Z_AXIS];
		out[Y_AXIS] = v1[Z_AXIS] * v2[X_AXIS] - v2[Z_AXIS] * v1[X_AXIS];
		out[Z_AXIS] = v1[X_AXIS] * v2[Y_AXIS] - v2[X_AXIS] * v1[Y_AXIS];
	}

	public static double total(double[] arr) {
		double sum = 0;
		for (int i = 0; i < arr.length; i++)
			sum += arr[i];
		return sum;
	}

	/**
	 * Normalize given vector *arr* in-place and return length of original
	 * vector *arr*. Note that vector *arr* can be of any dimension.
	 *
	 * @param arr
	 *            double[], not normalized vector on input, will be normalized
	 *            upon the exit .
	 *
	 * @return double, length of orinal vector *arr*
	 */
	public static double normalize(double[] arr) {
		double sum = 0;
		for (int i = 0; i < arr.length; i++)
			sum += arr[i] * arr[i];
		double norm = Math.sqrt(sum); // norm = length of vector
		for (int i = 0; i < arr.length; i++)
			arr[i] /= norm;
		return norm;
	}

	/**
	 * Having 3-dimensional vector *x*, construct vector *y* and *z* such that
	 * these three vectors will form orthonormal basis (vector *x* will be
	 * normalized to meet this criteria). Notes: 1. Of course, there exist
	 * continuum possibilities to do this. We just use one of them here. 2.
	 * Vector *x* should not be 0-vector.
	 *
	 * @param x
	 *            double[], given vector
	 * @param y
	 *            double[]
	 * @param z
	 *            double[]
	 */
	public static void con_basis_2(double[] x, double[] y, double[] z) {
		normalize(x);
		// Choose two coordinates of vector *x* so that at least one of them
		// is not equal to zero.
		int indexOfMaxElement = 0;
		double max = Math.abs(x[0]);
		for (int i = 1; i < 3; i++) {
			if (Math.abs(x[i]) > max) {
				max = Math.abs(x[i]);
				indexOfMaxElement = i;
			}
		}
		// construct vector *y* so that it is perpendicular to *x* (scalar
		// product
		// of x and y is equal to zero)
		y[(indexOfMaxElement + 1) % 3] = x[indexOfMaxElement];
		y[indexOfMaxElement] = -x[(indexOfMaxElement + 1) % 3];
		y[(indexOfMaxElement + 2) % 3] = 0;
		normalize(y);
		crossp(x, y, z);
		// we do not need normalize *z* as it is already normalized, because
		// length(z) = length(x) * length(y) sin(between x and y) =
		// 1 * 1 * sin(90deg) = 1
	}

	public static double quadMax(double a, double b, double c) {
		double d = b * b - 4. * a * c;
		if (d < 0)
			d = 1000.;
		double q = -0.5 * (b + sgn(b) * Math.sqrt(d));
		return Math.max(q / a, c / q);
	}

	public static double sgn(double x) {
		if (x > 0)
			return 1;
		return -1;
	}

	public static void addVectors(double[] v1, double[] v2, double[] result) {
		for (int i = 0; i < 3; i++)
			result[i] = v1[i] + v2[i];
	}

	public static void addMatrices3x3(double[][] m1, double[][] m2, double[][] result) {
		for (int i = 0; i < 3; i++)
			for (int j = 0; j < 3; j++)
				result[i][j] = m1[i][j] + m2[i][j];
	}

	public static void scaleVector(double v[], double scalar) {
		scaleVector(v, scalar, v);
	}

	public static void scaleVector(double in[], double scalar, double out[]) {
		for (int i = 0; i < 3; i++)
			out[i] = in[i] * scalar;
	}

	public static void scaleMatrix(double m[][], double scalar) {
		for (int i = 0; i < 3; i++)
			for (int j = 0; j < 3; j++)
				m[i][j] *= scalar;
	}

	// calculate directional cosines from vector components

	public static double vectorToAngles(double[] v, double[] cos, double[] angle_deg) {
		double length = Math.sqrt(v[X_AXIS] * v[X_AXIS] + v[Y_AXIS] * v[Y_AXIS] + v[Z_AXIS] * v[Z_AXIS]);

		cos[X_AXIS] = v[X_AXIS] / length;
		cos[Y_AXIS] = v[Y_AXIS] / length;
		cos[Z_AXIS] = v[Z_AXIS] / length;

		angle_deg[X_AXIS] = Math.toDegrees(Math.acos(cos[X_AXIS]));
		angle_deg[Y_AXIS] = Math.toDegrees(Math.acos(cos[Y_AXIS]));
		angle_deg[Z_AXIS] = Math.toDegrees(Math.acos(cos[Z_AXIS]));
		return length;
	}

	/**
	 * Read Geodipole moment coefficients from file if any.<br>
	 * <br>
	 * 
	 * File structure
	 * <ul>
	 * <li>this the text file</li>
	 * <li>if line starts with semicolon(;), or sharp(#) symbols, or line
	 * consists of only spaces then it is considered as comment line, otherwice
	 * it is considered as data line (exclusion is the 1st line)</li>
	 * <li>first line should start with "GeopackData "</li>
	 * <li>all subsequent lines are either comment or data lines</li>
	 * <li>data lines contains four items divided by commas:
	 * <ul>
	 * <code>TimeScale</code>, <code>G10</code>, <code>G11</code>,
	 * <code>H11</code>
	 * </ul>
	 * where <code>TimeScale</code> can be presented using one of the following
	 * forms:
	 * <ul>
	 * <li><code>yyyy</code></li>
	 * <li><code>yyyy/MM</code></li>
	 * <li><code>yyyy/MM/dd</code></li>
	 * <li><code>yyyy/MM/dd (DDD)</code></li>
	 * <li><code>yyyy/MM/dd (DDD) HH</code></li>
	 * <li><code>yyyy/MM/dd HH</code></li>
	 * <li><code>yyyy/MM/dd (DDD) HH:mm</code></li>
	 * <li><code>yyyy/MM/dd HH:mm</code></li>
	 * <li><code>yyyy/MM/dd (DDD) HH:mm:ss</code></li>
	 * <li><code>yyyy/MM/dd HH:mm:ss</code></li>
	 * </ul>
	 *
	 * instead of slash (/) symbol you can use also dot (.), or dash(-) symbols.
	 * <br>
	 * And <code>G10</code>, <code>G11</code>, and <code>H11</code> are
	 * geodipole moment coefficient, that correspond given TimeScale (for last
	 * data line it have a little different meaning, see later)</li>
	 * <li>all data line should be given in ascending time order</li>
	 * <li>last data line has slightly different meaning:
	 * <ul>
	 * <li><code>TimeScale</code> is the time from which it is not recommended to
	 * use GeoPack</li>
	 * <li><code>G10</code>, <code>G11</code>, and <code>H11</code> are linear
	 * extrapolation delta per year for <code>G10</code>, <code>G11</code>, and
	 * <code>H11</code> beginning from before last data line</li>
	 * </ul>
	 * </li>
	 * </ul>
	 */
	static private void readGeodipoleMomentCoeff() {
		datafileInUse = false;
		datafileErrorCode = DATAFILE_ERROR_OK;
		softwareObsolete = false;
		File datafile = new File(DATAFILE_NAME);
		if (!datafile.isFile())
			return;
		double version = 0;
		Vector<TimeScale> vTimes = new Vector<TimeScale>(50, 50);
		Vector<Double> vG10 = new Vector<Double>(50, 50);
		Vector<Double> vG11 = new Vector<Double>(50, 50);
		Vector<Double> vH11 = new Vector<Double>(50, 50);
		int qty = 0;
		BufferedReader br = null;
		String line;
		String[] lineData;
		int lineCount = 0;
		int[] dt = null;
		try {
			br = new BufferedReader(new FileReader(datafile));
			while (br.ready()) {
				line = br.readLine();
				if (line == null)
					break;
				lineCount++;
				line = StrUtil.trimRight(line, ' ');
				if (lineCount == 1) {
					if (!line.startsWith(FILE_SIGNATURE)) {
						Console.showError("Geopack file " + DATAFILE_NAME + ": bad signature");
						datafileErrorCode = DATAFILE_ERROR_SIGNATURE;
						break;
					} else {
						version = FC.StringToDouble(line.substring(FILE_SIGNATURE.length() + 1).trim());
						if (version < 2) {
							Console.showError("Geopack file " + DATAFILE_NAME + " is obsolete");
							datafileErrorCode = DATAFILE_ERROR_OBSOLETE;
							break;
						} else if (version > DATA_FILE_VERSION) {
							Console.showError("Geopack file " + DATAFILE_NAME + " has a newer version" + System.lineSeparator() + "you need to upgrade software");
							softwareObsolete = true;
							break;
						}
					}
					continue;
				}
				if (line.length() == 0 || line.startsWith("#") || line.startsWith(";"))
					continue;
				lineData = StrUtil.listToArray( line, ",");
				if (lineData.length != 4) {
					Console.showError("Geopack file " + DATAFILE_NAME + " line " + lineCount + ":" + System.lineSeparator()
							+ "  should be exactly four elements divided by comma");
					datafileErrorCode = DATAFILE_ERROR_ELEMENT_QTY;
					break;
				}
				vTimes.addElement(TimeScale.parse(lineData[0]));
				if (!StrUtil.isFloatingPointNumber(lineData[1])) {
					Console.showError("Geopack file " + DATAFILE_NAME + " line " + lineCount + ": G10 should be number");
					datafileErrorCode = DATAFILE_ERROR_NUMBER_FORMAT;
					break;
				}
				vG10.addElement(new Double(lineData[1]));
				if (!StrUtil.isFloatingPointNumber(lineData[2])) {
					Console.showError("Geopack file " + DATAFILE_NAME + " line " + lineCount + ": G11 should be number");
					datafileErrorCode = DATAFILE_ERROR_NUMBER_FORMAT;
					break;
				}
				vG11.addElement(new Double(lineData[2]));
				if (!StrUtil.isFloatingPointNumber(lineData[3])) {
					Console.showError("Geopack file " + DATAFILE_NAME + " line " + lineCount + ": H11 should be number");
					datafileErrorCode = DATAFILE_ERROR_NUMBER_FORMAT;
					break;
				}
				vH11.addElement(new Double(lineData[3]));
				if (qty > 0) {
					if (!vTimes.elementAt(qty - 1).before(vTimes.elementAt(qty))) {
						Console.showError("Geopack file " + DATAFILE_NAME + " line " + lineCount + ": date/time should increase");
						datafileErrorCode = DATAFILE_ERROR_TIME_DECREASE;
						break;
					}
				}
				qty++;
			}
		} catch (IOException e) {
			Console.showError("File " + DATAFILE_NAME + " line " + (++lineCount) + ":" + System.lineSeparator() + "  " + e.toString());
			datafileErrorCode = DATAFILE_ERROR_IO;
		} catch (Exception e) // TODO catch( IllegalTimeScaleStringException e )
		{
			Console.showError("File " + DATAFILE_NAME + " line " + lineCount + ":" + System.lineSeparator() + "  " + e.toString());
			datafileErrorCode = DATAFILE_ERROR_TIME_FORMAT;
		} finally {
			if (br != null)
				try {
					br.close();
				} catch (IOException ex) {
					// TODO Check Util.printThreadStackTrace( ex );
					ex.printStackTrace();
				}
		}

		if (datafileErrorCode == DATAFILE_ERROR_OK) {
			if (lineCount == 0) {
				datafileErrorCode = DATAFILE_ERROR_ZEROLENGTH;
				Console.showError("Geopack file " + DATAFILE_NAME + " of zero length");
			} else if (qty == 0) {
				Console.showError("Geopack file " + DATAFILE_NAME + " doesn't contain entries");
				datafileErrorCode = DATAFILE_ERROR_NO_ENTRIES;
			} else if (qty == 1) {
				Console.showError("Geopack file " + DATAFILE_NAME + " contains only one entry");
				datafileErrorCode = DATAFILE_ERROR_LACK_OF_ENTRIES;
			}
		}

		if (datafileErrorCode != DATAFILE_ERROR_OK || softwareObsolete)
			return;
		if (!MAX_DATE.before(vTimes.lastElement()))
			return;

		TIMES = new TimeScale[qty];
		G10 = new double[qty];
		G11 = new double[qty];
		H11 = new double[qty];
		for (int i = 0; i < qty; i++) {
			TIMES[i] = vTimes.elementAt(i);
			G10[i] = vG10.elementAt(i).doubleValue();
			G11[i] = vG11.elementAt(i).doubleValue();
			H11[i] = vH11.elementAt(i).doubleValue();
		}
		datafileInUse = true;
		MIN_DATE = TIMES[0];
		MAX_DATE = new TimeScale(TIMES[TIMES.length - 1].getTimeInMinutes());
		MAX_DATE.add(Calendar.MILLISECOND, -1);
		MIN_DATE_IN_MINUTES = MIN_DATE.getTimeInMinutes();
		MAX_DATE_IN_MINUTES = MAX_DATE.getTimeInMinutes();
	}
	
	public double[] getMagneticCoordinates(final double[] coords, double altitude, TimeScale time) {
		double geoLatLong[] = { coords[0], coords[1], altitude };
		double magLatLong[] = new double[3];
		recalc(time);
		geomapToMagmap(geoLatLong, magLatLong);
		return new double[] { magLatLong[0], magLatLong[1] };
	}

	public double[] getMagneticCoordinates(final double[] coords, TimeScale time) {
		return getMagneticCoordinates(coords, 0, time);
	}
	
}