package edu.uml.lgdc.geospace;

import java.util.Calendar;
import java.util.Vector;

import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.time.RealHours;
import edu.uml.lgdc.time.TimeScale;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
public class GroundSolar {
	static public final double ALWAYS_BELOW_HORIZON = -99;
	static public final double ALWAYS_ABOVE_HORIZON = 99;

	static private final int MAX_DAYS_TO_SKIP = 365;

	// local time of sunrise in hours
	private double sunriseInHoursLT;
	// local time of sunset in hours
	private double sunsetInHoursLT;

	// UT time of sunrise in hours
	private double sunriseInHoursUT;
	// UT time of sunset in hours
	private double sunsetInHoursUT;

	// local time of sunrise in hours
	private TimeScale sunriseTimeLT;
	// local time of sunset in hours
	private TimeScale sunsetTimeLT;

	// UT time of sunrise in hours
	private TimeScale sunriseTimeUT;
	// UT time of sunset in hours
	private TimeScale sunsetTimeUT;

	// sun declination angle at noon of this day
	private double noonSunDeclination;
	// sun zenith angle at noon of this day
	private double noonSunZenith;

	// sun declination angle at given time
	private double sunDeclination;
	// sun zenith angle at given time
	private double sunZenith;

	// Location for which it was calculated
	private double latitude; // Geographical north latitude (in Degrees)
	private double longitude; // Geographical east longitude (in Degrees)
	// Time for which it was calculated
	private TimeScale ut;
	// Height (in km) above the sea level
	private double height;

	// Local time offset in hours to UT
	private double offsetInHoursToUT = 0;

	// Local discrete time offset
	// Used only in string representation
	private double offset = 0;

	// Local (longitude) Time for which it was calculated.
	// Only important here Day Of Year (dayOfYear) and Time of Day (hours)
	@SuppressWarnings("unused")
	private int dayOfYear;
	@SuppressWarnings("unused")
	private double hours;

	/**
	 * Constructor Here: latitude - geographical north latitude (in Degrees)
	 * longitude - geographical east longitude (in Degrees) ut - Universal time
	 */
	public GroundSolar(double latitude, double longitude, TimeScale ut) {
		this(latitude, longitude, 0, ut);
	}

	/**
	 * Constructor Here: latitude - geographical north latitude (in Degrees)
	 * longitude - geographical east longitude (in Degrees) height - height
	 * above the sea level (in km) ut - Universal time
	 */
	public GroundSolar(double latitude, double longitude, double height, TimeScale ut) {
		this.latitude = latitude;
		this.longitude = longitude;
		this.height = height;
		this.ut = ut;
		setup();
	}

	/**
	 * Set time *ut* as UT time and recalculates.
	 */
	public void setTime(TimeScale ut) {
		this.ut = ut;
		setup();
	}

	/**
	 * Set height (in km) and recalculates.
	 */
	public void setHeight(double height) {
		this.height = height;
		setup();
	}

	/**
	 * Returns <true> if normal sunrise and sunset times, otherwise returns
	 * <false>.
	 */
	public boolean isNormalTrajectory() {
		return sunriseInHoursLT != ALWAYS_ABOVE_HORIZON && sunriseInHoursLT != ALWAYS_BELOW_HORIZON;
	}

	/**
	 * Returns <true> if position of sun always above horizon, otherwise returns
	 * <false>.
	 */
	public boolean isAlwaysAbove() {
		return sunriseInHoursLT == ALWAYS_ABOVE_HORIZON;
	}

	/**
	 * Returns <true> if position of sun always below horizon, otherwise returns
	 * <false>.
	 */
	public boolean isAlwaysBelow() {
		return sunriseInHoursLT == ALWAYS_BELOW_HORIZON;
	}

	/**
	 * Returns sunrise local time in hours at given day ( dayOfYear )
	 */
	public double getSunriseTimeInHoursLT() {
		return sunriseInHoursLT;
	}

	/**
	 * Returns sunset local time in hours at given day ( dayOfYear )
	 */
	public double getSunsetTimeInHoursLT() {
		return sunsetInHoursLT;
	}

	/**
	 * Returns sunrise UT time in hours at given day ( dayOfYear )
	 */
	public double getSunriseTimeInHoursUT() {
		return sunriseInHoursUT;
	}

	/**
	 * Returns sunset UT time in hours at given day ( dayOfYear )
	 */
	public double getSunsetTimeInHoursUT() {
		return sunsetInHoursUT;
	}

	/**
	 * Returns sunrise local time as TimeScale object.
	 */
	public TimeScale getSunriseTimeLT() {
		return sunriseTimeLT;
	}

	/**
	 * Returns sunset local time as TimeScale object.
	 */
	public TimeScale getSunsetTimeLT() {
		return sunsetTimeLT;
	}

	/**
	 * Returns sunrise UT time as TimeScale object.
	 */
	public TimeScale getSunriseTimeUT() {
		return sunriseTimeUT;
	}

	/**
	 * Returns sunset UT time as TimeScale object.
	 */
	public TimeScale getSunsetTimeUT() {
		return sunsetTimeUT;
	}

	/**
	 * Returns sunset offset in minutes
	 */
	public double getOffsetInHoursToUT() {
		return offsetInHoursToUT;
	}

	/**
	 * Returns declination angle at given time
	 */
	public double getDeclination() {
		return sunDeclination;
	}

	/**
	 * Returns zenith angle at given time
	 */
	public double getZenith() {
		return sunZenith;
	}

	/**
	 * Returns declination angle at noon time in given day ( dayOfYear )
	 */
	public double getNoonDeclination() {
		return noonSunDeclination;
	}

	/**
	 * Returns zenith angle at noon time in given day ( dayOfYear )
	 */
	public double getNoonZenith() {
		return noonSunZenith;
	}

	/**
	 * Sets local time offset.
	 */
	public void setLocalTimeOffset(double offset) {
		this.offset = offset;
	}

	/**
	 * Returns true if day Parameter Longitude Hours
	 */
	public boolean isDay(double lcHour) {
		return !isNight(lcHour);
	}

	/**
	 * Returns true if night Parameter Longitude Hours
	 */
	public boolean isNight(double lcHour) {
		boolean isNight = false;
		if (isAlwaysBelow())
			isNight = true;
		else if (isNormalTrajectory()) {
			if (lcHour > getSunsetTimeInHoursLT() || lcHour < getSunriseTimeInHoursLT())
				isNight = true;
		}
		return isNight;
	}

	/**
	 * Returns true if day
	 */
	public boolean isDay() {
		return !isNight();
	}

	/**
	 * Returns true if night
	 */
	public boolean isNight() {
		boolean isNight = false;
		if (isAlwaysBelow())
			isNight = true;
		else if (isNormalTrajectory())
			isNight = sunsetTimeUT.before(ut) || ut.before(sunriseTimeUT);
		return isNight;
	}

	/**
	 * Visual representation of values.
	 */
	public String toString() {
		double localSunrise = sunriseInHoursUT;
		if (localSunrise != ALWAYS_BELOW_HORIZON && localSunrise != ALWAYS_ABOVE_HORIZON) {
			localSunrise += offset;
			if (localSunrise >= 24)
				localSunrise -= 24;
			else if (localSunrise < 0)
				localSunrise += 24;
		}
		double localSunset = sunsetInHoursUT;
		if (localSunset != ALWAYS_BELOW_HORIZON && localSunset != ALWAYS_ABOVE_HORIZON) {
			localSunset += offset;
			if (localSunset >= 24)
				localSunset -= 24;
			else if (localSunset < 0)
				localSunset += 24;
		}
		return "latitude = " + FC.DoubleToString(latitude, 7, 2) + "\r\n" + "longitude = " + FC.DoubleToString(longitude, 7, 2) + "\r\n" + "ut = "
				+ ut.toHumanUT() + "\r\n" + "sunrise time = " + timeToHuman(sunriseInHoursLT) + "\r\n" + "sunset time = "
				+ timeToHuman(sunsetInHoursLT) + "\r\n" + "sunrise time local = " + timeToHuman(localSunrise) + "\r\n" + "sunset time local = "
				+ timeToHuman(localSunset) + "\r\n" + "sun declination = " + FC.DoubleToString(sunDeclination, 7, 2) + "\r\n" + "sun zenith = "
				+ FC.DoubleToString(sunZenith, 7, 2) + "\r\n" + "noon sun declination = " + FC.DoubleToString(noonSunDeclination, 7, 2) + "\r\n"
				+ "noon sun zenith = " + FC.DoubleToString(noonSunZenith, 7, 2) + "\r\n";
	}

	/**
	 * Return exact time as string with treating special cases for time values:
	 * ALWAYS_ABOVE_HORIZON and ALWAYS_BELOW_HORIZON
	 */
	private String timeToHuman(double time) {
		String timeStr = "";
		if (time == ALWAYS_ABOVE_HORIZON)
			timeStr = "ALWAYS ABOVE";
		else if (time == ALWAYS_BELOW_HORIZON)
			timeStr = "ALWAYS BELOW";
		else
			timeStr = RealHours.toHuman(time);
		return timeStr;
	}

	/**
	 * Calculates and Set up fields with correct values.
	 */
	private void setup() {
		setup(latitude, longitude, height, TimeScale.getLongitudeTime(longitude, ut));
	}

	/**
	 * Set up fields with correct values Here: latitude - geographical north
	 * latitude (in Degrees) longitude - geographical east longitude (in
	 * Degrees) height - height above the sea level (in km) longitudeTime -
	 * TimeScale object containing time related to this longitude
	 */
	private void setup(double latitude, double longitude, double height, TimeScale longitudeTime) {
		setup(latitude, longitude, height, longitudeTime.get(Calendar.DAY_OF_YEAR), RealHours.toRealHours(longitudeTime));
	}

	/**
	 * Set up fields with correct values Here: latitude - geographical north
	 * latitude longitude - geographical east longitude localDayOfYear - related
	 * to this longitude longitudeHour - exact hour (as double type) related to
	 * this longitude
	 */
	private void setup(double latitude, double longitude, double height, int localDayOfYear, double longitudeHour) {
		calc(latitude, longitude, height, localDayOfYear, longitudeHour);
		double tmpSunDeclination = sunDeclination;
		double tmpSunZenith = sunZenith;
		calc(latitude, longitude, height, localDayOfYear, 12.0);
		noonSunDeclination = sunDeclination;
		noonSunZenith = sunZenith;
		sunDeclination = tmpSunDeclination;
		sunZenith = tmpSunZenith;
		// Set sunrise and sunset time as TimeScale objects
		if (sunriseInHoursUT == ALWAYS_ABOVE_HORIZON || sunriseInHoursUT == ALWAYS_ABOVE_HORIZON) {
			sunriseTimeLT = null;
			sunsetTimeLT = null;
			sunriseTimeUT = null;
			sunsetTimeUT = null;
		} else {
			RealHours rh;
			rh = new RealHours(sunriseInHoursLT);
			sunriseTimeLT = TimeScale.getLongitudeTime(longitude, ut);
			sunriseTimeLT.set(Calendar.HOUR_OF_DAY, rh.getHours());
			sunriseTimeLT.set(Calendar.MINUTE, rh.getMinutes());
			sunriseTimeLT.set(Calendar.SECOND, rh.getSeconds());
			sunriseTimeLT.set(Calendar.MILLISECOND, rh.getMillisecs());
			rh = new RealHours(sunsetInHoursLT);
			sunsetTimeLT = TimeScale.getLongitudeTime(longitude, ut);
			sunsetTimeLT.set(Calendar.HOUR_OF_DAY, rh.getHours());
			sunsetTimeLT.set(Calendar.MINUTE, rh.getMinutes());
			sunsetTimeLT.set(Calendar.SECOND, rh.getSeconds());
			sunsetTimeLT.set(Calendar.MILLISECOND, rh.getMillisecs());
			sunriseTimeUT = TimeScale.getTimeUT(longitude, sunriseTimeLT);
			sunsetTimeUT = TimeScale.getTimeUT(longitude, sunsetTimeLT);
		}
	}

	/**
	 * Calculate the solar declination, zenith angle, and sunrise & sunset times
	 * - based on Newbern Smith's algorithm Sunset and sunrise time considered
	 * on sea level of Earth. [Leo McNamara, 1-sep-86, last modified 16-jun-87]
	 * {Dieter Bilitza, 30-oct-89, modified for IRI application}
	 *
	 * Parameters: latitude - geographical north latitude longitude -
	 * geographical east longitude localDayOfYear - related to this longitude
	 * longitudeHour - exact hour (as double type) related to this longitude
	 */
	@SuppressWarnings("unused")
	private void calc(double latitude, double longitude, int localDayOfYear, double longitudeHour) {
		calc(latitude, longitude, 0, localDayOfYear, longitudeHour);
	}

	/**
	 * Calculate the solar declination, zenith angle, and sunrise & sunset times
	 * - based on Newbern Smith's algorithm Sunset and sunrise time considered
	 * at the given height above sea level of Earth. [Leo McNamara, 1-sep-86,
	 * last modified 16-jun-87] {Dieter Bilitza, 30-oct-89, modified for IRI
	 * application}
	 *
	 * Parameters: latitude - geographical north latitude (in Degrees) longitude
	 * - geographical east longitude (in Degrees) height - height above the sea
	 * level (in km.) localDayOfYear - related to this longitude longitudeHour -
	 * exact hour (as double type) related to this longitude
	 */
	private void calc(double latitude, double longitude, double height, int localDayOfYear, double longitudeHour) {
		double a, b, ch, cosphi, cosx, dc, dcl, eqt, et, fa, phi, secphi, td, te, tf, wlon;

		// amplitudes of Fourier coefficients -- 1955 epoch.................
		double p1 = 0.017203534;
		double p2 = 0.034407068;
		double p3 = 0.051610602;
		double p4 = 0.068814136;
		double p6 = 0.103221204;

		// transform kilometers to meters
		height *= 1000;

		offsetInHoursToUT = longitude / 15.;
		if (offsetInHoursToUT > 12.) {
			offsetInHoursToUT = 24. - offsetInHoursToUT;
			offsetInHoursToUT = -offsetInHoursToUT;
		}

		// s/r is formulated in terms of WEST longitude.......................
		wlon = 360. - longitude;

		// time of equinox for 1980...........................................
		td = localDayOfYear + (longitudeHour + wlon / 15.) / 24.;
		te = td + 0.9369;

		// declination of the sun..............................................
		dcl = 23.256 * Math.sin(p1 * (te - 82.242)) + 0.381 * Math.sin(p2 * (te - 44.855)) + 0.167 * Math.sin(p3 * (te - 23.355)) - 0.013
				* Math.sin(p4 * (te + 11.97)) + 0.011 * Math.sin(p6 * (te - 10.41)) + 0.339137;
		sunDeclination = dcl;
		dc = Math.toRadians(dcl);

		// the equation of time................................................
		tf = te - 0.5;
		eqt = -7.38 * Math.sin(p1 * (tf - 4.)) - 9.87 * Math.sin(p2 * (tf + 9.)) + 0.27 * Math.sin(p3 * (tf - 53.)) - 0.2
				* Math.cos(p4 * (tf - 17.));
		et = Math.toRadians(eqt) / 4.0;

		fa = Math.toRadians(latitude);
		phi = 0.26179939 * (longitudeHour - 12.0) + et;

		a = Math.sin(fa) * Math.sin(dc);
		b = Math.cos(fa) * Math.cos(dc);
		cosx = a + b * Math.cos(phi);
		if (cosx > 1.0)
			cosx = 1.0;
		else if (cosx < -1.0)
			cosx = -1.0;
		sunZenith = Math.toDegrees(Math.acos(cosx));

		// calculate sunrise and sunset times -- at the ground...........
		// see Explanatory Supplement to the Ephemeris (1961) pg 401......
		// sunrise at height h metres is at...............................
		// chi(h) = 90.83 + 0.0347 * sqrt(h)........................
		// this includes corrections for horizontal refraction and........
		// semi-diameter of the solar disk................................
		ch = Math.cos(Math.toRadians(90.83 + 0.0347 * Math.sqrt(height)));
		cosphi = (ch - a) / b;

		// if abs(secphi) > 1., sun does not rise/set.....................
		// allow for sun never setting - high latitude summer.............
		secphi = 999999.;
		if (cosphi != 0)
			secphi = 1.0 / cosphi;
		sunsetInHoursLT = ALWAYS_ABOVE_HORIZON;
		sunriseInHoursLT = ALWAYS_ABOVE_HORIZON;
		if (secphi > -1.0 && secphi <= 0.) {
			sunsetInHoursUT = sunsetInHoursLT;
			sunriseInHoursUT = sunriseInHoursLT;
			return;
		}

		// allow for sun never rising - high latitude winter..............
		sunsetInHoursLT = ALWAYS_BELOW_HORIZON;
		sunriseInHoursLT = ALWAYS_BELOW_HORIZON;
		if (secphi > 0.0 && secphi < 1.0) {
			sunsetInHoursUT = sunsetInHoursLT;
			sunriseInHoursUT = sunriseInHoursLT;
			return;
		}

		if (cosphi > 1.0)
			cosphi = 1.0;
		if (cosphi < -1.0)
			cosphi = -1.0;
		phi = Math.acos(cosphi);
		et = et / 0.26179939;
		phi = phi / 0.26179939;

		sunriseInHoursLT = 12. - phi - et;
		sunsetInHoursLT = 12. + phi - et;

		if (sunriseInHoursLT < 0) {
			sunriseInHoursLT = (-sunriseInHoursLT) % 24;
			if (sunriseInHoursLT > 0)
				sunriseInHoursLT = 24 - sunriseInHoursLT;
		}
		if (sunsetInHoursLT >= 24)
			sunsetInHoursLT = sunsetInHoursLT % 24;

		sunriseInHoursUT = sunriseInHoursLT;
		sunsetInHoursUT = sunsetInHoursLT;

		sunriseInHoursUT = sunriseInHoursUT - longitude / 15.;
		if (sunriseInHoursUT < 0) {
			sunriseInHoursUT = (-sunriseInHoursUT) % 24;
			if (sunriseInHoursUT > 0)
				sunriseInHoursUT = 24 - sunriseInHoursUT;
		}
		if (sunriseInHoursUT >= 24)
			sunriseInHoursUT = sunriseInHoursUT % 24;

		sunsetInHoursUT = sunsetInHoursUT - longitude / 15.;
		if (sunsetInHoursUT < 0) {
			sunsetInHoursUT = (-sunsetInHoursUT) % 24;
			if (sunsetInHoursUT > 0)
				sunsetInHoursUT = 24 - sunsetInHoursUT;
		}
		if (sunsetInHoursUT >= 24)
			sunsetInHoursUT = sunsetInHoursUT % 24;
	}

	public TimeScale[][] getCoverage(TimeScale startTime, TimeScale endTime, boolean isNight, int maxNumberOfIntervals) {
		if (!isNight)
			return getDayCoverage(startTime, endTime, maxNumberOfIntervals);
		else
			return getNightCoverage(startTime, endTime, maxNumberOfIntervals);
	}

	public TimeScale[][] getDayCoverage(TimeScale startTime, TimeScale endTime, int maxNumberOfIntervals) {
		TimeScale[][] intervals = new TimeScale[0][2];
		Vector<TimeScale[]> vIntervals = new Vector<TimeScale[]>(50, 50);
		TimeScale startInt;
		TimeScale endInt;

		setTime(startTime);
		if (isDay()) {
			startInt = getStartOfDaytimeBefore(startTime);
			if (startInt == null)
				return new TimeScale[][] { { null, null } };
			TimeScale startInt1 = getSunsetBefore(startInt);
			if (startInt1 != null) {
				startInt1 = getSunriseBefore(startInt1);
				if (startInt1 != null) {
					TimeScale endInt1 = getEndOfDaytimeAfter(startInt1);
					vIntervals.addElement(new TimeScale[] { startInt1, endInt1 });
				}
			}
		} else {
			startInt = getSunriseBefore(startTime);
			if (startInt == null)
				return intervals;
			endInt = getEndOfDaytimeAfter(startInt);
			vIntervals.addElement(new TimeScale[] { startInt, endInt });
			startInt = getSunriseAfter(startTime, endTime);
			if (startInt == null)
				return new TimeScale[][] { { ((TimeScale[]) vIntervals.elementAt(0))[0], ((TimeScale[]) vIntervals.elementAt(0))[1] } };
		}

		while (true) {
			endInt = getEndOfDaytimeAfter(startInt);
			vIntervals.addElement(new TimeScale[] { startInt, endInt });
			if (vIntervals.size() >= maxNumberOfIntervals)
				break;
			if (!endInt.before(endTime))
				break;
			startInt = getSunriseAfter(endInt, endTime);
			if (startInt == null)
				break;
		}

		intervals = new TimeScale[vIntervals.size()][2];
		vIntervals.copyInto(intervals);
		return intervals;
	}

	public TimeScale[][] getNightCoverage(TimeScale startTime, TimeScale endTime, int maxNumberOfIntervals) {
		TimeScale[][] intervals = new TimeScale[0][2];
		Vector<TimeScale[]> vIntervals = new Vector<TimeScale[]>(50, 50);
		TimeScale startInt;
		TimeScale endInt;

		setTime(startTime);
		if (isNight()) {
			startInt = getStartOfNighttimeBefore(startTime);
			if (startInt == null)
				return new TimeScale[][] { { null, null } };
			TimeScale startInt1 = getSunriseBefore(startInt);
			if (startInt1 != null) {
				startInt1 = getSunsetBefore(startInt1);
				if (startInt1 != null) {
					TimeScale endInt1 = getEndOfNighttimeAfter(startInt1);
					vIntervals.addElement(new TimeScale[] { startInt1, endInt1 });
				}
			}
		} else {
			startInt = getSunsetBefore(startTime);
			if (startInt == null)
				return intervals;
			endInt = getEndOfNighttimeAfter(startInt);
			vIntervals.addElement(new TimeScale[] { startInt, endInt });
			startInt = getSunsetAfter(startTime, endTime);
			if (startInt == null)
				return new TimeScale[][] { { ((TimeScale[]) vIntervals.elementAt(0))[0], ((TimeScale[]) vIntervals.elementAt(0))[1] } };
		}
		while (true) {
			endInt = getEndOfNighttimeAfter(startInt);
			vIntervals.addElement(new TimeScale[] { startInt, endInt });
			if (vIntervals.size() >= maxNumberOfIntervals)
				break;
			if (!endInt.before(endTime))
				break;
			startInt = getSunsetAfter(endInt, endTime);
			if (startInt == null)
				break;
		}

		intervals = new TimeScale[vIntervals.size()][2];
		vIntervals.copyInto(intervals);
		return intervals;
	}

	private TimeScale getStartOfDaytimeBefore(TimeScale startTime) {
		TimeScale sunrise = null;
		TimeScale time = new TimeScale(startTime.getTimeInMinutes());
		TimeScale startOfLastFullDaylight = null;
		int numberOfSkippedDays = 0;
		while (true) {
			setTime(time);
			if (isNormalTrajectory()) {
				if (startOfLastFullDaylight == null) {
					if (!startTime.before(sunriseTimeUT)) {
						sunrise = sunriseTimeUT;
						break;
					}
				} else {
					if (!startTime.before(startOfLastFullDaylight)) {
						sunrise = startOfLastFullDaylight;
						break;
					}
				}
			} else if (isAlwaysBelow()) {
				sunrise = time.endOfDay();
				break;
			} else {
				startOfLastFullDaylight = time.startOfDay();
				if (!startTime.before(startOfLastFullDaylight)) {
					startOfLastFullDaylight = null;
				}
			}
			time.add(Calendar.DAY_OF_MONTH, -1);
			numberOfSkippedDays++;
			if (numberOfSkippedDays > MAX_DAYS_TO_SKIP)
				break;
		}
		return sunrise;
	}

	private TimeScale getStartOfNighttimeBefore(TimeScale startTime) {
		TimeScale sunset = null;
		TimeScale time = new TimeScale(startTime.getTimeInMinutes());
		TimeScale startOfLastFullNight = null;
		int numberOfSkippedDays = 0;
		while (true) {
			setTime(time);
			if (isNormalTrajectory()) {
				if (startOfLastFullNight == null) {
					if (!startTime.before(sunsetTimeUT)) {
						sunset = sunsetTimeUT;
						break;
					}
				} else {
					if (!startTime.before(startOfLastFullNight)) {
						sunset = startOfLastFullNight;
						break;
					}
				}
			} else if (isAlwaysAbove()) {
				sunset = time.endOfDay();
				break;
			} else {
				startOfLastFullNight = time.startOfDay();
				if (!startTime.before(startOfLastFullNight))
					startOfLastFullNight = null;
			}
			time.add(Calendar.DAY_OF_MONTH, -1);
			numberOfSkippedDays++;
			if (numberOfSkippedDays > MAX_DAYS_TO_SKIP)
				break;
		}
		return sunset;
	}

	private TimeScale getSunriseBefore(TimeScale startTime) {
		TimeScale sunrise = null;
		TimeScale time = new TimeScale(startTime.getTimeInMinutes());
		TimeScale startOfLastFullDay = null;
		int numberOfSkippedDays = 0;
		while (true) {
			setTime(time);
			if (isNormalTrajectory()) {
				if (sunriseTimeUT.before(startTime)) {
					if (startOfLastFullDay == null)
						sunrise = sunriseTimeUT;
					else
						sunrise = startOfLastFullDay;
					break;
				}
			} else if (isAlwaysAbove()) {
				startOfLastFullDay = time.startOfDay();
			} else {
				if (startOfLastFullDay != null) {
					sunrise = startOfLastFullDay;
					break;
				}
			}
			time.add(Calendar.DAY_OF_MONTH, -1);
			numberOfSkippedDays++;
			if (numberOfSkippedDays > MAX_DAYS_TO_SKIP)
				break;
		}
		return sunrise;
	}

	private TimeScale getSunsetBefore(TimeScale startTime) {
		TimeScale sunset = null;
		TimeScale time = new TimeScale(startTime.getTimeInMinutes());
		TimeScale startOfLastFullNight = null;
		int numberOfSkippedDays = 0;
		while (true) {
			setTime(time);
			if (isNormalTrajectory()) {
				if (sunsetTimeUT.before(startTime)) {
					if (startOfLastFullNight == null)
						sunset = sunsetTimeUT;
					else
						sunset = startOfLastFullNight;
					break;
				}
			} else if (isAlwaysBelow()) {
				startOfLastFullNight = time.startOfDay();
			} else {
				if (startOfLastFullNight != null) {
					sunset = startOfLastFullNight;
					break;
				}
			}
			time.add(Calendar.DAY_OF_MONTH, -1);
			numberOfSkippedDays++;
			if (numberOfSkippedDays > MAX_DAYS_TO_SKIP)
				break;
		}
		return sunset;
	}

	private TimeScale getSunriseAfter(TimeScale startTime, TimeScale endTime) {
		TimeScale sunrise = null;
		TimeScale maxDateForCheck = new TimeScale(endTime.getTimeInMinutes());
		maxDateForCheck.add(Calendar.DAY_OF_MONTH, 1);
		TimeScale time = new TimeScale(startTime.getTimeInMinutes());
		while (true) {
			setTime(time);
			if (isNormalTrajectory()) {
				if (startTime.before(sunriseTimeUT)) {
					if (sunriseTimeUT.before(endTime))
						sunrise = sunriseTimeUT;
					break;
				}
			} else if (isAlwaysAbove()) {
				sunrise = time.startOfDay();
				if (startTime.before(sunrise))
					break;
				sunrise = null;
				break;
			}
			time.add(Calendar.DAY_OF_MONTH, 1);
			if (!time.before(maxDateForCheck))
				break;
		}
		return sunrise;
	}

	private TimeScale getSunsetAfter(TimeScale startTime, TimeScale endTime) {
		TimeScale sunset = null;
		TimeScale maxDateForCheck = new TimeScale(endTime.getTimeInMinutes());
		maxDateForCheck.add(Calendar.DAY_OF_MONTH, 1);
		TimeScale time = new TimeScale(startTime.getTimeInMinutes());
		while (true) {
			setTime(time);
			if (isNormalTrajectory()) {
				if (startTime.before(sunsetTimeUT)) {
					if (sunsetTimeUT.before(endTime))
						sunset = sunsetTimeUT;
					break;
				}
			} else if (isAlwaysBelow()) {
				sunset = time.startOfDay();
				if (startTime.before(sunset))
					break;
				sunset = null;
			}
			time.add(Calendar.DAY_OF_MONTH, 1);
			if (!time.before(maxDateForCheck))
				break;
		}
		return sunset;
	}

	private TimeScale getEndOfDaytimeAfter(TimeScale startTime) {
		TimeScale sunset = null;
		TimeScale time = new TimeScale(startTime.getTimeInMinutes());
		TimeScale endOfLastFullDaylight = null;
		int numberOfSkippedDays = 0;
		while (true) {
			setTime(time);
			if (isNormalTrajectory()) {
				if (endOfLastFullDaylight == null) {
					if (startTime.before(sunsetTimeUT)) {
						sunset = sunsetTimeUT;
						break;
					}
				} else {
					sunset = endOfLastFullDaylight;
					break;
				}
			} else if (isAlwaysBelow()) {
				sunset = time.startOfDay();
				break;
			} else {
				endOfLastFullDaylight = time.endOfDay();
				if (!startTime.before(endOfLastFullDaylight))
					endOfLastFullDaylight = null;
			}
			time.add(Calendar.DAY_OF_MONTH, 1);
			numberOfSkippedDays++;
			if (numberOfSkippedDays > MAX_DAYS_TO_SKIP)
				break;
		}
		return sunset;
	}

	private TimeScale getEndOfNighttimeAfter(TimeScale startTime) {
		TimeScale sunrise = null;
		TimeScale time = new TimeScale(startTime.getTimeInMinutes());
		TimeScale endOfLastFullNight = null;
		int numberOfSkippedDays = 0;
		while (true) {
			setTime(time);
			if (isNormalTrajectory()) {
				if (endOfLastFullNight == null) {
					if (startTime.before(sunriseTimeUT)) {
						sunrise = sunriseTimeUT;
						break;
					}
				} else {
					sunrise = endOfLastFullNight;
					break;
				}
			} else if (isAlwaysAbove()) {
				sunrise = time.startOfDay();
				break;
			} else {
				endOfLastFullNight = time.endOfDay();
				if (!startTime.before(endOfLastFullNight))
					endOfLastFullNight = null;
			}
			time.add(Calendar.DAY_OF_MONTH, 1);
			numberOfSkippedDays++;
			if (numberOfSkippedDays > MAX_DAYS_TO_SKIP)
				break;
		}
		return sunrise;
	}

}
