/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
//
//     SUBROUTINE IGRF(IY,NM,R,T,F,BR,BT,BF)
//
//     CALCULATES COMPONENTS OF THE MAIN (INTERNAL) GEOMAGNETIC FIELD IN SPHERICAL
//     GEOGRAPHICAL COORDINATE SYSTEM, USING IAGA INTERNATIONAL GEOMAGNETIC REFERENCE
//     MODEL COEFFICIENTS (e.g., http://www.ngdc.noaa.gov/IAGA/wg8/igrf2000.html)
//
//     UPDATING THE COEFFICIENTS TO A GIVEN EPOCH IS MADE AUTOMATICALLY UPON THE FIRST
//     CALL AND AFTER EVERY CHANGE OF THE PARAMETER IY.
//
//-----INPUT PARAMETERS:
//
//     IY  -  YEAR NUMBER (FOUR-DIGIT; 1965 &LE IY &LE 2010)
//     NM  -  HIGHEST ORDER OF SPHERICAL HARMONICS IN THE SCALAR POTENTIAL (NM &LE 10)
//     R,T,F -  SPHERICAL COORDINATES (RADIUS R IN UNITS RE=6371.2 KM, GEOGRAPHIC
//                COLATITUDE  T  AND LONGITUDE  F  IN RADIANS)
//
//-----OUTPUT PARAMETERS:
//
//     BR,BT,BF - SPHERICAL COMPONENTS OF THE MAIN GEOMAGNETIC FIELD IN
//                NANOTESLA: BR IS THE RADIAL COMPONENT (BR > 0 CORRESPONDS
//                TO OUTWARD FIELD DIRECTION), BT IS THE SOUTHWARD COMPONENT,
//                AND BF IS THE EASTWARD COMPONENT.
//     LAST MODIFICATION:  JANUARY 5, 2001.
//     THE CODE WAS MODIFIED TO ACCEPT DATES THROUGH 2010.
//     IT HAS ALSO BEEN SLIGHTLY SIMPLIFIED BY TAKING OUT SOME REDUNDANT STATEMENTS,
//     AND A "SAVE" STATEMENT WAS ADDED, TO AVOID POTENTIAL PROBLEMS WITH SOME
//     FORTRAN COMPILERS.
//
//     WRITTEN BY: N. A. TSYGANENKO
//
//     SAVE MA,IYR,IPR
//
//  Jan 2005    added possibility to read IGRF coefficients from
//  Kozlov A.   data file igrfdata.txt. See format of this file in abstract
//              of method <readIGRFCoeff>.

package edu.uml.lgdc.geospace;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Calendar;
import java.util.Vector;

import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.format.StrUtil;
import edu.uml.lgdc.project.CommonConst;
import edu.uml.lgdc.project.Console;
import edu.uml.lgdc.time.TimeScale;

public class IGRF {
	// IGRF data file name
	static public final String IGRFDATA = "igrfdata.txt";
	static public final String NAME = "IGRF";

	static public final double RE = 6371.2;

	static private final String FILE_SIGNATURE = "IGRFData ";
	// historically, version 2.0 and 3.0 are identical format
	static private final double DATA_FILE_VERSION = 3.0;

	static private final int MAX_HARMONICS = 14;
	static private final int COEFF_DIM = MAX_HARMONICS * (MAX_HARMONICS + 1) / 2;
	static private final int LAST_COEFF_DIM = 45;

	static private final double REC[] = new double[COEFF_DIM];

	static private double version = 0;

	static {
		int n, m, n2, mn;

		for (n = 1; n <= MAX_HARMONICS; n++) {
			n2 = 2 * n - 1;
			n2 = n2 * (n2 - 2);
			for (m = 1; m <= n; m++) {
				mn = n * (n - 1) / 2 + m;
				REC[mn - 1] = ((n - m) * (n + m - 2)) / (double) n2;
			}
		}
	}

	/**
	 * Calculates spherical components of the main geomagnetic field in local cartesian orthogonal system, where X goes southward 
	 * (in another conventions X goes northward), Y goes eastward and Z goes outward the Earth (in another convention Z goes downward the Earth).<br>
	 * Values calculated in nanoteslas,
	 * <ul>
	 *   <li><code>bR</code> is the radial/Z-axis component (<code>bR &gt; 0</code> corresponds to outward field direction)</li> 
	 *   <li><code>bT</code> - the southward/X-axis</li>
	 *   <li><code>bF</code> goes eastward/Y-axis</li>
	 * </ul>
	 * Also for convenience, magnitude of this vector and angular components are derived from bR, bT and bF:
	 * <ul> 
	 *   <li><code>B</code> is the magnitude of geomagnetic field vector, in nanoteslas</li> 
	 *   <li><code>dec</code> is the magnetic declination, in degrees</li> 
	 *   <li><code>dip</code> is the magnetic inclination/dip, in degrees</li> 
	 *   <li><code>fh</code> is giro-frequency, in MHz</li>
	 * </ul>
	 */
	public double bR, bT, bF;
	public double B, dip, dec, fce;

	static private final String DATAFILE_NAME = new File(CommonConst.getShareResourcesDir(), IGRFDATA).getPath();

	static public final int DATAFILE_ERROR_OK = 0;
	static public final int DATAFILE_ERROR_ZEROLENGTH = 1;
	static public final int DATAFILE_ERROR_SIGNATURE = 2;
	static public final int DATAFILE_ERROR_ELEMENT_QTY = 3;
	static public final int DATAFILE_ERROR_TIME_FORMAT = 4;
	static public final int DATAFILE_ERROR_NUMBER_FORMAT = 5;
	static public final int DATAFILE_ERROR_TIME_DECREASE = 6;
	static public final int DATAFILE_ERROR_NO_ENTRIES = 7;
	static public final int DATAFILE_ERROR_LACK_OF_ENTRIES = 8;
	static public final int DATAFILE_ERROR_IO = 9;
	static public final int DATAFILE_ERROR_OBSOLETE = 10;

	static private final int MAX_INACCURATE_MES = 10;

	static private TimeScale[] TIMES = { new TimeScale("1965.01.01"), new TimeScale("1970.01.01"), new TimeScale("1975.01.01"),
			new TimeScale("1980.01.01"), new TimeScale("1985.01.01"), new TimeScale("1990.01.01"), new TimeScale("1995.01.01"),
			new TimeScale("2000.01.01"), new TimeScale("2005.01.01"), new TimeScale("2011.01.01"), new TimeScale("2015.01.01") };

	static private double[][] G = {
			// 1965
			{ 0., -30334., -2119., -1662., 2997., 1594., 1297., -2038., 1292., 856., 957., 804., 479., -390., 252., -219., 358., 254., -31.,
					-157., -62., 45., 61., 8., -228., 4., 1., -111., 75., -57., 4., 13., -26., -6., 13., 1., 13., 5., -4., -14., 0., 8., -1.,
					11., 4., 8., 10., 2., -13., 10., -1., -1., 5., 1., -2., -2., -3., 2., -5., -2., 4., 4., 0., 2., 2., 0., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0. },
			// 1970
			{ 0., -30220., -2068., -1781., 3000., 1611., 1287., -2091., 1278., 838., 952., 800., 461., -395., 234., -216., 359., 262., -42.,
					-160., -56., 43., 64., 15., -212., 2., 3., -112., 72., -57., 1., 14., -22., -2., 13., -2., 14., 6., -2., -13., -3., 5., 0.,
					11., 3., 8., 10., 2., -12., 10., -1., 0., 3., 1., -1., -3., -3., 2., -5., -1., 6., 4., 1., 0., 3., -1., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0. },
			// 1975
			{ 0., -30100., -2013., -1902., 3010., 1632., 1276., -2144., 1260., 830., 946., 791., 438., -405., 216., -218., 356., 264., -59.,
					-159., -49., 45., 66., 28., -198., 1., 6., -111., 71., -56., 1., 16., -14., 0., 12., -5., 14., 6., -1., -12., -8., 4., 0.,
					10., 1., 7., 10., 2., -12., 10., -1., -1., 4., 1., -2., -3., -3., 2., -5., -2., 5., 4., 1., 0., 3., -1., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0. },
			// 1980
			{ 0., -29992., -1956., -1997., 3027., 1663., 1281., -2180., 1251., 833., 938., 782., 398., -419., 199., -218., 357., 261., -74.,
					-162., -48., 48., 66., 42., -192., 4., 14., -108., 72., -59., 2., 21., -12., 1., 11., -2., 18., 6., 0., -11., -7., 4., 3.,
					6., -1., 5., 10., 1., -12., 9., -3., -1., 7., 2., -5., -4., -4., 2., -5., -2., 5., 3., 1., 2., 3., 0., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0. },
			// 1985
			{ 0., -29873., -1905., -2072., 3044., 1687., 1296., -2208., 1247., 829., 936., 780., 361., -424., 170., -214., 355., 253., -93.,
					-164., -46., 53., 65., 51., -185., 4., 16., -102., 74., -62., 3., 24., -6., 4., 10., 0., 21., 6., 0., -11., -9., 4., 4., 4.,
					-4., 5., 10., 1., -12., 9., -3., -1., 7., 1., -5., -4., -4., 3., -5., -2., 5., 3., 1., 2., 3., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0. },
			// 1990
			{ 0., -29775., -1848., -2131., 3059., 1686., 1314., -2239., 1248., 802., 939., 780., 325., -423., 141., -214., 353., 245., -109.,
					-165., -36., 61., 65., 59., -178., 3., 18., -96., 77., -64., 2., 26., -1., 5., 9., 0., 23., 5., -1., -10., -12., 3., 4., 2.,
					-6., 4., 9., 1., -12., 9., -4., -2., 7., 1., -6., -3., -4., 2., -5., -2., 4., 3., 1., 3., 3., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0. },
			// 1995
			{ 0., -29692., -1784., -2200., 3070., 1681., 1335., -2267., 1249., 759., 940., 780., 290., -418., 122., -214., 352., 235., -118.,
					-166., -17., 68., 67., 68., -170., -1., 19., -93., 77., -72., 1., 28., 5., 4., 8., -2., 25., 6., -6., -9., -14., 9., 6.,
					-5., -7., 4., 9., 3., -10., 8., -8., -1., 10., -2., -8., -3., -6., 2., -4., -1., 4., 2., 2., 5., 1., 0., 0., 0, 0., 0., 0.,
					0., 0., 0., 0, 0., 0., 0., 0., 0., 0., 0, 0., 0., 0., 0., 0., 0., 0, 0., 0., 0., 0., 0., 0., 0, 0., 0., 0., 0., 0., 0., 0,
					0., 0. },
			// 2000
			{ 0., -29619.4, -1728.2, -2267.7, 3068.4, 1670.9, 1339.6, -2288., 1252.1, 714.5, 932.3, 786.8, 250., -403., 111.3, -218.8, 351.4,
					222.3, -130.4, -168.6, -12.9, 72.3, 68.2, 74.2, -160.9, -5.9, 16.9, -90.4, 79.0, -74.0, 0., 33.3, 9.1, 6.9, 7.3, -1.2, 24.4,
					6.6, -9.2, -7.9, -16.6, 9.1, 7.0, -7.9, -7., 5., 9.4, 3., -8.4, 6.3, -8.9, -1.5, 9.3, -4.3, -8.2, -2.6, -6., 1.7, -3.1,
					-0.5, 3.7, 1., 2., 4.2, 0.3, -1.1, 2.7, -1.7, -1.9, 1.5, -0.1, 0.1, -0.7, 0.7, 1.7, 0.1, 1.2, 4.0, -2.2, -0.3, 0.2, 0.9,
					-0.2, 0.9, -0.5, 0.3, -0.3, -0.4, -0.1, -0.2, -0.4, -0.2, -0.9, 0.3, 0.1, -0.4, 1.3, -0.4, 0.7, -0.4, 0.3, -0.1, 0.4, 0.,
					0.1 },
			// 2005
			{ 0.00, -29554.63, -1669.05, -2337.24, 3047.69, 1657.76, 1336.30, -2305.83, 1246.39, 672.51, 920.55, 797.96, 210.65, -379.86,
					100.00, -227.00, 354.41, 208.95, -136.54, -168.05, -13.55, 73.60, 69.56, 76.74, -151.34, -14.58, 14.58, -86.36, 79.88,
					-74.46, -1.65, 38.73, 12.30, 9.37, 5.42, 1.94, 24.80, 7.62, -11.73, -6.88, -18.11, 10.17, 9.36, -11.25, -4.87, 5.58, 9.76,
					3.58, -6.94, 5.01, -10.76, -1.25, 8.76, -6.66, -9.22, -2.17, -6.12, 1.42, -2.35, -0.15, 3.06, 0.29, 2.06, 3.77, -0.21,
					-2.09, 2.95, -1.60, -1.88, 1.44, -0.31, 0.29, -0.79, 0.53, 1.80, 0.16, 0.96, 3.99, -2.15, -0.29, 0.21, 0.89, -0.38, 0.96,
					-0.30, 0.46, -0.35, -0.36, 0.08, -0.49, -0.08, -0.16, -0.88, 0.30, 0.28, -0.43, 1.18, -0.37, 0.75, -0.26, 0.35, -0.05, 0.41,
					-0.10, -0.18 },
			// 2010
			{ 0.0, -29496.5, -1585.9, -2396.6, 3026.0, 1668.6, 1339.7, -2326.3, 1231.7, 634.2, 912.6, 809.0, 166.6, -357.1, 89.7, -231.1, 357.2,
					200.3, -141.2, -163.1, -7.7, 72.8, 68.6, 76.0, -141.4, -22.9, 13.1, -77.9, 80.4, -75.0, -4.7, 45.3, 14.0, 10.4, 1.6, 4.9,
					24.3, 8.2, -14.5, -5.7, -19.3, 11.6, 10.9, -14.1, -3.7, 5.4, 9.4, 3.4, -5.3, 3.1, -12.4, -0.8, 8.4, -8.4, -10.1, -2.0, -6.3,
					0.9, -1.1, -0.2, 2.5, -0.3, 2.2, 3.1, -1.0, -2.8, 3.0, -1.5, -2.1, 1.6, -0.5, 0.5, -0.8, 0.4, 1.8, 0.2, 0.8, 3.8, -2.1,
					-0.2, 0.3, 1.0, -0.7, 0.9, -0.1, 0.5, -0.4, -0.4, 0.2, -0.8, 0.0, -0.2, -0.9, 0.3, 0.4, -0.4, 1.1, -0.3, 0.8, -0.2, 0.4,
					0.0, 0.4, -0.3, -0.3 },
			//2015
			{ 0.00, -29442.00, -1501.00, -2445.10, 3012.90, 1676.70, 1350.70, -2352.30, 1225.60, 582.00, 907.60, 813.70, 120.40, -334.90,
   					70.40, -232.60, 360.10, 192.40, -140.90, -157.50,  4.10, 70.00, 67.70, 72.70, -129.90, -28.90, 13.20, -70.90, 81.60, 
   					-76.10, -6.80, 51.80, 15.00,  9.40, -2.80, 6.80, 24.20,  8.80, -16.90, -3.20, -20.60, 13.40, 11.70, -15.90, -2.00,  5.40,  
   					8.80,  3.10, -3.30, 0.70, -13.30, -0.10,  8.70, -9.10, -10.50, -1.90, -6.30,  0.10,  0.50, -0.50,  1.80, -0.70,  2.10, 
   					2.40, -1.80, -3.60,  3.10, -1.50, -2.30,  2.00, -0.80,  0.60, -0.70,  0.20,  1.70, -0.20,  0.40, 3.50, -1.90, -0.20,  
   					0.40,  1.20, -0.80,  0.90, 0.10,  0.50, -0.30, -0.40,  0.20, -0.90,  0.00, 0.00, -0.90,  0.40,  0.50, -0.50,  1.00, -0.20,
   					0.80, -0.10,  0.30,  0.10,  0.50, -0.40, -0.30} 			
			};
	static private double[][] H = {
			// 1965
			{ 0., 0., 5776., 0., -2016., 114., 0., -404., 240., -165., 0., 148., -269., 13., -269., 0., 19., 128., -126., -97., 81., 0., -11.,
					100., 68., -32., -8., -7., 0., -61., -27., -2., 6., 26., -23., -12., 0., 7., -12., 9., -16., 4., 24., -3., -17., 0., -22.,
					15., 7., -4., -5., 10., 10., -4., 1., 0., 2., 1., 2., 6., -4., 0., -2., 3., 0., -6., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0. },
			// 1970
			{ 0., 0., 5737., 0., -2047., 25., 0., -366., 251., -196., 0., 167., -266., 26., -279., 0., 26., 139., -139., -91., 83., 0., -12.,
					100., 72., -37., -6., 1., 0., -70., -27., -4., 8., 23., -23., -11., 0., 7., -15., 6., -17., 6., 21., -6., -16., 0., -21.,
					16., 6., -4., -5., 10., 11., -2., 1., 0., 1., 1., 3., 4., -4., 0., -1., 3., 1., -4., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0. },
			// 1975
			{ 0., 0., 5675., 0., -2067., -68., 0., -333., 262., -223., 0., 191., -265., 39., -288., 0., 31., 148., -152., -83., 88., 0., -13.,
					99., 75., -41., -4., 11., 0., -77., -26., -5., 10., 22., -23., -12., 0., 6., -16., 4., -19., 6., 18., -10., -17., 0., -21.,
					16., 7., -4., -5., 10., 11., -3., 1., 0., 1., 1., 3., 4., -4., -1., -1., 3., 1., -5., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0. },
			// 1980
			{ 0., 0., 5604., 0., -2129., -200., 0., -336., 271., -252., 0., 212., -257., 53., -297., 0., 46., 150., -151., -78., 92., 0., -15.,
					93., 71., -43., -2., 17., 0., -82., -27., -5., 16., 18., -23., -10., 0., 7., -18., 4., -22., 9., 16., -13., -15., 0., -21.,
					16., 9., -5., -6., 9., 10., -6., 2., 0., 1., 0., 3., 6., -4., 0., -1., 4., 0., -6., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0. },
			// 1985
			{ 0., 0., 5500., 0., -2197., -306., 0., -310., 284., -297., 0., 232., -249., 69., -297., 0., 47., 150., -154., -75., 95., 0., -16.,
					88., 69., -48., -1., 21., 0., -83., -27., -2., 20., 17., -23., -7., 0., 8., -19., 5., -23., 11., 14., -15., -11., 0., -21.,
					15., 9., -6., -6., 9., 9., -7., 2., 0., 1., 0., 3., 6., -4., 0., -1., 4., 0., -6., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0. },
			// 1990
			{ 0., 0., 5406., 0., -2279., -373., 0., -284., 293., -352., 0., 247., -240., 84., -299., 0., 46., 154., -153., -69., 97., 0., -16.,
					82., 69., -52., 1., 24., 0., -80., -26., 0., 21., 17., -23., -4., 0., 10., -19., 6., -22., 12., 12., -16., -10., 0., -20.,
					15., 11., -7., -7., 9., 8., -7., 2., 0., 2., 1., 3., 6., -4., 0., -2., 3., -1., -6., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0. },
			// 1995
			{ 0., 0., 5306., 0., -2366., -413., 0., -262., 302., -427., 0., 262., -236., 97., -306., 0., 46., 165., -143., -55., 107., 0., -17.,
					72., 67., -58., 1., 36., 0., -69., -25., 4., 24., 17., -24., -6., 0., 11., -21., 8., -23., 15., 11., -16., -4., 0., -20.,
					15., 12., -6., -8., 8., 5., -8., 3., 0., 1., 0., 4., 5., -5., -1., -2., 1., -2., -7., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
					0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0. },
			// 2000
			{ 0., 0., 5186.1, 0., -2481.6, -458.0, 0., -227.6, 293.4, -491.1, 0., 272.6, -231.9, 119.8, -303.8, 0., 43.8, 171.9, -133.1, -39.3,
					106.3, 0., -17.4, 63.7, 65.1, -61.2, 0.7, 43.8, 0., -64.6, -24.2, 6.2, 24., 14.8, -25.4, -5.8, 0.0, 11.9, -21.5, 8.5, -21.5,
					15.5, 8.9, -14.9, -2.1, 0.0, -19.7, 13.4, 12.5, -6.2, -8.4, 8.4, 3.8, -8.2, 4.8, 0.0, 1.7, 0.0, 4.0, 4.9, -5.9, -1.2, -2.9,
					0.2, -2.2, -7.4, 0.0, 0.1, 1.3, -0.9, -2.6, 0.9, -0.7, -2.8, -0.9, -1.2, -1.9, -0.9, 0.0, -0.4, 0.3, 2.5, -2.6, 0.7, 0.3,
					0.0, 0.0, 0.3, -0.9, -0.4, 0.8, 0.0, -0.9, 0.2, 1.8, -0.4, -1.0, -0.1, 0.7, 0.3, 0.6, 0.3, -0.2, -0.5, -0.9 },
			// 2005
			{ 0.00, 0.00, 5077.99, 0.00, -2594.50, -515.43, 0.00, -198.86, 269.72, -524.72, 0.00, 282.07, -225.23, 145.15, -305.36, 0.00, 42.72,
					180.25, -123.45, -19.57, 103.85, 0.00, -20.33, 54.75, 63.63, -63.53, 0.24, 50.94, 0.00, -61.14, -22.57, 6.82, 25.35, 10.93,
					-26.32, -4.64, 0.00, 11.20, -20.88, 9.83, -19.71, 16.22, 7.61, -12.76, -0.06, 0.00, -20.11, 12.69, 12.67, -6.72, -8.16,
					8.10, 2.92, -7.73, 6.01, 0.00, 2.19, 0.10, 4.46, 4.76, -6.58, -1.01, -3.47, -0.86, -2.31, -7.93, 0.00, 0.26, 1.44, -0.77,
					-2.27, 0.90, -0.58, -2.69, -1.08, -1.58, -1.90, -1.39, 0.00, -0.55, 0.23, 2.38, -2.63, 0.61, 0.40, 0.01, 0.02, 0.28, -0.87,
					-0.34, 0.88, 0.00, -0.76, 0.33, 1.72, -0.54, -1.07, -0.04, 0.63, 0.21, 0.53, 0.38, -0.22, -0.57, -0.82 },
			// 2010
			{ 0.0, 0.0, 4945.1, 0.0, -2707.7, -575.4, 0.0, -160.5, 251.7, -536.8, 0.0, 286.4, -211.2, 164.4, -309.2, 0.0, 44.7, 188.9, -118.1,
					0.1, 100.9, 0.0, -20.8, 44.2, 61.5, -66.3, 3.1, 54.9, 0.0, -57.8, -21.2, 6.6, 24.9, 7.0, -27.7, -3.4, 0.0, 10.9, -20.0,
					11.9, -17.4, 16.7, 7.1, -10.8, 1.7, 0.0, -20.5, 11.6, 12.8, -7.2, -7.4, 8.0, 2.2, -6.1, 7.0, 0.0, 2.8, -0.1, 4.7, 4.4, -7.2,
					-1.0, -4.0, -2.0, -2.0, -8.3, 0.0, 0.1, 1.7, -0.6, -1.8, 0.9, -0.4, -2.5, -1.3, -2.1, -1.9, -1.8, 0.0, -0.8, 0.3, 2.2, -2.5,
					0.5, 0.6, 0.0, 0.1, 0.3, -0.9, -0.2, 0.8, 0.0, -0.8, 0.3, 1.7, -0.6, -1.2, -0.1, 0.5, 0.1, 0.5, 0.4, -0.2, -0.5, -0.8 },
			// 2015
			{ 0.0, 0.0, 4797.1, 0.0, -2845.6, -641.9, 0.0, -115.3, 244.9, -538.4, 0.0, 283.3, -188.7, 180.9, -329.5, 0.0, 47.3, 197.0, -119.3, 
					16.0, 100.2, 0.0, -20.8, 33.2, 58.9, -66.7, 7.3, 62.6, 0.0, -54.1, -19.5, 5.7, 24.4, 3.4, -27.4, -2.2, 0.0, 10.1, -18.3, 
					13.3, -14.6, 16.2, 5.7, -9.1, 2.1, 0.0, -21.6, 10.8, 11.8, -6.8, -6.9, 7.8, 1.0, -4.0, 8.4, 0.0, 3.2, -0.4, 4.6, 4.4, -7.9, 
					-0.6, -4.2, -2.8, -1.2, -8.7, 0.0, -0.1, 2.0, -0.7, -1.1, 0.8, -0.2, -2.2, -1.4, -2.5, -2.0,-2.4, 0.0, -1.1, 0.4, 1.9, -2.2, 
					0.3, 0.7, -0.1, 0.3, 0.2, -0.9, -0.1, 0.7, 0.0, -0.9, 0.4, 1.6, -0.5, -1.2, -0.1, 0.4, -0.1, 0.4, 0.5, -0.3, -0.4, -0.8 }					
			};
	static private double[] DG = { 0.0, 11.4, 16.7, -11.3, -3.9, 2.7, 1.3, -3.9, -2.9, -8.1, -1.4, 2.0, -8.9, 4.4, -2.3, -0.5, 0.5, -1.5, -0.7,
			1.3, 1.4, -0.3, -0.3, -0.3, 1.9, -1.6, -0.2, 1.8, 0.2, -0.1, -0.6, 1.4, 0.3, 0.1, -0.8, 0.4, -0.1, 0.1, -0.5, 0.3, -0.3, 0.3, 0.2,
			-0.5, 0.2 };
	static private double[] DH = { 0.0, 0.0, -28.8, 0.0, -23.0, -12.9, 0.0, 8.6, -2.9, -2.1, 0.0, 0.4, 3.2, 3.6, -0.8, 0.0, 0.5, 1.5, 0.9, 3.7,
			-0.6, 0.0, -0.1, -2.1, -0.4, -0.5, 0.8, 0.5, 0.0, 0.6, 0.3, -0.2, -0.1, -0.8, -0.3, 0.2, 0.0, 0.0, 0.2, 0.5, 0.4, 0.1, -0.1, 0.4,
			0.4 };

	static private boolean datafileInUse = false;
	static private boolean softwareObsolete = false;
	static private int datafileErrorCode = DATAFILE_ERROR_OK;

	// Time range where this model is working
	static public TimeScale MIN_DATE = TIMES[0];
	static public TimeScale MAX_DATE = new TimeScale(TIMES[TIMES.length - 1].getTimeInMinutes());
	static {
		MAX_DATE.add(Calendar.MILLISECOND, -1);
	}
	static public double MIN_DATE_IN_MINUTES = MIN_DATE.getTimeInMinutes();
	static public double MAX_DATE_IN_MINUTES = MAX_DATE.getTimeInMinutes();

	private double g[] = new double[COEFF_DIM];
	private double h[] = new double[COEFF_DIM];

	// cache for time coefficients

	// for compatible mode
	private int prevYear = 0;
	private int prevDayOfYear = 0;
	// for advanced mode
	private TimeScale prevUT = null;

	// cache for position and calculation precision
	private int prevHarmonicsQty = -1;
	private double prevRadius = 0;
	private double prevTheta = -1000;
	private double prevPhi = -1000;

	static private boolean extrapolationInaccurateMes = false;
	static private int countInaccurateMes;

	static private boolean compatibleMode;
	static {
		readIGRFCoeff();
		compatibleMode = true;
		for (int i = 0; i < TIMES.length - 1; i++) {
			if (!TIMES[i].equals(TIMES[i].startOfYear())) {
				compatibleMode = false;
				break;
			}
		}
	}
	
	static private IGRF myIGRF = null;

	public IGRF() {
	}

	/**
	 * Remains here for compatibility with previous version
	 */
	public void calc(int year, int harmonicsQty, double radius, double theta, double phi) {
		calc(year, 1, harmonicsQty, radius, theta, phi);
	}

	/**
	 * Calculates components of the Main (internal) Geomagnetic Field in Spherical Geographical coordinate system, using IAGA International
	 * Geomagnetic Reference model coefficients (e.g., http://www.ngdc.noaa.gov/iaga/wg8/igrf2000.html)<br><br>
	 *
	 * Updating the coefficients to a given epoch is made automatically upon the first call and after every change of the parameter <code>year</code>.
	 *
	 * Parameters:
	 * <ul>
	 *   <li>year is year number (four-digit; work accurately from MIN_YEAR (=1965) to MAX_YEAR (now = 2010)</li>
	 *   <li>dayOfYear - Day in this year (counting from 1)</li>
	 *   <li>harmonicsQty is the highest order of spherical harmonics in the scalar potential (no more than 10)</li>
	 *   <li>radius is the radius component of point in Spherical Geographical coordinate system. Given in radii of the Earth (RE) ( 1 RE = 6371.2 km )</li>
	 *   <li>theta is polar angle component of point in* Spherical Geographical coordinate system. Given in radians. Often also referred to as colatitude.</li>
	 *   <li>phi is azimuthal angle component of point in Spherical Geographical coordinate system. Given in radians. Often also  referred to as longitude.</li>
	 * </ul>
	 * Calculates spherical components of the main geomagnetic field in local local cartesian orthogonal system, where X goes southward 
	 * (in another conventions X goes northward), Y goes eastward and Z goes outward the Earth (in another convention Z goes downward the Earth).<br>
	 * Values calculated in nanoteslas.<br>
	 * Statistics: written by: N. A. Tsyganenko, converted to Java code by
	 * Alexander Kozlov, March 2002
	 */
	public void calc(int year, int dayOfYear, int harmonicsQty, double radius, double theta, double phi) {
		double a[] = new double[MAX_HARMONICS];
		double b[] = new double[MAX_HARMONICS];

		double prorateCoeff = 0;
		double dt;
		TimeScale ut = TimeScale.createFromFields(year, TimeScale.FIELD_NOT_SET, TimeScale.FIELD_NOT_SET, dayOfYear, 0, 0, 0);

		if (ut.before(MIN_DATE))
			throw new RuntimeException("IGRF: date is below " + MIN_DATE + ", - no coefficients exist");
		if (MAX_DATE.before(ut) && countInaccurateMes < MAX_INACCURATE_MES) {
			String msg = "IGRF: date is above " + MAX_DATE + ", extrapolated coefficients may be inaccurate";
			countInaccurateMes++;
			if (!extrapolationInaccurateMes && !softwareObsolete) {
				if (!datafileInUse) {
					if (datafileErrorCode == DATAFILE_ERROR_OK)
						msg += System.lineSeparator() + "  you do not use IGRF data file";
					else
						msg += System.lineSeparator() + "  your IGRF data file is obsolete or corrupted";
				}
				msg += System.lineSeparator() + "  try to get up-to-date IGRF data file";
				extrapolationInaccurateMes = true;
			}
			//Console.showWarn(msg);
		}

		if (prevYear == year && prevDayOfYear == dayOfYear && prevHarmonicsQty == harmonicsQty && prevRadius == radius && prevTheta == theta
				&& prevPhi == phi)
			return;

		if (compatibleMode && (year != prevYear || dayOfYear != prevDayOfYear) || !compatibleMode && !ut.equals(prevUT)) {
			if (compatibleMode) {
				if (year >= TIMES[TIMES.length - 2].get(Calendar.YEAR)) {
					dt = ((year - TIMES[TIMES.length - 2].get(Calendar.YEAR)) * 365 + dayOfYear) / 365.0;
					for (int i = 0; i < COEFF_DIM; i++) {
						g[i] = G[TIMES.length - 2][i];
						h[i] = H[TIMES.length - 2][i];
						if (i < LAST_COEFF_DIM) {
							g[i] += DG[i] * dt;
							h[i] += DH[i] * dt;
						}
					}
				} else {
					for (int i = 0; i < TIMES.length - 2; i++) {
						if (year < TIMES[i + 1].get(Calendar.YEAR)) {
							prorateCoeff = ((year - TIMES[i].get(Calendar.YEAR)) * 365 + dayOfYear) / (365.0 * 5);
							// Prorate between i and (i+1)
							prorate(prorateCoeff, G[i], G[i + 1], H[i], H[i + 1]);
							break;
						}
					}
				}
			} else {
				if (!ut.before(TIMES[TIMES.length - 2])) {
					dt = ut.diffIn(Calendar.YEAR, TIMES[TIMES.length - 2]);
					for (int i = 0; i < COEFF_DIM; i++) {
						g[i] = G[TIMES.length - 2][i];
						h[i] = H[TIMES.length - 2][i];
						if (i < LAST_COEFF_DIM) {
							g[i] += DG[i] * dt;
							h[i] += DH[i] * dt;
						}
					}
				} else {
					for (int i = 0; i < TIMES.length - 2; i++) {
						if (ut.before(TIMES[i + 1])) {
							prorateCoeff = ((year - TIMES[i].get(Calendar.YEAR)) * 365 + dayOfYear) / (365.0 * 5);
							prorateCoeff = ut.diffIn(Calendar.MINUTE, TIMES[i]) / TIMES[i + 1].diffIn(Calendar.MINUTE, TIMES[i]);
							// Prorate between i and (i+1)
							prorate(prorateCoeff, G[i], G[i + 1], H[i], H[i + 1]);
							break;
						}
					}
				}
			}
			//
			// Coefficients for a given year have been calculated;
			// Now multiply them by Schmidt Normalization Factors:
			//
			double s = 1.;
			int mn, mnn;
			double p, aa;
			for (int i = 2; i <= MAX_HARMONICS; i++) {
				mn = i * (i - 1) / 2 + 1;
				s *= (double) (2 * i - 3) / (double) (i - 1);
				g[mn - 1] *= s;
				h[mn - 1] *= s;
				p = s;
				for (int j = 2; j <= i; j++) {
					aa = 1;
					if (j == 2)
						aa = 2;
					p *= Math.sqrt(aa * (double) (i - j + 1) / (double) (i + j - 2));
					mnn = mn + j - 1;
					g[mnn - 1] *= p;
					h[mnn - 1] *= p;
				}
			}
		}

		//
		// Now calculate the field components ( in case of multiple invocations
		// with the same values of iy and nm, calculations start right here ):
		//
		double p = 1. / radius;
		int k = harmonicsQty + 1;

		for (int i = 1; i <= k; i++) {
			p /= radius;
			a[i - 1] = p;
			b[i - 1] = p * i;
		}
		p = 1.;
		double d = 0.;
		double bbr = 0.;
		double bbt = 0.;
		double bbf = 0.;
		double cosPhi = Math.cos(phi);
		double sinPhi = Math.sin(phi);
		double cosTheta = Math.cos(theta);
		double sinTheta = Math.sin(theta);
		double x = 0;
		double y = 1;
		int mm = 0, mn;
		double w;
		double q, z, bi, p2, d2, an, e, hh, qq, xk, dp, pm;
		for (int m = 1; m <= k; m++) {
			if (m > 1) {
				mm = m - 1;
				w = x;
				x = w * cosPhi + y * sinPhi;
				y = y * cosPhi - w * sinPhi;
			}
			q = p;
			z = d;
			bi = 0.;
			p2 = 0.;
			d2 = 0.;
			for (int n = m; n <= k; n++) {
				an = a[n - 1];
				mn = n * (n - 1) / 2 + m;
				e = g[mn - 1];
				hh = h[mn - 1];
				w = e * y + hh * x;
				bbr += b[n - 1] * w * q;
				bbt -= an * w * z;
				if (m > 1) {
					qq = q;
					if (sinTheta < 1.E-5)
						qq = z;
					bi += an * (e * x - hh * y) * qq;
				}
				xk = REC[mn - 1];
				dp = cosTheta * z - sinTheta * q - xk * d2;
				pm = cosTheta * q - xk * p2;
				d2 = z;
				p2 = q;
				z = dp;
				q = pm;
			}
			d = sinTheta * d + cosTheta * p;
			p = sinTheta * p;
			if (m > 1) {
				bi *= mm;
				bbf += bi;
			}
		}

		bR = bbr;
		bT = bbt;
		if (sinTheta >= 1.E-5) {
			bF = bbf / sinTheta;
		} else {
			if (cosTheta < 0.)
				bbf = -bbf;
			bF = bbf;
		}

		// Calculate magnitude
		B = Math.sqrt(bR * bR + bT * bT + bF * bF);

		// fce [MHz] = 2.8011 * 10^-5 B [nT]
		// fce [MHz] = 2.8011 B [G]
		// 0.000028011 - constant to calculate gyrofrequency in MHz
		// from magnetic magnitude in nanoTesla (nT)
		// 100,000 nT = 1 Gauss
		fce = 0.000028011 * B;

		dip = -(90 - Math.toDegrees(Math.acos(bR / B)));
		dec = Math.toDegrees(Math.atan2(bF, -bT));

		prevYear = year;
		prevDayOfYear = dayOfYear;
		prevHarmonicsQty = harmonicsQty;
		prevRadius = radius;
		prevTheta = theta;
		prevPhi = phi;
		return;
	}

	/**
	 * This is another call which use geomap latitude, longitude and altidute.
	 */
	public void calc(TimeScale time, double[] geomap) {
		calc(time, geomap[Geopack.LATITUDE], geomap[Geopack.LONGITUDE], geomap[Geopack.RADIUS]);
	}

	/**
	 * This is another call which use latitude, longitude and altitude. Latitude
	 * is given in degrees ( from 90 on north pole to -90 on south pole ).
	 * Longitude is given in degrees (from 0 on greenwich to 360 going
	 * eastward). Altitude is given as elevation in km.
	 */
	public void calc(TimeScale time, double latitude_DEG, double longitude_DEG, double altitude_KM) {
		double theta = Math.toRadians(90 - latitude_DEG);
		double phi = Math.toRadians(longitude_DEG);
		calc(time.get(Calendar.YEAR), time.get(Calendar.DAY_OF_YEAR), 10, 1 + altitude_KM / RE, theta, phi);
	}

	private void prorate(double coeff, double g1[], double g2[], double h1[], double h2[]) {
		double f1 = 1. - coeff;
		for (int i = 0; i < COEFF_DIM; i++) {
			g[i] = g1[i] * f1 + g2[i] * coeff;
			h[i] = h1[i] * f1 + h2[i] * coeff;
		}
	}

	/**
	 * Read IGRF G and H-coefficients from file if any File structure 1) This
	 * the text file. 2) If line starts with semicolon(;) or sharp(#) symbols or
	 * line consist of only spaces then it considered as comment line otherwise
	 * it considered as data line (exclusion 1st line) 3) First line should
	 * start with "IGRFData " 4) All subsequent lines are either comment or data
	 * lines 5) All data consists of several data arrays 6) Data array occupies
	 * one line or several data lines; 7) Elements of array are separated by
	 * comma 8) If last element on the data line is followed by comma then
	 * continuation of this array will be the next data line 9) First element of
	 * each array Date/Time and the next 210 ( 105 G-coeff and then 105 H-coeff)
	 * are the Legendre coefficients for given Date/time (Last array has
	 * slightly different meaning) 10) Date/Time can be presented using one of
	 * the following forms: yyyy yyyy/MM yyyy/MM/dd yyyy/MM/dd (DDD) yyyy/MM/dd
	 * (DDD) HH yyyy/MM/dd HH yyyy/MM/dd (DDD) HH:mm yyyy/MM/dd HH:mm yyyy/MM/dd
	 * (DDD) HH:mm:ss yyyy/MM/dd HH:mm:ss
	 *
	 * instead slash (/) symbol you can use also dot (.) or dash(-) symbols.
	 *
	 * and <G10>, <G11> and <H11> are geodipole moment coefficient that
	 * correspond given TimeScale (for last data line it have a little different
	 * meaning, see later) 11) All data arrays should be given in ascending time
	 * order. 12) Last data array has slightly different meaning: Date/Time is
	 * the time from which it is not recommended to use IGRF. First 45 G
	 * delta-values and first 45 H delta-values are linear extrapolation delta
	 * per year for G- and H-coeff beginning from previous array. Discarding the
	 * changes of the rest 105 - 45 = 60 G and H coefficients.
	 *
	 */
	static private void readIGRFCoeff() {
		datafileInUse = false;
		datafileErrorCode = DATAFILE_ERROR_OK;
		softwareObsolete = false;
		File datafile = new File(DATAFILE_NAME);
		if (!datafile.isFile())
			return;
		version = 0;
		final int minNoOfElements = 2 * COEFF_DIM;
		final int minNoOfElementsLast = 2 * LAST_COEFF_DIM;
		Vector<TimeScale> vTimes = new Vector<TimeScale>(50, 50);
		Vector<double[]> vG = new Vector<double[]>(50, 50);
		Vector<double[]> vH = new Vector<double[]>(50, 50);
		double[] g = null;
		double[] h = null;
		int qty = 0;
		BufferedReader br = null;
		String line;
		String arrayString = "";
		String[] stringData;
		int lineCount = 0;
		int qtyOfCoeff = 0;
		try {
			br = new BufferedReader(new FileReader(datafile));
			while (br.ready()) {
				line = br.readLine();
				if (line == null)
					break;
				lineCount++;
				line = StrUtil.trimRight(line, ' ');
				if (lineCount == 1) {
					if (!line.startsWith(FILE_SIGNATURE)) {
						Console.showError("IGRF file " + DATAFILE_NAME + ": bad signature");
						datafileErrorCode = DATAFILE_ERROR_SIGNATURE;
						break;
					} else {
						version = FC.StringToDouble(line.substring(FILE_SIGNATURE.length() + 1).trim());
						if (version < 2) {
							Console.showError("IGRF file " + DATAFILE_NAME + " is obsolete");
							datafileErrorCode = DATAFILE_ERROR_OBSOLETE;
							break;
						} else if (version > DATA_FILE_VERSION) {
							Console.showError("IGRF file " + DATAFILE_NAME + " has a newer version" + System.lineSeparator() + "you need to upgrade software");
							softwareObsolete = true;
							break;
						}
					}
					continue;
				}
				if (line.length() == 0 || line.startsWith("#") || line.startsWith(";"))
					continue;
				arrayString += line;
				if (line.endsWith(","))
					continue;
				stringData = StrUtil.listToArray(arrayString, ",");
				if (stringData.length < minNoOfElementsLast + 1) {
					Console.showError("IGRF file " + DATAFILE_NAME + " line " + stringData[0] + System.lineSeparator() + "  should be more than "
							+ minNoOfElementsLast + " elements separated by comma");
					datafileErrorCode = DATAFILE_ERROR_ELEMENT_QTY;
					break;
				}
				if (qty > 0 && qtyOfCoeff < COEFF_DIM) {
					Console.showError("IGRF file " + DATAFILE_NAME + " line " + vTimes.elementAt(qty - 1).toShortDateUT() + ":" + System.lineSeparator()
							+ "  not last array, " + qtyOfCoeff + ", should have not less than " + minNoOfElements + " elements");
					datafileErrorCode = DATAFILE_ERROR_ELEMENT_QTY;
					break;
				}
				vTimes.addElement(TimeScale.parse(stringData[0]));
				qtyOfCoeff = (stringData.length - 1) / 2;
				int dim = COEFF_DIM;
				if (qtyOfCoeff < COEFF_DIM)
					dim = LAST_COEFF_DIM;
				g = new double[dim];
				h = new double[dim];
				for (int i = 0; i < dim; i++) {
					if (!StrUtil.isFloatingPointNumber(stringData[1 + i]) || !StrUtil.isFloatingPointNumber(stringData[1 + qtyOfCoeff + i])) {
						Console.showError("IGRF file " + DATAFILE_NAME + " line " + lineCount + ", element index " + i
								+ ": G and H should be a number");
						datafileErrorCode = DATAFILE_ERROR_NUMBER_FORMAT;
						break;
					}
					g[i] = FC.StringToDouble(stringData[1 + i]);
					h[i] = FC.StringToDouble(stringData[1 + qtyOfCoeff + i]);
				}
				if (datafileErrorCode != DATAFILE_ERROR_OK)
					break;
				vG.addElement(g);
				vH.addElement(h);
				arrayString = "";
				if (qty > 0) {
					if (!vTimes.elementAt(qty - 1).before(vTimes.elementAt(qty))) {
						Console.showError("IGRF file " + DATAFILE_NAME + " line " + lineCount + ":  date/time should increase");
						datafileErrorCode = DATAFILE_ERROR_TIME_DECREASE;
						break;
					}
				}
				qty++;
			}
		} catch (IOException e) {
			Console.showError("File " + DATAFILE_NAME + " line " + (++lineCount) + ":" + System.lineSeparator() + "  " + e.toString());
			datafileErrorCode = DATAFILE_ERROR_IO;
		} catch (Exception e) // TODO catch( IllegalTimeScaleStringException e )
		{
			Console.showError("File " + DATAFILE_NAME + " line " + lineCount + ":" + System.lineSeparator() + "  " + e.toString());
			datafileErrorCode = DATAFILE_ERROR_TIME_FORMAT;
		} finally {
			if (br != null)
				try {
					br.close();
				} catch (IOException ex) {
					Console.printThreadStackTrace( ex );
				}
		}

		if (softwareObsolete)
			return;

		if (datafileErrorCode == DATAFILE_ERROR_OK) {
			if (lineCount == 0) {
				datafileErrorCode = DATAFILE_ERROR_ZEROLENGTH;
				Console.showError("IGRF file " + DATAFILE_NAME + " of zero length");
			} else if (arrayString.length() > 0) {
				Console.showError("IGRF file " + DATAFILE_NAME + ": last array is incomplete");
				datafileErrorCode = DATAFILE_ERROR_ELEMENT_QTY;
			} else if (qty == 0) {
				Console.showError("IGRF file " + DATAFILE_NAME + " doesn't contain entries");
				datafileErrorCode = DATAFILE_ERROR_NO_ENTRIES;
			} else if (qty == 1) {
				Console.showError("IGRF file " + DATAFILE_NAME + " contains only one entry");
				datafileErrorCode = DATAFILE_ERROR_LACK_OF_ENTRIES;
			} else if (qtyOfCoeff < LAST_COEFF_DIM) {
				Console.showError("IGRF file " + DATAFILE_NAME + ": last array should " + "contain not less than " + (2 * LAST_COEFF_DIM)
						+ " elements");
				datafileErrorCode = DATAFILE_ERROR_ELEMENT_QTY;
			}
		}

		if (datafileErrorCode != DATAFILE_ERROR_OK)
			return;

		TIMES = new TimeScale[qty];
		G = new double[qty - 1][COEFF_DIM];
		H = new double[qty - 1][COEFF_DIM];
		DG = new double[LAST_COEFF_DIM];
		DH = new double[LAST_COEFF_DIM];
		for (int i = 0; i < qty - 1; i++) {
			TIMES[i] = vTimes.elementAt(i);
			G[i] = vG.elementAt(i);
			H[i] = vH.elementAt(i);
		}
		TIMES[qty - 1] = vTimes.elementAt(qty - 1);
		DG = vG.elementAt(qty - 1);
		DH = vH.elementAt(qty - 1);
		datafileInUse = true;
		MIN_DATE = TIMES[0];
		MAX_DATE = new TimeScale(TIMES[TIMES.length - 1].getTimeInMinutes());
		MAX_DATE.add(Calendar.MILLISECOND, -1);
		MIN_DATE_IN_MINUTES = MIN_DATE.getTimeInMinutes();
		MAX_DATE_IN_MINUTES = MAX_DATE.getTimeInMinutes();
	}

	public String getModelName() {
		return NAME + " " + FC.doubleToString(version, 1);
	}

	/**
	 * Returns instance of {@link IGRF} for this date and location; this is specific to ionosondelib convenion.
	 */
	synchronized static public IGRF get(TimeScale date, double latitude_DEG, double longitude_DEG, double altitude_KM ) {
		
		double theta = Math.toRadians(90 - latitude_DEG);
		double phi = Math.toRadians(longitude_DEG);
		double radius = 1 + altitude_KM/RE;
		int year = date.get( Calendar.YEAR );
		int dayOfYear = date.get( Calendar.DAY_OF_YEAR );

		if (myIGRF != null 
				&& myIGRF.prevYear == year
				&& myIGRF.prevDayOfYear == dayOfYear 
				&& Math.abs(myIGRF.prevPhi - phi) < 0.01
				&& Math.abs(myIGRF.prevTheta - theta) < 0.01
				&& Math.abs(myIGRF.prevRadius - radius) < 0.01 )
			return myIGRF;
		
		myIGRF = new IGRF();
		myIGRF.calc(year, dayOfYear, 10, radius, theta, phi);
		return myIGRF;
	}
	
}