package edu.uml.lgdc.geospace;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Calendar;

import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.project.CommonConst;
import edu.uml.lgdc.time.RealHours;
import edu.uml.lgdc.time.TimeScale;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
public class IRIModel {

	// URSI-model file prefix
	static public final String URSI_MODEL_PREF = "URSI";
	// URSI-model file extension, without dot
	static public final String URSI_MODEL_EXT = "ASC";

	public static final double NO_FOF1 = 999.9;
	public static final double ALTITUDE_KM = 300.0;

	/**
	 * The following properties are calculated for given time and location on
	 * the Earth
	 */

	// Sunspot number for this date (really with precision of month)
	private SunSpots sunspots;
	// private float sunspots;
	// private int sunspotsStatus = SSN_MISSING;

	// Ground-solar time/location properties
	private GroundSolar groundSolar;

	// critical frequency of E-layer at given time
	private double foEP;
	// critical frequency of F1-layer at given time, and its probability
	private double foF1P;
	private double foF1probability;
	// critical frequency of F2-layer at given time
	private double foF2P;

	// D-region peak density
	private double NmDP;
	
	// Location in geographical latitude/longitude
	private double latitude; // Geographical north latitude, -90 to 90
	private double longitude; // Geographical east longitude, 0 to 360

	// Location in dipole magnetic coordinate which is derived from
	// latitude/longitude and time (time is taken in account for precision
	// purpose as magnetic pole constantly change its location near around
	// geographical pole)
	private double magLatitude; // Geographical north magnetic latitude
	private double magLongitude; // Geographical east magnetic longitude

	// Time for which it was calculated
	private TimeScale ut;
	private int month;
	private int dayOfYear;
	private RealHours realHours;

	private Geopack geopack = new Geopack();
	public IGRF igrf = new IGRF();

	// some things for optimization
	private int monthPrev = -1;
	private double rgPrev = -1;
	private double f2[][][] = new double[2][76][13];

	// output from fieldFromIGRF:
	// bet - magnitude (in Gauss) of magnetic field at altitude 300 km
	// xma - x component (in Gauss) of magnetic vector
	// Axis X goes to the north
	// yma - y component (in Gauss) of magnetic vector
	// Axis Y goes to the east
	// zma - z component (in Gauss) of magnetic vector
	// Axis Z goes downwards to the center of Earth
	// dip - dip/inclination angle (in degrees) of magnetic vector
	// dec - declination angle (in degrees) of magnetic vector
	// modip - modified dip angle (in degrees)
	public double xma, yma, zma, bet, dip, dec, modip, fce;

	/**
	 * Constructor
	 */
	public IRIModel() {
	}

	/**
	 * Visual representation of values.
	 */
	public String toString() {
		return groundSolar.toString() + "\r\n" + "sun spot number = " + FC.DoubleToString(sunspots.getNumber(), 5, 1) + "\r\n" + "latitude = "
				+ FC.DoubleToString(latitude, 7, 2) + "\r\n" + "longitude = " + FC.DoubleToString(longitude, 7, 2) + "\r\n" + "ut = " + ut.toHumanUT() + "\r\n"
				+ "critical frequency of E-layer = " + foEP + "\r\n" + "critical frequency of F1-layer = " + foF1P + "\r\n"
				+ "critical frequency of F2-layer = " + foF2P + "\r\n" + "Dec = " + dec + "\r\n" + "Dip = " + dip + "\r\n" + "fce = " + fce;
	}

	/**
	 * Parameter <code>latitude</code> is given as geographical north latitude.<br>
	 * Parameter <code>longitude</code> is given as geographical east longitude.<br>
	 * Dipole magnetic coordinates (and other time-dependent stuff) are
	 * calculated using current UT time.
	 */
	public void calculate(double latitude, double longitude) {
		calculate(latitude, longitude, new TimeScale());
	}

	/**
	 * Parameter <code>latitude</code> is given as geographical north latitude.<br>
	 * Parameter <code>longitude</code> is given as geographical east longitude.<br>
	 * Time, <code>ut</code>, is given as UT.
	 */
	public void calculate(double latitude, double longitude, TimeScale ut) {
		setTime(ut);
		setLocation(latitude, longitude);
		setSolarActivity(); // call this once time is set
		setMagneticField();
		loadCoefficients(); // call that once time and solar indices are set
		calcPrediction();
	}

	public void setTime(TimeScale ut) {
		this.ut = ut;
		month = ut.get(Calendar.MONTH) + 1;
		dayOfYear = ut.get(Calendar.DAY_OF_YEAR);
		ut.get(Calendar.HOUR_OF_DAY);
		ut.get(Calendar.MINUTE);
		ut.get(Calendar.SECOND);
		ut.get(Calendar.MILLISECOND);
		realHours = new RealHours(ut);
	}

	public void setLocation(double latitude, double longitude) {
		this.latitude = latitude;
		this.longitude = longitude;
	}

	/**
	 * Calculates:<br>
	 * <ul>
	 * <li>betIGRF - total field (gauss)</li>
	 * <li>zmaIGRF - downward vertical component</li>
	 * <li>xmaIGRF - x-component in the equatorial plane (x to zero longitude)</li>
	 * <li>ymaIGRF - y-component in the equatorial plane (x to zero longitude).</li>
	 * <li>dipIGRF - inclination angle (degree).</li>
	 * <li>decIGRF - declination angle (degree).</li>
	 * <li>modipIGRF - rawer's modified dip.</li>
	 * </ul>
	 * make sure location is set
	 */
	public void setMagneticField() {
		igrf.calc(ut, latitude, longitude, ALTITUDE_KM);
		xma = -igrf.bT / 100000;
		yma = igrf.bF / 100000;
		zma = -igrf.bR / 100000;
		bet = igrf.B / 100000;
		dip = igrf.dip;
		dec = igrf.dec;
		fce = igrf.fce;
		// calculate modipIGRF
		double dipInRadian = Math.toRadians(dip);
		double dipdiv = dipInRadian / Math.sqrt(dipInRadian * dipInRadian + Math.cos(Math.toRadians(latitude)));
		if (dipdiv > 1.0)
			dipdiv = 1.0;
		else if (dipdiv < -1.0)
			dipdiv = -1.0;
		modip = Math.asin(dipdiv);
		modip = Math.toDegrees(modip);
		//
		// Calculation of geom. coordinates in deg.............................
		// Calculation of magnetic inclination (dip), declination (dec)........
		// dip latitude (magbr) and modified dip (modip). all in degree......
		//
		double geoLatLong[] = new double[3];
		double magLatLong[] = new double[3];
		geoLatLong[Geopack.LATITUDE] = latitude;
		geoLatLong[Geopack.LONGITUDE] = longitude;
		geoLatLong[Geopack.ALTITUDE] = 0;
		geopack.recalc(ut);
		geopack.geomapToMagmap(geoLatLong, magLatLong);
		magLatitude = magLatLong[Geopack.LATITUDE];
		magLongitude = magLatLong[Geopack.LONGITUDE];
	}

	public void setSolarActivity() {
		sunspots = SunSpots.get(ut);
	}

	/**
	 * Returns latitude for which IRI model have been calculated.
	 */
	public double getLatitude() {
		return latitude;
	}

	/**
	 * Returns longitude for which IRI model have been calculated.
	 */
	public double getLongitude() {
		return longitude;
	}

	/**
	 * Returns dipole magnetic latitude for which IRI model have been
	 * calculated.
	 */
	public double getMagLatitude() {
		return magLatitude;
	}

	/**
	 * Returns dipole magnetic longitude for which IRI model have been
	 * calculated.
	 */
	public double getMagLongitude() {
		return magLongitude;
	}

	/**
	 * Returns time UT for which IRI model have been calculated.
	 */
	public TimeScale getTimeUT() {
		return ut;
	}

	/**
	 * Returns calculated sun spot number.
	 */
	public float getSunspotNumber() {
		return sunspots.getNumber();
	}

	/**
	 * Methods for sunspot number status
	 */
	public boolean isSunspotNumberPredicted() {
		return sunspots.isPredicted();
	}

	public boolean isSunspotNumberAvailable() {
		return sunspots.isAvailable();
	}

	/**
	 * Returns calculated Ground-solar time/location properties (GroundSolar
	 * object).
	 */
	public GroundSolar getGroundSolar() {
		return groundSolar;
	}

	/**
	 * Returns calculated critical frequency of E-layer.
	 */
	public double getFoEP() {
		return foEP;
	}

	/**
	 * Returns calculated critical frequency of F1-layer.
	 */
	public double getFoF1P() {
		return foF1P;
	}

	/**
	 * Returns calculated probability of F1-layer existence.
	 */
	public double getFoF1Probability() {
		return foF1probability;
	}

	/**
	 * Returns calculated critical frequency of F2-layer.
	 */
	public double getFoF2P() {
		return foF2P;
	}

	/**
	 * Returns calculated peak density of D layer.
	 */
	public double getNmDP() {
		return NmDP;
	}

	/**
	 * Returns <true> if auroral layer is possible
	 */
	public boolean isAuroralPossible() {
		return (magLatitude >= 60);
	}

	public double[][] getCoefs_foF2() {
		return f2[0];
	}

	public double[][] getCoefs_hmF2() {
		return f2[0]; // TODO
	}

	public void loadCoefficients() {
		double rr1, rr2;
		double rg = sunspots.getNumber() > 150.0 ? 150.0 : sunspots.getNumber();
		if (month != monthPrev || rg != rgPrev) {
			// .....only for a new month and/or a new sunspot number,
			// read two ursi coefficient sets for given month and low/high solar
			// activity
			// then interpolate for the given solar activity rg
			File file = new File(CommonConst.getShareResourcesDir(), URSI_MODEL_PREF + (10 + month) + "." + URSI_MODEL_EXT);
			String fileURSI = file.getPath();
			BufferedReader in = null;
			String line;
			int i = 0, j = 0, k = 0;
			try {
				in = new BufferedReader(new FileReader(file));
				while (in.ready()) {
					line = in.readLine();
					if (line == null)
						break;
					for (int m = 0; m < 4; m++) {
						f2[i][j][k] = FC.StringToDouble(line.substring(1 + m * 15, 1 + (m + 1) * 15));
						k++;
						if (k >= 13) {
							k = 0;
							j++;
							if (j >= 76) {
								j = 0;
								i++;
							}
						}
					}
					if (i == 2)
						break;
				}
			} catch (IOException ex) {
				System.out.println("Can't read file " + fileURSI + " !");
				System.out.println(ex.toString());
			} finally {
				if (in != null)
					try {
						in.close();
					} catch (IOException ex) {
						// Util.printThreadStackTrace( ex );
						ex.printStackTrace();
					}
			}
			//
			// Linear interpolation in solar activity, for a new value of rg
			//
			rr2 = rg / 100.0;
			rr1 = 1.0 - rr2;
			for (i = 0; i < 76; i++) {
				for (j = 0; j < 13; j++) {
					f2[0][i][j] = f2[0][i][j] * rr1 + f2[1][i][j] * rr2;
				}
			}
			monthPrev = month;
			rgPrev = rg;
		}
	}

	public void setCoefficients(double coeff[][]) {
		for (int i = 0; i < 76; i++)
			for (int j = 0; j < 13; j++)
				f2[0][i][j] = coeff[i][j];
	}

	private void calcPrediction() {
		double abslat, absmbr, magbr, ph, utHour, lcHour, rg;

		utHour = realHours.getRealHours();
		lcHour = TimeScale.getLongitudeHours((float) longitude, utHour).getRealHours();
		groundSolar = new GroundSolar(latitude, longitude, ut);
		rg = sunspots.getNumber() > 150.0 ? 150.0 : sunspots.getNumber();

		//
		// -----------------------foE
		// PREDICTION----------------------------------
		//
		ph = 63.75 + rg * (0.728 + rg * 0.00089);
		abslat = Math.abs(latitude);
		foEP = foEpIRI(ph, groundSolar.getZenith(), groundSolar.getNoonZenith(), abslat);
		// foEP = Math.round( foEP * 1000.0 ) / 1000.0;
		//
		// -----------------------foF1
		// prediction---------------------------------
		//
		// Calculation of season (spring=1, summer=2, autumn=3, winter=4)
		// According to the following calculations, in the north hemisphere
		// spring: iday= 47 - 138 (feb 16 - may 18)
		// summer: iday=139 - 230 (may 19 - aug 18)
		// autumn: iday=231 - 322 (aug 19 - nov 18)
		// winter: iday=323 - 46 (nov 19 - feb 15)
		int season = (dayOfYear + 45) / 92;
		if (season < 1)
			season = 4;
		if (latitude < 0.0) {
			season -= 2;
			if (season < 1)
				season += 4;
		}

		// .....gfq/MHz=gyro-frequency at 300 km is
		// gfq=anint(2.84*bet*1000.0)/1000.0
		// write(*,'(2(A,F8.3))') " fH=",gfq," dip=",dip

		magbr = Math.atan(0.5 * Math.tan(Math.toRadians(dip)));
		magbr = Math.toDegrees(magbr);
		absmbr = Math.abs(magbr);
		// Predicted foF1p
		// .....note: If foF1p=999.9 (MHz), it means no F1 layer.
		foF1P = NO_FOF1;
		foF1probability = 0.;
		if (!groundSolar.isNight(lcHour)) {
			foF1P = foF1ED(absmbr, sunspots.getNumber(), groundSolar.getZenith());
			if (foF1P < foEP)
				foF1P = NO_FOF1;
			else
				foF1probability = getF1Probability(magLatitude, sunspots.getNumber(), groundSolar.getZenith());
		}

		NmDP = XmDED( groundSolar.getZenith(), sunspots.getNumber(), 4.E8 );
		
		//
		// -----------------------foF2  prediction---------------------------------
		//

		foF2P = f2Out(modip, latitude, longitude, f2[0], utHour); // make sure
																	// f2[] are
																	// loaded...
	}

	// ----------------------------------------------------------------
	// ....THE FOLLOWING FUNCTION IS COPIED FROM IRI-90 TO PREDICT foE.
	//
	static private double foEpIRI(double cov, double zenith, double noonZenith, double absLatitude) {
		double foEpIRI;
		// ----------------------------------------------------------------
		// Calculates FOE/MHz by the Edinburgh-method.
		// Input: mean 10.7CM solar ratio flux (cov), absolute geographic
		// latitude in degree (absLatitude), solar zenith angle ( zenith )
		// solar zenith angle at noon ( noonZenith ).
		// Reference:
		// Kouris-Muggeleton, CCIR doc. 6/3/07, 1973
		// TROST, J. GEOPHYS. RES. 84, 2736, 1979 (was used
		// to improve the nighttime variation)
		// D.Bilitza --------------------------------- August 1986.
		//
		double a, b, c, d, smin, sl, sm, sp, xhic;
		//
		// variation with solar activity (factor A) ...............
		a = 1.0 + 0.0094 * (cov - 66.0);
		// variation with noon solar zenith angle (B) and with latitude (C)
		sl = Math.cos(Math.toRadians(absLatitude));
		if (absLatitude < 32.0) {
			sm = -1.93 + 1.92 * sl;
			c = 23.0 + 116.0 * sl;
		} else {
			sm = 0.11 - 0.49 * sl;
			c = 92.0 + 35.0 * sl;
		}
		if (noonZenith >= 90.0)
			noonZenith = 89.999;
		b = Math.pow(Math.cos(Math.toRadians(noonZenith)), sm);
		// variation with solar zenith angle (D) ..........................
		if (absLatitude > 12.0)
			sp = 1.2;
		else
			sp = 1.31;
		// adjusted solar zenith angle during nighttime (XHIC) .............
		xhic = zenith - 3. * Math.log(1. + Math.exp((zenith - 89.98) / 3.));
		d = Math.pow(Math.cos(Math.toRadians(xhic)), sp);
		// determine foE**4 ................................................
		foEpIRI = a * b * c * d;
		// minimum allowable foE (sqrt[SMIN])...............................
		smin = 0.121 + 0.0015 * (cov - 60.);
		smin = smin * smin;
		if (foEpIRI < smin)
			foEpIRI = smin;
		foEpIRI = Math.pow(foEpIRI, 0.25);
		return foEpIRI;
	}

	// --------------------------------------------------------------
	// CALCULATES THE F1 PEAK PLASMA FREQUENCY (FOF1/MHZ)
	// FOR DIP-LATITUDE (YLATI/DEGREE)
	// SMOOTHED ZURICH SUNSPOT NUMBER (R)
	// SOLAR ZENITH ANGLE (CHI/DEGREE)
	// REFERENCE:
	// E.D.DUCHARME ET AL., RADIO SCIENCE 6, 369-378, 1971
	// AND 8, 837-839, 1973
	// HOWEVER WITH MAGNETIC DIP LATITUDE INSTEAD OF GEOMAGNETIC
	// DIPOLE LATITUDE, EYFRIG, 1979
	// --------------------------------------------- D. BILITZA, 1988.
	static private double foF1ED(double yLati, double r, double chi) {
		double foF1ED = 0.0;
		double chi0, chi100, chim, dla, f0, f100, fs, xmue;

		dla = yLati;
		chi0 = 49.84733 + 0.349504 * dla;
		chi100 = 38.96113 + 0.509932 * dla;
		chim = (chi0 + (chi100 - chi0) * r / 100.);
		if (chi <= chim) {
			f0 = 4.35 + dla * (0.0058 - 1.2E-4 * dla);
			f100 = 5.348 + dla * (0.011 - 2.3E-4 * dla);
			fs = f0 + (f100 - f0) * r / 100.0;
			xmue = 0.093 + dla * (0.0046 - 5.4E-5 * dla) + 3.0E-4 * r;
			foF1ED = fs * Math.pow(Math.cos(Math.toRadians(chi)), xmue);
		}
		return foF1ED;
	}

	// -----------------------------------------------------------------------
	//
	// c Occurrence probability of F1 layer after Scotto et al., Advances in
	// c Space Research, Volume 20, Number 9, 1773-1775, 1997.
	// c Input: solar zenith angle (sza) in degrees, geomagnetic latitude
	// c (glat) in degrees, 12-month running mean of sunspot number (Rz12).
	// c Output: F1 occurrence probability without L-condition cases (f1prob)
	// c and with L-condition cases (f1probl)

	static private double getF1Probability(double magLat, double ssn, double zenith_deg) {

		double xarg = 0.5 + 0.5 * Math.cos(Math.toRadians(zenith_deg));
		double a = 2.98 + 0.0854 * ssn;
		double b = 0.0107 - 0.0022 * ssn;
		double c = -0.000256 + 0.0000147 * ssn;
		double gamma = a + (b + c * magLat) * magLat;
		double f1pr = Math.pow(xarg, gamma);
		if (f1pr < 1.e-3)
			f1pr = 0.;
		return f1pr;
		// double f1prl = Math.pow( xarg, 2.36 );
		// if( f1prl < 1.e-3)
		// f1prl = 0.;
		// f1probl=f1prl;
	}

	// --------------------------------------------------------------
	// D. BILITZA, 1978, CALCULATES ELECTRON DENSITY OF D MAXIMUM.                   
	// XHI/DEG. IS SOLAR ZENITH ANGLE, R SMOOTHED ZURICH SUNSPOT NUMBER              
	// AND YW/M-3 THE ASSUMED CONSTANT NIGHT VALUE.     
	// [REF.: D.BILITZA, WORLD DATA CENTER A REPORT UAG-82,7,BOULDER,1981]
	// corrected 4/25/97 - D. Bilitza
	//
	static private double XmDED(double chi, double r, double nightValue ) {

		if( chi > 90 ) return nightValue;
		double y = 6.05E8 + 0.088E8 * r;
		double yy = Math.cos ( Math.toRadians(chi));
		double yyy = -0.1 / ( Math.pow(yy, 2.7) ); 
        double ymd = yyy < -40? 0 : y * Math.exp(yyy); 
        return Math.max(ymd, nightValue );
	}


	/**
	 * Calculates fof2/mhz and m3000 using the CCIR-MAPS.<br>
	 * Input:<br>
	 * <ul>
	 * <li>modified dip latitude, <code>modip</code>, in deg.</li>
	 * <li>geographical latitude, <code>latitude</code>, in deg.</li>
	 * <li><code>longitude</code>, in deg.</li>
	 * <li>smoothed sun spot numbers, <code>ff0</code></li>
	 * <li>universal time of day <code>utHour</code>, in decimal hours</li>
	 * </ul>
	 * 
	 * @author D.Bilitza, July 85.
	 */
	static private double f2Out(double modip, double latitude, double longitude, double ff0[][], double utHour) {
		int qf[] = { 11, 11, 8, 4, 1, 0, 0, 0, 0 };
		return gamma1(modip, latitude, longitude, utHour, 6, qf, 9, 76, 13, 988, ff0);
	}

	public double foF2NewLocation(double latitude, double longitude) {
		this.latitude = latitude;
		this.longitude = longitude;
		setMagneticField();
		return f2Out(modip, latitude, longitude, f2[0], realHours.getRealHours());
	}

	/**
	 * Calculate new foF2 value using the same coefficients and new time
	 * 
	 * @param ut_hours
	 *            - UT of the day ranging from 0.0 to 24.00
	 * @return
	 */
	public double foF2NewTime(double ut_hours) {
		return f2Out(modip, latitude, longitude, f2[0], ut_hours);
	}

	/**
	 * Calculates gamma1=fof2 or M3000 using CCIR numerical map coefficients
	 * <code>sfe</code> for modified dip latitude, <code>smodip</code> (in deg),
	 * geographic latitude, <code>latitude</code> (in deg),
	 * <code>longitude</code> (in deg), and universal time of day,
	 * <code>hour</code> (in decimal hours).<br>
	 * <code>nq</code> is an integer array giving the highest degrees in
	 * latitude for each longitude harmonics.<br>
	 * m=1+nq1+2(nq2+1)+2(nq3+1)+... .<br>
	 * 
	 * @author Sheikh, 4.3.77.
	 */
	static private double gamma1(double smodip, double latitude, double longitude, double hour, int TIME_ORDER6, int nq[], int k1, int M76,
			int M13, int m3, double sfe[][]) {
		int index;
		double sum;
		double s0, s1, s2, s3, ss, xsinx[] = new double[13];
		int j, l, np;

		double[] coef = getSphericalCoefficients(hour, TIME_ORDER6, M76, sfe);

		// now do spherical restoration using 76 coefficients in coef[]
		// first use ss variable to compute function terms for sin**n(modip)

		sum = coef[0]; // start with zero degree zero order constant
		ss = Math.sin(Math.toRadians(smodip));
		s3 = ss;
		xsinx[0] = 1.0;
		index = nq[0];
		for (j = 0; j < index; j++) {
			sum = sum + coef[j + 1] * ss;
			xsinx[j + 1] = ss;
			ss = ss * s3;
		}
		xsinx[nq[0] + 1] = ss;
		np = nq[0] + 1;
		ss = Math.cos(Math.toRadians(latitude));
		s3 = ss;
		for (j = 1; j < k1; j++) {
			s0 = Math.toRadians(longitude * j);
			s1 = Math.cos(s0);
			s2 = Math.sin(s0);
			index = nq[j] + 1;
			for (l = 0; l < index; l++) {
				np++;
				sum += coef[np - 1] * xsinx[l] * ss * s1;
				np++;
				sum += coef[np - 1] * xsinx[l] * ss * s2;
			}
			ss = ss * s3;
		}
		return sum;
	}

	/**
	 * Returns array of spherical expansion coefficients for given time of day
	 * ("hour")
	 * 
	 */
	static public double[] getSphericalCoefficients(double hour, int TIME_ORDER6, int M76, double sfe[][]) {
		double[] c = new double[12], s = new double[12], coef = new double[M76];
		double hou;
		int i, j;

		// calculate base functions of diurnal expansion (sinT and cosT) -- up
		// to 6th order
		hou = Math.toRadians(15.0 * hour - 180.0);
		s[0] = Math.sin(hou);
		c[0] = Math.cos(hou);
		for (i = 1; i < TIME_ORDER6; i++) {
			c[i] = c[0] * c[i - 1] - s[0] * s[i - 1];
			s[i] = c[0] * s[i - 1] + s[0] * c[i - 1];
		}

		// now calculate 76 coefficients U for given time "hour"

		for (i = 0; i < M76; i++) {
			coef[i] = sfe[i][0];
			for (j = 0; j < TIME_ORDER6; j++) {
				coef[i] = coef[i] + sfe[i][2 * j + 1] * s[j] + sfe[i][2 * j + 2] * c[j];
			}
		}
		return coef;
	}

	/**
	 * 
	 * Returns coefficients of diurnal expansion for a given location
	 * 
	 * @param smodip
	 *            -- modified modip for the location
	 * @param latitude
	 *            -- GEO coordinates of the location
	 * @param longitude
	 * 
	 * @param nq
	 *            -- array of degrees for each order of spherical expansions
	 * @param k1
	 *            -- number of orders (usualy 9)
	 * @param M13
	 *            -- total number of output coefficients
	 * @param sfe
	 *            -- big matrix of all expansions coefficients
	 * @return
	 */
	static public double[] getDiurnalCoefficients(double smodip, double latitude, double longitude, int nq[], int k1, int M13, double sfe[][]) {
		int index;
		double[] coef = new double[M13];
		double sum;
		double s0, s1, s2, s3, ss, xsinx[] = new double[13];
		int i, j, l, np;

		for (i = 0; i < M13; i++) {
			sum = sfe[0][i]; // start with zero degree zero order constant
			ss = Math.sin(Math.toRadians(smodip));
			s3 = ss;
			xsinx[0] = 1.0;
			index = nq[0];
			for (j = 0; j < index; j++) {
				sum = sum + sfe[j + 1][i] * ss;
				xsinx[j + 1] = ss;
				ss = ss * s3;
			}
			xsinx[nq[0] + 1] = ss;
			np = nq[0] + 1;
			ss = Math.cos(Math.toRadians(latitude));
			s3 = ss;
			for (j = 1; j < k1; j++) {
				s0 = Math.toRadians(longitude * j);
				s1 = Math.cos(s0);
				s2 = Math.sin(s0);
				index = nq[j] + 1;
				for (l = 0; l < index; l++) {
					np++;
					sum += sfe[np - 1][i] * xsinx[l] * ss * s1;
					np++;
					sum += sfe[np - 1][i] * xsinx[l] * ss * s2;
				}
				ss = ss * s3;
			}
			coef[i] = sum;
		}
		return coef;
	}
	
	
}
