package edu.uml.lgdc.geospace;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Calendar;

import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.project.CommonConst;
import edu.uml.lgdc.time.TimeScale;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
public class SunSpots {
	// Sun spots file
	static public final String SUNSPOT = "Sunspt.asc";
	static private final String SUNSPOT_FILENAME = new File(CommonConst.getShareResourcesDir(), SUNSPOT).getPath();
	static private final int MAX_NUMBER_OF_YEARS = 300;
	static private final String SPACES = "                              " + "                              " + "                              ";
	// file name ( including path if necessary ) where sunspot data are located
	static private String fileName = SUNSPOT_FILENAME;

	static public final int SUNSPOT_UNKNOWN = -1;

	// = true, if we already called method load
	static private boolean loaded = false;

	/**
	 * These variables and array constitute internal structure and control of
	 * sunspot data
	 */

	// This array contains sunspot numbers. One number per month.
	// First dimension is for years, second dimension (of length 12) is for
	// months in this year. So element ssNumbers[2][3] contains sunspot number
	// for month April of year <firstYear> + 2
	static private float[][] ssNumbers = null;

	// Year which correspond to the first year in array ssNumbers
	static private int firstYear = 0;

	// Number of years in array ssNumbers - usually this number is less than
	// actual dimension of array ssNumbers.
	static private int qtyOfYears = 0;

	// This array contains start of solar activities cycles counted as
	// offset in month beginning from the start of year <firstYear>.
	static private int[] startOfCycles = null;

	// Number of cycles in array startOfCycles - usually this number is less
	// than
	// actual dimension of array startOfCycles
	static private int qtyOfCycles = 0;

	// = true if data already have been extrapolated
	static private boolean extrapolated = false;

	static private int lastActualDataYear = 0;
	static private int lastActualDataMonth = 0;

	/**
	 * cache
	 */
	static private SunSpots sunspots;

	static public final int SSN_PREDICTED = 0;
	static public final int SSN_ACTUAL = 1;
	static public final int SSN_MISSING = 2;

	private int year;
	private int month; // goes from 1 to 12

	private float number;
	private int status;

	private SunSpots(int year, int month, float number, int status) {
		if (status != SSN_PREDICTED && status != SSN_ACTUAL && status != SSN_MISSING)
			throw new IllegalArgumentException("illegal status, " + status + ", should be " + SSN_PREDICTED + ", " + SSN_ACTUAL + ", or "
					+ SSN_MISSING);
		if (month < 1 || month > 12)
			throw new IllegalArgumentException("illegal month number, " + month);
		if (year < 1 || year > 3000)
			throw new IllegalArgumentException("illegal year, " + year);
		if (status != SSN_MISSING && number < 1 || number > 200)
			throw new IllegalArgumentException("illegal spot number, " + number);
		if (status == SSN_MISSING && number != SUNSPOT_UNKNOWN || number == SUNSPOT_UNKNOWN && status != SSN_MISSING)
			throw new IllegalArgumentException("incompatible values for status and number, " + status + " and " + number);
		this.year = year;
		this.month = month;
		this.number = number;
		this.status = status;
	}

	public float getNumber() {
		return number;
	}

	public int getStatus() {
		return status;
	}

	public boolean isAvailable() {
		return status != SSN_MISSING;
	}

	public boolean isPredicted() {
		return status == SSN_PREDICTED;
	}

	public int getYear() {
		return year;
	}

	public int getMonth() {
		return month;
	}

	/**
	 * Returns instance of {@link SunSpots} for this date.
	 */
	static public SunSpots get(TimeScale date) {
		return get(date.get(Calendar.YEAR), date.get(Calendar.MONTH) + 1);
	}

	/**
	 * Returns instance of {@link SunSpots} for this date.
	 */
	synchronized static public SunSpots get(int year, int month) {
		if (!loaded)
			load();
		if (sunspots != null && year == sunspots.year && month == sunspots.month)
			return sunspots;
		sunspots = getIt(year, month);
		return sunspots;
	}

	/**
	 * This method is used for loading sunspot data from file into memory. !!
	 * inside of this method months counting is based on zero
	 */
	synchronized static private void load() {
		boolean error = false;
		BufferedReader in = null;
		int cycle = 0, year = 0, month = 0;
		int prevYear = 0;
		int mo, pos;
		boolean blankField = false;
		String line;

		firstYear = 0;
		qtyOfYears = 0;
		qtyOfCycles = 0;
		try {
			in = new BufferedReader(new FileReader(fileName));
			initMemory();
			loaded = true;
			extrapolated = false;
			while (in.ready()) {
				line = in.readLine();
				if (line == null)
					break;
				if (line.length() == 0)
					continue;
				if (!Character.isDigit(line.charAt(0)))
					continue;
				if (line.length() < 83) {
					line += SPACES.substring(0, 83 - line.length());
				}
				cycle = (int) FC.StringToInteger(line.substring(79, 83));
				year = (int) FC.StringToInteger(line.substring(0, 4));
				month = (int) FC.StringToInteger(line.substring(5, 7));
				if (qtyOfCycles == 0 && cycle > 0)
					qtyOfCycles = cycle - 1;
				if (prevYear == 0 && year > 0)
					prevYear = year - 1;
				if (prevYear != year - 1 || year < 1749) {
					System.out.println("File " + fileName + " constructed wrong, data missing before year = " + year);
					error = true;
					break;
				}
				prevYear = prevYear + 1;

				blankField = false;
				pos = 7;
				for (mo = 0; mo < 12; mo++) {
					if (line.substring(pos, pos + 6).equals("      ")) {
						if (mo != 0)
							qtyOfYears++;
						blankField = true;
						break;
					}
					ssNumbers[qtyOfYears][mo] = (float) FC.StringToDouble(line.substring(pos, pos + 6));
					pos += 6;
				}
				if (!blankField) {
					qtyOfYears++;
					if (firstYear == 0)
						firstYear = year;
					if (month == 0)
						continue;
				}
				if (!blankField || month != 0) {
					qtyOfCycles++;
					if (qtyOfCycles != cycle) {
						System.out.println("File " + fileName + " constructed wrong. Cycles do not match before cycle " + cycle);
						error = true;
						break;
					} else {
						startOfCycles[qtyOfCycles - 1] = (year - firstYear) * TimeScale.MONTHS_PER_YEAR + month - 1;
					}
				} else {
					// considered as end of data
					break;
				}
			}
		} catch (Exception ex) {
			System.out.println("Can't calculate sunspot number due the following:");
			System.out.println(ex.toString());
			error = true;
		} finally {
			if (in != null)
				try {
					in.close();
				} catch (IOException ex) {
					// TODO Util.printThreadStackTrace( ex );
					ex.printStackTrace();
				}
		}
		if (error) {
			ssNumbers = null;
			startOfCycles = null;
		}
	}

	/**
	 * Gets sunspot data. Extrapolate if necessary.<br>
	 * Returns instance of {@link SunSpots} for this date.
	 */
	synchronized static private SunSpots getIt(int year, int month) {
		float sunspotNumber = SUNSPOT_UNKNOWN;
		int status = SSN_MISSING;
		if (ssNumbers != null) {
			if (year >= firstYear && year < firstYear + MAX_NUMBER_OF_YEARS) {
				sunspotNumber = ssNumbers[year - firstYear][month - 1];
				if (sunspotNumber < 0) {
					if (!extrapolated) {
						extrapolate();
						sunspotNumber = ssNumbers[year - firstYear][month - 1];
						extrapolated = true;
					}
				}
				status = SSN_PREDICTED;

				if (!extrapolated)
					status = SSN_ACTUAL;
				else {
					if (year < lastActualDataYear || (year == lastActualDataYear && month <= lastActualDataMonth + 1))
						status = SSN_ACTUAL;
				}
			}
		}
		return new SunSpots(year, month, sunspotNumber, status);
	}

	/**
	 * This function extrapolates sunspot data 11 years ahead. This function is
	 * called only once. !! months is based on zero inside of this method.
	 */
	static private void extrapolate() {
		int nbeg = firstYear + startOfCycles[qtyOfCycles - 1] / TimeScale.MONTHS_PER_YEAR;
		int nend = nbeg + TimeScale.MONTHS_PER_YEAR - 1;
		int int1 = 0;
		int ibmo = startOfCycles[qtyOfCycles - 1] % 12; // starting month of
														// last cycle
		int idxyr;
		int ix;
		float ssn = 0f;
		float[] x = new float[qtyOfCycles], y = new float[qtyOfCycles];

		for (int yr = nbeg; yr <= nend; yr++) {
			idxyr = yr - firstYear;
			for (int mo = 0; mo < TimeScale.MONTHS_PER_YEAR; mo++) {
				if (yr == nbeg && mo < ibmo)
					continue;
				if (ssNumbers[idxyr][mo] >= 0) {
					// sunspot is known, no prediction necessary
					ssn = ssNumbers[idxyr][mo];
					lastActualDataYear = yr;
					lastActualDataMonth = mo;
				} else {
					int1 = int1 + 1; // # of months into prediction
					for (int j = 0; j < qtyOfCycles - 1; j++) {
						ix = startOfCycles[j] + (lastActualDataYear - nbeg) * 12 + (lastActualDataMonth - ibmo);
						x[j] = ssNumbers[ix / TimeScale.MONTHS_PER_YEAR][ix % TimeScale.MONTHS_PER_YEAR];
						y[j] = ssNumbers[(ix + int1) / TimeScale.MONTHS_PER_YEAR][(ix + int1) % TimeScale.MONTHS_PER_YEAR];
					}
					ssNumbers[idxyr][mo] = (float) predictSeries(x, y, 7, qtyOfCycles - 1, ssn);
				}
			}
		}
	}

	/**
	 * This method predict next value ( nextY ) of series y if we know next
	 * value ( nextX ) of correlated series x.
	 */
	static private double predictSeries(float[] x, float[] y, int skip, int qty, float nextX) {
		int n = qty - skip;
		double nextY = 0;
		double sumX, sumX2, sumY, sumY2, sumXY, avgX, avgY;
		double corr;

		sumX = 0.0;
		sumY = 0.0;
		sumXY = 0.0;
		sumX2 = 0.0;
		sumY2 = 0.0;
		for (int j = skip; j < qty; j++) {
			sumX = sumX + x[j];
			sumY = sumY + y[j];
			sumXY = sumXY + x[j] * y[j];
			sumY2 = sumY2 + y[j] * y[j];
			sumX2 = sumX2 + x[j] * x[j];
		}
		avgX = sumX / n;
		avgY = sumY / n;
		corr = (n * sumXY - sumX * sumY) / (n * sumX2 - sumX * sumX);
		nextY = avgY + corr * (nextX - avgX);
		return nextY;
	}

	/**
	 * Initialize containers and control parameters before loading
	 */
	static private void initMemory() {
		if (ssNumbers == null)
			ssNumbers = new float[MAX_NUMBER_OF_YEARS][TimeScale.MONTHS_PER_YEAR];
		if (startOfCycles == null)
			startOfCycles = new int[MAX_NUMBER_OF_YEARS * TimeScale.MONTHS_PER_YEAR / 10];
		for (int i = 0; i < ssNumbers.length; i++) {
			for (int j = 0; j < TimeScale.MONTHS_PER_YEAR; j++)
				ssNumbers[i][j] = SUNSPOT_UNKNOWN;
		}
		for (int i = 0; i < startOfCycles.length; i++) {
			startOfCycles[i] = -1;
		}
	}
}