/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.geospace;

import java.util.Calendar;

import edu.uml.lgdc.time.TimeScale;

public class TimeGrid {
	static public final int DEFAULT_TIME_STEP = 15; // in minutes
	static public final int MAX_TIME_STEP = 2; // in minutes

	private int timeStep = DEFAULT_TIME_STEP; // in munites
	private int nDiurnalSteps = TimeScale.MINUTES_PER_DAY / timeStep;

	private TimeScale timesUT[] = null;

	public TimeGrid() {
		//
	}
	
	public void setNewTimeStep(int timeStep_min) {
		if (checkTimeStep(TimeScale.MINUTES_PER_DAY, timeStep_min) == false) {
			throw new RuntimeException("Check time grid step");
		}
		timeStep = timeStep_min;
		nDiurnalSteps = TimeScale.MINUTES_PER_DAY / timeStep;
	}

	public int getTimeStep_min() {
		return timeStep;
	}

	public int getNumDiurnalSteps() {
		return nDiurnalSteps;
	}

	static public boolean checkTimeStep(int range, int step) {
		return range % step == 0 && step >= MAX_TIME_STEP;
	}

	/**
	 * Calculates time shifted by iTime steps from startTime
	 * 
	 * @param startTime
	 * @param iTime
	 * @return
	 */
	public TimeScale getTime(final TimeScale startTime, int iTime) {
		return new TimeScale(startTime.getTimeInMinutes() + iTime * timeStep);
	}

	public int findClosestTimeStep(final TimeScale time) {
//		double minuteOfDay = time.getTimeSinceMidnightIn(DateTime.MINUTE);
		double minutesSinceStart = time.diffIn(Calendar.MINUTE, timesUT[0] );
		return (int) Math.round(minutesSinceStart / timeStep);
	}

	public int timeToDiurnalPhaseIndex(final TimeScale time) {
		double minuteOfDay = time.getTimeSinceMidnightIn(TimeScale.MINUTE);
		int iTime = (int) Math.round(minuteOfDay / timeStep);
		return (iTime + nDiurnalSteps) % nDiurnalSteps;
	}

	public int findTimeNode(final TimeScale time, final TimeScale startTime) {
		double minutes = time.getTimeInMinutes() - startTime.getTimeInMinutes();
		return (int) Math.round(minutes / timeStep);
	}

	/**
	 * Search for the nearest time before/after given <code>time</code> that
	 * satisfies conditions: minutes(<code>time</code>) ==
	 * <code>TIME_STEP</code>*n + <code>offsetInMinutes</code> where n is
	 * integer. This method changes original <code>time</code>. This method is
	 * needed since sites (ionosondes) send data to GIRO with some delay and
	 * most data are available at specific times. For example - 15:07, 15:22,
	 * 15:37 etc.
	 * 
	 * <p>
	 * Seconds and milliseconds are irrelevant for this method. If input
	 * <code>time</code> has minutes satisfy relation <code>TIME_STEP</code>*n +
	 * <code>offsetInMinutes</code>, method will NOT change it neither for
	 * <code>before == true</code> nor for <code>before == false</code>, see
	 * below:
	 * 
	 * <li>Initially 2014.06.13 19:06:02 before 2014.06.13 16:52:00 after 2014.06.13 19:07:00
	 * <li>Initially 2014.06.13 19:07:02 before 2014.06.13 19:07:00 after 2014.06.13 19:07:00
	 * <li>Initially 2014.06.13 19:08:02 before 2014.06.13 19:07:00 after 2014.06.13 19:22:00
	 * <li>Initially 2014.06.13 19:09:02 before 2014.06.13 19:07:00 after 2014.06.13 19:22:00
	 * 
	 * @param time
	 *            - original time
	 * @param timeStep
	 * @param offsetInMinutes
	 *            - offset from TIME_STEPs
	 * @param before
	 *            - specify before or after given time
	 */
	public void findOffsetedTime(TimeScale time, int offsetInMinutes, boolean before) {
		time.set(TimeScale.MILLISECOND, 0);
		time.set(TimeScale.SECOND, 0);
		int minutes = timeStep * (int) Math.floor(time.getTimeSinceMidnightIn(TimeScale.MINUTE) / timeStep);
		TimeScale temp = time.startOfDay();
		temp.add(TimeScale.MINUTE, minutes + offsetInMinutes);
		while (before == true && temp.after(time)) {
			temp.add(TimeScale.MINUTE, -timeStep);
		}
		while (before == false && temp.before(time)) {
			temp.add(TimeScale.MINUTE, timeStep);
		}
		time.setTimeInMillis(temp.getTimeInMillis());
	}

	public void findClosestOffsetedTime(TimeScale time, int offsetInMinutes) {
		TimeScale beforeTime = new TimeScale(time.getTimeInMinutes());
		TimeScale afterTime = new TimeScale(time.getTimeInMinutes());
		findOffsetedTime(beforeTime, offsetInMinutes, true);
		findOffsetedTime(afterTime, offsetInMinutes, false);
		double delayBefore = Math.abs(time.getTimeInMinutes() - beforeTime.getTimeInMinutes());
		double delayAfter = Math.abs(time.getTimeInMinutes() - afterTime.getTimeInMinutes());
		time.setTimeInMinutes(delayAfter < delayBefore ? afterTime.getTimeInMinutes() : beforeTime.getTimeInMinutes());
	}

	public boolean isTimeOffseted(TimeScale time, int offsetInMinutes) {
		return time.get(TimeScale.MINUTE) % timeStep == offsetInMinutes;
	}

	/**
	 * Calculates number of time steps in interval. Both ends are included.
	 * 
	 * @param startTime
	 * @param endTime
	 * @return
	 */
	public int calculateNumberTimeSteps(TimeScale startTime, TimeScale endTime) {
		return (int) ((endTime.getTimeInMinutes() - startTime.getTimeInMinutes()) / timeStep) + 1;
	}

	/**
	 * Returns time that is closest to <code>time</code> and has integer number
	 * of <code>timeStep</code> since midnight.
	 * 
	 * @param time
	 * @param timeStep
	 * @return
	 */
	public TimeScale snapToClosestGrid(final TimeScale time) {
		TimeScale timeAdjusted = new TimeScale( timesUT[0].getTimeInMillis());
		timeAdjusted.addMillis(TimeScale.MILLISECONDS_PER_MINUTE * (long) timeStep * findClosestTimeStep(time));
		return timeAdjusted;
	}

	public TimeScale snapToEarlierGrid(final TimeScale time) {
		TimeScale timeAdjusted = snapToClosestGrid(time);
		if( timeAdjusted.after(time)) {
			timeAdjusted.add(Calendar.MINUTE, -getTimeStep_min());
		}
		return timeAdjusted;
	}
	
	public TimeScale lowerBoundSnapped(final TimeScale startTime) {
		TimeScale fromTime = snapToClosestGrid(startTime);
		return new TimeScale(fromTime.getTimeInMinutes() - timeStep / 2. + 1. / TimeScale.MILLISECONDS_PER_MINUTE);
	}

	public TimeScale upperBoundSnapped(final TimeScale endTime) {
		TimeScale toTime = snapToClosestGrid(endTime);
		return new TimeScale(toTime.getTimeInMinutes() + timeStep / 2. - 1. / TimeScale.MILLISECONDS_PER_MINUTE);
	}
	
	public TimeScale getOneDayWindowStart(TimeScale tov ) {
		return new TimeScale(tov.getTimeInMinutes() - TimeScale.MINUTES_PER_DAY + getTimeStep_min());
	}

	public TimeScale getOneDayWindowStartQuery(TimeScale tov ) {
		return lowerBoundSnapped(getOneDayWindowStart(tov));
	}

	public void setUTs(TimeScale tov, int forecastDuration_min) {
		// Fill UT times for each of the time grid steps
		final double start = tov.getTimeInMinutes() - TimeScale.MINUTES_PER_DAY + timeStep;
		final int nTimeNodes = getNumDiurnalSteps();
		final int totalTimeNodes = nTimeNodes + forecastDuration_min/timeStep;
		timesUT = new TimeScale[totalTimeNodes];
		for (int iTime = 0; iTime < totalTimeNodes; iTime++) {
			timesUT[iTime] = new TimeScale(start + timeStep * iTime);
		}
	}

	public TimeScale[] getUTs() {
		return timesUT;
	}

	public TimeScale getTov() {
		return timesUT[timesUT.length-1];
	}
}
