package edu.uml.lgdc.instrument;

import java.io.IOException;

import edu.uml.lgdc.time.TimeScale;
import edu.uml.lgdc.time.TimeType;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
public interface AbstractStation extends Comparable<AbstractStation> {
	/**
	 * Returns name of this station.
	 */
	abstract public String getName();
	/**
	 * Returns URSI code of this station.<br>
	 * This method has a sense for URSI registered stations, which the most of ground stations are.
	 */
	abstract public String getUrsi();
	/**
	 * Returns <b>true</b> if this station is mobile. For example, satellite is always considered mobile, some ground station also can be
	 * mobile.<br>
	 * In short, not mobile station is a ground station that has the same position on Earth all the time.<br>
	 * For satellite methods {@link #getLatitude(TimeScale)} and {@link #getLongitude(TimeScale)} might not have much sense, but rather
	 * method {@link #getGEIPosition(TimeScale time)} has to be carefully implemented with prediction in near future.
	 */
	abstract public boolean isMobile();
	/**
	 * Returns Geographic coordinates for this satellite/station for specified UT <code>time</code>.<br>
	 * This method does not have much sense for satellites.
	 */
	abstract public double[] getGEOPosition(TimeScale time);
	/**
	 * Returns Geocentric Equatorial Inertial coordinates for this station for specified UT <code>time</code>.<br>
	 * This method does not have much sense for ground stations.
	 */
	abstract public double[] getGEIPosition(TimeScale time);
	/**
	 * Returns geographic latitude of this station.<br>
	 * Notes:
	 * <ul>
	 *   <li>For mobile station it calls method {@link #getLatitude(TimeScale time)} using current time</li>
	 *   <li>This method does not have much sense for satellite</li>
	 * </ul>
	 */
	abstract public double getLatitude();
	/**
	 * Returns geographic latitude of this station for given <code>time</code>.<br>
	 * Note:
	 * <ul>
	 *   <li>For not mobile station parameter <code>time</code> is not used</li>
	 *   <li>This method does not have much sense for satellite</li>
	 * </ul>
	 */
	abstract public double getLatitude(TimeScale time);
	/**
	 * Returns geographic longitude of this station.<br>
	 * Notes:
	 * <ul>
	 *   <li>For mobile station it calls method {@link #getLongitude(TimeScale time)} using current time</li>
	 *   <li>This method does not have much sense for satellite</li>
	 * </ul>
	 */
	abstract public double getLongitude();
	/**
	 * Returns geographic longitude of this station for given <code>time</code>.<br>
	 * Notes:
	 * <ul>
	 *   <li>For not mobile station parameter <code>time</code> is not used</li>
	 *   <li>This method does not have much sense for satellite</li>
	 * </ul>
	 */
	abstract public double getLongitude(TimeScale time);
	/**
	 * Compares ursi codes of this station with ursi code of given <code>station</code>.<br>
	 * Returns:
	 * <ul>
	 *   <li>-1 - if this ursi code is less than given <code>station</code> ursi code</li>
	 *   <li>0 - if they are equaled</li>
	 *   <li>1 - if this ursi code is greater than <code>station</code> ursi code</li>
	 * </ul>
	 */
	@Override abstract public int compareTo(AbstractStation station);
	/**
	 * For specified Type of Local Time, <code>localTimeType</code>, calculates time shift in minutes, which is the amount we have to add to
	 * the <i>Universal</i> Time to get <i>Local Time</i> for station longitude.<br>
	 * This method has sense only for not mobile stations.
	 */
	abstract public double getLocalTimeShiftInMin(TimeType localTimeType);
	/**
	 * Returns unique name of this station.<br>
	 * Usually the realization of this method should return URSI code for not mobile stations and station name for mobile stations. But it
	 * might be implemented differently.
	 */
	abstract public String getUniqueName();
	/**
	 * Returns unique code of this station.<br>
	 * It should be not negative number that uniquely identifies station. It also possible that this method returns -1, it may occurred only
	 * in the case when station class is still not constructed completely.<br>
	 * Note:
	 * <ul>
	 *   All negative numbers other than -1 are prohibited.
	 * </ul>
	 */
	abstract public int getUniqueCode();
	/**
	 * Returns number of antennas being installed for this station at given <code>time</code>
	 */
	abstract public int getNumberOfAntennas(TimeScale time);
	/**
	 * Sets unique code for this station.<br>
	 * It should be positive number that uniquely identifies station.<br>
	 * Note:
	 * <ul>
	 *   This method is usually used for building station class from databases.
	 * </ul>
	 */
	abstract public void setUniqueCode(int value);
	/**
	 * Sets station by unique code.<br>
	 * Contents of station object will be changed with correspondence to given unique code.<br>
	 * Before calling, object station can be created with minimal constructor.
	 */
	abstract public void setStation(int uniqueCode) throws IOException;
}