package edu.uml.lgdc.math;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;

import edu.uml.lgdc.datatype.DoubleArray;
import edu.uml.lgdc.datatype.OrderedMetricable;
import edu.uml.lgdc.datatype.TimeOrderedMetric;
import edu.uml.lgdc.fileio.FileRW;
import edu.uml.lgdc.project.Console;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
public final class Search {
	/**
	 * Tries to find object <code>whatToFind</code> in array of objects <code>objects</code>.<br>
	 * Returns index in array if matched otherwise returns -1.<br>
	 * Two objects, <code>a</code> and <code>b</code>, are considering here "the same", if method <code>a.equals( b )</code> returns <b>true</b>.
	 */
	static public int scan(Object objects[], Object whatToFind) {
		return scan(objects, whatToFind, 0, objects.length - 1);
	}
	/**
	 * Tries to find object <code>whatToFind</code> in array of objects, <code>objects</code>.<br>
	 * Returns index in array if matched otherwise returns -1.<br>
	 * Two objects, <code>a</code> and <code>b</code>, are considering here "the same", if method <code>a.equals( b )</code> returns <b>true</b>.
	 */
	static public int scan(Object objects[], Object whatToFind, int left, int right) {
		for (int i = left; i <= right; i++) {
			if (whatToFind.equals(objects[i]))
				return i;
		}
		return -1;
	}
	/**
	 * Tries to find object <code>whatToFind</code> in the list of objects, <code>objects</code>.<br>
	 * Returns the index in the list if matched, otherwise returns -1.<br>
	 * Two objects, <code>a</code> and <code>b</code>, are considering here "the same", if method <code>a.equals( b )</code> returns <b>true</b>.
	 */
	static public int scan(List<?> objects, Object whatToFind) {
		return scan(objects, whatToFind, 0, objects.size() - 1);
	}
	/**
	 * Tries to find object <code>whatToFind</code> in the list of objects, <code>objects</code>.<br>
	 * Returns the index in the list if matched, otherwise returns -1.<br>
	 * Two objects, <code>a</code> and <code>b</code>, are considering here "the same", if method <code>a.equals( b )</code> returns <b>true</b>.
	 */
	static public int scan(List<?> objects, Object whatToFind, int left, int right) {
		int index = objects.subList(left, right + 1).indexOf(whatToFind);
		if (index >= 0)
			return left + index;
		return -1;
	}
	/**
	 * Tries to find object <code>whatToFind</code> in array of <code>objects</code>.<br>
	 * Returns index in array if matched, otherwise returns -1.<br>
	 * Two object are considering here "the same" if they are really the same, i.e. they reference the same object.
	 */
	static public int strictScan(Object objects[], Object whatToFind) {
		return strictScan(objects, whatToFind, 0, objects.length - 1);
	}
	/**
	 * Tries to find object <code>whatToFind</code> in array of <code>objects</code> starting from index <code>left</code> and up to index
	 * <code>right</code>.<br>
	 * Returns index in array if matched, otherwise returns -1.<br>
	 * Two object are considering here "the same" if they are really the same, i.e. they reference the same object.
	 */
	static public int strictScan(Object objects[], Object whatToFind, int left, int right) {
		for (int i = left; i <= right; i++) {
			if (whatToFind == objects[i])
				return i;
		}
		return -1;
	}
	/**
	 * Tries to find object <code>whatToFind</code> in {@link List} of <code>objects</code>.<br>
	 * Returns index in {@link List} if matched, otherwise returns -1.<br>
	 * Two object are considering here "the same" if they are really the same, i.e. they reference the same object.
	 */
	static public int strictScan(List<?> objects, Object whatToFind) {
		for (int i = 0; i < objects.size(); i++) {
			if (whatToFind == objects.get(i))
				return i;
		}
		return -1;
	}
	/**
	 * Tries to find object <code>whatToFind</code> in {@link List} of <code>objects</code> starting from index <code>left</code> and up to index
	 * <code>right</code>.<br>
	 * Returns index in {@link List} if matched, otherwise returns -1.<br>
	 * Two object are considering here "the same" if they are really the same, i.e. they reference the same object.
	 */
	static public int strictScan(List<?> objects, Object whatToFind, int left, int right) {
		int index = strictScan(objects.subList(left, right + 1), whatToFind);
		if (index >= 0)
			return left + index;
		return -1;
	}
	/**
	 * Tries to find integer <code>whatToFind</code> in array of integer <code>values</code>.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public int scan(int values[], int whatToFind) {
		return scan(values, whatToFind, 0, values.length - 1);
	}
	/**
	 * Tries to find integer <code>whatToFind</code> in array of integer <code>values</code>.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public int scan(int values[], int whatToFind, int left, int right) {
		for (int i = left; i <= right; i++) {
			if (whatToFind == values[i])
				return i;
		}
		return -1;
	}
	/**
	 * Tries to find char <code>whatToFind</code> in array of char <code>values</code>.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public int scan(char values[], char whatToFind) {
		return scan(values, whatToFind, 0, values.length - 1);
	}
	/**
	 * Tries to find char <code>whatToFind</code> in array of char <code>values</code>.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public int scan(char values[], char whatToFind, int left, int right) {
		for (int i = left; i <= right; i++) {
			if (whatToFind == values[i])
				return i;
		}
		return -1;
	}
	/**
	 * Tries to find long <code>whatToFind</code> in array of long <code>values</code>.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public int scan(long values[], long whatToFind) {
		return scan(values, whatToFind, 0, values.length - 1);
	}
	/**
	 * Tries to find long <code>whatToFind</code> in array of long <code>values</code>.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public int scan(long values[], long whatToFind, int left, int right) {
		for (int i = left; i <= right; i++) {
			if (whatToFind == values[i])
				return i;
		}
		return -1;
	}
	/**
	 * Tries to find double <code>whatToFind</code> in array of double <code>values</code>.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public int scan(double values[], double whatToFind) {
		return scan(values, whatToFind, 0, values.length - 1);
	}
	/**
	 * Tries to find double <code>whatToFind</code> in array of double <code>values</code>.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public int scan(double values[], double whatToFind, int left, int right) {
		for (int i = left; i <= right; i++) {
			if (whatToFind == values[i])
				return i;
		}
		return -1;
	}
	/**
	 * Tries to find String <code>str</code> in array of strings <code>strings</code>.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public int scanStr(String strings[], String str) {
		return scan(strings, str);
	}
	/**
	 * Tries to find String <code>str</code> in array of strings <code>strings</code>. Letter case will be ignored in comparison algorithm.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public int scanStrIgnoreCase(String strings[], String str) {
		return scanStr(strings, str, true);
	}
	/**
	 * Tries to find String <code>str</code> in array of strings <code>strings</code>. If parameter <code>ignorCase</code> is <b>true</b> then
	 * letter case will be ignored in comparison algorithm.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public int scanStr(String strings[], String str, boolean ignoreCase) {
		if (ignoreCase) {
			for (int i = 0; i < strings.length; i++) {
				if (str.equalsIgnoreCase(strings[i]))
					return i;
			}
		} else
			return scan(strings, str);
		return -1;
	}
	/**
	 * Tries to find object <code>whatToFind</code> in array of ordered objects, <code>objects</code>.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public <T extends Comparable<? super T>> int binarySearch(T[] objects, T whatToFind) {
		return binarySearch(objects, whatToFind, 0, objects.length - 1);
	}
	/**
	 * Tries to find object <code>whatToFind</code> in array of ordered objects, <code>objects</code> between <code>left</code> and
	 * <code>right</code> elements.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public <T extends Comparable<? super T>> int binarySearch(T[] objects, T whatToFind, int left, int right) {
		if (left > right)
			return -1;
		if (whatToFind.compareTo(objects[left]) < 0 || objects[right].compareTo(whatToFind) < 0)
			return -1;
		if (whatToFind.compareTo(objects[right]) == 0)
			return right;
		while (right - left > 1) {
			int middle = (left + right) / 2;
			if (whatToFind.compareTo(objects[middle]) < 0)
				right = middle;
			else
				left = middle;
		}
		if (whatToFind.compareTo(objects[left]) == 0)
			return left;
		return -1;
	}
	/**
	 * Tries to find object <code>whatToFind</code> in the List of String objects, <code>objects</code>.<br>
	 * Returns index in the List if matched, otherwise returns -1.
	 */
	static public int binarySearch(List<String> objects, String whatToFind) {
		return binarySearch(objects, whatToFind, 0, objects.size() - 1);
	}
	/**
	 * Tries to find object <code>whatToFind</code> in the List of String objects, <code>objects</code>, between <code>left</code> and
	 * <code>right</code> indexes inclusively<br>
	 * Returns index in the List if matched, otherwise returns -1.
	 */
	static public int binarySearch(List<String> objects, String whatToFind, int left, int right) {
		if (left > right)
			return -1;
		if (whatToFind.compareTo(objects.get(left)) < 0 || objects.get(right).compareTo(whatToFind) < 0)
			return -1;
		if (whatToFind.equals(objects.get(right)))
			return right;
		while (right - left > 1) {
			int middle = (left + right) / 2;
			if (whatToFind.compareTo(objects.get(middle)) < 0)
				right = middle;
			else
				left = middle;
		}
		if (whatToFind.equals(objects.get(left)))
			return left;
		return -1;
	}
	/**
	 * Tries to find object <code>whatToFind</code> in List of comparable objects, <code>objects</code>.<br>
	 * Returns index in array if matched otherwise returns -1.
	 */
	static public <T extends Comparable<? super T>> int binarySearch(List<T> objects, T whatToFind) {
		return binarySearch(objects, whatToFind, 0, objects.size() - 1);
	}
	/**
	 * Tries to find object <code>whatToFind</code> in List of comparable objects, <code>objects</code> between <code>left</code> and
	 * <code>right</code> elements.<br>
	 */
	static public <T extends Comparable<? super T>> int binarySearch(List<T> objects, T whatToFind, int left, int right) {
		if (left > right)
			return -1;
		if (whatToFind.compareTo(objects.get(left)) < 0 || objects.get(right).compareTo(whatToFind) < 0)
			return -1;
		if (whatToFind.compareTo(objects.get(right)) == 0)
			return right;
		while (right - left > 1) {
			int middle = (left + right) / 2;
			if (whatToFind.compareTo(objects.get(middle)) < 0)
				right = middle;
			else
				left = middle;
		}
		if (whatToFind.compareTo(objects.get(left)) == 0)
			return left;
		return -1;
	}
	/**
	 * Tries to find left nearest to integer number <code>whatToFind</code> in monotonically increasing array of integers <code>arr</code>.<br>
	 * Returns index of left nearest element in array <code>arr</code>.<br>
	 * Returns -1 if <code>whatToFind</code> is less than 1st element of array <code>arr</code>.
	 */
	static public int leftNearest(int[] arr, int whatToFind) {
		return leftNearest(arr, whatToFind, 0, arr.length - 1);
	}
	/**
	 * Tries to find left nearest to integer number <code>whatToFind</code> in monotonically increasing array of integers <code>arr</code>.<br>
	 * Returns index of left nearest element in array <code>arr</code> between <code>first</code> and <code>last</code> indexes inclusively.<br>
	 * Returns -1 if <code>whatToFind</code> is less than <code>first</code> element of array <code>arr</code>.
	 */
	static public int leftNearest(int[] arr, int whatToFind, int first, int last) {
		if (arr == null || arr.length == 0)
			return -1;
		if (first > last)
			return -1;
		int index = Arrays.binarySearch(arr, first, last + 1, whatToFind);
		if (index >= 0)
			return index;
		index = -(index + 1);
		if (index > first)
			return index - 1;
		return -1;
	}
	/**
	 * Tries to find left nearest to double number <code>whatToFind</code> in monotonically increasing array of doubles <code>arr</code>.<br>
	 * Returns index of left nearest element in array <code>arr</code>.<br>
	 * Returns -1 if <code>whatToFind</code> is less than 1st element of array <code>arr</code>.
	 */
	static public int leftNearest(double[] arr, double whatToFind) {
		return leftNearest(arr, whatToFind, 0, arr.length - 1);
	}
	/**
	 * Tries to find left nearest to double number <code>whatToFind</code> in monotonically increasing array of doubles <code>arr</code>.<br>
	 * Returns index of left nearest element in array <code>arr</code> between <code>first</code> and <code>last</code> indexes inclusively.<br>
	 * Returns -1 if <code>whatToFind</code> is less than <code>first</code> element of array <code>arr</code>.
	 */
	static public int leftNearest(double[] arr, double whatToFind, int first, int last) {
		if (arr == null || arr.length == 0)
			return -1;
		if (first > last)
			return -1;
		int index = Arrays.binarySearch(arr, first, last + 1, whatToFind);
		if (index >= 0)
			return index;
		index = -(index + 1);
		if (index > first)
			return index - 1;
		return -1;
	}
	/**
	 * Tries to find left nearest to double number <code>whatToFind</code> in monotonically increasing DoubleArray <code>arr</code>.<br>
	 * Returns index of left nearest element in array <code>arr</code>.<br>
	 * Returns -1 if <code>whatToFind</code> is less than 1st element of array <code>arr</code>.
	 */
	static public int leftNearest(DoubleArray arr, double whatToFind) {
		return leftNearest(arr, whatToFind, 0, arr.size() - 1);
	}
	/**
	 * Tries to find left nearest to double number <code>whatToFind</code> in monotonically increasing DoubleArray <code>arr</code>.<br>
	 * Returns index of left nearest element in array <code>arr</code> between <code>first</code> and <code>last</code> indexes inclusively.<br>
	 * Returns -1 if <code>whatToFind</code> is less than <code>first</code> element of array <code>arr</code>.
	 */
	static public int leftNearest(DoubleArray arr, double whatToFind, int first, int last) {
		if (arr == null || arr.size() == 0)
			return -1;
		if (first > last)
			return -1;
		if (whatToFind < arr.element(first))
			return -1;
		int right = last;
		if (arr.element(right) < whatToFind || arr.element(right) == whatToFind)
			return right;
		int left = first;
		while (right - left > 1) {
			int middle = (left + right) / 2;
			if (whatToFind < arr.element(middle))
				right = middle;
			else
				left = middle;
		}
		return left;
	}
	/**
	 * Tries to find left nearest to object <code>whatToFind</code> in monotonically increasing array of objects <code>array</code>.<br>
	 * Returns index of left nearest element in array <code>array</code>.<br>
	 * Returns -1 if <code>whatToFind</code> is less than 1st element of array <code>array</code>. Note,
	 * <ul>
	 * Natural ordering is considered here, i.e. objects of array has to implements interface Comparable&lt;E&gt;, otherwise behavior of this
	 * method is unpredictable.
	 * </ul>
	 */
	static public int leftNearest(Object[] array, Object whatToFind) {
		return leftNearest(array, whatToFind, 0, array.length - 1);
	}
	/**
	 * Tries to find left nearest to object <code>whatToFind</code> in monotonically increasing array of objects <code>array</code>.<br>
	 * Returns index of left nearest element in array <code>array</code> between <code>first</code> and <code>last</code> indexes inclusively.<br>
	 * Returns -1 if <code>whatToFind</code> is less than <code>first</code> element of array <code>array</code>. Note,
	 * <ul>
	 * Natural ordering is considered here, i.e. objects of array has to implements interface Comparable&lt;E&gt;, otherwise behavior of this
	 * method is unpredictable.
	 * </ul>
	 */
	static public int leftNearest(Object[] array, Object whatToFind, int first, int last) {
		if (array == null || array.length == 0)
			return -1;
		if (first > last)
			return -1;
		int index = Arrays.binarySearch(array, first, last + 1, whatToFind);
		if (index >= 0)
			return index;
		index = -(index + 1);
		if (index > first)
			return index - 1;
		return -1;
	}
	/**
	 * Tries to find the left nearest element to given string <code>whatToFind</code> in monotonically increasing comparable elements of the
	 * <i>List</i> <code>list</code>.<br>
	 * Returns index of the left nearest element in the List <code>list</code>.<br>
	 * Returns -1 if <code>whatToFind</code> is less than 1st element of the <i>List</i> <code>list</code>.
	 */
	static public <T extends Comparable<? super T>> int leftNearest(List<T> list, T whatToFind) {
		return leftNearest(list, whatToFind, 0, list.size() - 1);
	}
	/**
	 * Tries to find the left nearest element to given string <code>whatToFind</code> in monotonically increasing comparable elements of the
	 * <i>List</i> <code>list</code>.<br>
	 * Returns index of the left nearest element in the <i>List</i> <code>list</code> which occurred between <code>first</code> and
	 * <code>last</code> indexes inclusively.<br>
	 * Returns -1 if <code>whatToFind</code> is less than <code>first</code> element of the <i>List</i> <code>list</code>.
	 */
	static public <T extends Comparable<? super T>> int leftNearest(List<T> list, T whatToFind, int first, int last) {
		if (list == null || list.size() == 0)
			return -1;
		if (first > last)
			return -1;
		if (whatToFind.compareTo(list.get(first)) < 0)
			return -1;
		int right = last;
		if (list.get(right).compareTo(whatToFind) <= 0)
			return right;
		int left = first;
		while (right - left > 1) {
			int middle = (left + right) / 2;
			if (whatToFind.compareTo(list.get(middle)) < 0)
				right = middle;
			else
				left = middle;
		}
		return left;
	}
	/**
	 * Tries to find right nearest to integer number <code>whatToFind</code> in monotonically increasing array of integers <code>array</code>.<br>
	 * Returns index of right nearest element in array <code>array</code>.<br>
	 * Returns -1 if <code>whatToFind</code> is greater than the last element of array <code>array</code>
	 */
	static public int rightNearest(int[] array, int whatToFind) {
		return rightNearest(array, whatToFind, 0, array.length - 1);
	}
	/**
	 * Tries to find right nearest to integer number <code>whatToFind</code> in monotonically increasing array of integers <code>array</code>.<br>
	 * Returns index of right nearest element in array <code>array</code> between <code>first</code> and <code>last</code> indexes inclusively.<br>
	 * Returns -1 if <code>whatToFind</code> is greater than the <code>last</code> element of array <code>array</code>.
	 */
	static public int rightNearest(int[] array, int whatToFind, int first, int last) {
		if (array == null || array.length == 0)
			return -1;
		if (first > last)
			return -1;
		int index = Arrays.binarySearch(array, first, last + 1, whatToFind);
		if (index >= 0)
			return index;
		index = -(index + 1);
		if (index < last + 1)
			return index;
		return -1;
	}
	/**
	 * Tries to find right nearest to double number <code>whatToFind</code> in monotonically increasing array of double values <code>arr</code>.<br>
	 * Returns index of right nearest element in array <code>arr</code>.<br>
	 * Returns -1 if <code>whatToFind</code> is greater than the last element of array <code>arr</code>
	 */
	static public int rightNearest(double[] arr, double whatToFind) {
		return rightNearest(arr, whatToFind, 0, arr.length - 1);
	}
	/**
	 * Tries to find right nearest to double number <code>whatToFind</code> in monotonically increasing array of double values <code>arr</code>.<br>
	 * Returns index of right nearest element in array <code>arr</code> between <code>first</code> and <code>last</code> indexes inclusively.<br>
	 * Returns -1 if <code>whatToFind</code> is greater than the <code>last</code> element of array <code>arr</code>.
	 */
	static public int rightNearest(double[] arr, double whatToFind, int first, int last) {
		if (arr == null || arr.length == 0)
			return -1;
		if (first > last)
			return -1;
		int index = Arrays.binarySearch(arr, first, last + 1, whatToFind);
		if (index >= 0)
			return index;
		index = -(index + 1);
		if (index < last + 1)
			return index;
		return -1;
	}
	/**
	 * Tries to find the left nearest element to given string <code>whatToFind</code> in monotonically increasing comparable elements of the
	 * <i>List</i> <code>list</code>.<br>
	 * Returns index of the left nearest element in the List <code>list</code>.<br>
	 * Returns -1 if <code>whatToFind</code> is less than 1st element of the <i>List</i> <code>list</code>.
	 */
	static public <T extends Comparable<? super T>> int rightNearest(List<T> list, T whatToFind) {
		return rightNearest(list, whatToFind, 0, list.size() - 1);
	}
	/**
	 * Tries to find the right nearest element to given string <code>whatToFind</code> in monotonically increasing comparable elements of the
	 * <i>List</i> <code>list</code>.<br>
	 * Returns index of the right nearest element in the <i>List</i> <code>list</code> which occurred between <code>first</code> and
	 * <code>last</code> indexes inclusively.<br>
	 * Returns -1 if <code>whatToFind</code> greater than the <code>last</code> element of the <i>List</i> <code>list</code>.
	 */
	static public <T extends Comparable<? super T>> int rightNearest(List<T> list, T whatToFind, int first, int last) {
		if (list == null || list.size() == 0)
			return -1;
		if (first > last)
			return -1;
		int right = last;
		if (list.get(right).compareTo(whatToFind) < 0)
			return -1;
		if (whatToFind.compareTo(list.get(first)) <= 0)
			return first;
		int left = first;
		while (right - left > 1) {
			int middle = (left + right) / 2;
			if (list.get(middle).compareTo(whatToFind) < 0)
				left = middle;
			else
				right = middle;
		}
		return right;
	}
	/**
	 * Tries to find right nearest to object <code>whatToFind</code> in monotonically increasing array of objects <code>array</code>.<br>
	 * Returns index of right nearest element in array <code>array</code><br>
	 * Returns -1 if object <code>whatToFind<code> is greater than the last element of array <code>array</code>. Note,
	 * <ul>
	 * Natural ordering is considered here, i.e. objects of array has to implements interface Comparable&lt;E&gt;, otherwise behavior of this
	 * method is unpredictable.
	 * </ul>
	 */
	static public int rightNearest(Object[] array, Object whatToFind) {
		return leftNearest(array, whatToFind, 0, array.length - 1);
	}
	/**
	 * Tries to find right nearest to object <code>whatToFind</code> in monotonically increasing array of objects <code>array</code>.<br>
	 * Returns index of right nearest element in array <code>array</code> between <code>first</code> and <code>last</code> indexes inclusively.<br>
	 * Returns -1 if String <code>whatToFind</code> is less greater than element of array with index <code>last</code>. Note,
	 * <ul>
	 * Natural ordering is considered here, i.e. objects of array has to implements interface Comparable&lt;E&gt;, otherwise behavior of this
	 * method is unpredictable.
	 * </ul>
	 */
	static public int rightNearest(Object[] array, Object whatToFind, int first, int last) {
		if (array == null || array.length == 0)
			return -1;
		if (first > last)
			return -1;
		int index = Arrays.binarySearch(array, first, last + 1, whatToFind);
		if (index >= 0)
			return index;
		index = -(index + 1);
		if (index < last + 1)
			return index;
		return -1;
	}
	/**
	 * Tries to find right nearest to String <code>whatToFind</code> in monotonically increasing strings of List <code>list</code>.<br>
	 * Returns index of right nearest element in List <code>list</code><br>
	 * Returns -1 if String <code>whatToFind<code> is greater than the last element of List <code>list</code>.
	 */
	static public int rightNearest(List<String> list, String whatToFind) {
		return rightNearest(list, whatToFind, 0, list.size() - 1);
	}
	/**
	 * Tries to find right nearest to String <code>whatToFind</code> in monotonically increasing array of strings of List <code>list</code>.<br>
	 * Returns index of right nearest element in List <code>list</code> is between <code>first</code> and <code>last</code> indexes inclusively.<br>
	 * Returns -1 if String <code>whatToFind</code> is greater than element of List with index <code>last</code>.
	 */
	static public int rightNearest(List<String> list, String whatToFind, int first, int last) {
		if (list == null || list.size() == 0)
			return -1;
		if (first > last)
			return -1;
		int right = last;
		if (list.get(right).compareTo(whatToFind) < 0)
			return -1;
		if (whatToFind.compareTo(list.get(first)) <= 0)
			return first;
		int left = first;
		while (right - left > 1) {
			int middle = (left + right) / 2;
			if (whatToFind.compareTo(list.get(middle)) < 0)
				right = middle;
			else
				left = middle;
		}
		return right;
	}
	/**
	 * Tries to find closest to double value <what> in array of ordered double values <code>arr</code>. Returns index of the nearest element in
	 * array <code>arr</code>. Returns -1 if array <code>arr</code> is empty.
	 */
	static public int closest(double[] arr, double what) {
		return closest(arr, what, 0, arr.length - 1);
	}
	/**
	 * Tries to find closest to double value <what> in array of ordered double values <code>arr</code>. Returns index of the nearest element in
	 * array <code>arr</code> between <code>first</code> and <code>last</code> indexes inclusively. Returns -1 if <code>first</code> is greater
	 * than <code>last</code>.
	 */
	static public int closest(double[] arr, double what, int first, int last) {
		int index = leftNearest(arr, what, first, last);
		if (index >= 0) {
			if (index < last) {
				if (arr[index + 1] - what < what - arr[index])
					index++;
			}
		} else if (first <= last)
			index = first;

		return index;
	}
	/**
	 * Tries to find closest to double value <what> in DoubleArray of ordered values <code>arr</code>. Returns index of the nearest element in
	 * DoubleArray <code>arr</code>. Returns -1 if array <code>arr</code> is empty.
	 */
	static public int closest(DoubleArray arr, double what) {
		return closest(arr, what, 0, arr.size() - 1);
	}
	/**
	 * Tries to find closest to double value <what> in DoubleArray of ordered values <code>arr</code>.<br>
	 * Returns index of the nearest element in array <code>arr</code> between <code>first</code> and <code>last</code> indexes inclusively.<br>
	 * Returns -1 if <code>first</code> is greater than <code>last</code>.
	 */
	static public int closest(DoubleArray arr, double what, int first, int last) {
		int index = leftNearest(arr, what, first, last);
		if (index >= 0) {
			if (index < last) {
				if (arr.element(index + 1) - what < what - arr.element(index))
					index++;
			}
		} else if (first <= last)
			index = first;

		return index;
	}
	/**
	 * Tries to find closest to object <what> in array of Ordered Metric objects <code>objects</code>.<br>
	 * Returns index of the nearest element in Vector <code>objects</code>. Returns -1 if vector <code>objects</code> is empty.
	 */
	static public <T extends OrderedMetricable<T>> int closest(T[] objects, T what) {
		return closest(objects, what, 0, objects.length - 1);
	}
	/**
	 * Tries to find closest to object <what> in array of Ordered Metric objects <code>objects</code>.<br>
	 * Returns index of the nearest element in Vector <code>objects</code> between <code>first</code> and <code>last</code> indexes inclusively.<br>
	 * Returns -1 if <code>first</code> is greater than <code>last</code>.
	 */
	static public <T extends OrderedMetricable<T>> int closest(T[] objects, T what, int first, int last) {
		int index = leftNearest(objects, what, first, last);
		if (index >= 0) {
			if (index < last) {
				if (objects[index + 1].dist(what) < objects[index].dist(what))
					index++;
			}
		} else if (first <= last)
			index = first;

		return index;
	}
	/**
	 * Tries to find closest to object <what> in List of Ordered Metric objects <code>objects</code>.<br>
	 * Returns index of the nearest element in List <code>objects</code>.<br>
	 * Returns -1 if vector <code>objects</code> is empty.
	 */
	static public <T extends OrderedMetricable<T>> int closest(List<T> objects, T what) {
		return closest(objects, what, 0, objects.size() - 1);
	}
	/**
	 * Tries to find closest to object <what> in List of Ordered Metric objects <code>objects</code>.<br>
	 * Returns index of the nearest element in List <code>objects</code> between <code>first</code> and <code>last</code> indexes inclusively.<br>
	 * Returns -1 if <code>first</code> is greater than <code>last</code>.
	 */
	static public <T extends OrderedMetricable<T>> int closest(List<T> objects, T what, int first, int last) {
		int index = leftNearest(objects, what, first, last);
		if (index >= 0) {
			if (index < last) {
				if (objects.get(index + 1).dist(what) < objects.get(index).dist(what))
					index++;
			}
		} else if (first <= last)
			index = first;
		return index;
	}
	static public <T extends TimeOrderedMetric> int closest(List<T> objects, T what) {
		return closest(objects, what, 0, objects.size() - 1);
	}
	static public <T extends TimeOrderedMetric> int closest(List<T> objects, T what, int first, int last) {
		int index = leftNearest(objects, what, first, last);
		if (index >= 0) {
			if (index < last) {
				if (objects.get(index + 1).dist(what) < objects.get(index).dist(what))
					index++;
			}
		} else if (first <= last)
			index = first;
		return index;
	}
	static public long searchBack(byte[] what, String filename, byte[] buffer) {
		return searchBack(what, filename, buffer, -1);
	}
	static public long searchBack(byte[] what, String filename, byte[] buffer, long maxOffsetFromEnd) {
		if (buffer == null)
			throw new IllegalArgumentException("buffer is null");
		if (buffer.length < what.length)
			throw new IllegalArgumentException("buffer.length < what.length");
		long offset = -1;
		FileRW file = null;
		try {
			file = new FileRW(filename, "r");
			offset = searchBack(what, file, buffer, -1);
		} catch (IOException ex) {
			Console.printThreadStackTrace(ex);
		} finally {
			if (file != null)
				try {
					file.close();
				} catch (IOException ex) {
					Console.printThreadStackTrace(ex);
				}
		}
		return offset;
	}
	/**
	 * Starting from <code>current_position - 1</code> in given instance of <code>FileRW</code>, <code>file</code>, searches given bytes pattern,
	 * <code>what</code>, in backward direction.<br>
	 * Returns offset or -1, if not found.
	 */
	static public long searchBack(byte[] what, FileRW file, byte[] buffer, long maxOffsetFromEnd) throws IOException {
		long offset = -1;
		long pos = file.getFilePointer();
		long minOffset = 0;
		if (maxOffsetFromEnd >= 0)
			minOffset = Math.max(pos - 1 - maxOffsetFromEnd, 0);
		long startPos = Math.max(pos - buffer.length, minOffset);
		int numberOfBytesToRead = (int)(pos - startPos);
		if (numberOfBytesToRead <= what.length)
			return -1;
		int off;
		int tailLength = 0;
		long nextPos;
		while (true) {
			file.seek(startPos);
			file.read(buffer, 0, numberOfBytesToRead);
			off = searchBack(what, buffer, numberOfBytesToRead + tailLength);
			if (off >= 0) {
				offset = file.getFilePointer() - numberOfBytesToRead + off;
				break;
			}
			if (startPos == minOffset)
				break;
			tailLength = what.length;
			nextPos = Math.max(startPos - buffer.length + tailLength, minOffset);
			numberOfBytesToRead = (int)(startPos - nextPos);
			if (numberOfBytesToRead + tailLength < buffer.length) {
				for (int i = 0; i < what.length - 1; i++)
					buffer[numberOfBytesToRead + i] = buffer[buffer.length - what.length + i];
			}
			startPos = nextPos;
		}
		return offset;
	}
	static public int searchBack(byte[] what, byte[] searchBuffer, int length) {
		int first = what.length - 1;
		for (int i = length - 1; i >= first; i--) {
			boolean found = true;
			int k = i - what.length + 1;
			for (int j = 0; j < what.length; j++) {
				if (searchBuffer[k + j] != what[j]) {
					found = false;
					break;
				}
			}
			if (found)
				return k;
		}
		return -1;
	}
	static public int search(byte[] what, byte[] searchBuffer) {
		return search(what, searchBuffer, 0, searchBuffer.length);
	}
	static public int search(byte[] what, byte[] searchBuffer, int start, int length) {
		if (start < 0 || length < 0 || start + length > searchBuffer.length)
			throw new IllegalArgumentException("some of the arguments, start or length, or their combination is illegal");
		for (int i = start; i <= start + length - what.length; i++) {
			boolean found = true;
			for (int j = 0; j < what.length; j++) {
				if (searchBuffer[i + j] != what[j]) {
					found = false;
					break;
				}
			}
			if (found)
				return i;
		}
		return -1;
	}
}