package edu.uml.lgdc.multithread;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
public class Semaphore {
	protected int counter;

	/**
	 * Constructor with given initial <code>counter</code>
	 */
	public Semaphore(int counter) {
		this.counter = counter;
	}

	/**
	 * This method simply increases <code>counter</code> (which you can consider as a counter of some resource) by one and awakes, if any, 
	 * exactly one thread of those waiting on this object. The awaken thread is put into concurrent queue of threads trying to get control 
	 * over monitor of this object and it is no guarantee that awaken thread will get this control first (and in this case, it is no guarantee 
	 * that counter is not zero).
	 */
	synchronized public void post() {
		counter++;
		notify(); // wake up one of the threads waited on this object
	}

	/**
	 * Pends without surrendering until counter is not zero and so it is possible to decrease it by one (and you can consider it as a taking 
	 * one unit of resource).<br>
	 * Returns <b>true</b> (never <b>false</b>)<br>
	 * 
	 * Note:
	 * <ul>
	 *   Infinite loop on wait is necessary as it is not guaranteed that awaken thread gets control over monitor first among the possible 
	 *   concurrent threads, and in this case counter can be zero upon gaining control by this thread. We should anticipate this unpleasant 
	 *   possibility.
	 * </ul>
	 */
	synchronized public boolean pend() throws InterruptedException {
		waitCountNotZero();
		// should be always TRUE, and, actually, caller anticipates TRUE by not checking the result
		return decreaseCounter();
	}

	/**
	 * If parameter <code>milliSec</code> is zero then this method is equivalent to the same method without parameters.<br>
	 * Otherwise it pends either until counter is not zero and so it is possible to decrease it by one (and you can consider it as a taking 
	 * one unit of resource) or until wait time is expired (<b>false</b> will be returned in this case)
	 */
	synchronized public boolean pend(int milliSec) throws InterruptedException {
		if (milliSec < 0)
			throw new IllegalArgumentException("milliSec should be >= 0");
		if (milliSec == 0)
			return pend();
		long start = System.currentTimeMillis();
		long end = start + milliSec;
		while (counter <= 0 && end - start > 0) {
			wait(end - start); // put itself in wait status on this object
			start = System.currentTimeMillis();
		}
		return decreaseCounter(); // might be FALSE
	}

	/**
	 * If parameters <code>milliSec</code> and <code>nanoSec</code> are zeroes then this method is equivalent to the same method without parameters.<br>
	 * Otherwise it pends either until counter is not zero and so it is possible to decrease it by one (and you can consider it as a taking 
	 * one unit of resource) or until wait time is expired (<b>false</b> will be returned in this case)
	 */
	synchronized public boolean pend(int milliSec, int nanoSec)	throws InterruptedException {
		if (milliSec < 0)
			throw new IllegalArgumentException("milliSec should be >= 0");
		if (nanoSec < 0)
			throw new IllegalArgumentException("nanoSec should be >= 0");
		if (milliSec == 0) {
			if (nanoSec == 0)
				return pend();
			else
				return pend(1);
		}
		if (nanoSec < 500)
			return pend(milliSec);
		else
			return pend(milliSec + 1);
	}

	/**
	 * Pends without surrendering until counter is not zero and the returns.<br>
	 * 
	 * Note:
	 * <ul>
	 *   It is no guarantee that other thread will nullify counter just a moment after this method finished.
	 * </ul>
	 */
	synchronized public void waitCountNotZero() throws InterruptedException {
		while (counter <= 0)
			wait();
	}

	/**
	 * This method has to be used with care as it potentially can leave some threads waiting on this semaphore unawaken forever
	 */
	synchronized public void clear() {
		counter = 0;
	}

	/**
	 * This method has to be used with care as it potentially can leave some threads waiting on this semaphore unawaken forever
	 */
	synchronized public void set(int value) {
		counter = value;
		if (value > 0) {
			if (value == 1)
				notify(); // wake up one of the threads waited on this object
			else
				notifyAll(); // wake up all the threads waited on this object
		}
	}

	private boolean decreaseCounter() {
		if (counter > 0) {
			counter--;
			return true;
		} else
			return false;
	}

	/**
	 * Returns semaphore counter as a tentative value. This method works fast but value of counter can be used just as an approximated/tentative one.
	 */
	public int ask() {
		return counter;
	}
}