package edu.uml.lgdc.project;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
public class CommandLineParam {
	public enum KeyType {
		KEY_ONLY,
		VALUE_ONLY,
		KEY_OR_VALUE;
	}
	static private final String[] SEP_1 = {"--", "="};
	static private final String[] SEP_2 = {"-", ":"};

	private final String keyPrefix;
	private final String key_value_separator;
	private final Set<String> paramKeys = new HashSet<>();
	private final Map<String,String> paramValues = new HashMap<>();
	private final List<String> notKeyParams = new ArrayList<>();
	private final Set<String> duplicatedKeyOnly = new HashSet<>();
	private final Set<String> duplicatedKeyValue = new HashSet<>();

	public CommandLineParam(String[] params) {
		this(getSeps(params)[0], getSeps(params)[1], params);
	}

	static private String[] getSeps(String[] params) {
		for (String param : params) {
			if (param.startsWith("--")) {
				return SEP_1;
			} else if (param.startsWith("-")) {
				return SEP_2;
			}
		}
		return SEP_1;
	}
	
	public CommandLineParam(String keyPrefix, String key_value_separator, String[] params) {
		assert keyPrefix != null && keyPrefix.length() > 0;
		assert key_value_separator != null && key_value_separator.length() > 0;
		this.keyPrefix = keyPrefix;
		this.key_value_separator = key_value_separator;
		for (String param : params) {
			if (isStartAsKeyParam(param)) {
				String[] keyAndValue = getKeyAndValue(param);
				if (keyAndValue[1] == null) {
					if (paramKeys.add(keyAndValue[0])) {
						duplicatedKeyOnly.add(keyAndValue[0]);
					}
				} else {
					if (!paramValues.containsKey(keyAndValue[0])) {
						paramValues.put(keyAndValue[0], keyAndValue[1]);
					} else {
						duplicatedKeyValue.add(keyAndValue[0]);
					}
				}
			} else {
				notKeyParams.add(param);
			}
		}
	}
	public String[] getNotKeyParams() {
		return notKeyParams.toArray(new String[notKeyParams.size()]);
	}
	/**
	 * Returns <i>value</i> for given <code>key</code>, or <b>null</b> if <code>key</code> is not found<br>
	 * If we suggest that {@link #keyPrefix} is -- (two dash symbols) and {@link #key_value_separator} is = (equal sign) then 
	 * <code>key</code> / <i>value</i> pair is presented in the form,
	 * <ul>
	 *   <b>--</b><code>key</code><b>=</b><i>value</i>
	 * </ul>
	 * Important,
	 * <ul>
	 *   <strong>There should be NO space or any whitespace characters in <code>key=value</code> representation.</strong><br>
	 *   If you really need it, you have put it in double quote, like "count=tinker tailor soldier sailor"
	 * </ul>
	 */
	public String getKeyValue(String key) {
		assert key != null && key.length() > 0;
		return paramValues.get(key);
	}
	/**
	 * Returns <i>value</i> for given set of key synonyms. Semantically, set of key synonyms represents possible names of the same 
	 * semantical <i>key</i>.<br>
	 * Returns <b>null</b> if <i>key</i> is not found. 
	 */
	public String getKeyValue(Set<String> keySynonyms) {
		assert keySynonyms != null && keySynonyms.size() > 0;
		for (String synonym : keySynonyms) {
			if (paramValues.containsKey(synonym)) {
				return paramValues.get(synonym);
			}
		}
		return null;
	}
	/**
	 * Returns <b>true</b> if given <code>key</code> with or without <i>value</i> exists in parameter call string, otherwise returns <b>false</b>.<br>
	 * If we suggest that {@link #keyPrefix} is -- (two dashes) and {@link #key_value_separator} is = (equal sign) then 
	 * <code>key</code> with or without value can be presented in one of the following forms,
	 * <ul>
	 *   <li><b>--</b><code>key</code>=<i>value</i></li>
	 *   <li><b>--</b><code>key</code></li>
	 * </ul>
	 * Parameter call string given as array of strings delivered by java main program parameter.
	 */
	public boolean isKeyExists(String key) {
		assert key != null && key.length() > 0;
		return paramKeys.contains(key) || paramValues.get(key) != null;
	}
	/**
	 * Returns <b>true</b> if given set of key synonyms presented in parameter call string in any form, <i>with value</i> or 
	 * <i>without value</i>, otherwise returns <b>false</b>. 
	 */
	public boolean isKeyExists(Set<String> keySynonyms) {
		assert keySynonyms != null && keySynonyms.size() > 0;
		if (isKeyOnlyExists(keySynonyms)) {
			return true;
		}
		return getKeyValue(keySynonyms) != null;
	}
	/**
	 * Returns <b>true</b> if given <code>key</code>presented in parameter call string <i>without value</i>, otherwise returns <b>false</b>.<br>
	 * If we suggest that {@link #keyPrefix} is -- (two dashes) and {@link #key_value_separator} is = (equal sign) then 
	 * <code>key</code> without <i>value</i> is presented in the form,
	 * <ul>
	 *   <b>--</b><code>key</code><b>>
	 * </ul>
	 * Parameter call string given as array of strings delivered by java main program parameter.
	 */
	public boolean isKeyOnlyExists(String key) {
		assert key != null && key.length() > 0;
		return paramKeys.contains(key);
	}
	/**
	 * Returns <b>true</b> if given set of key synonyms presented in parameter call string <i>without value</i>, otherwise returns <b>false</b>. 
	 */
	public boolean isKeyOnlyExists(Set<String> keySynonyms) {
		assert keySynonyms != null && keySynonyms.size() > 0;
		for (String synonym : keySynonyms) {
			if (paramKeys.contains(synonym)) {
				return true;
			}
		}
		return false;
	}
	/**
	 * Checks that parameter call string does not contain any item except given key-only parameters, <code>keys</code><br>
	 * Returns <b>null</b> if check is OK, otherwise returns first encountered illegal keyword.
	 */
	public String checkKeyOnlyParams(String[] keys) {
		KeyType[] keyTypes = new KeyType[keys.length];
		Arrays.fill(keyTypes, KeyType.KEY_ONLY);
		return checkKeyParams(keys, keyTypes);
	}

	/**
	 * Checks that parameter call string does not contain any item except given key parameters, <code>keys</code><br>. 
	 * Array <code>keyTypes</code> contains type codes of these keys as follows (if we suggest that {@link #keyPrefix} 
	 * is -- and {@link #key_value_separator} is =), 
	 * <ul>
	 *   <li>KEY_ONLY - key can be used only without value, i.e. in the form --<i>key</i></li>
	 *   <li>VALUE_ONLY - key can be used only with value, i.e. in the form --<i>key</i>=<i>value</i></li>
	 *   <li>KEY_OR_VALUE - key can be used either without or with value</li>
	 * </ul>
	 * Returns <b>null</b> if check is OK, otherwise returns error message
	 */
	public String checkKeyParams(String[] keys, KeyType[] keyTypes) {
		Map<String,KeyType> keyTypeMap = new HashMap<>();
		for (int i = 0; i < keys.length; i++) {
			keyTypeMap.put(keys[i], keyTypes[i]);
		}
		return checkParams(keyTypeMap, false);
	}

	public String checkKeyParams(Map<String,KeyType> keyTypeMap) {
		return checkParams(keyTypeMap, false);
	}

	/**
	 * Checks that parameter call string does not contain any item except given key parameters, <code>keys</code>, and, probably, 
	 * several not-key parameters (if <code>notKeyParamsAllowed</code> is <b>true</b>)<br>
	 * Array <code>keyTypes</code> contains type codes of these keys as follows (if we suggest that {@link #keyPrefix} 
	 * is -- and {@link #key_value_separator} is =), 
	 * <ul>
	 *   <li>KEY_ONLY - key can be used only without value, i.e. in the form --<i>key</i></li>
	 *   <li>VALUE_ONLY - key can be used only with value, i.e. in the form --<i>key</i>=<i>value</i></li>
	 *   <li>KEY_OR_VALUE - key can be used either without or with value</li>
	 * </ul>
	 * Returns <b>null</b> if check is OK, otherwise returns error message
	 */
	public String checkParams(String[] keys, KeyType[] keyTypes, boolean notKeyParamsAllowed) {
		Map<String,KeyType> keyTypeMap = new HashMap<>();
		for (int i = 0; i < keys.length; i++) {
			keyTypeMap.put(keys[i], keyTypes[i]);
		}
		return checkParams(keyTypeMap, notKeyParamsAllowed);
	}

	public String checkParams(Map<String,KeyType> keyTypeMap, boolean notKeyParamsAllowed) {
		if (!notKeyParamsAllowed && notKeyParams.size() > 0) {
			return "illegal parameters: " + Arrays.toString(notKeyParams.toArray(new String[notKeyParams.size()]));
		}
		for (String key : paramKeys) {
			KeyType keyType = keyTypeMap.get(key);
			if (keyType != null) {
				if (keyType == KeyType.VALUE_ONLY) {
					return "key " + key + " has to be accompanied with value";
				}
			} else {
				return "illegal key " + key;
			}
		}
		for (String key : paramValues.keySet()) {
			KeyType keyType = keyTypeMap.get(key);
			if (keyType != null) {
				if (keyType == KeyType.KEY_ONLY) {
					return "key " + key + " doesn't have to be accompanied with value";
				}
			} else {
				return "illegal key " + key;
			}
		}
		return null;
	}

	public String checkParamsWithSynonyms(Map<Set<String>,KeyType> keyTypeMap, boolean notKeyParamsAllowed) {
		if (!notKeyParamsAllowed && notKeyParams.size() > 0) {
			return "illegal parameters: " + Arrays.toString(notKeyParams.toArray(new String[notKeyParams.size()]));
		}
		Map<String,Set<String>> keyToSynonyms = new HashMap<>();
		Map<Set<String>,List<String>> presentedAsKeyOnly = new HashMap<>();
		Map<Set<String>,List<String>> presentedAsKeyValue = new HashMap<>();
		for (Set<String> keySynonyms : keyTypeMap.keySet()) {
			for (String key : keySynonyms) {
				keyToSynonyms.put(key, keySynonyms);
			}
			presentedAsKeyOnly.put(keySynonyms, new ArrayList<>());
			presentedAsKeyValue.put(keySynonyms, new ArrayList<>());
		}
		for (String key : paramKeys) {
			Set<String> keySynonyms = keyToSynonyms.get(key);
			if (keySynonyms != null) {
				KeyType keyType = keyTypeMap.get(keySynonyms);
				if (keyType == KeyType.VALUE_ONLY) {
					return "key " + key + " has to be accompanied with value";
				} else {
					presentedAsKeyOnly.get(keySynonyms).add(key);
				}
			} else {
				return "illegal key " + key;
			}
		}
		for (String key : paramValues.keySet()) {
			Set<String> keySynonyms = keyToSynonyms.get(key);
			if (keySynonyms != null) {
				KeyType keyType = keyTypeMap.get(keySynonyms);
				if (keyType == KeyType.KEY_ONLY) {
					return "key " + key + " doesn't have to be accompanied with value";
				} else {
					presentedAsKeyValue.get(keySynonyms).add(key);
				}
			} else {
				return "illegal key " + key;
			}
		}
		String errMsg = "";
		String dup = checkDuplicationOfSynonyms(presentedAsKeyOnly);
		if (!dup.isEmpty()) {
			errMsg += "Duplication of key-only synonyms of the same 'keyword'\n" + dup;
		}
		dup = checkDuplicationOfSynonyms(presentedAsKeyValue);
		if (!dup.isEmpty()) {
			errMsg += "Duplication of key-value synonyms of the same 'keyword'\n" + dup;
		}
		return errMsg.isEmpty() ? null : errMsg;
	}

	/**
	 * Extracts string array of arguments, but only for given synonyms, <code>synonym2keyTypeMap</code>.
	 */
	public String[] extractArgs(Map<Set<String>,KeyType> synonym2keyTypeMap) {
		List<String> argList = new ArrayList<String>();
		for (Set<String> keySynonyms : synonym2keyTypeMap.keySet()) {
			KeyType keyType = synonym2keyTypeMap.get(keySynonyms);
			if (keyType != KeyType.VALUE_ONLY) {
				boolean found = false;
				for (String key : keySynonyms) {
					if (paramKeys.contains(key)) {
						argList.add("--" + key);
						found = true;
						break;
					}
				}
				if (found) {
					continue;
				}
			}
			if (keyType != KeyType.KEY_ONLY) {
				for (String key : keySynonyms) {
					String value = paramValues.get(key);
					if (value != null) {
						argList.add("--" + key + "=" + value);
						break;
					}
				}
			}
		}
		return argList.toArray(new String[]{});
	}

	private String checkDuplicationOfSynonyms(Map<Set<String>,List<String>> presented) {
		String dupMsg = "";
		for (List<String> list : presented.values()) {
			if (list.size() > 1) {
				dupMsg += Arrays.toString(list.toArray(new String[list.size()])) + "\n";
			}
		}
		return dupMsg;
	}

	private String[] getKeyAndValue(String par) {
		String parWithoutPrefix = par.substring(keyPrefix.length());
		int sepPos = parWithoutPrefix.indexOf(key_value_separator);
		if (sepPos < 0) {
			return new String[] {parWithoutPrefix, null};
		} else {
			return new String[] {parWithoutPrefix.substring(0, sepPos), parWithoutPrefix.substring(sepPos + key_value_separator.length())};
		}
	}

	private boolean isStartAsKeyParam(String par) {
		return (par.startsWith(keyPrefix));
	}
}