package edu.uml.lgdc.project;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.Calendar;
import java.util.TimeZone;

import edu.uml.lgdc.fileio.FileRW;
import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.format.StrUtil;
import edu.uml.lgdc.time.TimeScale;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
abstract public class Console {
	static public final String TIME_FORMAT = "yyyy.MM.dd HH:mm:ss.SSS";
	static public final int TIME_STR_LEN = TIME_FORMAT.length();
	static public final String MSGERR_PREFIX_START = "*** ";
	static public final String MSGERR_PREFIX_END = "ERROR: ";
	static public final String MSGERR_PREFIX = MSGERR_PREFIX_START + MSGERR_PREFIX_END;
	static public final String MSGWARN_PREFIX = "+++ WARNING: ";

	static private final String EMPTY_STRING = "";

	static private final String PREV_LOG_SUFFIX = "_prev";
	static private final int MB = 1024 * 1024;

	/**
	 * this memory space supposed to be reserved by specific application-aware subclass of this class.<br>
	 * It is supposed to be released and used by JVM when OutOfMemoryError occurred.<br>
	 * To release this space for JVM, you need to call static method releaseReservedSpace()
	 */
	static protected byte[] RESERVED_SPACE;
	static protected Object printSync = new Object();
	protected ControlPar controlPar = null;
	private String line = EMPTY_STRING;

	public Console() {
		//
	}

	/**
	 * Releases reserved space for JVM.<br>
	 * This method is supposed to be called no more than once per application life, when OutOfMemoryError occurred and you need to finish 
	 * important stuff before terminating the application.
	 */
	static public void releaseReservedSpace() {
		if (RESERVED_SPACE != null) {
			RESERVED_SPACE = null;
			System.runFinalization();
			System.gc();
		}
	}

	/**
	 * Terminate application in case of unpredictable error occurred.
	 */
	final public void terminate(final Throwable error, final TimeScale time) {
		final Thread threadInError = Thread.currentThread();
		Thread thread = new Thread(new Runnable() {
			@Override public void run() {
				releaseReservedSpace();
				crashErrorReport(threadInError, error, time);
				System.exit(1);
			}
		}, "CrashReport");
		thread.setPriority(Thread.MAX_PRIORITY);
		thread.start();
		try {
			thread.join();
		} catch (InterruptedException e) {}
	}

	/**
	 * Prepare Crash Error Report.<br>
	 * Specific application-aware subclass will redefine this method.
	 */
	protected void crashErrorReport(Thread crashedThread, Throwable error, TimeScale time) {
		// do nothing
	}

	public String getCrashReportFilename(String threadName, TimeScale time) {
		return "fe_" + CommonConst.getApplicationName() + "_" + FC.padLeft("" + (time.get(Calendar.MONTH) + 1), 2, '0')
		+ FC.padLeft("" + time.get(Calendar.DAY_OF_MONTH), 2, '0') + FC.padLeft("" + time.get(Calendar.HOUR_OF_DAY), 2, '0') + "_"
		+ threadName;
	}

	public void print(String str) {
		if (str.length() == 0) {
			return;
		}
		String[] result;
		do {
			result = StrUtil.cutLine(str);
			str = result[1];
			if (result[0] != null) {
				addToLine(result[0]);
				logIt_impl(line, true, false);
				line = EMPTY_STRING;
			} else if (result[1].length() > 0) {
				addToLine(result[1]);
			}
		} while (result[0] != null && str.length() > 0);
	}
	
	private void addToLine(String str) {
		if (line == EMPTY_STRING) {
			line = str;
		} else {
			line += str;
		}
	}

	/**
	 * Logs the message as is (i.e. without embedded current datetime at the start)
	 */
	public void logAsIs(String message) {logIt(message, false);}

	public void printStackTrace(Throwable e) {
		e.printStackTrace();
		if (controlPar.logOutput != null) {
			e.printStackTrace(controlPar.logOutput);
		}
	}

	/**
	 * Logs the message with embedded current datetime at the start.
	 */
	public void logIt(String message) {logIt(message, true, true);}

	/**
	 * Logs the message
	 */
	public void logIt(String message, boolean embedTime) {logIt(message, embedTime, true);}

	/**
	 * Logs the message
	 */
	public void logIt(String message, boolean embedTime, boolean writeToStandardOutput) {
		logIt_impl(message, embedTime, writeToStandardOutput);
	}

	/**
	 * Logs the message
	 */
	private void logIt_impl(String message, boolean embedTime, boolean writeToStandardOutput) {
		if (embedTime) {
			message = TimeScale.prefixWithLocalTime(message, "  ");
		}
		if (writeToStandardOutput) {
			System.out.println(message);
		}
		if (controlPar.logOutput != null) {
			controlPar.logOutput.println(message);
		}
		if (!controlPar.isConsoleMode()) {
			printToScreen(message);
		}
	}

	public void printToScreen(String message) {
		if (controlPar.taLogOutput != null) {
			synchronized (controlPar.taLogOutputLock) {
				controlPar.taLogOutput.append(message + "\n");
			}
		}
	}

	static public void printThreadStackTrace(Throwable e) {
		StackTraceElement[] ste = e.getStackTrace();
		String[] lines = new String[ste.length + 1];
		lines[0] = e.toString();
		for (int i = 0; i < ste.length; i++) {
			lines[i + 1] = ste[i].toString();
		}
		showError(lines);
		if (e.getCause() != null) {
			showError("Caused by:");
			printThreadStackTrace(e.getCause());
		}
	}

	static public void showMsg(String s) {
		showMsg(s, false);
	}
	static public void showMsg(String[] lines) {
		showMsg(lines, false);
	}
	static public void showError(String s, String where) {
		showMsg(MSGERR_PREFIX_START + where + " " + MSGERR_PREFIX_END + s, true);
	}
	static public void showError(String s) {
		showMsg(MSGERR_PREFIX + s, true);
	}
	static public void showError(String[] lines) {
		if (lines.length > 0) {
			lines[0] = MSGERR_PREFIX + lines[0];
		}
		showMsg(lines, true);
	}
	static public void showWarn(String s) {
		showMsg(MSGWARN_PREFIX + s);
	}
	static public void showWarn(String[] lines) {
		if (lines.length > 0) {
			lines[0] = MSGWARN_PREFIX + lines[0];
		}
		showMsg(lines, false);
	}
	static public void showMsg(String s, boolean err) {
		showMsg(StrUtil.textToLines(s), err);
	}
	static public void showMsg(String[] lines, boolean err) {
		synchronized (printSync) {
			PrintStream ps = err ? System.err : System.out;
			String timeStr = new TimeScale(TimeZone.getDefault()).toFormatUT(TIME_FORMAT);
			String tmp;
			if (lines.length > 0) {
				tmp = timeStr + ": " + lines[0];
				ps.println(tmp);
				for (int i = 1; i < lines.length; i++) {
					tmp = FC.STR_BLANKS.substring(0, 2) + lines[i];
					ps.println(tmp);
				}
			} else {
				tmp = timeStr + ": ";
				ps.println(tmp);
			}
		}
	}
	static public void showMsgAsIs(String s, boolean err) {
		synchronized (printSync) {
			(err ? System.err : System.out).print(s);
		}
	}
	static public void printThreadStackTrace(Throwable e, FileRW fileRW) throws IOException {
		StackTraceElement[] ste = e.getStackTrace();
		String[] lines = new String[ste.length + 1];
		lines[0] = e.toString();
		for (int i = 0; i < ste.length; i++) {
			lines[i + 1] = ste[i].toString();
		}
		showError(lines, fileRW);
		if (e.getCause() != null) {
			showError("Caused by:", fileRW);
			printThreadStackTrace(e.getCause(), fileRW);
		}
	}
	static public String getErrMsgPrefix() {
		return new TimeScale(TimeZone.getDefault()).toFormatUT(TIME_FORMAT) + ": " + MSGERR_PREFIX;
	}
	static public void showError(String[] lines, FileRW fileRW)	throws IOException {
		if (lines.length > 0) {
			lines[0] = MSGERR_PREFIX + lines[0];
		}
		showMsg(lines, fileRW);
	}
	static public void showMsg(String s, FileRW fileRW) throws IOException {
		showMsg(StrUtil.textToLines(s), fileRW);
	}
	static public void showError(String s, FileRW fileRW) throws IOException {
		showError(StrUtil.textToLines(s), fileRW);
	}
	static public void showMsg(String[] lines, FileRW fileRW) throws IOException {
		writeMsg(new TimeScale(TimeZone.getDefault()), lines, fileRW);
	}
	static public void writeMsg(TimeScale time, String s, FileRW fileRW) throws IOException {
		writeMsg(time, StrUtil.textToLines(s), fileRW);
	}
	static public void writeMsg(TimeScale time, String[] lines, FileRW fileRW) throws IOException {
		synchronized (printSync) {
			String timeStr = time.toFormatUT(TIME_FORMAT);
			if (lines.length > 0) {
				fileRW.write(timeStr + ": " + lines[0] + "\n");
				for (int i = 1; i < lines.length; i++) {
					fileRW.write(FC.STR_BLANKS.substring(0, 2) + lines[i] + "\n");
				}
			} else {
				fileRW.write(timeStr + ":\n");
			}
		}
	}
	static public void writeMsgNoEx(TimeScale time, String s, FileRW fileRW) {
		writeMsgNoEx(time, StrUtil.textToLines(s), fileRW);
	}
	static public void writeMsgNoEx(TimeScale time, String[] lines,	FileRW fileRW) {
		synchronized (printSync) {
			String timeStr = (time != null) ? time.toFormatUT(TIME_FORMAT) : FC.padRight("UNKNOWN_TIME", TIME_FORMAT.length());
			try {
				if (lines.length > 0) {
					fileRW.write(timeStr + ": " + lines[0] + "\n");
					for (int i = 1; i < lines.length; i++) {
						fileRW.write(FC.STR_BLANKS.substring(0, 2) + lines[i] + "\n");
					}
				} else {
					fileRW.write(timeStr + ":\n");
				}
			} catch (IOException e) {
				printThreadStackTrace(e);
			}
		}
	}
	static public void writeMsgNoEx(String s, FileRW fileRW) {
		synchronized (printSync) {
			try {
				fileRW.write(s + "\n");
			} catch (IOException e) {
				printThreadStackTrace(e);
			}
		}
	}
	static public String getStampedMsgLine(String line) {
		StringBuilder strBuf = new StringBuilder(200);
		strBuf.append(new TimeScale(TimeZone.getDefault()).toFormatUT(TIME_FORMAT)).append(": ").append(line).append(line).append("\n");
		return strBuf.toString();
	}
	static public String getStampedErrLine(String line) {
		StringBuilder strBuf = new StringBuilder(200);
		strBuf.append(new TimeScale(TimeZone.getDefault()).toFormatUT(TIME_FORMAT)).append(": ").append(MSGERR_PREFIX).append(line).append(line).append("\n");
		return strBuf.toString();
	}

	/**
	 * @see #prepareLogFile(String logFilename, long maxSize_MB )
	 */
	static public void prepareLogFile(String logFilename) {
		prepareLogFile(logFilename, 1);
	}

	/**
	 * This simple method is intended for utilities log file and supposed to be called from method <code>main</code> before redirecting 
	 * output/error to log file.<br>
	 * The idea is to continue to append output to existing log-file, <code>logFilename</code>, if it is not bigger than
	 * <code>maxSize_MB</code> megabytes, otherwise rename existing log-file to <i>previous log-file name</i>.
	 */
	static public void prepareLogFile(String logFilename, long maxSize_MB) {
		String logExt = FileRW.getExt(logFilename);
		String logFileBasename = FileRW.getNameButExt(logFilename);
		if (logExt.length() > 0) {
			logExt = "." + logExt;
		}
		File logFile = new File(logFilename);
		if (logFile.length() > maxSize_MB * MB) {
			File prevLogFile = new File(logFileBasename + PREV_LOG_SUFFIX + logExt);
			prevLogFile.delete();
			logFile.renameTo(prevLogFile);
		}
	}
	/**
	 * Redirects <i>output</i> and <i>error</i> streams to given file, <code>filename</code>.<br>
	 * If file <code>filename</code> does not exist, then it will be created.<br>
	 * If file <code>filename</code> exists, then new content will be appended if parameter <code>append</code> is <b>true</b>, otherwise 
	 * this file will be truncated to zero, so deleting previous content.<br>
	 * If parameter <code>autoflush</code> is <b>true</b>, then auto flush technique will be applied.
	 */
	static public void redirectToFile(String filename, boolean append, boolean autoflush) {
		OutputStream os = null;
		try {
			os = new FileOutputStream(filename, append);
			PrintStream ps = new PrintStream(os, autoflush);
			System.setOut(ps);
			System.setErr(ps);
		} catch (IOException e) {
			showError(e.toString());
		}
	}
}