/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.project.settings;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import edu.uml.lgdc.project.Console;
import edu.uml.lgdc.project.settings.FromStrHelper.Result;

/**
 * Abstract class that describes option's section.
 * 
 * @author Alex Kozlov
 */
public abstract class OptSecDescriptor <E extends OptSecDescriptor<?>> implements Settings {
	protected final int INT_DONT_USE = Integer.MIN_VALUE;

	protected boolean parsingOrIllegalValueError;
	/**
	 * <code>rangeOfOccurrences[0]</code> is minimum number of occurrences, <code>&ge; 0</code>.<br>
	 * <code>rangeOfOccurrences[1]</code> is maximum number of occurrences, <code>&ge; 0</code>.<br>
	 * 0 is a special value for <code>rangeOfOccurrences[1]</code> that means,
	 * <ul>
	 *   the only restriction on number of occurrences is it has to be <code>&ge; rangeOfOccurrences[0]</code>
	 * </ul>
	 * if <code>rangeOfOccurrences[1]</code> is not zero then it has to be <code>&ge; rangeOfOccurrences[0]</code> 
	 * and restriction on number of occurrences is
	 * <ul>
	 *    <code>&ge; rangeOfOccurrences[0]</code> and <code>&le; rangeOfOccurrences[1]</code> 
	 * </ul>
	 */
	private int[] rangeOfOccurrences = {0, 1};
	
	/**
	 * Returns name of this section
	 */
	abstract public String getName();

	/**
	 * Returns options description as a map of Option Name (instance of {@link String}) to Option Descriptor (instance of {@link OptionDesc}). 
	 */
	abstract public Map<String,OptionDesc> getOptionsDescriptor();

	public int[] getRangeOfOccurrences() {
		return rangeOfOccurrences;
	}
	
	public void setRangeOfOccurrences(int[] rangeOfOccurrences) {
		Objects.requireNonNull(rangeOfOccurrences);
		assert rangeOfOccurrences.length == 2 : "range has to be array of 2 elements, but is " + rangeOfOccurrences.length;
		assert rangeOfOccurrences[0] >= 0;
		assert rangeOfOccurrences[1] >= 0;
		assert rangeOfOccurrences[1] == 0 || rangeOfOccurrences[1] >= rangeOfOccurrences[0];
		this.rangeOfOccurrences = rangeOfOccurrences;
	}
	
	public boolean isNumberOfOccurrencesIsInRange(int numberOfOccurrences) {
		if (rangeOfOccurrences[1] > 0) {
			return numberOfOccurrences >= rangeOfOccurrences[0] && numberOfOccurrences <= rangeOfOccurrences[1]; 
		} else {
			return numberOfOccurrences >= rangeOfOccurrences[0];
		}
	}

	public String ranegAsString() {
		if (rangeOfOccurrences[1] > 0) {
			return ">= " + rangeOfOccurrences[0] + " and <= " + rangeOfOccurrences[1];
		} else {
			return ">= " + rangeOfOccurrences[0];
		}
	}
	
	/**
	 * Returns default value of section as a list of instances.<br>
	 * This implementation returns list containing from one instance (= this instance) of section if range of occurrences is [0,1],
	 * otherwise returns <b>null</b>. 
	 */
	public List<E> getDefaultValue() {
		if (rangeOfOccurrences[0] == 0 && rangeOfOccurrences[1] == 1) {
			List<E> list = new ArrayList<E>();
			list.add(getThis());
			return  list;
		} else {
			return null;
		}
	}
	
	/**
	 * Returns this instance
	 */
	abstract public E getThis();
	
	/**
	 * Returns <b>true</b> if there were parsing or illegal values for some options of these section, otherwise returns <b>false</b>
	 */
	public boolean isParsingOrIllegalValueError() {
		return parsingOrIllegalValueError;
	}
	
	/**
	 * Returns error message if it was registered but not issued/logged.<br>
	 * So, if <b>null</b> is returned that still not means error was not not occurred, because it might be occurred was issued/logged.<br>
	 * So always use method {@link #isParsingOrIllegalValueError()} to check whether error occurred or not.
	 */
	public String getErrMsg() {
		return null;
	}
	
	/**
	 * Returns,
	 * <ul>
	 *   <li><b>null</b> if <code>unknowmOptionName</code> is not the option name that was renamed</li>
	 *   <li>String array of two elements, new option name and new value of this option</li>
	 * </ul>
	 * DEfault implementation returns <b>null</b>.
	 */
	public String[] optionRenameHook(String unknowmOptionName, String valueOfUnknownOptionName) {
		return null;
	}

	/**
	 * Returns map Option Name --&gt; Option Value, restricted to the options which are tagged as <i>Processing</i>.<br>
	 * <i>Processing</i> tag is entirely related with application semantic.<b> 
	 * Sometimes you may need to override this method to filter out some inactive processing options.<br>
	 */
	public Map<String,String> getProcessingOptions(OptionsIO optionsIO) {
		putTo(optionsIO);
		Map<String,OptionDesc> descriptor = getOptionsDescriptor();
		Map<String,String> nameValPairs = new HashMap<>();
		Set<String> optionNames = descriptor.keySet();
		for (String optionName: optionNames) {
			OptionDesc optionDesc = descriptor.get(optionName);
			if (optionDesc.isProcessingOption()) {
				nameValPairs.put(optionName, optionsIO.get(getName(), optionName));
			}
		}
		return nameValPairs;
	}
	protected int getStrictPositiveInt(OptionsIO optionsIO, String optionName) {
		Result result = FromStrHelper.getStrictPositiveInt(optionsIO.get(getName(), optionName));
		return checkIntResult(result, optionName, -1);
	}
	protected int getStrictPositiveInt(OptionsIO optionsIO, String optionName, int index) {
		Result result = FromStrHelper.getStrictPositiveInt(optionsIO.get(getName(), optionName, index));
		return checkIntResult(result, optionName, index);
	}
	protected int getPositiveInt(OptionsIO optionsIO, String optionName) {
		Result result = FromStrHelper.getPositiveInt(optionsIO.get(getName(), optionName));
		return checkIntResult(result, optionName, -1);
	}
	protected int getPositiveInt(OptionsIO optionsIO, String optionName, int index) {
		Result result = FromStrHelper.getPositiveInt(optionsIO.get(getName(), optionName, index));
		return checkIntResult(result, optionName, index);
	}
	private int checkIntResult(Result result, String optionName, int index) {
		if (result.errMsg == null) {
			if (!result.isDefaultValue) {
				return (Integer)result.value;
			} else {
				return INT_DONT_USE;
			}
		} else {
			parsingOrIllegalValueError = true;
			Console.showError("Section [" + getName() + "]" + (index >= 0 ? "(" + index + ")" : "") + " option " + optionName + ": " + result.errMsg);
			return INT_DONT_USE;
		}
	}
	/**
	 * It is simple utility-method used in sub-classes
	 */
	protected String removeEnclosure(String str) {
		if (str != null && str.length() >= 2 && str.charAt(0) == '[' && str.charAt(str.length() - 1) == ']') {
			return str.substring(1, str.length() - 1);
		} else {
			return str;
		}
	}
}
