/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.project.settings;

import java.util.Objects;

/**
 * Contains simple description of anonymous option, which currently consists of two properties,
 * <ul>
 *   <li>Default value, {@link #defaultValue}</li>
 *   <li>Processing related flag, {@link #processingOption}</li>
 * </ul> 
 * 
 * @author Alex Kozlov
 */
public class OptionDesc {
	/**
	 * Default value. Will be used if bound option is not specified by user.
	 */
	private final Object defaultValue;
	/**
	 * <b>true</b> if bound option is processing related option, otherwise is <b>false</b> 
	 */
	private final boolean processingOption;
	/**
	 * <code>rangeOfOccurrences[0]</code> is minimum number of occurrences, <code>&ge; 0</code>.<br>
	 * <code>rangeOfOccurrences[1]</code> is maximum number of occurrences, <code>&ge; 0</code>.<br>
	 * 0 is a special value for <code>rangeOfOccurrences[1]</code> that means,
	 * <ul>
	 *   the only restriction on number of occurrences is it has to be <code>&ge; rangeOfOccurrences[0]</code>
	 * </ul>
	 * if <code>rangeOfOccurrences[1]</code> is not zero then it has to be <code>&ge; rangeOfOccurrences[0]</code> 
	 * and restriction on number of occurrences is
	 * <ul>
	 *    <code>&ge; rangeOfOccurrences[0]</code> and <code>&le; rangeOfOccurrences[1]</code> 
	 * </ul>
	 */
	private int[] rangeOfOccurrences = {0, 1};

	public OptionDesc(Object defaultValue, boolean processingOption) {
		this(defaultValue, new int[] {0, 1}, processingOption);
	}
	public OptionDesc(Object defaultValue, int[] rangeOfOccurrences) {
		this(defaultValue, rangeOfOccurrences, false);
	}
	public OptionDesc(Object defaultValue, int[] rangeOfOccurrences, boolean processingOption) {
		this.defaultValue = defaultValue;
		setRangeOfOccurrences(rangeOfOccurrences);
		this.processingOption = processingOption;
	}
	public Object getDefaultValue() {
		return defaultValue;
	}
	public boolean isProcessingOption() {
		return processingOption;
	}
	public int[] getRangeOfOccurrences() {
		return rangeOfOccurrences;
	}
	public void setRangeOfOccurrences(int[] rangeOfOccurrences) {
		Objects.requireNonNull(rangeOfOccurrences);
		assert rangeOfOccurrences.length == 2 : "range has to be array of 2 elements, but is " + rangeOfOccurrences.length;
		assert rangeOfOccurrences[0] >= 0;
		assert rangeOfOccurrences[1] >= 0;
		assert rangeOfOccurrences[1] == 0 || rangeOfOccurrences[1] >= rangeOfOccurrences[0];
		this.rangeOfOccurrences = rangeOfOccurrences;
	}
	public boolean isMultiOption() {
		return rangeOfOccurrences[1] != 1; 
	}
	public boolean isNumberOfOccurrencesIsInRange(int numberOfOccurrences) {
		if (rangeOfOccurrences[1] > 0) {
			return numberOfOccurrences >= rangeOfOccurrences[0] && numberOfOccurrences <= rangeOfOccurrences[1]; 
		} else {
			return numberOfOccurrences >= rangeOfOccurrences[0];
		}
	}
	public String ranegAsString() {
		if (rangeOfOccurrences[1] > 0) {
			return ">= " + rangeOfOccurrences[0] + " and <= " + rangeOfOccurrences[1];
		} else {
			return ">= " + rangeOfOccurrences[0];
		}
	}
}
