/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.project.settings;

import java.io.OutputStream;
import java.util.Map;

/**
 * Interface describes not-typed content of options grouped into sections, and I/O operations with them.<br>
 * Not-typed content means all option' values are interpreted as string.  
 * 
 * @author Alexander Kozlov
 */
public interface OptionsIO {

	/**
	 * Returns user's section descriptor of given <code>sectionName</code>.<br>
	 * Returns <b>null</b> if given <code>sectionName</code> is not present.
	 */
	OptSecDescriptor<?> getOptSecDescriptor(String sectionName);

	/**
	 * Returns only those options which are tagged "related with Processing".<br>
	 * Returned as map of section names to map of option name to option value.  
	 */
	Map<String,Map<String,String>> getProcessingOptions();

	/**
	 * Saves content to the same file it was read from.
	 */
	void save();

	/**
	 * Writes content to given instance <code>outputStream</code> of {@link OutputStream} 
	 */
	void save(OutputStream outputStream);
	
	/**
	 * Removes given section <code>sectionName</code>. Obviously, if this section has multiple instances, all of them will be removed as well.  
	 */
	void removeSection(String sectionName);

	/**
	 * Returns <b>true</b> if there were errors in options, otherwise <b>false</b>
	 */
	boolean isError();
	
	/**
	 * Returns <b>true</b> if given {@code sectionName} is present in use settings data, or set by default, otherwise returns <b>false</b>.<br>
	 * To catch this event you have to not allow default settings for this section. To do this you have to override method
	 * {@code isAbsenceAsSpecialCase()} of class {@link OptSEcDescriptor} which by default returns <b>false</b>, and make it returns <b>true</b>.
	 */
	boolean isSectionPresent(String sectionName);
	
	/**
	 * Returns <b>true</b> if given {@code sectionName}.{@code optionName} is present in settings data, or set by default, 
	 * otherwise returns <b>false</b>.<br>
	 * To catch this event you have to not allow default settings for given {@code sectionName}.{@code optionName}. To do this you 
	 * have to assign <b>null</b> default value to given {@code sectionName}.{@code optionName}.
	 */
	boolean isOptionPresent(String sectionName, String optionName);

	//
	// methods for not-multiple section (which is usual case)
	//
	
	/**
	 * Returns value for given <code>optionName</code> of given <code>sectionName</code>.<br>
	 * Throws {@link IllegalArgumentException} if,
	 * <ul>
	 *   <li>given <code>sectionName</code> does not exist</li>
	 *   <li>given <code>optionName</code> does not exist</code>
	 * </ul>
	 */
	String get(String sectionName, String optionName);

	/**
	 * Sets given value <code>optionValue</code> for given <code>optionName</code> of given <code>sectionName</code>.<br>
	 * Throws {@link IllegalArgumentException} if,
	 * <ul>
	 *   <li>given <code>sectionName</code> is not present in user's sections descriptor {@link #parSecDescriptors}</li> 
	 *   <li>given <code>optionName</code> of <code>sectionName</code> is not present in user's sections descriptor {@link #parSecDescriptors}</li> 
	 * </ul>
	 * Note,
	 * <ul>
	 *   If instance with given <code>index</code> of given <code>sectionName</code> is not present, it will be added
	 * </ul>
	 */
	void put(String sectionName, String optionName, String optionValue);

	//
	// methods for not-multiple section (which is usual case)
	//
	
	/**
	 * This method is intended for multiple section mode.<br>
	 * It returns number of occurrences/instances of given section <code>sectionName</code>.<br>
	 * Note,
	 * <ul>
	 *   returns 0 if given section <code>sectionName</code> does not exist
	 * </ul>  
	 */
	int getNumberOfMultipleSections(String sectionName);
	
	/**
	 * This method is intended for multiple section mode.<br>
	 * Returns value for given <code>optionName</code> of given instance <code>index</code> of given <code>sectionName</code>.<br>
	 * Throws {@link IllegalArgumentException} if,
	 * <ul>
	 *   <li>given <code>sectionName</code> does not exist</li>
	 *   <li>instance of section with given <code>index</code> does not exist</li>
	 *   <li>given <code>optionName</code> does not exist</code>
	 * </ul>
	 */
	String get(String sectionName, String optionName, int index);
	
	/**
	 * This method is intended for multiple section mode.<br>
	 * Sets given value <code>optionValue</code> for given <code>optionName</code> of given instance <code>index</code> of 
	 * given <code>sectionName</code>.<br>
	 * Note,
	 * <ul>
	 *   If instance with given <code>index</code> of given <code>sectionName</code> is not present, it will be added
	 * </ul>
	 */
	void put(String sectionName, String optionName, String optionValue, int index);

}