/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.project.settings;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import org.ini4j.Config;
import org.ini4j.Ini;
import org.ini4j.Profile.Section;

import edu.uml.lgdc.project.Console;

/**
 * Implementation of interface {@link OptionsIO} based on library ini4j.<br>
 * @see <a href="http://ini4j.sourceforge.net">http://ini4j.sourceforge.net</a>
 * 
 * @author Alexander Kozlov
 */
public class OptionsIOini implements OptionsIO {
	/**
	 * User's sections descriptor. So it is possible to implement more semantic check of ini-file content, as well as more tunings. 
	 */
	private final Map<String,OptSecDescriptor<?>> optSecDescriptors;
	private Ini ini;
	private final boolean treatUnknownNamesAsError; 
	private final File file;
	/**
	 * <b>true</b> if there were errors in options, otherwise <b>false</b>
	 */
	private boolean error;

	transient private Map<String,List<Map<String,String>>> changedNameOptions = new HashMap<>();
	transient private boolean optionsAcquired;

	public OptionsIOini(Map<String,OptSecDescriptor<?>> optSecDescriptors, String fileName) {
		this(optSecDescriptors, fileName, true);
	}

	public OptionsIOini(Map<String,OptSecDescriptor<?>> optSecDescriptors, String fileName, boolean treatUnknownnamesAsError) {
		this(optSecDescriptors, fileName, treatUnknownnamesAsError, false);
	}

	public OptionsIOini(Map<String,OptSecDescriptor<?>> optSecDescriptors, String fileName, boolean treatUnknownNamesAsError, 
					boolean noIniFileAbsentMsg) {
		this.optSecDescriptors = optSecDescriptors;
		File file = new File(fileName);
		this.file = file;
		this.treatUnknownNamesAsError = treatUnknownNamesAsError;
		ini = new Ini();
		Config config = new Config();
		config.setMultiSection(true);
		config.setMultiOption(true);
		ini.setConfig(config);
		if (!file.isFile()) {
			if (!noIniFileAbsentMsg) {
				Console.showMsg("ini-file \"" + file + "\" doesn't exist. Default values will be used.");
			}
		} else {
			try {
				ini.setFile(file);
				ini.load(file);
			} catch (IOException ex) {
				throw new RuntimeException(ex);
			}
		}
		if (checkNames()) {
			get();
		} else {
			error = true;
		}
	}

	public OptionsIOini(Map<String,OptSecDescriptor<?>> optSecDescriptors, InputStream is) {
		this(optSecDescriptors, is, true);
	}

	public OptionsIOini(Map<String,OptSecDescriptor<?>> optSecDescriptors, InputStream is, boolean treatUnknownNamesAsError) {
		this.optSecDescriptors = optSecDescriptors;
		ini = new Ini();
		file = null;
		this.treatUnknownNamesAsError = treatUnknownNamesAsError;
		Config config = new Config();
		config.setMultiSection(true);
		config.setMultiOption(true);
		ini.setConfig(config);
		try {
			ini.load(is);
		} catch (IOException e) {
			e.printStackTrace();
			ini = new Ini();
		}
		if (checkNames()) {
			get();
		} else {
			error = true;
		}
	}
	
	private boolean checkNames() {
		boolean noDuplication = checkForDuplication();
		boolean allNamesAreKnown = checkForUnknownName();
		if (noDuplication && (allNamesAreKnown || !treatUnknownNamesAsError) && addDefaultValues()) {
			if (!treatUnknownNamesAsError) {
				setValuesForRenamedOptions();
			}
			if (!allNamesAreKnown && file != null) {
				save();
			}
			changedNameOptions = null;
			return true;
		} else {
			Iterator<String> iterSections = ini.keySet().iterator();
			while (iterSections.hasNext()) {
				iterSections.next();
				iterSections.remove();
			}
			return false;
		}
	}

	public File getFile() {
		return ini.getFile();
	}

	/**
	 * Reads options from raw ini-content to user type-aware content.
	 */
	private void get() {
		for (OptSecDescriptor<?> settings: optSecDescriptors.values()) {
			if (settings.getDefaultValue() != null || isSectionPresent(settings.getName())) {
				settings.getFrom(this);
			}
		}
		optionsAcquired = true;
	}

	/**
	 * Saves options from user type-aware content to raw ini-content.
	 */
	private void put() {
		for (OptSecDescriptor<?> settings: optSecDescriptors.values()) {
			settings.putTo(this);
		}
	}

	/**
	 * Returns user's section descriptor of given <code>sectionName</code>.<br>
	 * Returns <b>null</b> if given <code>sectionName</code> is not present.
	 */
	@Override public OptSecDescriptor<?> getOptSecDescriptor(String sectionName) {
		return optSecDescriptors.get(sectionName);
	}

	@Override public boolean equals(Object obj) {
		if (obj == this) {
			return true;
		} else if (obj != null && obj.getClass() == getClass()) {
			OptionsIOini optionsIO = (OptionsIOini)obj;
			return optSecDescriptors.equals(optionsIO.optSecDescriptors) && ini.entrySet().equals(optionsIO.ini.entrySet());
		} else {
			return false;
		}
	}

	public void fill(String sectionName, OptionsIOini other) {
		Section thisSection = ini.get(sectionName);
		Section otherSection = other.ini.get(sectionName);
		for (String optionName: otherSection.keySet()) {
			thisSection.put(optionName, otherSection.get(optionName));
		}
	}
	
	@Override public boolean isSectionPresent(String sectionName) {
		return ini.get(sectionName) != null;
	}
	
	@Override public boolean isOptionPresent(String sectionName, String optionName) {
		Section section = ini.get(sectionName);
		if (section != null) {
			return section.get(optionName) != null;
		} else {
			throw new RuntimeException("Design error: section " + sectionName + " is not present");
		}
	}

	/**
	 * Returns only those options which are tagged "related with Processing".<br>
	 * Returned as map of section names to map of option name to option value.  
	 */
	@Override public Map<String,Map<String,String>> getProcessingOptions() {
		Map<String,Map<String,String>> processingOptions = new HashMap<>();
		Collection<OptSecDescriptor<?>> optSecDescs = optSecDescriptors.values();
		for (OptSecDescriptor<?> optSecDescriptor: optSecDescs) {
			Map<String,String> nameValPairs = optSecDescriptor.getProcessingOptions(this);
			if (!nameValPairs.isEmpty()) {
				processingOptions.put(optSecDescriptor.getName(), nameValPairs);
			}
		}
		return processingOptions;
	}

	/**
	 * Returns <b>true</b> if no illegal duplications were found, otherwise returns <b>false</b>
	 */
	private boolean checkForDuplication() {
		boolean noDuplication = true;
		Iterator<String> iterSections = ini.keySet().iterator();
		while (iterSections.hasNext()) {
			String sectionName = iterSections.next();
			OptSecDescriptor<?> optSecDescriptor = optSecDescriptors.get(sectionName);
			if (optSecDescriptor == null) {
				// ignore unknown names here
				continue;
			}
			List<Section> sections = ini.getAll(sectionName);
			if (!optSecDescriptor.isNumberOfOccurrencesIsInRange(sections.size())) {
				Console.showError("Section [" + sectionName + "] occured " + sections.size() + " times in ini-file"
								+ ", but allowed range is " + optSecDescriptor.ranegAsString());
				Console.showError("Section [" + sectionName + "] is not multiple-section, but occurred more than once in ini-file.");
				noDuplication = false;
				continue;
			}
			Map<String,OptionDesc> options = optSecDescriptor.getOptionsDescriptor();
			for (int i = 0; i < sections.size(); i++) {
				Section section = sections.get(i);
				Iterator<String> iterOptions = section.keySet().iterator();
				while (iterOptions.hasNext()) {
					String optionName = iterOptions.next();
					OptionDesc optionDesc = options.get(optionName);
					if (optionDesc != null) {
						int numberOfOccurrences = section.length(optionName);
						if (!optionDesc.isNumberOfOccurrencesIsInRange(numberOfOccurrences)) {
							Console.showError("Option " + optionName + " of section [" + sectionName + "]" 
								+ (sections.size() == 1 ? "" : "(" + (i + 1) + ")") + " occured " + numberOfOccurrences 
								+ " times in ini-file\n  But allowed range is " + optionDesc.ranegAsString());
							noDuplication = false;
						}
					}
				}
			}
		}
		return noDuplication;
	}
	
	/**
	 * Checks for unknown section and option names. If detected, log message and delete unknown section.<br>
	 * Returns <b>true</b> if unknown names were not detected.
	 */
	private boolean checkForUnknownName() {
		boolean allNamesAreKnown = true;
		Iterator<String> iterSections = ini.keySet().iterator();
		while (iterSections.hasNext()) {
			String sectionName = iterSections.next();
			OptSecDescriptor<?> optSecDescriptor = optSecDescriptors.get(sectionName);
			if (optSecDescriptor == null) {
				if (treatUnknownNamesAsError) {
					Console.showError("Section [" + sectionName + "] found in ini-file is unknown.");
				} else {
					Console.showWarn("Section [" + sectionName + "] found in ini-file is unknown, will be deleted.");
				}
				iterSections.remove();
				allNamesAreKnown = false;
				continue;
			}
			Map<String,OptionDesc> options = optSecDescriptor.getOptionsDescriptor();
			List<Section> sections = ini.getAll(sectionName);
			if (!optSecDescriptor.isNumberOfOccurrencesIsInRange(sections.size())) {
				// ignore duplication here
				continue;
			}
			List<Map<String,String>> sectionChangedNameOptions = new ArrayList<>(sections.size());
			boolean isChangedNameOption = false;
			for (int i = 0; i < sections.size(); i++) {
				Section section = sections.get(i);
				Map<String,String> optionsName2ValueMap = new HashMap<>();
				sectionChangedNameOptions.add(optionsName2ValueMap);
				Iterator<String> iterOptions = section.keySet().iterator();
				while (iterOptions.hasNext()) {
					String optionName = iterOptions.next();
					OptionDesc optionDesc = options.get(optionName);
					if (optionDesc == null) {
						if (!treatUnknownNamesAsError) {
							String valueOfUnknownOptionName = section.get(optionName);
							String[] newOptionNameValue = optSecDescriptor.optionRenameHook(optionName, valueOfUnknownOptionName);
							if (newOptionNameValue == null) {
								Console.showWarn("Option " + optionName + " of section [" + sectionName 
												+ "] found in ini-file is unknown, will be deleted.");
							} else {
								String msg = "The name of option " + optionName + " of section [" + sectionName 
												+ "] will be changed to " + newOptionNameValue[0];
								if (!Objects.equals(valueOfUnknownOptionName, newOptionNameValue[1])) {
									msg += ", and value " + valueOfUnknownOptionName + " will be changed to " + newOptionNameValue[1];
								}
								Console.showWarn(msg);
								isChangedNameOption = true;
								optionsName2ValueMap.put(newOptionNameValue[0], newOptionNameValue[1]);
							}
						} else {
							Console.showError("Option " + optionName + " of section [" + sectionName + "] found in ini-file is unknown");
						}
						iterOptions.remove();
						allNamesAreKnown = false;
					}
				}
			}
			if (isChangedNameOption) {
				changedNameOptions.put(sectionName, sectionChangedNameOptions);
			}
		}
		return allNamesAreKnown;
	}
	private boolean addDefaultValues() {
		boolean valuesAreSpecified = true;
		for (String sectionName: optSecDescriptors.keySet()) {
			List<Section> sections = ini.getAll(sectionName);
			if (sections != null) {
				for (Section section : sections) {
					if (!addDefaultValuesForSection(section)) {
						valuesAreSpecified = false;
					}
				}
			} else if (!addDefaultForMissedSection(sectionName)) {
				valuesAreSpecified = false;
			}
		}
		return valuesAreSpecified;
	}
	
	private boolean addDefaultValuesForSection(Section section) {
		boolean valuesAreSpecified = true;
		OptSecDescriptor<?> optSecDescriptor = optSecDescriptors.get(section.getName());
		Map<String,OptionDesc> optionsMap = optSecDescriptor.getOptionsDescriptor();
		for (String optionName: optionsMap.keySet()) {
			if (section.get(optionName) == null) {
				OptionDesc optionDesc = optionsMap.get(optionName);
				Object defaultValueList = optionDesc.getDefaultValue();
				if (defaultValueList != null) {
					if (defaultValueList instanceof Object[]) {
						Object[] list = (Object[])defaultValueList;
						for (int i = 0; i < list.length; i++) {
							section.put(optionName, String.valueOf(list[i]), i);
						}
					} else {
						section.put(optionName, String.valueOf(defaultValueList));
					}
				} else {
					int[] rangeOfOccurence = optionDesc.getRangeOfOccurrences();
					if (rangeOfOccurence[0] > 0) {
						Console.showError("Section [" + section.getName() + "] option " + optionName + ": mandatory option is not specified");
						valuesAreSpecified = false;
					}
				}
			}
		}
		return valuesAreSpecified;
	}

	private boolean addDefaultForMissedSection(String sectionName) {
		boolean valuesAreSpecified = true;
		OptSecDescriptor<?> optSecDescriptor = optSecDescriptors.get(sectionName);
		if (optSecDescriptor.getDefaultValue() != null) {
			optSecDescriptor.putTo(this);
		} else {
			int[] rangeOfOccurence = optSecDescriptor.getRangeOfOccurrences();
			if (rangeOfOccurence[0] > 0) {
				Console.showError("Section [" + sectionName + "]: mandatory section is not specified");
				valuesAreSpecified = false;
			}
		}
		return valuesAreSpecified;
	}
	
	private void setValuesForRenamedOptions() {
		Set<String> keySet = changedNameOptions.keySet();
		for (String sectionName: keySet) {
			List<Map<String,String>> listOfMaps = changedNameOptions.get(sectionName);
			List<Section> sections = ini.getAll(sectionName);
			for (int i = 0; i < listOfMaps.size(); i++) {
				Map<String,String> optionValues = listOfMaps.get(i); 
				if (!optionValues.isEmpty()) {
					Section section = sections.get(i);
					Set<String> optionKeySet = optionValues.keySet();
					for (String optionName: optionKeySet) {
						section.put(optionName, optionValues.get(optionName));
					}
				}
			}
		}
	}

	/**
	 * Saves content to the same file it was read from.
	 */
	@Override public void save() {
		file.delete();
		OutputStream outputStream = null;
		try {
			outputStream = new FileOutputStream(file);
			save(outputStream);
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (outputStream != null) {
				try {
					outputStream.close();
				} catch (IOException ex) {
					ex.printStackTrace();
				}
			}
		}
	}

	/**
	 * Writes content to given instance <code>outputStream</code> of {@link OutputStream} 
	 */
	@Override public void save(OutputStream outputStream) {
		if (optionsAcquired) {
			put();
		}
		try {
			ini.store(outputStream);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Returns value for given <code>optionName</code> of given <code>sectionName</code>.<br>
	 * Throws {@link IllegalArgumentException} if,
	 * <ul>
	 *   <li>given <code>sectionName</code> does not exist</li>
	 *   <li>given <code>optionName</code> does not exist</code>
	 * </ul>
	 */
	@Override public String get(String sectionName, String optionName) {
		Section section = ini.get(sectionName);
		if (section == null) {
			throw new IllegalArgumentException("unknown section: " + sectionName);
		}
		if (!section.containsKey(optionName)) {
			throw new IllegalArgumentException("unknown option " + optionName + " for section: " + sectionName);
		} else {
			return section.get(optionName);
		}
	}

	/**
	 * Sets given value <code>optionValue</code> for given <code>optionName</code> of given <code>sectionName</code>.<br>
	 * Throws {@link IllegalArgumentException} if,
	 * <ul>
	 *   <li>given <code>sectionName</code> is not present in user's sections descriptor {@link #optSecDescriptors}</li> 
	 *   <li>given <code>optionName</code> of <code>sectionName</code> is not present in user's sections descriptor {@link #optSecDescriptors}</li> 
	 * </ul>
	 * Note,
	 * <ul>
	 *   If instance with given <code>index</code> of given <code>sectionName</code> is not present, it will be added
	 * </ul>
	 */
	@Override public void put(String sectionName, String optionName, String optionValue) {
		Map<String,OptionDesc> optionsDescMap = optSecDescriptors.get(sectionName).getOptionsDescriptor();
		if (optionsDescMap == null) {
			throw new IllegalArgumentException("unknown section name, " + sectionName);
		}
		OptionDesc optionDesc = optionsDescMap.get(optionName);
		if (optionDesc == null) {
			throw new IllegalArgumentException("unknown option " + optionName + ", for section name, " + sectionName);
		}
		Section section = ini.get(sectionName);
		if (section == null) {
			ini.add(sectionName);
			section = ini.get(sectionName);
		}
		section.put(optionName, optionValue);
	}

	/**
	 * Removes given section <code>sectionName</code>. Obviously, if this section has multiple instances, all of them will be removed as well.  
	 */
	@Override public void removeSection(String sectionName) {
		ini.remove(sectionName);
	}

	/**
	 * This method is intended for multiple section mode.<br>
	 * It returns number of occurrences/instances of given section <code>sectionName</code>.<br>
	 * Note,
	 * <ul>
	 *   returns 0 if given section <code>sectionName</code> does not exist
	 * </ul>  
	 */
	@Override public int getNumberOfMultipleSections(String sectionName) {
		return ini.length(sectionName);
	}
	
	/**
	 * This method is intended for multiple section mode.<br>
	 * Returns value for given <code>optionName</code> of given instance <code>index</code> of given <code>sectionName</code>.<br>
	 * Throws {@link IllegalArgumentException} if,
	 * <ul>
	 *   <li>given <code>sectionName</code> does not exist</li>
	 *   <li>instance of section with given <code>index</code> does not exist</li>
	 *   <li>given <code>optionName</code> does not exist</code>
	 * </ul>
	 */
	@Override public String get(String sectionName, String optionName, int index) {
		Section section = ini.get(sectionName, index);
		if (section == null) {
			throw new IllegalArgumentException("unknown section: " + sectionName);
		}
		String optionValue = section.get(optionName);
		if (optionValue == null) {
			throw new IllegalArgumentException("unknown option " + optionName + " for section: " + sectionName);
		}
		return optionValue;
	}
	
	/**
	 * This method is intended for multiple section mode.<br>
	 * Sets given value <code>optionValue</code> for given <code>optionName</code> of given instance <code>index</code> of 
	 * given <code>sectionName</code>.<br>
	 * Throws {@link IllegalArgumentException} if,
	 * <ul>
	 *   <li>given <code>sectionName</code> is not present in user's sections descriptor {@link #optSecDescriptors}</li> 
	 *   <li>given <code>optionName</code> of <code>sectionName</code> is not present in user's sections descriptor {@link #optSecDescriptors}</li> 
	 * </ul>
	 * Note,
	 * <ul>
	 *   If instance with given <code>index</code> of given <code>sectionName</code> is not present, it will be added
	 * </ul>
	 */
	@Override public void put(String sectionName, String optionName, String optionValue, int index) {
		Map<String,OptionDesc> optionsDescMap = optSecDescriptors.get(sectionName).getOptionsDescriptor();
		if (optionsDescMap == null) {
			throw new IllegalArgumentException("unknown section name, " + sectionName);
		}
		OptionDesc optionDesc = optionsDescMap.get(optionName);
		if (optionDesc == null) {
			throw new IllegalArgumentException("unknown option " + optionName + ", for section name, " + sectionName);
		}
		Section section = ini.get(sectionName, index);
		if (section == null) {
			ini.add(sectionName, section, index);
			section = ini.get(sectionName, index);
		}
		section.put(optionName, optionValue);
	}

	/**
	 * Returns <b>true</b> if there were errors in options, otherwise <b>false</b>
	 */
	@Override public boolean isError() {
		return error;
	}
}
