/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
package edu.uml.lgdc.project.settings;

import edu.uml.lgdc.datatype.Cast;

public class Setting<T> {
	static public class SettingAttributes<T> {
		private final String label;
		private final Class<T> c;
		private final String description;
		private final boolean isSelectable;
		private final boolean hasValue;
		private final T minValue;
		private final T maxValue;
		private final boolean parseable;

		public SettingAttributes(String label, Class<T> c, String description, boolean isSelectable, boolean hasValue, T min, T max) {
			this.parseable = checkClassForParse(c.getName());
			this.label = label;
			this.c = c;
			this.description = description;
			this.isSelectable = isSelectable;
			this.hasValue = hasValue;
			this.minValue = min;
			this.maxValue = max;
		}

		public SettingAttributes(String label, Class<T> c, String description, boolean isSelectable, boolean hasValue) {
			this(label, c, description, isSelectable, hasValue, null, null);
		}

		public boolean isSelectable() {
			return isSelectable;
		}

		public String getLabel() {
			return label;
		}

		public Class<T> getC() {
			return c;
		}

		public String getDescription() {
			return description;
		}

		public boolean isHasValue() {
			return hasValue;
		}

		public T getMinValue() {
			return minValue;
		}

		public T getMaxValue() {
			return maxValue;
		}

		static public boolean checkClassForParse(String className) {
			return (className.equals(Integer.class.getName()) || className.equals(Double.class.getName()) || className.equals(String.class
					.getName()));
		}
	}

	private T val;
	private boolean isSelected;
	private final SettingAttributes<T> attr;

	public Setting(T val, boolean isSelected, SettingAttributes<T> attr) {
		this.attr = attr;
		this.setVal(val);
		this.setSelected(isSelected);
	}

	public T getVal() {
		return val;
	}

	/**
	 * Return value if selected and null otherwise.
	 * 
	 * @return
	 */
	public T getValIfSelected() {
		if (isSelected == true)
			return val;
		else
			return null;
	}

	private void setVal(T val) {
		if (val != null) {
			this.val = val;
			if (attr.maxValue != null)
				this.val = this.compareVal(attr.maxValue) <= 0 ? val : attr.maxValue;
			if (attr.minValue != null)
				this.val = this.compareVal(attr.minValue) >= 0 ? val : attr.minValue;
		}
	}

	public void setVal_o(Object val) {
		T valCast = Cast.cast(val, attr.getC());
		if (valCast != null)
			setVal(valCast);
	}

	public void setVal(String str) {
		if (!attr.parseable) {
			System.out.println("Cannot parse class " + attr.getC().getName() + " from a string. Use direct set(T val).");
			return;
		}
		T val = this.val;
		val = Cast.cast(parse(str), attr.getC());
		setVal(val);
	}

	public boolean isSelected() {
		return isSelected;
	}

	public void setSelected(boolean isSelected) {
		this.isSelected = isSelected;
	}

	public SettingAttributes<T> getAttr() {
		return attr;
	}

	/**
	 * Parse only three types, for now: Integer Double and String.
	 * 
	 * @param str
	 * @param className
	 * @return
	 */
	private Object parse(String str) {
		String cn = attr.getC().getName();
		try {
			if (cn.equals(Integer.class.getName()))
				return Integer.parseInt(str);
			else if (cn.equals(Double.class.getName()))
				return Double.parseDouble(str);
			else
				// if(cn.equals(String.class.getName()))
				return str;
		} catch (NumberFormatException e) {
			System.out.println("Cannot parse class " + cn + " from \"" + str + "\". Set to previous valid value.");
			return null;
		}
	}

	/**
	 * Parse only three types, for now: Integer Double and String.
	 * 
	 * @param str
	 * @param className
	 * @return
	 */
	private int compareVal(T val) {
		String cn = attr.getC().getName();
		if (cn.equals(Integer.class.getName()))
			return Double.compare(((Integer) this.val), ((Integer) val));
		else if (cn.equals(Double.class.getName()))
			return Double.compare(((Double) this.val), ((Double) val));
		else
			// if(cn.equals(String.class.getName()))
			return ((String) this.val).compareTo((String) val);
	}
}