/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
//Title:        class RealHours
//Version:      1.0
//Copyright:    Copyright (c) 2001
//Author:       Alexander Kozlov
//Company:      UMLCAR
//Description:  To transfer time given as real number of hours to
//              hours, minutes, seconds, milliseconds and date offset

package edu.uml.lgdc.time;

import java.util.*;

import edu.uml.lgdc.format.FC;

public class RealHours
{
  private double realHours;
  private double longitude = 0;
  private int dateOffset = 0;

  private int hours;
  private int minutes;
  private int seconds;
  private int millisecs;

 /**
  *  Constructor
  */
  public RealHours( int hours )
  {
    this( hours, 0 );
  }

 /**
  *  Constructor
  */
  public RealHours( int hours, int minutes )
  {
    this( hours, minutes, 0 );
  }

 /**
  *  Constructor
  */
  public RealHours( int hours, int minutes, int seconds )
  {
    this( hours, minutes, seconds, 0 );
  }

 /**
  *  Constructor
  */
  public RealHours( int hours, int minutes, int seconds, int millisecs )
  {
    this( 0, hours, minutes, seconds, millisecs );
  }

 /**
  *  Constructor
  */
  public RealHours( double lon, int hours, int minutes, int seconds, int millisecs )
  {
    setup( lon, toRealHours( hours, minutes, seconds, millisecs ) );
  }

 /**
  *  Constructor
  */
  public RealHours( TimeScale time )
  {
    this( 0, time );
  }

 /**
  *  Constructor
  */
  public RealHours( double longitude, TimeScale time )
  {
    this( longitude, toRealHours( time ) );
  }

 /**
  *  Constructor
  */
  public RealHours( double realHours )
  {
    this( 0, realHours );
  }

 /**
  *  Constructor
  */
  public RealHours( double longitude, double realHours )
  {
    setup( longitude, realHours );
  }

 /**
  *  Returns hours as real value
  */
  public double getRealHours()
  {
    return realHours;
  }


 /**
  *  Returns day offset
  */
  public int getDateOffset()
  {
    return dateOffset;
  }


 /**
  *  Returns hours
  */
  public int getHours()
  {
    return hours;
  }


 /**
  *  Returns minutes
  */
  public int getMinutes()
  {
    return minutes;
  }


 /**
  *  Returns seconds
  */
  public int getSeconds()
  {
    return seconds;
  }


 /**
  *  Returns milliseconds
  */
  public int getMillisecs()
  {
    return millisecs;
  }

 /**
  *  Getting real value of hours fills fields of this object with
  *  correct values.
  */
  private void setup( double longitude, double utRealHours )
  {
    dateOffset = 0;

    if( longitude >= 360 )
      longitude %= 360;
    else if( longitude < 0 )
    {
      longitude = -longitude;
      longitude %= 360;
      longitude = 360 - longitude;
      if( longitude >= 360 )
        longitude = 0;
    }

    this.longitude = longitude;

    realHours = utRealHours + (longitude / 360) * TimeScale.HOURS_PER_DAY;
    if( realHours >= 24 )
      realHours -= 24;

    if( utRealHours < 12 )
    {
      if( realHours > utRealHours + 12 )
        dateOffset--;
    }
    else
    {
      if( realHours <= utRealHours - 12 )
        dateOffset++;
    }

    double temp = realHours;

    hours     = (int)temp;
    temp      = ( temp - hours ) * TimeScale.MINUTES_PER_HOUR;
    minutes   = (int)temp;
    temp      = ( temp - minutes ) * TimeScale.SECONDS_PER_MINUTE;
    seconds   = (int)temp;
    temp      = ( temp - seconds ) * TimeScale.MILLISECONDS_PER_SECOND;
    millisecs = (int)temp;
  }

 /**
  *  Having TimeScale object return its "time part" as one number
  *  of hours ( as double value ).
  */
  static public double toRealHours( TimeScale time )
  {
    return toRealHours( time.get( Calendar.HOUR_OF_DAY ),
                        time.get( Calendar.MINUTE ),
                        time.get( Calendar.SECOND ),
                        time.get( Calendar.MILLISECOND ) );
  }

 /**
  *  Having hours, minutes, seconds and milliseconds return it as one number
  *  of hours ( as double value ).
  *  !!! No overflowing is checked here (i.e. hours may be >= 24 and so on ).
  */
  static public double toRealHours( int hours, int minutes, int seconds,
                                    int millisecs )
  {
    return hours +
      (
        minutes +
        (
          seconds + (double)millisecs / TimeScale.MILLISECONDS_PER_SECOND
        ) / TimeScale.SECONDS_PER_MINUTE
      ) / TimeScale.MINUTES_PER_HOUR;
  }

 /**
  *  Having real hours return it as human string, i.e. as hh:mm:ss
  *  !!! No overflowing is checked here (i.e. hours should be < 24).
  */
  static public String toHuman( double realHours )
  {
    String out = "";
    int hours = (int)realHours;
    realHours = ( realHours - hours ) * TimeScale.MINUTES_PER_HOUR;
    int minute = (int)realHours;
    realHours = ( realHours - minute ) * TimeScale.SECONDS_PER_MINUTE;
    int sec = (int)realHours;
    out = FC.padLeft( FC.IntegerToString( hours ), 2, '0' ) + ":" +
          FC.padLeft( FC.IntegerToString( minute ), 2, '0' ) + ":" +
          FC.padLeft( FC.IntegerToString( sec ), 2, '0' );
    return out;
  }

}

