package edu.uml.lgdc.time;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Objects;
import java.util.TimeZone;

import edu.uml.lgdc.datatype.Metricable;
import edu.uml.lgdc.format.FC;
import edu.uml.lgdc.format.StrUtil;
import edu.uml.lgdc.math.Search;

/*******************************************************************************
 * Copyright (c) 2004, 2018 University of Massachusetts Lowell
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Creative Commons 
 * Attribution-ShareAlike 2.0 (CC BY-SA 2.0) available at
 * https://creativecommons.org/licenses/by-sa/2.0/ with the 
 * attribution clause "Lowell GIRO Data Center Rules of the Road"
 * available at http://umlcar.uml.edu/DIDBase/RulesOfTheRoadForDIDBase.htm
 *******************************************************************************/
public class TimeScale extends GregorianCalendar implements TimeMarked, Metricable<TimeScale> {
	static final long serialVersionUID = 8506009477987519770L;
	/**
	 * don't change this, as it is used in database conversion
	 */
	static public final String TIME_FORMAT_FOR_TO_STRING = "yyyy.MM.dd HH:mm:ss.SSS";

	static public final int FIELD_NOT_SET = -1;
	static public final int FIELD_ERROR = -999;

	static final public int MILLISECONDS_PER_SECOND = 1000;
	static final public int SECONDS_PER_MINUTE = 60;
	static final public int MINUTES_PER_HOUR = 60;
	static final public int HOURS_PER_DAY = 24;
	static final public int MONTHS_PER_YEAR = 12;
	static final public double DAYS_PER_YEAR = 365.2425;

	static final public int MINUTES_PER_DAY = HOURS_PER_DAY * MINUTES_PER_HOUR;
	static final public int SECONDS_PER_DAY = MINUTES_PER_DAY * SECONDS_PER_MINUTE;
	static final public int MILLISECONDS_PER_DAY = SECONDS_PER_DAY * MILLISECONDS_PER_SECOND;
	static final public int MILLISECONDS_PER_MINUTE = SECONDS_PER_MINUTE * MILLISECONDS_PER_SECOND;
	static final public int MILLISECONDS_PER_HOUR = MINUTES_PER_HOUR * MILLISECONDS_PER_MINUTE;

	static final protected int[] daysPerMonths = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	static final protected String[] shortNamesOfMonths = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	static final protected String[] fullNamesOfMonths = {"January", "February", "March", "April", "May", "June", "July", "August",
		"September", "October", "November", "December"};

	static final public TimeZone tzUTC = TimeZone.getTimeZone("UTC");

	static private final int[] maxDaysPerMonth = {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	static public final TimeScale TM = new TimeScale();

	static public final String DIGITS = "0123456789";
	
	public TimeScale(int year, int month, int day, int hour, int minute, int sec) {
		this(year, month, day, hour, minute, sec, 0);
	}
	public TimeScale(int year, int month, int day, int hour, int minute, int sec, int millisec) {
		super(year, month, day, hour, minute, sec);
		setTimeZone(tzUTC);
		set(Calendar.MILLISECOND, millisec);
	}
	public TimeScale() {
		this(tzUTC);
	}
	public TimeScale(TimeZone tz) {
		super(tz);
	}
	public TimeScale(long dateTimeInMillis) {
		setTimeZone(tzUTC);
		setTimeInMillis(dateTimeInMillis);
	}
	public TimeScale(double dateTimeInMinute) {
		setTimeZone(tzUTC);
		setTimeInMillis(Math.round(dateTimeInMinute * SECONDS_PER_MINUTE * MILLISECONDS_PER_SECOND));
	}
	public TimeScale(String dateStr, String formatString) throws ParseException {
		this(dateStr, formatString, false);
	}
	public TimeScale(String dateStr, String formatString, boolean strict) throws ParseException {
		setTimeZone(tzUTC);
		set(dateStr, formatString, strict);
	}
	/**
	 * Constructor from String.
	 * @see #decode(String, boolean)
	 */
	public TimeScale(String timeString) {
		timeString = Objects.requireNonNull(timeString).trim();
		setTimeZone(tzUTC);
		try {
			int[] datetime = decode(timeString);
			// year month day hour minute second
			set(datetime[0], datetime[1], datetime[2], datetime[3], datetime[4], datetime[5]);
			set(MILLISECOND, datetime[6]);
		} catch (IllegalDateTimeStringException ex) {
			throw new RuntimeException(ex);
		}
	}
	/**
	 * Parses given time string and returns TimeScale instance.<br>
	 * Note, 
	 * <ul>
	 *   returns <b>null</b> if parsing exception occurred
	 * </ul>
	 * @see #decode(String, boolean)
	 */
	static public TimeScale parseNoException(String timeString) {
		try {
			return parse(timeString, false);
		} catch (IllegalDateTimeStringException e) {
			return null;
		}
	}
	/**
	 * Parses given time string and returns TimeScale instance.<br>
	 * Note, 
	 * <ul>
	 *   returns <b>null</b> if parsing exception occurred
	 * </ul>
	 * @see #decode(String, boolean)
	 */
	static public TimeScale parseNoException(String timeString, boolean allowAnythingAfterMilliseconds) {
		try {
			return parse(timeString, allowAnythingAfterMilliseconds);
		} catch (IllegalDateTimeStringException e) {
			return null;
		}
	}
	/**
	 * Parses given time string and returns TimeScale instance.<br>
	 * @throws IllegalDateTimeStringException if parsing error occurred, so never returns <b>null</b>
	 * @see #decode(String, boolean)
	 */
	static public TimeScale parse(String timeString) throws IllegalDateTimeStringException {
		return parse(timeString, false);
	}
	/**
	 * Parses given time string and returns TimeScale instance.<br>
	 * @throws IllegalDateTimeStringException if parsing error occurred, so never returns <b>null</b>
	 * @see #decode(String, boolean)
	 */
	static public TimeScale parse(String timeString, boolean allowAnythingAfterMilliseconds) throws IllegalDateTimeStringException {
		int[] t = decode(timeString, allowAnythingAfterMilliseconds);
		return new TimeScale(t[0], t[1], t[2], t[3], t[4], t[5], t[6]);
	}
	/**
	 * This is equivalent to call <code>decode(timeString, false)</code>.<br>
	 * @see #decode(String, boolean)
	 */
	static public int[] decode(String timeString) throws IllegalDateTimeStringException {
		return decode(timeString, false);
	}
	/**
	 * Parses given time string and returns integer array which contains elements of date,
	 * <ul>
	 *   <li>0 = year,  &gt;0 and &le; 9999</li>
	 *   <li>1 = month, &ge;0 and &le; 11</li>
	 *   <li>2 = day, &ge;1 and &le; 31</li>
	 *   <li>3 = hour, &ge;0 and &le; 23</li>
	 *   <li>4 = minute, &ge;0 and &le; 59</li>
	 *   <li>5 = second, &ge;0 and &le; 59</li>
	 *   <li>6 = millisecond, &ge;0 and &le; 999</li>
	 * </ul>
	 * Datetime string can be in one of two forms,
	 * <ol>
	 *   <li><code>yyyy/MM/dd HH:mm:ss.SSS</code> or <code>yyyy.MM.ddTHH:mm:ss.SSS</code></li>
	 *   <li><code>yyyy/MM/dd (DDD) HH:mm:ss.SSS</code></li>
	 * </ol>
	 * Notes,
	 * <ul>
	 *   <li>in both forms, delimiter / which is used between date parts (year/month/day) can be replaced either with . (dot) or - (dash)</li>
	 *   <li>note that you have specified exact number of digits for any item that indicated in format for that item. So, if you have 
	 *      day number 6, you have put it as 06, and so on</li>
	 *   <li>both forms can be truncated at any position just right before delimiter, for example <code>1946.08.06</code></li>
	 *   <li>if parameter <code>allowAnythingAfterMilliseconds</code> is <b>true</b> then any characters after milliseconds are allowed,
	 *       otherwise, if milliseconds are presented, the <code>timeString</code> has to be ended after them</li>
	 * </ul> 
	 */
	static public int[] decode(String timeString, boolean allowAnythingAfterMilliseconds) throws IllegalDateTimeStringException {
		final char[] dateDelimiters = {'.', '/', '-' };
		final char timeDelim = ':'; 
		timeString = Objects.requireNonNull(timeString).trim();
		int year, month = 1, day = 1, hour = 0, minute = 0, sec = 0, msec = 0;
		int dayOfYear = -1;
		if (timeString.length() < 4) {
			throw new IllegalDateTimeStringException("string is too short");
		}
		int i = 0;
		while (true) {
			String tmpString = timeString.substring(i, i + 4);
			try {
				year = Integer.parseInt(tmpString);
			} catch (NumberFormatException ex) {
				throw new IllegalDateTimeStringException("bad year, " + tmpString);
			}
			if ( year <= 0) {
				throw new IllegalDateTimeStringException("bad year, " + tmpString);
			}
			if (timeString.length() < 7) {
				if (timeString.length() != 4) {
					throw new IllegalDateTimeStringException("garbage at the end " + timeString);
				}
				break;
			}
			i = 5;
			char dateDelim = timeString.charAt(i - 1);
			if (Search.scan(dateDelimiters, dateDelim) < 0) {
				throw new IllegalDateTimeStringException("illegal date delimiter, " + dateDelim);
			}
			tmpString = timeString.substring(i, i + 2);
			try {
				month = Integer.parseInt(tmpString);
			} catch (NumberFormatException ex) {
				throw new IllegalDateTimeStringException("bad month, " + tmpString);
			}
			if (!checkMonth(month)) {
				throw new IllegalDateTimeStringException("bad month, " + tmpString);
			}
			if (timeString.length() < 10) {
				if (timeString.length() != 7) {
					throw new IllegalDateTimeStringException("garbage at the end " + timeString);
				}
				break;
			}
			i = 8;
			if (dateDelim != timeString.charAt(i - 1)) {
				throw new IllegalDateTimeStringException("illegal or inconsistent date delimiter(s), " + timeString);
			}
			tmpString = timeString.substring(i, i + 2);
			try {
				day = Integer.parseInt(tmpString);
			} catch (NumberFormatException ex) {
				throw new IllegalDateTimeStringException("bad day, " + tmpString);
			}
			if (!checkDay(day)) {
				throw new IllegalDateTimeStringException("bad day, " + tmpString);
			}
			// is day of year (DDD) specified?
			if (timeString.length() > i + 3 && timeString.charAt(i + 3) == '(') {
				if (timeString.length() < i + 8 || timeString.charAt(i + 7) != ')') {
					throw new IllegalDateTimeStringException("garbage at the end " + timeString);
				}
				tmpString = timeString.substring(i + 4, i + 7);
				try {
					dayOfYear = Integer.parseInt(tmpString);
				} catch (NumberFormatException ex) {
					throw new IllegalDateTimeStringException("bad day of year, " + tmpString);
				}
				if (!checkDayOfYear(year, dayOfYear)) {
					throw new IllegalDateTimeStringException("bad day of year, " + tmpString);
				}
				if (timeString.length() == i + 8) {
					break;
				}
				i += 6;
			}
			if (timeString.length() < i + 5) {
				if (timeString.length() != i + 2) {
					throw new IllegalDateTimeStringException("garbage at the end " + timeString);
				}
				break;
			}
			i += 3;
			if (dayOfYear > 0) {
				if (timeString.charAt(i - 1) != ' ' || timeString.charAt(i - 7) != ' ') {
					throw new IllegalDateTimeStringException("before and after (DDD) has to be one space symbol " + timeString);
				}
			} else {
				if (timeString.charAt(i - 1) != ' ' && timeString.charAt(i - 1) != 'T') {
					throw new IllegalDateTimeStringException("between date and time parts has to be either space or letter T, " + timeString);
				}
			}
			tmpString = timeString.substring(i, i + 2);
			try {
				hour = Integer.parseInt(tmpString);
			} catch (NumberFormatException ex) {
				throw new IllegalDateTimeStringException("bad hour, " + tmpString);
			}
			if (!checkHour(hour)) {
				throw new IllegalDateTimeStringException("bad hour, " + tmpString);
			}
			if (timeString.length() < i + 5) {
				if (timeString.length() != i + 2) {
					throw new IllegalDateTimeStringException("garbage at the end " + timeString);
				}
				break;
			}
			i += 3;
			if (timeDelim != timeString.charAt(i - 1)) {
				throw new IllegalDateTimeStringException("illegal time delimiter, " + timeString);
			}
			tmpString = timeString.substring(i, i + 2);
			try {
				minute = Integer.parseInt(tmpString);
			} catch (NumberFormatException ex) {
				throw new IllegalDateTimeStringException("bad minute, " + tmpString);
			}
			if (!checkMinute(minute)) {
				throw new IllegalDateTimeStringException("bad minute, " + tmpString);
			}
			if (timeString.length() < i + 5) {
				if (timeString.length() != i + 2) {
					throw new IllegalDateTimeStringException("garbage at the end " + timeString);
				}
				break;
			}
			i += 3;
			if (timeDelim != timeString.charAt(i - 1)) {
				throw new IllegalDateTimeStringException("illegal time delimiter, " + timeString);
			}
			tmpString = timeString.substring(i, i + 2);
			try {
				sec = Integer.parseInt(tmpString);
			} catch (NumberFormatException ex) {
				throw new IllegalDateTimeStringException("bad second, " + tmpString);
			}
			if (!checkSecond(sec)) {
				throw new IllegalDateTimeStringException("bad second, " + tmpString);
			}
			if (timeString.length() < i + 6) {
				if (timeString.length() != i + 2) {
					throw new IllegalDateTimeStringException("garbage at the end " + timeString);
				}
				break;
			}
			i += 3;
			if (timeString.charAt(i - 1) != '.') {
				throw new IllegalDateTimeStringException("illegal time delimiter between seconds and milliseconds, " + timeString);
			}
			tmpString = timeString.substring(i, i + 3);
			try {
				msec = Integer.parseInt(tmpString);
			} catch (NumberFormatException ex) {
				throw new IllegalDateTimeStringException("bad millisecond, " + tmpString);
			}
			if (!checkMilliSecond(msec)) {
				throw new IllegalDateTimeStringException("bad millisecond, " + tmpString);
			}
			if (!allowAnythingAfterMilliseconds && timeString.length() > i + 3) {
				throw new IllegalDateTimeStringException("garbage at the end " + timeString);
			}
			break;
		}
		if (!checkDate(year, month, day)) {
			throw new IllegalDateTimeStringException("year, month and day are inconsistent");
		}
		if (dayOfYear > 0) {
			if (month != getMonth(year, dayOfYear) || day != getDayInMonth(year, dayOfYear)) {
				throw new IllegalDateTimeStringException("year, month, day, and day of year are inconsistent");
			}
		}
		month--;
		return new int[] {year, month, day, hour, minute, sec, msec};
	}
	@Override public boolean before(Object obj) {
		boolean result = false;
		if (obj instanceof TimeScale) {
			result = super.before(obj);
		}
		return result;
	}
	/*@Override public boolean equiv(Object obj) {
		return equals(obj);
	}*/
	@Override public double dist(TimeScale time) {
		if (time != null) {
			return Math.abs(diffIn(Calendar.MINUTE, time));
		}
		return Double.NaN;
	}
	@Override public void complete() {
		super.complete();
	}
	public void setTimeInMilliSeconds(long dateTimeInMillis) {
		setTimeInMillis(dateTimeInMillis);
	}
	public void setTimeInSeconds(long dateTimeInSeconds) {
		setTimeInMillis(dateTimeInSeconds * MILLISECONDS_PER_SECOND);
	}
	public void setTimeInMinutes(double dateTimeInMinute) {
		setTimeInMillis(Math.round(dateTimeInMinute * SECONDS_PER_MINUTE * MILLISECONDS_PER_SECOND));
	}
	public void setTime(TimeScale dateTime) {
		setTimeInMillis(dateTime.getTimeInMillis());
	}
	public long getTimeInMilliSeconds() {
		return getTimeInMillis();
	}
	public double getTimeInSeconds() {
		return getTimeIn(Calendar.SECOND);
	}
	@Override public double getTimeInMinutes() {
		return getTimeIn(Calendar.MINUTE);
	}
	public double getWholeTimeInSeconds() {
		return getWholeTimeIn(Calendar.SECOND);
	}
	public double getWholeTimeInMinutes() {
		return getWholeTimeIn(Calendar.MINUTE);
	}
	public String getAsString() {
		int year = get(Calendar.YEAR);
		int month = get(Calendar.MONTH) + 1;
		String str = "" + year;
		str += "." + FC.IntegerToString(month, 2).replace(' ', '0');
		str += "." + FC.IntegerToString(get(Calendar.DAY_OF_MONTH), 2).replace(' ', '0');
		str += " " + FC.IntegerToString(get(Calendar.HOUR_OF_DAY), 2).replace(' ', '0');
		str += ":" + FC.IntegerToString(get(Calendar.MINUTE), 2).replace(' ', '0');
		str += ":" + FC.IntegerToString(get(Calendar.SECOND), 2).replace(' ', '0');
		return str;
	}
	public String toFormatUT(String formatString) {
		SimpleDateFormat formatter = new SimpleDateFormat(formatString);
		formatter.setTimeZone(getTimeZone());
		return formatter.format(getTime());
	}
	public String toFormatISO8601() {
		TimeZone timeZone = getTimeZone();
		if (timeZone == tzUTC) {
			return toFormatUT("yyyy-MM-dd'T'HH:mm:ss.SSS'Z'");
		} else {
			int offset_min = timeZone.getOffset(getTimeInMillis()) / 60000;
			String sign = (offset_min >= 0) ? "+" : "-";
			offset_min = Math.abs(offset_min);
			return toFormatUT("yyyy-MM-dd'T'HH:mm:ss.SSS") + sign + FC.padLeft(FC.IntegerToString(offset_min / 60), 2, '0') + ":"
				+ FC.padLeft(FC.IntegerToString(offset_min % 60), 2, '0');
		}
	}
	
	public String toIonexEpoch() {
		return FC.IntegerToString(get(TimeScale.YEAR), 6) 
				+ FC.IntegerToString(get(TimeScale.MONTH) + 1, 6)
				+ FC.IntegerToString(get(TimeScale.DAY_OF_MONTH), 6) 
				+ FC.IntegerToString(get(TimeScale.HOUR), 6)
				+ FC.IntegerToString(get(TimeScale.MINUTE), 6) 
				+ FC.IntegerToString(get(TimeScale.SECOND), 6) 
				+ FC.nC2S(' ', 24);
	}
	
	public void set(String dateStr, String formStr) throws ParseException {
		set(dateStr, formStr, false);
	}
	public void set(String dateStr, String formStr, boolean strict) throws ParseException {
		dateStr = dateStr.replace('/', '.').replace('-', '.');
		formStr = formStr.replace('/', '.').replace('-', '.');
		SimpleDateFormat formatter = new SimpleDateFormat(formStr);
		formatter.setLenient(!strict);
		formatter.setTimeZone(tzUTC);
		setTimeZone(tzUTC);
		this.setTime(formatter.parse(dateStr));
	}
	static public TimeScale parse(String dateStr, String formStr) throws ParseException {
		return parse(dateStr, formStr, false);
	}
	static public TimeScale parse(String dateStr, String formStr, boolean strict) throws ParseException {
		TimeScale time = new TimeScale();
		time.set(dateStr, formStr, strict);
		return time;
	}
	
	static private final String[] DATE_TIME_POPULAR_FORMATS = {
			"yyyy.MM.dd (DDD) HH:mm:ss.SSS", 
			"yyyy.MM.dd (DDD) HH:mm:ss", 
			"yyyy.MM.dd HH:mm:ss.SSS", 
			"yyyy.MM.dd HH:mm:ss", 
			"yyyy.MM.dd HH:mm", 
			"yyyy.MM.dd HH", 
			"yyyy.MM.dd" 
		};
	
	static public TimeScale parseSmart(String dateStr) throws ParseException {
		TimeScale time = new TimeScale();
		dateStr = dateStr.replace('/', '.').replace('-', '.').replaceFirst("T", " ");
		for (int i = 0; i < DATE_TIME_POPULAR_FORMATS.length; i++) {
			String formStr = DATE_TIME_POPULAR_FORMATS[i];
			SimpleDateFormat formatter = new SimpleDateFormat(formStr);
			formatter.setLenient(false);
			formatter.setTimeZone(tzUTC);
			if (i < DATE_TIME_POPULAR_FORMATS.length - 1) {
				try {
					time.setTime(formatter.parse(dateStr));
					break; // successful parse, return
				} 
				catch (ParseException ex) 
				{
					// do nothing
				}
			} else {
				time.setTime(formatter.parse(dateStr));
			}
		}
		return time;
	}
	public String toUT() {
		return toFormatUT("yyyy-MMM-dd (DDD) HH:mm:ss");
	}
	public String toShortUT() {
		return toFormatUT("yyyy.MM.dd (DDD) HH:mm:ss");
	}
	public String toHumanUT() {
		return toFormatUT("yyyy.MM.dd HH:mm:ss");
	}
	public String toDateUT() {
		return toFormatUT("yyyy.MM.dd (DDD)");
	}
	public String toShortDateUT() {
		return toFormatUT("yyyy.MM.dd");
	}
	public String toTimeUT() {
		return toFormatUT("HH:mm:ss");
	}
	public String toTimestamp() {
		return toFormatUT("yyyy/MM/dd HH:mm:ss.SSS");
	}
	public String toFilename() {
		return toFormatUT("yyyyMMddHHmmss");
	}
	public String toFilenameAternative() {
		return toFormatUT("yyyyMMdd_HHmmss");
	}
	@Override public String toString() {
		return toFormatUT(TIME_FORMAT_FOR_TO_STRING);
	}
	public String toFilename_ms() {
		return toFormatUT("yyyyMMddHHmmssSSS");
	}
	public String toFilenameAlternative_ms() {
		return toFormatUT("yyyyMMdd_HHmmssSSS");
	}
	/**
	 * Returns number of days in this month.
	 */
	public int numberOfDaysInMonth() {
		return numberOfDaysInMonth(get(Calendar.MONTH) + 1, get(Calendar.YEAR));
	}
	/**
	 * Returns number of days in given month for given year.
	 */
	public int numberOfDaysInMonth(int month, int year) {
		return isLeapYear(year) && (month == 2) ? 29 : daysPerMonths[month - 1];
	}
	/**
	 * Adds given number of milliseconds, <code>ms</code> to this instance of <code>TimeScale</code><br>
	 */
	public void addMillis(long ms) {
		setTimeInMillis(ms + getTimeInMillis());
	}
	/**
	 * Returns short name of month, 3 first letters of month with the 1st capital letter, like Jan (for January).
	 */
	public String getShortNameOfMonth() {
		return shortNamesOfMonths[get(MONTH)];
	}
	/**
	 * Returns short name of month by month number (1,...,12).
	 */
	static public String shortNameOfMonth(int numberOfMonth) {
		return shortNamesOfMonths[numberOfMonth - 1];
	}
	/**
	 * Returns full name of month by month number (1,...,12).
	 */
	static public String fullNameOfMonth(int numberOfMonth) {
		return fullNamesOfMonths[numberOfMonth - 1];
	}
	/**
	 * Returns number of month by month name.
	 */
	static public int monthToNumber(String month) {
		month = month.toUpperCase();
		for (int i = 0; i < shortNamesOfMonths.length; i++) {
			if (month.equals(shortNamesOfMonths[i].toUpperCase())) {
				return i + 1;
			}
			if (month.equals(fullNamesOfMonths[i].toUpperCase())) {
				return i + 1;
			}
		}
		return -1;
	}
	/**
	 * Returns <b>true</b> if year of this object is leap year, otherwise returns <b>false</b>
	 */
	public boolean isLeapYear() {
		return super.isLeapYear(get(Calendar.YEAR));
	}
	/**
	 * Returns number of days in year
	 */
	public int numberOfDaysInYear() {
		return isLeapYear() ? 366 : 365;
	}
	/**
	 * Returns number of days in given year.<br>
	 */
	public int numberOfDaysInYear(int year) {
		return isLeapYear(year) ? 366 : 365;
	}
	/**
	 * Returns times as double value in units specified by field number "field". Only the next fields are admitted: MILLISECOND, SECOND,
	 * MINUTE, HOUR_OF_DAY, DAY_OF_MONTH, MONTH and YEAR. In all other cases this function raises IllegalArgumentException exception.<br>
	 * Note:
	 * <ul>
	 *   usage MONTH and YEAR doesn't have exact estimation
	 * </ul>
	 */
	public double getTimeIn(int field) {
		double time = 0;
		switch (field) {
		case MILLISECOND:
			time = getTimeInMillis();
			break;
		case SECOND:
			time = (double)getTimeInMillis() / MILLISECONDS_PER_SECOND;
			break;
		case MINUTE:
			time = (double)getTimeInMillis() / MILLISECONDS_PER_MINUTE;
			break;
		case HOUR_OF_DAY:
			time = (double)getTimeInMillis() / MILLISECONDS_PER_HOUR;
			break;
		case DAY_OF_MONTH:
			time = (double)getTimeInMillis() / MILLISECONDS_PER_DAY;
			break;
		case MONTH:
			time =
				MONTHS_PER_YEAR * get(Calendar.YEAR) + get(Calendar.MONTH)
					+ (get(Calendar.DAY_OF_MONTH) - 1 + getTimeSinceMidnightIn(Calendar.DAY_OF_MONTH)) / numberOfDaysInMonth();
			break;
		case YEAR:
			time = getTimeIn(Calendar.HOUR_OF_DAY) / DAYS_PER_YEAR;
			break;
		default:
			throw new IllegalArgumentException();
		}
		return time;
	}
	/**
	 * Returns times as whole long number (with fractional part thrown away) in units specified by given field number, <code>field</code>.<br> 
	 * Only the next fields are admitted,
	 * <ul>
	 *   MILLISECOND, SECOND, MINUTE, HOUR_OF_DAY, DAY_OF_MONTH, MONTH and YEAR. In all other cases this function raises 
	 *   {@link IllegalArgumentException} exception.
	 * </ul>  
	 * Note:
	 * <ul>
	 *   usage MONTH and YEAR doesn't have exact estimation
	 * </ul>
	 */
	public long getWholeTimeIn(int field) {
		return (long)(Math.floor(getTimeIn(field)));
	}
	/**
	 * Returns times as whole long number in units specified by given field number, <code>field</code>, that is minimal and not less.<br>
	 * Only the next fields are admitted,
	 * <ul>
	 *   MILLISECOND, SECOND, MINUTE, HOUR_OF_DAY, DAY_OF_MONTH, MONTH and YEAR. In all other cases this function raises 
	 *   {@link IllegalArgumentException} exception.
	 * </ul>  
	 * Note:
	 * <ul>
	 *   usage MONTH and YEAR doesn't have exact estimation
	 * </ul>
	 */
	public long getNextWholeTimeIn(int field) {
		return (long)(Math.ceil(getTimeIn(field)));
	}
	/**
	 * Returns times of day, i.e. how much time elapsed since last midnight up to this time, as double value in units specified by field
	 * number <field>. Only the next fields are admitted: MILLISECOND, SECOND, MINUTE, HOUR_OF_DAY, DAY_OF_MONTH. In all other cases this
	 * function raises IllegalArgumentException exception.
	 */
	public double getTimeSinceMidnightIn(int field) {
		double time = 0;
		switch (field) {
		case MILLISECOND:
			time = get(Calendar.MINUTE) + MINUTES_PER_HOUR * get(Calendar.HOUR_OF_DAY);
			time = get(Calendar.SECOND) + SECONDS_PER_MINUTE * time;
			time = get(Calendar.MILLISECOND) + MILLISECONDS_PER_SECOND * time;
			break;
		case SECOND:
			time = getTimeSinceMidnightIn(Calendar.MILLISECOND) / MILLISECONDS_PER_SECOND;
			break;
		case MINUTE:
			time = getTimeSinceMidnightIn(Calendar.MILLISECOND) / MILLISECONDS_PER_MINUTE;
			break;
		case HOUR_OF_DAY:
			time = getTimeSinceMidnightIn(Calendar.MILLISECOND) / MILLISECONDS_PER_HOUR;
			break;
		case DAY_OF_MONTH:
			time = getTimeSinceMidnightIn(Calendar.MILLISECOND) / MILLISECONDS_PER_DAY;
			break;
		default:
			throw new IllegalArgumentException();
		}
		return time;
	}
	/**
	 * Returns difference between this time and given time "otherTime" as double value in units specified by field number "field". Only the
	 * next fields are admitted: MILLISECOND, SECOND, MINUTE, HOUR_OF_DAY, DAY_OF_MONTH, MONTH and YEAR. In all other cases this function
	 * raises IllegalArgumentException exception. You should be aware of calculation difference in months and years because approximate
	 * result in these cases.
	 */
	public double diffIn(int field, TimeScale otherTime) {
		return getTimeIn(field) - otherTime.getTimeIn(field);
	}
	/**
	 * Returns between this time and given time "otherTime" as long value in units specified by field number "field". Only the next fields
	 * are admitted: MILLISECOND, SECOND, MINUTE, HOUR_OF_DAY, DAY_OF_MONTH, MONTH and YEAR. In all other cases this function raises
	 * IllegalArgumentException exception. You should be aware of calculation difference in months and years because approximate result in
	 * these cases. Examples: Let we try to apply this function to times fallen in the same day and want to know difference in hours. Here
	 * is two examples. 1) 2:00:00 - 1:59:59 = 1; ( because = 2 - 1 = 1 ) 2) 1:59:59 - 1:00:00 = 0; ( because = 1 - 1 = 0 )
	 */
	public long diffWholeIn(int field, TimeScale otherTime) {
		return getWholeTimeIn(field) - otherTime.getWholeTimeIn(field);
	}
	/**
	 * Returns date as January 1, 00:00:00.000 and with year equaled "this" TimeScale object.
	 */
	public TimeScale startOfYear() {
		TimeScale startOfYear = (TimeScale)clone();
		startOfYear.set(Calendar.MONTH, 0);
		startOfYear.set(Calendar.DAY_OF_MONTH, 1);
		startOfYear.set(Calendar.HOUR_OF_DAY, 0);
		startOfYear.set(Calendar.MINUTE, 0);
		startOfYear.set(Calendar.SECOND, 0);
		startOfYear.set(Calendar.MILLISECOND, 0);
		return startOfYear;
	}
	/**
	 * Returns date as December 31, 23:59:59.999 and with year equaled "this" TimeScale object.
	 */
	public TimeScale endOfYear() {
		TimeScale endOfYear = (TimeScale)clone();
		endOfYear.add(Calendar.YEAR, 1);
		endOfYear.set(Calendar.MONTH, 0);
		endOfYear.set(Calendar.DAY_OF_MONTH, 1);
		endOfYear.set(Calendar.HOUR_OF_DAY, 0);
		endOfYear.set(Calendar.MINUTE, 0);
		endOfYear.set(Calendar.SECOND, 0);
		endOfYear.set(Calendar.MILLISECOND, 0);
		endOfYear.add(Calendar.MILLISECOND, -1);
		return endOfYear;
	}
	/**
	 * Returns date as 1st of this month and year, and time 00:00:00.000.
	 */
	public TimeScale startOfMonth() {
		TimeScale startOfMonth = (TimeScale)clone();
		startOfMonth.set(Calendar.DAY_OF_MONTH, 1);
		startOfMonth.set(Calendar.HOUR_OF_DAY, 0);
		startOfMonth.set(Calendar.MINUTE, 0);
		startOfMonth.set(Calendar.SECOND, 0);
		startOfMonth.set(Calendar.MILLISECOND, 0);
		return startOfMonth;
	}
	/**
	 * Returns date as with day equaled to last day of this month and time equaled to 23:59:59.999.
	 */
	public TimeScale endOfMonth() {
		TimeScale endOfMonth = (TimeScale)clone();
		endOfMonth.add(Calendar.MONTH, 1);
		endOfMonth.set(Calendar.DAY_OF_MONTH, 1);
		endOfMonth.set(Calendar.HOUR_OF_DAY, 0);
		endOfMonth.set(Calendar.MINUTE, 0);
		endOfMonth.set(Calendar.SECOND, 0);
		endOfMonth.set(Calendar.MILLISECOND, 0);
		endOfMonth.add(Calendar.MILLISECOND, -1);
		return endOfMonth;
	}
	/**
	 * Returns date with this month, day and year, but time is 00:00:00.000.
	 */
	public TimeScale startOfDay() {
		TimeScale startOfDay = (TimeScale)clone();
		startOfDay.set(Calendar.HOUR_OF_DAY, 0);
		startOfDay.set(Calendar.MINUTE, 0);
		startOfDay.set(Calendar.SECOND, 0);
		startOfDay.set(Calendar.MILLISECOND, 0);
		return startOfDay;
	}
	/**
	 * Returns date as with day, month and year equaled to this day, month, and year, and time equaled to 23:59:59.999.
	 */
	public TimeScale endOfDay() {
		TimeScale endOfDay = (TimeScale)clone();
		endOfDay.add(Calendar.DAY_OF_MONTH, 1);
		endOfDay.set(Calendar.HOUR_OF_DAY, 0);
		endOfDay.set(Calendar.MINUTE, 0);
		endOfDay.set(Calendar.SECOND, 0);
		endOfDay.set(Calendar.MILLISECOND, 0);
		endOfDay.add(Calendar.MILLISECOND, -1);
		return endOfDay;
	}
	/**
	 * Returns date with this month, day, year and hour, but time minute, seconds and milliseconds are zeroes.
	 */
	public TimeScale startOfHour() {
		TimeScale startOfHour = (TimeScale)clone();
		startOfHour.set(Calendar.MINUTE, 0);
		startOfHour.set(Calendar.SECOND, 0);
		startOfHour.set(Calendar.MILLISECOND, 0);
		return startOfHour;
	}
	/**
	 * Returns date with this day, month, year and hour, but mmss.SSS is equaled to 59:59.999.
	 */
	public TimeScale endOfHour() {
		TimeScale endOfHour = (TimeScale)clone();
		endOfHour.add(Calendar.HOUR_OF_DAY, 1);
		endOfHour.set(Calendar.MINUTE, 0);
		endOfHour.set(Calendar.SECOND, 0);
		endOfHour.set(Calendar.MILLISECOND, 0);
		endOfHour.add(Calendar.MILLISECOND, -1);
		return endOfHour;
	}
	/**
	 * Returns date with this month, day, year, hour and minute, but seconds and milliseconds are zeroes.
	 */
	public TimeScale startOfMinute() {
		TimeScale startOfMinute = (TimeScale)clone();
		startOfMinute.set(Calendar.SECOND, 0);
		startOfMinute.set(Calendar.MILLISECOND, 0);
		return startOfMinute;
	}
	/**
	 * Returns date with this day, month, year and hour, but ss.SSS is equaled to 59.999.
	 */
	public TimeScale endOfMinute() {
		TimeScale endOfMinute = (TimeScale)clone();
		endOfMinute.add(Calendar.MINUTE, 1);
		endOfMinute.set(Calendar.SECOND, 0);
		endOfMinute.set(Calendar.MILLISECOND, 0);
		endOfMinute.add(Calendar.MILLISECOND, -1);
		return endOfMinute;
	}
	public static TimeScale getTimestampFromString(String strTimeStamp, String formatStr) throws ParseException {
		return getTimestampFromString(strTimeStamp, formatStr, false);
	}
	public static TimeScale getTimestampFromString(String strTimeStamp, String formatStr, boolean strict) throws ParseException {
		TimeScale timeStamp;
		String mSecStr;
		int pos, mSecs = 0;
		pos = strTimeStamp.lastIndexOf('.');
		if (pos >= 0) {
			mSecStr = StrUtil.trimRight(strTimeStamp.substring(pos + 1));
			if (mSecStr.length() > 0) {
				if (!StrUtil.isOnlyDigits(mSecStr)) {
					if (!strict) {
						mSecStr = mSecStr.substring(0, StrUtil.sniffOthers(mSecStr, DIGITS, 0));
						if (mSecStr.length() == 0) {
							mSecStr = "000";
						}
					} else {
						throw new ParseException("illegal date format: " + strTimeStamp, pos + 1);
					}
				}
				if (mSecStr.length() < 3) {
					mSecStr = FC.padRight(mSecStr, 3, '0');
				}
				try {
					mSecs = Integer.parseInt(mSecStr);
				} catch (NumberFormatException ex) {
					if (strict) {
						throw new ParseException("illegal date format, " + strTimeStamp + ", " + ex.getMessage(), pos + 1);
					} else {
						mSecs = 0;
					}
				}
				if (strict && mSecs > 999) {
					throw new ParseException("illegal number of milliseconds, " + strTimeStamp, pos + 1);
				}
			} else if (strict) {
				throw new ParseException("illegal number of milliseconds, " + strTimeStamp, pos + 1);
			}
			strTimeStamp = strTimeStamp.substring(0, pos);
		}
		timeStamp = new TimeScale(strTimeStamp, formatStr, strict);
		timeStamp.set(Calendar.MILLISECOND, mSecs);
		return timeStamp;
	}

	/**
	 * Creates TimeScale object using field values.<br>
	 * Special values:
	 * <ul>
	 *   <li>FIELD_NOT_SET means <i>value not set for this field</i></li>
	 *   <li>FIELD_ERROR means <i>this field value is erroneous</i></li>
	 * </ul>
	 * Notes,
	 * <ul>
	 *   <li>Value for <code>month</code> starts from 1</li>
	 *   <li>If some error encountered then returns <b>null</b>.</li> 
	 *   <li>If value for some field is not set then default value for this field is used as follows:
	 *     <ul>
	 *       <li><code>year</code> assigned to current year</li>
	 *       <li><code>month</code> assigned to 1 (if <code>dayOfYear</code> is not set)</li> 
	 *       <li><code>dayOfMonth</code> assigned to 1 (if <code>dayOfYear</code> is not set)</li> 
	 *       <li><code>dayOfYear</code assigned to 1< (if <code>month</code> and <code>dayOfMonth</code> are not set)</li>
	 *       <li><code>hour</code> assigned to 0</li> 
	 *       <li><code>minute</code> assigned to 0</li>
	 *       <li><code>second</code> assigned to 0</li>
	 *     </ul>  
	 *   </li>
	 * </ul>
	 */
	static public TimeScale createFromFields(int year, int month, int dayOfMonth, int dayOfYear, int hour, int minute, int second) {
		TimeScale time = null;
		if (year != FIELD_ERROR && month != FIELD_ERROR && dayOfMonth != FIELD_ERROR && dayOfYear != FIELD_ERROR && hour != FIELD_ERROR
			&& minute != FIELD_ERROR && second != FIELD_ERROR) {
			time = new TimeScale();
			if (year != FIELD_NOT_SET) {
				time.set(YEAR, year);
			} else {
				year = time.get(YEAR);
				if (dayOfYear != FIELD_NOT_SET) {
					if (dayOfYear > time.get(DAY_OF_YEAR)) {
						year--;
						time.set(YEAR, year);
					}
				} else if (dayOfMonth != FIELD_NOT_SET && month != FIELD_NOT_SET) {
					int dayInYear = 0;
					for (int i = 0; i < month - 1; i++) {
						dayInYear += daysPerMonths[i];
					}
					dayInYear += dayOfMonth;
					if (month > 2 && time.isLeapYear()) {
						dayInYear++;
					}
					if (dayInYear > time.get(DAY_OF_YEAR)) {
						year--;
						time.set(YEAR, year);
					}
				}
			}
			if (hour == FIELD_NOT_SET) {
				hour = 0;
			}
			if (minute == FIELD_NOT_SET) {
				minute = 0;
			}
			if (second == FIELD_NOT_SET) {
				second = 0;
			}
			time.set(HOUR_OF_DAY, hour);
			time.set(MINUTE, minute);
			time.set(SECOND, second);
			time.set(MILLISECOND, 0);
			if (dayOfYear == FIELD_NOT_SET) {
				if (month == FIELD_NOT_SET) {
					month = 1;
				}
				if (dayOfMonth == FIELD_NOT_SET) {
					dayOfMonth = 1;
				}
				time.set(Calendar.MONTH, month - 1);
				time.set(Calendar.DAY_OF_MONTH, dayOfMonth);
			} else {
				time.set(Calendar.DAY_OF_YEAR, dayOfYear);
			}
			time.complete();
			boolean result = false;
			while (true) {
				if (time.get(YEAR) != year) {
					break;
				}
				if (month != FIELD_NOT_SET && time.get(MONTH) != month - 1) {
					break;
				}
				if (dayOfMonth != FIELD_NOT_SET && time.get(DAY_OF_MONTH) != dayOfMonth) {
					break;
				}
				if (dayOfYear != FIELD_NOT_SET && time.get(DAY_OF_YEAR) != dayOfYear) {
					break;
				}
				if (time.get(HOUR_OF_DAY) != hour) {
					break;
				}
				if (time.get(MINUTE) != minute) {
					break;
				}
				if (time.get(SECOND) != second) {
					break;
				}
				result = true;
				break;
			}
			if (!result) {
				time = null;
			}
		}
		return time;
	}
	/**
	 * Returns time as string YYYYMMDDHHmmSS
	 */
	public String getTimeString() {
		String timeStr = null;
		String year = "" + get(Calendar.YEAR);
		String month = "" + (get(Calendar.MONTH) + 1);
		String day = "" + get(Calendar.DAY_OF_MONTH);
		String hour = "" + get(Calendar.HOUR_OF_DAY);
		String min = "" + get(Calendar.MINUTE);
		String sec = "" + get(Calendar.SECOND);
		timeStr = FC.padLeft(year, 4, '0') + FC.padLeft(month, 2, '0') + FC.padLeft(day, 2, '0') + FC.padLeft(hour, 2, '0')
			+ FC.padLeft(min, 2, '0') + FC.padLeft(sec, 2, '0');
		return timeStr;
	}
	static public String millisToHHMMSS(long millis) {
		int sec = (int)Math.round(millis / 1000.);
		return FC.padLeft("" + sec / 3600, 2, '0') + ":" + FC.padLeft("" + sec % 3600 / 60, 2, '0') + ":" + FC.padLeft("" + sec % 60, 2, '0');
	}
	/**
	 * Convert milliseconds to HH:mm:ss.SSS string. Doesn't show hour if 0 hours and millis if 0 millis
	 */
	static public String millisToShortHHMMSS(long millis) {
		int sec = (int)(millis / 1000);
		millis = (int)(millis % 1000);
		int hour = sec / 3600;
		int min = (sec % 3600) / 60;
		sec = sec % 60;
		String str = "";
		if (hour > 0) {
			str += FC.padLeft("" + hour, 2, '0') + ":";
		}
		str += FC.padLeft("" + min, 2, '0') + ":" + FC.padLeft("" + sec, 2, '0');
		if (millis > 0) {
			str += "." + FC.padLeft("" + millis, 3, '0');
		}
		return str;
	}

	static private final double COEF_DEG_TO_MIN = 720.0 / 180.0;

	/**
	 * Taking longitude <longitude> in degrees given as eastern longitude ( from 0 to 360 deg ), recalculate time shift in minutes, which is
	 * the amount we have to add to the Universal Time to get specified Local Time for this longitude.
	 */
	static public double getLocalTimeShiftInMin(double longitude, TimeType localTimeType) {
		if (localTimeType == TimeType.AZT) {
			longitude = getAZTReferenceLogitude(longitude);
		} else if (localTimeType != TimeType.LMT) {
			throw new IllegalArgumentException("Parameter typeOfLocalTime is illegal (" + localTimeType + ")");
		}
		return getLongitudeTimeShiftInMin(longitude);
	}
	/**
	 * Having longitude calculate reference longitude for it in terms of Abstract Zone Time concept.
	 */
	static public double getAZTReferenceLogitude(double longitude) {
		longitude = 15 * Math.round(longitude / 15);
		if (longitude >= 360) {
			longitude = 0;
		}
		return longitude;
	}
	/**
	 * For specified longitude, <longitude>, and Type of Local Time, <typeOfLocalTime>, calculate time shift in minutes, which is the amount
	 * we have to add to the Universal Time to get specified Local Time for this longitude.
	 */
	static public double getLongitudeTimeShiftInMin(double longitude) {
		if (longitude < 180) {
			return longitude * COEF_DEG_TO_MIN;
		} else {
			return (longitude - 360) * COEF_DEG_TO_MIN;
		}
	}
	/**
	 * For given longitude and UT time given as TimeScale object returns corresponding Local Mean Time as TimeScale object. Longitude is
	 * given in degrees as geographical eastern longitude.
	 */
	static public TimeScale getLongitudeTime(double longitude, TimeScale ut) {
		return getLocalMeanTime(longitude, ut);
	}
	/**
	 * For given longitude and UT time given as TimeScale object returns corresponding Local Mean Time as TimeScale object. Longitude is
	 * given in degrees as geographical eastern longitude.
	 */
	static public TimeScale getLocalMeanTime(double longitude, TimeScale ut) {
		return new TimeScale(ut.getTimeInMinutes() + getLongitudeTimeShiftInMin(longitude));
	}
	/**
	 * For given longitude and UT time given as TimeScale object returns corresponding Local Time as TimeScale object. Longitude is given in
	 * degrees as geographical eastern longitude.
	 */
	static public TimeScale convertUT2Local(double longitude, TimeScale ut, TimeType localTimeType) {
		if (localTimeType == TimeType.AZT) {
			longitude = getAZTReferenceLogitude(longitude);
		} else if (localTimeType != TimeType.LMT) {
			throw new IllegalArgumentException("Parameter typeOfLocalTime is illegal (" + localTimeType + ")");
		}
		return getLocalMeanTime(longitude, ut);
	}
	/**
	 * For given longitude and Local Time for this longitude given as TimeScale object return corresponding Universal Time as TimeScale
	 * object. Longitude is given in degrees as geographical eastern longitude.
	 */
	static public TimeScale convertLocal2UT(double longitude, TimeScale lt, TimeType localTimeType) {
		if (localTimeType == TimeType.AZT) {
			longitude = getAZTReferenceLogitude(longitude);
		} else if (localTimeType != TimeType.LMT) {
			throw new IllegalArgumentException("Parameter typeOfLocalTime is illegal (" + localTimeType + ")");
		}
		return convertLMT2UT(longitude, lt);
	}
	/**
	 * For given longitude and Local Mean Time for this longitude given as TimeScale object return corresponding Universal Time as TimeScale
	 * object. Longitude is given in degrees as geographical eastern longitude.
	 */
	static public TimeScale getTimeUT(double longitude, TimeScale lmt) {
		return convertLMT2UT(longitude, lmt);
	}
	/**
	 * For given longitude and Local Mean Time for this longitude given as TimeScale object return corresponding Universal Time as TimeScale
	 * object. Longitude is given in degrees as geographical eastern longitude.
	 */
	static public TimeScale convertLMT2UT(double longitude, TimeScale lmt) {
		RealHours lmtRealHours = new RealHours(0, lmt);
		RealHours utRealHours = new RealHours(360 - longitude, lmt);
		TimeScale ut = new TimeScale(lmt.getTimeInMinutes());
		ut.set(Calendar.HOUR_OF_DAY, utRealHours.getHours());
		ut.set(Calendar.MINUTE, utRealHours.getMinutes());
		ut.set(Calendar.SECOND, utRealHours.getSeconds());
		ut.set(Calendar.MILLISECOND, utRealHours.getMillisecs());
		double lmtHours = lmtRealHours.getRealHours();
		double utHours = utRealHours.getRealHours();
		if (utHours < 12) {
			if (lmtHours > utHours + 12) {
				ut.add(Calendar.DAY_OF_MONTH, 1);
			}
		} else {
			if (lmtHours <= utHours - 12) {
				ut.add(Calendar.DAY_OF_MONTH, -1);
			}
		}
		ut.complete();
		return ut;
	}
	/**
	 * Convert one of the following types of time:
	 * <ol>
	 *   <li>Universal Time</li>
	 *   <li>Local Mean Time</li>
	 *   <li>Abstract Zone Time, as follows,
	 *     <ul>
	 *       <li>7.5deg to -7.5deg is 0 hours offset</li>
	 *       <li>352.5 to 345.5(=352.5-15) is -1 hours offset</li>
	 *       <li>..................</li> 
	 *       <li>7.5 to 22.5(=7.5+15) is 1 hours offset</li>
	 *     </ul>
	 *   </li>  
	 * </ol>
	 * 
	 * to the one of the same set.
	 */
	static public TimeScale convert(double longitude, TimeScale time, TimeType sourceTimeType, TimeType targetTimeType) {
		if (!checkTimeType(sourceTimeType, true)) {
			throw new IllegalArgumentException("sourceTypeOfTime == " + sourceTimeType);
		}
		if (!checkTimeType(targetTimeType, true)) {
			throw new IllegalArgumentException("targetTypeOfTime == " + targetTimeType);
		}
		if (sourceTimeType == targetTimeType) {
			return new TimeScale(time.getTimeInMinutes());
		}
		if (sourceTimeType == TimeType.UT) {
			return convertUT2Local(longitude, time, targetTimeType);
		} else if (targetTimeType == TimeType.UT) {
			return convertLocal2UT(longitude, time, sourceTimeType);
		} else {
			double longitudeDifference;
			if (sourceTimeType == TimeType.LMT) {
				longitudeDifference = 15 * Math.round(longitude / 15) - longitude;
			} else {
				longitudeDifference = longitude - 15 * Math.round(longitude / 15);
			}
			return new TimeScale(time.getTimeInMinutes() + longitudeDifference * COEF_DEG_TO_MIN);
		}
	}
	/**
	 * Checks Type of Time index for legality.
	 */
	static private boolean checkTimeType(TimeType timeType, boolean dontThrowException) {
		if (timeType != null) {
			return true;
		} else {
			if (dontThrowException) {
				return false;
			} else {
				throw new IllegalArgumentException("timeType is null");
			}
		}
	}
	/**
	 * For given longitude and UT time given as TimeScale object returns corresponding exact (longitude) local time as RealHours object.
	 * Longitude is given in degrees as geographical eastern longitude.
	 */
	static public RealHours getLongitudeHours(double longitude, TimeScale ut) {
		int utHours = ut.get(Calendar.HOUR_OF_DAY);
		int utMinutes = ut.get(Calendar.MINUTE);
		int utSeconds = ut.get(Calendar.SECOND);
		int utMillisecs = ut.get(Calendar.MILLISECOND);
		return getLongitudeHours(longitude, utHours, utMinutes, utSeconds, utMillisecs);
	}
	/**
	 * For given longitude and UT time given as whole hours of day returns corresponding exact (longitude) local time as RealHours object.
	 * Longitude is given in degrees as geographical eastern longitude.
	 */
	static public RealHours getLongitudeHours(double longitude, int utHours) {
		return getLongitudeHours(longitude, utHours, 0);
	}
	/**
	 * For given longitude and UT time given as hours and minutes of day returns corresponding exact (longitude) local time as RealHours
	 * object. Longitude is given in degrees as geographical eastern longitude.
	 */
	static public RealHours getLongitudeHours(double longitude, int utHours, int utMinutes) {
		return getLongitudeHours(longitude, utHours, utMinutes, 0);
	}
	/**
	 * For given longitude and UT time given as hours, minutes and seconds of day returns corresponding exact (longitude) local time as
	 * RealHours object. Longitude is given in degrees as geographical eastern longitude.
	 */
	static public RealHours getLongitudeHours(double longitude, int utHours, int utMinutes, int utSeconds) {
		return getLongitudeHours(longitude, utHours, utMinutes, utSeconds, 0);
	}
	/**
	 * For given longitude and UT time given as hours, minutes, seconds and milliseconds of day returns corresponding exact (longitude)
	 * local time as RealHours object. Longitude is given in degrees as geographical eastern longitude.
	 */
	static public RealHours getLongitudeHours(double longitude, int utHours, int utMinutes, int utSeconds, int utMillisecs) {
		double utRealHours =
			utHours + (utMinutes + (utSeconds + (double)utMillisecs / TimeScale.MILLISECONDS_PER_SECOND) / TimeScale.SECONDS_PER_MINUTE)
				/ TimeScale.MINUTES_PER_HOUR;
		return getLongitudeHours(longitude, utRealHours);
	}
	/**
	 * For given longitude and UT time given in real (fractional) hours of day returns corresponding exact (longitude) local time as
	 * RealHours object. Longitude is given in degrees as geographical eastern longitude.
	 */
	static public RealHours getLongitudeHours(double longitude, double utRealHours) {
		return new RealHours(longitude, utRealHours);
	}
	static public boolean checkYear(int year) {
		return year > 0 && year < 10000;
	}
	static public boolean checkMonth(int month) {
		return month >= 1 && month <= 12;
	}
	static public boolean checkDay(int day) {
		return day >= 1 && day <= 31;
	}
	static public boolean checkDay(int month, int day) {
		return day >= 1 && day <= getMaxDaysPerMonth(month);
	}
	static public boolean checkDay(int year, int month, int day) {
		return day >= 1 && day <= getMaxDaysPerMonth(year, month);
	}
	static public boolean checkDayOfYear(int year, int dayOfYear) {
		return dayOfYear >= 1 && dayOfYear <= (TM.isLeapYear(year) ? 366 : 365);
	}
	static public boolean checkDate(int year, int month, int day) {
		boolean ok = false;
		while (true) {
			if (!checkYear(year)) {
				break;
			}
			if (!checkMonth(month)) {
				break;
			}
			if (!checkDay(day)) {
				break;
			}
			if (day > getMaxDaysPerMonth(year, month)) {
				break;
			}
			ok = true;
			break;
		}
		return ok;
	}
	static public int getMaxDaysPerMonth(int year, int month) {
		int qty = 28;
		if (month != 2) {
			qty = maxDaysPerMonth[month - 1];
		} else if (TM.isLeapYear(year)) {
			qty = 29;
		}
		return qty;
	}
	static public int getMaxDaysPerMonth(int month) {
		return maxDaysPerMonth[month - 1];
	}
	static public boolean checkTime(int hr, int min, int sec) {
		return checkHour(hr) && checkMinute(min) && checkSecond(sec);
	}
	static public boolean checkHour(int hour) {
		return hour >= 0 && hour < TimeScale.HOURS_PER_DAY;
	}
	static public boolean checkMinute(int min) {
		return min >= 0 && min < TimeScale.MINUTES_PER_HOUR;
	}
	static public boolean checkSecond(int sec) {
		return sec >= 0 && sec < TimeScale.SECONDS_PER_MINUTE;
	}
	static public boolean checkMilliSecond(int msec) {
		return msec >= 0 && msec < TimeScale.MILLISECONDS_PER_SECOND;
	}
	static public int getMonth(int year, int dayInYear) {
		int month = 1;
		int day = dayInYear;
		for (int i = 0; i < TimeScale.MONTHS_PER_YEAR; i++) {
			day -= TM.numberOfDaysInMonth(i + 1, year);
			if (day <= 0) {
				break;
			}
			month++;
		}
		if (month > TimeScale.MONTHS_PER_YEAR) {
			month = -1;
		}
		return month;
	}
	static public int getDayInMonth(int year, int dayInYear) {
		int month = 1;
		int day = dayInYear;
		for (int i = 0; i < TimeScale.MONTHS_PER_YEAR; i++) {
			if (day <= TM.numberOfDaysInMonth(i + 1, year)) {
				break;
			} else {
				day -= TM.numberOfDaysInMonth(i + 1, year);
			}
			if (day <= 0) {
				break;
			}
			month++;
		}
		if (month > TimeScale.MONTHS_PER_YEAR) {
			day = -1;
		}
		return day;
	}
	/**
	 * Returns host local time as TimeScale object.
	 */
	static public TimeScale getHostLocalTime() {
		TimeScale time = new TimeScale();
		time.setTimeZone(TimeZone.getDefault());
		return time;
	}
	/**
	 * Prefixes message with local time followed by two spaces and returns combined string.
	 */
	static public String prefixWithLocalTime(String message) {
		return prefixWithLocalTime(message, "  ");
	}
	/**
	 * Prefixes message with local time followed by <code>separator</code>, and returns combined string.
	 */
	static public String prefixWithLocalTime(String message, String separator) {
		TimeScale time = getHostLocalTime();
		return time.toShortDateUT() + " " + time.toTimeUT() + separator + message;
	}
	
	static public boolean checkTimesUniformity(TimeScale[] times, int interval_sec) {
		for (int i = 1; i < times.length; i++) {
			if ((int) (times[i].getTimeInSeconds() - times[i - 1].getTimeInSeconds()) != interval_sec) {
				return false;
			}
		}
		return true;
	}

	static public void checkTimeIntervals(TimeScale[][] seTime) {
		if (seTime[0][0].after(seTime[0][1])) {
			throw new RuntimeException("Time period number " + 0 + " has start time after end time.");
		}
		System.out.println(seTime[0][0].toHumanUT() + " " + seTime[0][1].toHumanUT());
		for (int iTime = 1; iTime < seTime.length; iTime++) {
			if (seTime[iTime - 1][0].after(seTime[iTime][0]) || seTime[iTime - 1][1].after(seTime[iTime][0])) {
				throw new RuntimeException("Time periods should be ordered.");
			}
			if (seTime[iTime][0].after(seTime[iTime][1])) {
				throw new RuntimeException("Time period number " + iTime + " has start time after end time.");
			}
			System.out.println(seTime[iTime][0].toHumanUT() + " " + seTime[iTime][1].toHumanUT());
		}
	}

	
}